#include "ba0_exception.h"
#include "ba0_stack.h"
#include "ba0_basic_io.h"
#include "ba0_analex.h"
#include "ba0_global.h"

#define ba0_analex_save_full ba0_global.analex.analex_save_full
#define ba0_analex	     ba0_global.analex.analex
#define ba0_analex_save	     ba0_global.analex.analex_save
#define ba0_nbtokens	     ba0_initialized_global.analex.nb_tokens
#define ba0_token_lmax	     ba0_initialized_global.analex.token_lmax
#define ba0_context_analex   ba0_global.analex.context

/*
 * A zero value restores the setting
 */

BA0_DLL void ba0_set_settings_analex (ba0_int_p nbtokens, ba0_int_p token_lmax)
{
    ba0_nbtokens = nbtokens ? nbtokens : BA0_NBTOKENS;
    if (token_lmax == 0)
	ba0_token_lmax = BA0_TOKEN_LMAX;
    else if (BA0_SIZE_CELL_ANALEX_STACK <= 2*(unsigned ba0_int_p)token_lmax)
	ba0_token_lmax = BA0_SIZE_CELL_ANALEX_STACK / 2 - 1;
}

BA0_DLL void 
	ba0_get_settings_analex (ba0_int_p* nbtokens, ba0_int_p* token_lmax)
{
    if (nbtokens) *nbtokens = ba0_nbtokens;
    if (token_lmax) *token_lmax = ba0_token_lmax;
}

BA0_DLL char* ba0_get_context_analex (void)
{
    return ba0_context_analex;
}

/*
 * Writes the content of the fifo in ba0_context_analex
 * The current token is enclosed between --> and <--
 */

BA0_DLL void ba0_write_context_analex (void)
{   ba0_int_p l, l_unget, n_unget, n_get;

    ba0_context_analex [0] = '\0';

    if (ba0_type_token_analex () == ba0_no_token)
	return;

    l = strlen (ba0_value_token_analex ()) + strlen (" -->  <-- ");
    if (l >= BA0_CONTEXT_LMAX - 1)
	return;

    n_unget = 0;
    l_unget = 0;
    for (;;)
    {	if (ba0_analex.first == (ba0_analex.last + 1) % ba0_nbtokens)
	    break;
	ba0_unget_token_analex ();
	n_unget += 1;
	if (ba0_type_token_analex () == ba0_no_token)
	{   ba0_get_token_analex ();
	    n_unget -= 1;
	    break;
	}
	l_unget += strlen (ba0_value_token_analex ());
	if (ba0_analex.fifo [ba0_analex.first].spaces_before)
	    l_unget += 1;
	if (l + l_unget >= 2 * BA0_CONTEXT_LMAX / 3 - 1)
	{   ba0_get_token_analex ();
	    n_unget -= 1;
	    break;
	}
    }
    while (n_unget > 0)
    {	if (ba0_analex.fifo [ba0_analex.first].spaces_before)
	    strcat (ba0_context_analex, " ");
	strcat (ba0_context_analex, ba0_value_token_analex ());
	ba0_get_token_analex ();
	n_unget -= 1;
    }
    strcat (ba0_context_analex, " --> ");
    strcat (ba0_context_analex, ba0_value_token_analex ());
    strcat (ba0_context_analex, " <-- ");
    if (ba0_isatty_input ())
	return;
    n_get = 0;
    for (;;)
    {   if (n_get == ba0_nbtokens - 1)
	    break;
	ba0_get_token_analex ();
	n_get += 1;
	if (ba0_type_token_analex () == ba0_no_token)
	{   ba0_unget_token_analex ();
	    n_get -= 1;
	    break;
	}
	if (strlen (ba0_context_analex) + 
		strlen (ba0_value_token_analex ()) + 1 >= BA0_CONTEXT_LMAX - 1)
	{   ba0_unget_token_analex ();
	    n_get -= 1;
	    break;
	}
	if (ba0_spaces_before_token_analex ())
	    strcat (ba0_context_analex, " ");
	strcat (ba0_context_analex, ba0_value_token_analex ());
    }
    while (n_get > 0)
    {	ba0_unget_token_analex ();
	n_get -= 1;
    }
}

BA0_DLL void ba0_record_analex (void)
{
    if (ba0_analex_save_full)
	BA0_RAISE_EXCEPTION (BA0_ERRSOV);
    ba0_record_input ();
    ba0_analex_save.first = ba0_analex.first;
    ba0_analex_save.last = ba0_analex.last;
    memcpy (ba0_analex_save.fifo, ba0_analex.fifo, 
				ba0_nbtokens * sizeof (struct ba0_token));
    ba0_analex_save_full = true;
}

BA0_DLL void ba0_restore_analex (void)
{
    if (! ba0_analex_save_full)
	BA0_RAISE_EXCEPTION (BA0_ERRSOV);
    ba0_restore_input ();
    ba0_analex.first = ba0_analex_save.first;
    ba0_analex.last = ba0_analex_save.last;
    memcpy (ba0_analex.fifo, ba0_analex_save.fifo, 
				ba0_nbtokens * sizeof (struct ba0_token));
    ba0_analex_save_full = false;
}

/*
 * nb_tokens must not be modified between init and clear.
 */

BA0_DLL void ba0_init_analex (void)
{   ba0_int_p analex;

    ba0_get_settings_stack (0, 0, &analex, 0, 0);
    ba0_analex.fifo = (struct ba0_token*)ba0_malloc 
				(ba0_nbtokens * sizeof (struct ba0_token));
    ba0_analex_save.fifo = (struct ba0_token*)ba0_malloc
				(ba0_nbtokens * sizeof (struct ba0_token));
    ba0_reset_analex ();
}

BA0_DLL void ba0_clear_analex (void)
{
    ba0_free (ba0_analex.fifo);
    ba0_free (ba0_analex_save.fifo);
    ba0_malloc_counter -= 2 * ba0_nbtokens * sizeof (struct ba0_token);
}

static void ba0_empty_analex (void)
{   ba0_int_p i;

    for (i = 0; i < ba0_nbtokens; i++)
    {   ba0_analex.fifo [i].type = ba0_no_token;
	ba0_analex.fifo [i].spaces_before = true;
        ba0_analex.fifo [i].value = (char*)0;
    }
    ba0_analex.first = 0;
    ba0_analex.last = 0;
    ba0_context_analex [0] = '\0';
}

BA0_DLL void ba0_reset_analex (void)
{
    ba0_reset_input ();
    ba0_empty_analex ();
    ba0_analex_save_full = false;
    ba0_reset_stack (&ba0_analex_stack);
}

BA0_DLL void ba0_set_analex_FILE (FILE* f)
{
    ba0_set_input_FILE (f);
    ba0_empty_analex ();
}

BA0_DLL void ba0_set_analex_string (char* s)
{
    ba0_set_input_string (s);
    ba0_empty_analex ();
}

/*
   Looks for a cell in ba0_analex_stack.cells which involves none of the
   token present in ba0_analex and ba0_analex_save. -1 if not found.
*/

static ba0_int_p first_free_cell (void)
{   ba0_int_p k, i, j;
    bool free;

    free = false;
    k = 0;
    while (k < ba0_analex_stack.cells.size && ! free)
    {	free = true;
	for (i = 0; free && i < ba0_nbtokens; i++)
	{   if (ba0_analex.fifo [i].type != ba0_no_token)
	    {	j = ba0_cell_index_mark 
			(ba0_analex.fifo [i].value, &ba0_analex_stack.free);
		if (j == -1)
		    BA0_RAISE_EXCEPTION (BA0_ERRALG);
		free = j != k;
	    }
	}
	if (ba0_analex_save_full)
	{   for (i = 0; free && i < ba0_nbtokens; i++)
	    {	if (ba0_analex_save.fifo [i].type != ba0_no_token)
		{   j = ba0_cell_index_mark 
				(ba0_analex_save.fifo [i].value, 
				 &ba0_analex_stack.free);
		    if (j == -1)
			BA0_RAISE_EXCEPTION (BA0_ERRALG);
		    free = j != k;
		}
	    }
	}
	if (! free)
	    k += 1;
    }
    return free ? k : -1;
}

BA0_DLL void ba0_get_token_analex (void)
{   ba0_int_p i;
    bool spaces_before;
    int chrin;
    struct ba0_mark M;

    if (ba0_analex.first != ba0_analex.last)
	ba0_analex.first = (ba0_analex.first + 1) % ba0_nbtokens;
    else
    {   ba0_push_stack (&ba0_analex_stack);

/* Rotate cells if needed */

	if (ba0_alloc_would_malloc_a_cell (ba0_token_lmax))
	{   i = first_free_cell ();
	    if (i != -1)
		ba0_rotate_cells (i);
	}
	ba0_analex.first = (ba0_analex.first + 1) % ba0_nbtokens;
	ba0_analex.last = ba0_analex.first;

/* The mark will be used to shrink the allocated value */

	ba0_record (&M);
	ba0_analex.fifo [ba0_analex.last].value = 
			(char*)ba0_alloc (ba0_token_lmax * sizeof (char));

/* Read a char (skip spaces) */

	spaces_before = false;
	do
	{   do 
	    {	chrin = ba0_get_char (); 
		if (isspace (chrin))
		    spaces_before = true;
	    } while (chrin != EOF && isspace (chrin));
	    if (chrin == '#')
	    {	spaces_before = true;
		do
		    chrin = ba0_get_char ();
		while (chrin != EOF && chrin != '\n');
	    }
	} while (isspace (chrin));

	if (chrin == EOF)
	{   ba0_analex.fifo [ba0_analex.last].value [0] = '\0';
	    ba0_analex.fifo [ba0_analex.last].spaces_before = spaces_before;
            ba0_analex.fifo [ba0_analex.last].type = ba0_no_token;
	} else if (isdigit (chrin))
	{   for (i = 0; i < ba0_token_lmax-1 && isdigit (chrin); i++)
	    {	ba0_analex.fifo [ba0_analex.last].value [i] = chrin;
		chrin = ba0_get_char ();
	    }
	    if (i == ba0_token_lmax-1)
		BA0_RAISE_EXCEPTION (BA0_ERROOM);
	    ba0_unget_char (chrin);
	    ba0_analex.fifo [ba0_analex.last].value [i] = '\0';
	    ba0_analex.fifo [ba0_analex.last].spaces_before = spaces_before;
	    ba0_analex.fifo [ba0_analex.last].type = ba0_integer_token;
	} else if (isalpha (chrin) || chrin == '_')
	{   for (i = 0; i < ba0_token_lmax-1 && 
				(isalnum(chrin) || chrin == '_'); i++)
	    {	ba0_analex.fifo [ba0_analex.last].value [i] = chrin;
		chrin = ba0_get_char ();
	    }
	    if (i == ba0_token_lmax-1)
		BA0_RAISE_EXCEPTION (BA0_ERROOM);
	    ba0_unget_char (chrin);
	    ba0_analex.fifo [ba0_analex.last].value [i] = '\0';
	    ba0_analex.fifo [ba0_analex.last].spaces_before = spaces_before;
	    ba0_analex.fifo [ba0_analex.last].type = ba0_string_token;
	} else
	{   ba0_analex.fifo [ba0_analex.last].value [0] = chrin;
	    ba0_analex.fifo [ba0_analex.last].value [1] = '\0';
	    ba0_analex.fifo [ba0_analex.last].spaces_before = spaces_before;
	    ba0_analex.fifo [ba0_analex.last].type = ba0_sign_token;
	}

	ba0_pull_stack ();
    }
}

BA0_DLL void ba0_unget_token_analex (void)
{
/*
    if (ba0_type_token_analex () != ba0_no_token)
*/
	ba0_analex.first = 
			(ba0_analex.first + ba0_nbtokens - 1) % ba0_nbtokens;
}

BA0_DLL void ba0_unget_given_token_analex 
		(char* value, enum ba0_typeof_token type, bool spaces_before)
{   ba0_int_p i;

    if ((ba0_int_p)strlen (value) > ba0_token_lmax-1)
        BA0_RAISE_EXCEPTION (BA0_ERROOM);

    ba0_push_stack (&ba0_analex_stack);

/* Rotate cells if needed */

    if (ba0_alloc_would_malloc_a_cell (ba0_token_lmax))
    {   i = first_free_cell ();
        if (i != -1)
            ba0_rotate_cells (i);
    }

    ba0_analex.fifo [ba0_analex.first].value = 
                                (char*)ba0_alloc (strlen (value) + 1);
    strcpy (ba0_analex.fifo [ba0_analex.first].value, value);
    ba0_analex.fifo [ba0_analex.first].type = spaces_before;
    ba0_analex.fifo [ba0_analex.first].type = type;

    ba0_analex.first = (ba0_analex.first + ba0_nbtokens - 1) % ba0_nbtokens;

    ba0_pull_stack ();
}

BA0_DLL bool ba0_sign_token_analex (char* s)
{
    if (ba0_analex.fifo [ba0_analex.first].type != ba0_sign_token)
	return false;
    else
	return strcmp (s, ba0_analex.fifo [ba0_analex.first].value) == 0;
}

BA0_DLL bool ba0_spaces_before_token_analex (void)
{
    return ba0_analex.fifo [ba0_analex.first].spaces_before;
}

BA0_DLL enum ba0_typeof_token ba0_type_token_analex (void)
{
    return ba0_analex.fifo [ba0_analex.first].type;
}

BA0_DLL char* ba0_value_token_analex (void)
{
    return ba0_analex.fifo [ba0_analex.first].value;
}

