#include "ba0_exception.h"
#include "ba0_basic_io.h"
#include "ba0_global.h"

/*
 * OUTPUT
 */

#define ba0_output		ba0_global.basic_io.output
#define ba0_input		ba0_global.basic_io.input

#define ba0_output_stack	ba0_global.basic_io.output_stack
#define ba0_output_sp		ba0_global.basic_io.output_sp

#define ba0_input_stack		ba0_global.basic_io.input_stack
#define ba0_input_sp            ba0_global.basic_io.input_sp

BA0_DLL void ba0_record_output (void)
{
    if (ba0_output_sp >= BA0_BASIC_IO_SIZE_STACK)
	BA0_RAISE_EXCEPTION (BA0_ERRSOV);
    ba0_output_stack [ba0_output_sp++] = ba0_output;
}

BA0_DLL void ba0_restore_output (void)
{
    if (ba0_output_sp <= 0)
	BA0_RAISE_EXCEPTION (BA0_ERRSOV);
    ba0_output = ba0_output_stack [--ba0_output_sp];
}

BA0_DLL void ba0_set_output_FILE (FILE* f)
{   ba0_output.vers = ba0_file_device;
    ba0_output.file_flux = f;
    ba0_output.counter = 0;
}

BA0_DLL void ba0_set_output_string (char* s)
{   ba0_output.vers = ba0_string_device;
    ba0_output.string_flux = s;
    ba0_output.indice = 0;
    s [0] = '\0';
}

BA0_DLL void ba0_set_output_counter (void)
{
    ba0_output.vers = ba0_counter_device;
    ba0_output.counter = 0;
}

BA0_DLL ba0_int_p ba0_output_counter (void)
{
    if (ba0_output.vers != ba0_counter_device)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    return ba0_output.counter;
}

BA0_DLL void ba0_reset_output (void)
{
    ba0_set_output_FILE (stdout);
    ba0_output_sp = 0;
    ba0_output_line_length = BA0_DEFAULT_OUTPUT_LINE_LENGTH;
}

BA0_DLL void ba0_put_char (char c)
{
    switch (ba0_output.vers)
    {	case ba0_string_device:
	    ba0_output.string_flux [ba0_output.indice++] = c;
	    ba0_output.string_flux [ba0_output.indice] = '\0';
	    break;
	case ba0_counter_device:
	    ba0_output.counter++;
	    break;
	case ba0_file_device:
        {   if (c == '\n')
            {   ba0_output.counter = 0;
                fputc ('\n', ba0_output.file_flux);
            } else if (ba0_output_line_length != 0 &&
			ba0_output.counter + 1 >= ba0_output_line_length)
            {   ba0_output.counter = 1;
                fputc ('\n', ba0_output.file_flux);
                fputc (c, ba0_output.file_flux);
            } else
            {   ba0_output.counter++;
                fputc (c, ba0_output.file_flux);
            }
        }
    }
}

BA0_DLL void ba0_put_string (void* s)
{   char* z = (char*)s;

    switch (ba0_output.vers)
    {	case ba0_string_device:
	    strcpy (ba0_output.string_flux + ba0_output.indice, z);
	    ba0_output.indice += strlen (z);
	    break;
	case ba0_counter_device:
	    ba0_output.counter += strlen (z);
	    break;
	case ba0_file_device:
        {   int l = (int)strlen (z);
	    int i;
	    if (ba0_output_line_length != 0 &&
			l + ba0_output.counter >= ba0_output_line_length)
	    {   ba0_output.counter = 0;
	        fputc ('\n', ba0_output.file_flux);
	    }
	    fputs (z, ba0_output.file_flux);
	    for (i = 0; z [i] != '\0'; i++)
	        if (z [i] == '\n')
		    ba0_output.counter = 0;
	        else
		    ba0_output.counter++;
        }
    }
}

BA0_DLL void ba0_put_int_p (ba0_int_p n)
{   char buffer [4 * BA0_NBBITS_INT_P];
    sprintf (buffer, BA0_FORMAT_INT_P, n);
    ba0_put_string (buffer);
}

BA0_DLL void ba0_put_hexint_p (ba0_int_p n)
{   char buffer [4 * BA0_NBBITS_INT_P];
    sprintf (buffer, BA0_FORMAT_HEXINT_P, n);
    ba0_put_string (buffer);
}

/*
 * INPUT
 */

BA0_DLL void ba0_record_input (void)
{
    if (ba0_input_sp >= BA0_BASIC_IO_SIZE_STACK)
	BA0_RAISE_EXCEPTION (BA0_ERRSOV);
    ba0_input_stack [ba0_input_sp++] = ba0_input;
}

BA0_DLL void ba0_restore_input (void)
{
    if (ba0_input_sp <= 0)
        BA0_RAISE_EXCEPTION (BA0_ERRSOV);
    ba0_input = ba0_input_stack [--ba0_input_sp];
}

BA0_DLL void ba0_set_input_FILE (FILE* f)
{
    ba0_input.from = ba0_file_device;
    ba0_input.file_flux = f;
}

BA0_DLL void ba0_set_input_string (char* s)
{
    ba0_input.from = ba0_string_device;
    ba0_input.string_flux = s;
    ba0_input.indice = 0;
}

BA0_DLL void ba0_reset_input (void)
{
    ba0_set_input_FILE (stdin);
    ba0_input_sp = 0;
}

static int _mygetc (void)
{   int c = '\0';

    switch (ba0_input.from)
    {   case ba0_file_device:
	    c = fgetc (ba0_input.file_flux);
	    break;
	case ba0_counter_device:
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
	    break;
	case ba0_string_device:
            if (ba0_input.string_flux [ba0_input.indice] == '\0')
		c = EOF;
	    else
                c = (unsigned int)ba0_input.string_flux [ba0_input.indice++];
    }
    return c;
}

BA0_DLL bool ba0_isatty_input (void)
{
    return (ba0_input.from == ba0_file_device &&
				ba0_isatty (ba0_fileno (ba0_input.file_flux)));
}

/*
   Reads a character.
   A trailing slash at the end of the line is skipped together with all
	the following spaces.
*/

BA0_DLL int ba0_get_char (void)
{   int c = _mygetc ();
    if (c == '\\')
        do c = _mygetc (); while (c != EOF && isspace (c));
    return c;
}

/*
   Replaces c on the input stream.
   If the input stream is a string, this operation must not modify
   the stream (i.e. one cannot unget a character which was not read).
   Exception BA0_ERRALG is raised otherwise.
   This exception is also raised if too many characters are ungot.
*/

BA0_DLL void ba0_unget_char (int c)
{
    switch (ba0_input.from)
    {   case ba0_file_device:
	    if (c != EOF)
                ungetc (c, ba0_input.file_flux);
	    break;
	case ba0_counter_device:
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
	    break;
	case ba0_string_device:
            if (ba0_input.indice == 0)
		BA0_RAISE_EXCEPTION (BA0_ERRALG);
	    else if (c != EOF)
	    {	if (c != ba0_input.string_flux [ba0_input.indice - 1])
		    BA0_RAISE_EXCEPTION (BA0_ERRALG);
		else
		    ba0_input.indice--;
	    }
    }
}


