#include "ba0_double.h"
#include "ba0_mesgerr.h"
#include "ba0_basic_io.h"
#include "ba0_analex.h"
#include "ba0_garbage.h"
#include "ba0_exception.h"

BA0_DLL ba0_double ba0_new_double (void)
{
    return (ba0_double)ba0_alloc (sizeof (double));
}

BA0_DLL void* ba0_scanf_double (void* z)
{   struct ba0_mark M;
    char *buf1;
    char *buf2;
    ba0_double e;
    ba0_int_p token_lmax;
    bool mantissa;
    bool minus_sign;

    ba0_get_settings_analex (0, &token_lmax);
    ba0_push_another_stack ();
    ba0_record (&M);
    buf1 = (char*)ba0_alloc (token_lmax);
    buf2 = (char*)ba0_alloc (token_lmax);
    ba0_pull_stack ();

    if (z == (void*)0)
	e = (ba0_double)ba0_alloc (sizeof (double));
    else
	e = (ba0_double)z;
/*
   Possibly start with a minus sign.
*/
    if (ba0_sign_token_analex ("-"))
    {	ba0_get_token_analex ();
	minus_sign = true;
    } else
    {	if (ba0_sign_token_analex ("+"))
	    ba0_get_token_analex ();
	minus_sign = false;
    }
/*
   Possibly a non empty integer part
   Spaces are authorized between the minus sign and what follows.
*/
    mantissa = false;
    buf1 [0] = '\0';
    if (ba0_type_token_analex () == ba0_integer_token)
    {	strcat (buf1, ba0_value_token_analex ());
	ba0_get_token_analex ();
	mantissa = true;
    }
/*
   Possibly a dot plus a possible fractional part
*/
    if ((! mantissa || ! ba0_spaces_before_token_analex ()) && 
						ba0_sign_token_analex ("."))
    {	strcat (buf1, ".");
	ba0_get_token_analex ();
	if (! ba0_spaces_before_token_analex () && 
			ba0_type_token_analex () == ba0_integer_token)
	{   strcat (buf1, ba0_value_token_analex ());
	    ba0_get_token_analex ();
	}
	mantissa = true;
    }
/*
   There must be a mantissa
*/
    if (! mantissa)
        BA0_RAISE_PARSER_EXCEPTION (BA0_ERRFLT);
/*
   Possibly an exponent
*/
    strcpy (buf2, ba0_value_token_analex ());
    if (! ba0_spaces_before_token_analex () &&
		(buf2 [0] == 'e' || buf2 [0] == 'E' || buf2 [0] == '@') &&
		(buf2 [1] == '\0' || isdigit ((int)buf2 [1])))
    {   strcat (buf1, "e");
	if (isdigit ((int)buf2 [1]))
	{   ba0_int_p i;
	    i = 2;
	    while (isdigit ((int)buf2 [i]))
		i++;
	    if (buf2 [i] == '\0')
		ba0_unget_given_token_analex 
					(buf2 + 1, ba0_integer_token, false);
	    else
	    {	ba0_unget_given_token_analex 
					(buf2 + i, ba0_string_token, false);
		buf2 [i] = '\0';
		ba0_unget_given_token_analex 
					(buf2 + 1, ba0_integer_token, false);
	    }
	} 

	ba0_get_token_analex ();
	if (! ba0_spaces_before_token_analex () && 
			ba0_type_token_analex () == ba0_integer_token)
	{   strcat (buf1, ba0_value_token_analex ());
	} else if (! ba0_spaces_before_token_analex () && 
					(ba0_sign_token_analex ("+") || 
					 ba0_sign_token_analex ("-")))
	{   strcat (buf1, ba0_value_token_analex ());
	    ba0_get_token_analex ();
	    if (ba0_spaces_before_token_analex () ||
				ba0_type_token_analex () != ba0_integer_token)
		BA0_RAISE_PARSER_EXCEPTION (BA0_ERRFLT);
	    strcat (buf1, ba0_value_token_analex ());
	} else
	    BA0_RAISE_PARSER_EXCEPTION (BA0_ERRFLT);
    } else
	ba0_unget_token_analex ();

    *e = atof (buf1);
    if (minus_sign)
	*e = - *e;
    ba0_restore (&M);
    return e;
}

BA0_DLL void ba0_printf_double (void* z)
{   char buffer [64];
    sprintf (buffer, "%f", *(ba0_double)z);
    ba0_put_string (buffer);
}

/*
 * Readonly static data
 */

static char _double [] = "double";

BA0_DLL ba0_int_p ba0_garbage1_double (void* z, enum ba0_garbage_code code)
{
    if (code == ba0_isolated)
    {	ba0_new_gc_info (z, sizeof (double), _double);
	return 1;
    } else
	return 0;
}

BA0_DLL void* ba0_garbage2_double (void* z, enum ba0_garbage_code code)
{
    if (code == ba0_isolated)
	return ba0_new_addr_gc_info (z, _double);
    else
	return z;
}

BA0_DLL void* ba0_copy_double (void* z)
{   ba0_double e;

    e = ba0_new_double ();
    *e = *(ba0_double)z;
    return e;
}

        

