#include "ba0_garbage.h"
#include "ba0_copy.h"
#include "ba0_exception.h"
#include "ba0_global.h"

#define ba0_exception_extra_stack ba0_global.exception.exception_extra_stack
#define ba0_exception_extra_sp    ba0_global.exception.exception_extra_sp
#define ba0_within_push_exception ba0_global.exception.within_push_exception
#define ba0_buffer		ba0_global.exception.buffer
#define ba0_exception_log	ba0_global.exception.exception_log
#define ba0_exception_log_qp	ba0_global.exception.exception_log_qp

BA0_DLL void ba0_reset_exception_extra_stack (void)
{
    ba0_exception_extra_sp = -1;
}

BA0_DLL void ba0_push_exception_extra_stack 
			(ba0_int_p* p, ba0_extra_restore_function* r)
{
    ba0_exception_extra_sp++;

    if (ba0_exception_extra_sp < 0 || 
		ba0_exception_extra_sp >= BA0_SIZE_EXCEPTION_EXTRA_STACK)
	BA0_RAISE_EXCEPTION (BA0_ERRSOV);
    ba0_exception_extra_stack [ba0_exception_extra_sp].pointer = p;
    ba0_exception_extra_stack [ba0_exception_extra_sp].restore = r;
}

BA0_DLL void ba0_pull_exception_extra_stack (void)
{
    if (ba0_exception_extra_sp < 0 || 
		ba0_exception_extra_sp >= BA0_SIZE_EXCEPTION_EXTRA_STACK)
        BA0_RAISE_EXCEPTION (BA0_ERRALG);
    ba0_exception_extra_sp--;
}

BA0_DLL void ba0_reset_exception (void)
{   ba0_int_p i;

    ba0_exception_sp = -1;
    ba0_exception_log_qp = -1;
    ba0_within_push_exception = false;
    for (i = 0; i < BA0_SIZE_EXCEPTION_LOG; i++)
	ba0_exception_log [i].file = (char*)0;
}

BA0_DLL void ba0_push_exception (char* file, int line)
{   char* _file;
    int _line;
    int i;

    _file = file;
    _line = line;

    if (ba0_within_push_exception)
    {   ba0_within_push_exception = false;
	BA0_CERR (BA0_ERRALG);
    }

    ba0_within_push_exception = true;

    ba0_exception_log_qp = -1;

    if (ba0_exception_sp >= BA0_SIZE_EXCEPTION_STACK - 1)
    {   ba0_within_push_exception = false;
        BA0_CERR (BA0_ERRSOV);
    }

    ba0_push_stack (&ba0_main_stack);
    ba0_record (&ba0_exception_stack [ba0_exception_sp+1].main);
    ba0_pull_stack ();
    ba0_push_stack (&ba0_second_stack);
    ba0_record (&ba0_exception_stack [ba0_exception_sp+1].second);
    ba0_pull_stack ();
    ba0_exception_stack [ba0_exception_sp+1].stack_sp = ba0_stack_sp;
    for (i = 0; i <= ba0_exception_extra_sp; i++)
	ba0_exception_stack [ba0_exception_sp+1].extra [i] = 
					*ba0_exception_extra_stack [i].pointer;
    ba0_exception_sp++;

    ba0_within_push_exception = false;
}


BA0_DLL void ba0_pull_exception (struct ba0_exception_code code)
{
    if (ba0_exception_sp != code.exception_sp)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    ba0_exception_sp--;
}

BA0_DLL bool ba0_exception_is_set (struct ba0_exception_code code)
{
    return code.jmp_code == 0;
}

BA0_DLL bool ba0_exception_is_raised (struct ba0_exception_code code)
{
    return code.jmp_code != 0;
}

BA0_DLL void ba0_raise_exception (char* file, int line, char* s)
{   ba0_int_p i;

    ba0_exception_log_qp = (ba0_exception_log_qp + 1) % BA0_SIZE_EXCEPTION_LOG;
    ba0_exception_log [ba0_exception_log_qp].file = file;
    ba0_exception_log [ba0_exception_log_qp].line = line;

    if (s == BA0_ERRALG || s == BA0_ERRNYP)
	fprintf (stderr, "trapped '%s' at file: %s:%d\n", s, file, line);

    if (ba0_exception_sp < 0)
    {	for (i = ba0_exception_log_qp + 1; i < BA0_SIZE_EXCEPTION_LOG; i++)
	    ba0_exception_log [i].file = (char*)0;
	sprintf (ba0_buffer, "%s (%s)", BA0_ERRNCE, s);
	BA0_CERR (ba0_buffer);
    }

    ba0_mesgerr = s;
    ba0_restore (&ba0_exception_stack [ba0_exception_sp].main);
    ba0_restore (&ba0_exception_stack [ba0_exception_sp].second);
    ba0_stack_sp = ba0_exception_stack [ba0_exception_sp].stack_sp;
    for (i = 0; i <= ba0_exception_extra_sp; i++)
    {	ba0_extra_restore_function* r = ba0_exception_extra_stack [i].restore;
	if (r == (ba0_extra_restore_function*)0)
	    *ba0_exception_extra_stack [i].pointer = 
			ba0_exception_stack [ba0_exception_sp].extra [i];
	else
	    (*r) (ba0_exception_stack [ba0_exception_sp].extra [i]);
    }

    ba0_exception_sp--;

    ba0_longjmp (ba0_exception_stack [ba0_exception_sp + 1].jmp_b, 1);
}

BA0_DLL void ba0_raise_exception2 
			(char* file, int line, char* s, char* f, void** o)
{   struct ba0_stack* H;
    ba0_int_p i;

    ba0_exception_log_qp = (ba0_exception_log_qp + 1) % BA0_SIZE_EXCEPTION_LOG;
    ba0_exception_log [ba0_exception_log_qp].file = file;
    ba0_exception_log [ba0_exception_log_qp].line = line;
/*
    if (s == BA0_ERRALG)
	fprintf (stderr, 
		"trapped BA0_ERRALG at file: %s, line: %d\n", file, line);
*/
    if (ba0_exception_sp < 0)
    {   for (i = ba0_exception_log_qp + 1; i < BA0_SIZE_EXCEPTION_LOG; i++)
            ba0_exception_log [i].file = (char*)0;
        sprintf (ba0_buffer, "%s (%s)", BA0_ERRNCE, s);
        BA0_CERR (ba0_buffer);
    }

    ba0_mesgerr = s;

    H = ba0_which_stack (*o);

    if (H == (struct ba0_stack*)0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    else if (H != ba0_stack_of_stacks 
			[ba0_exception_stack [ba0_exception_sp].stack_sp])
    {	ba0_restore (&ba0_exception_stack [ba0_exception_sp].main);
	ba0_restore (&ba0_exception_stack [ba0_exception_sp].second);
	ba0_stack_sp = ba0_exception_stack [ba0_exception_sp].stack_sp;
	*o = ba0_copy (f, *o);
    } else if (H == &ba0_main_stack)
    {	ba0_garbage (f, &ba0_exception_stack [ba0_exception_sp].main, o);
	ba0_restore (&ba0_exception_stack [ba0_exception_sp].second);
    } else if (H == &ba0_second_stack)
    {   ba0_garbage (f, &ba0_exception_stack [ba0_exception_sp].second, o);
        ba0_restore (&ba0_exception_stack [ba0_exception_sp].main);
    } else
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
/*
    if (ba0_stack_of_stacks [ba0_stack_sp] == 
	    ba0_stack_of_stacks 
		[ba0_exception_stack [ba0_exception_sp].stack_sp])
    {	if (ba0_stack_of_stacks [ba0_stack_sp] == 
			ba0_exception_stack [ba0_exception_sp].main.stack)
	{   ba0_garbage (f, &ba0_exception_stack [ba0_exception_sp].main, o);
	    ba0_restore (&ba0_exception_stack [ba0_exception_sp].second);
	} else
	{   ba0_garbage (f, &ba0_exception_stack [ba0_exception_sp].second, o);
	    ba0_restore (&ba0_exception_stack [ba0_exception_sp].main);
	}
	ba0_stack_sp = ba0_exception_stack [ba0_exception_sp].stack_sp;
    } else
    {	ba0_restore (&ba0_exception_stack [ba0_exception_sp].main);
	ba0_restore (&ba0_exception_stack [ba0_exception_sp].second);
	ba0_stack_sp = ba0_exception_stack [ba0_exception_sp].stack_sp;
	*o = ba0_copy (f, *o);
    } 
*/
    for (i = 0; i <= ba0_exception_extra_sp; i++)
    {	ba0_extra_restore_function* r = ba0_exception_extra_stack [i].restore;
	if (r == (ba0_extra_restore_function*)0)
	    *ba0_exception_extra_stack [i].pointer = 
			ba0_exception_stack [ba0_exception_sp].extra [i];
	else
	    (*r) (ba0_exception_stack [ba0_exception_sp].extra [i]);
    }

    ba0_exception_sp--;

    ba0_longjmp (ba0_exception_stack [ba0_exception_sp + 1].jmp_b, 1);
}

BA0_DLL void ba0_cerr (char* fname, int line, char* s)
{   ba0_int_p i;

    fprintf (stderr, "Error in file %s:%d\n%s\n", fname, line, s);
    fprintf (stderr, "Exception log:\n");
    for (i = 0; i < BA0_SIZE_EXCEPTION_LOG && ba0_exception_log [i].file; i++)
    {	if (i == ba0_exception_log_qp)
	    fprintf (stderr, "--> ");
	else
	    fprintf (stderr, "    ");
	fprintf (stderr, "file: %s:%d\n",
		ba0_exception_log [i].file, ba0_exception_log [i].line);
    }
    exit (1);
}

