#ifndef BA0_EXCEPTION_H
#define BA0_EXCEPTION_H 

#include "ba0_common.h"
#include "ba0_stack.h"
#include "ba0_analex.h"
#include "ba0_mesgerr.h"

BEGIN_C_DECLS

/*
 * jmp_b    : used by setjmp/longjmp
 * main     : mark used for restoring the free pointer of the main stack
 * second   : mark used used for restoring the free pointer of the second stack
 * stack_sp : provides the current stack
 * extra_sp : index for both extra and extra_rf
 * extra    : values used for restoring further global variables
 */ 

#define	BA0_SIZE_EXCEPTION_EXTRA_STACK 10
struct ba0_exception {
    ba0_jmp_buf jmp_b;
    struct ba0_mark main;
    struct ba0_mark second;
    ba0_int_p stack_sp;
    ba0_int_p extra [BA0_SIZE_EXCEPTION_EXTRA_STACK];
};

/*
 * The global variables concerned by extra are stored in a stack.
 * The entries of this stack are as follows.
 * 
 * pointer : a pointer to the global variable to be restored.
 * restore : a pointer to the restoring function (possibly zero).
 */

typedef void ba0_extra_restore_function (ba0_int_p);

struct ba0_exception_extra {
    ba0_int_p* pointer;
    ba0_extra_restore_function *restore;
};

/*
 * The data type for local variables handling exceptions
 */

struct ba0_exception_code {
    ba0_int_p exception_sp;
    int jmp_code;
};

/*
 * The current exception
 */

#define ba0_mesgerr		ba0_global.exception.mesgerr

/*
 * The exception stack.
 */

#define BA0_SIZE_EXCEPTION_STACK	100
#define ba0_exception_stack	ba0_global.exception.exception_stack
#define ba0_exception_sp	ba0_global.exception.exception_sp

/*
 * The size of the exception log fifo.
 * And the type of log fifo entry.
 */

#define BA0_SIZE_EXCEPTION_LOG	10
struct ba0_exception_log {
    char *file;
    int line;
};

/*
 * The macros for throwing/catching exceptions
 */

#define BA0_PUSH_EXCEPTION(code) do { \
    ba0_push_exception (__FILE__, __LINE__);			\
    code.exception_sp = ba0_exception_sp;	\
    code.jmp_code = ba0_setjmp (ba0_exception_stack [ba0_exception_sp].jmp_b,1);\
    } while (0)

#define BA0_RAISE_EXCEPTION(msg) ba0_raise_exception (__FILE__, __LINE__, msg)
#define BA0_RAISE_PARSER_EXCEPTION(msg) do { \
    ba0_write_context_analex ();	      \
    ba0_raise_exception (__FILE__, __LINE__, msg); \
    } while (0)

#define BA0_RAISE_EXCEPTION2(msg,f,o) \
	ba0_raise_exception2 (__FILE__, __LINE__, msg, f, o)
#define BA0_CERR(msg) ba0_cerr (__FILE__, __LINE__, msg)

extern BA0_DLL void ba0_reset_exception_extra_stack (void);
extern BA0_DLL void ba0_push_exception_extra_stack 
				(ba0_int_p*, ba0_extra_restore_function*);
extern BA0_DLL void ba0_pull_exception_extra_stack (void);

extern BA0_DLL void ba0_reset_exception (void);
extern BA0_DLL void ba0_push_exception (char*, int);
extern BA0_DLL void ba0_pull_exception (struct ba0_exception_code);
extern BA0_DLL bool ba0_exception_is_raised (struct ba0_exception_code);
extern BA0_DLL bool ba0_exception_is_set (struct ba0_exception_code);
extern BA0_DLL void ba0_raise_exception (char*, int, char*);
extern BA0_DLL void ba0_raise_exception2 (char*, int, char*, char*, void**);
extern BA0_DLL void ba0_cerr (char*, int, char*);

END_C_DECLS

#endif /* !BA0_EXCEPTION_H */
