#include "ba0_exception.h"
#include "ba0_stack.h"
#include "ba0_format.h"
#include "ba0_global.h"

#define BA0_SUBFORMAT_REALLOC_VALUE	512

#define ba0_leaf_subformat 	ba0_global.format.leaf_subformat
#define ba0_htable		ba0_global.format.htable
#define ba0_nbelem_htable	ba0_global.format.nbelem_htable
#define ba0_input_string	ba0_global.format.input_string
#define ba0_index		ba0_global.format.index
#define ba0_buffer		ba0_global.format.buffer

/*
 * Initializes the table of the atomic formats and the H-table of the formats.
 */

BA0_DLL void ba0_initialize_format (void)
{   ba0_int_p i;

    ba0_push_stack (&ba0_format_stack);
    
    ba0_init_table ((ba0_table)&ba0_leaf_subformat);
    ba0_realloc_table 
		((ba0_table)&ba0_leaf_subformat, BA0_SUBFORMAT_REALLOC_VALUE);

    ba0_init_table ((ba0_table)&ba0_htable);
    ba0_realloc_table ((ba0_table)&ba0_htable, BA0_SIZE_HTABLE_FORMAT);
    ba0_htable.size = ba0_htable.alloc;
    for (i = 0; i < ba0_htable.size; i++)
	ba0_htable.tab [i] = (struct ba0_pair*)0;
    ba0_nbelem_htable = 0;

    ba0_pull_stack ();
}

static ba0_int_p ba0_not_yet_implemented 
				(void* objet, enum ba0_garbage_code code)
{   objet = (void*)0;
    code = ba0_isolated;

    BA0_RAISE_EXCEPTION (BA0_ERRNYP);
    return 0;	/* to avoid an annoying warning */
}

BA0_DLL ba0_int_p ba0_empty_garbage1 (void* objet, enum ba0_garbage_code code)
{   objet = (void*)0;
    code = ba0_isolated;

    return 0; /* to avoid an annoying warning */
}

BA0_DLL void* ba0_empty_garbage2 (void* objet, enum ba0_garbage_code code)
{   code = ba0_isolated;

    return objet;
}

static void* identity (void* objet)
{
    return objet;
}

BA0_DLL void ba0_define_format (char* id, ba0_scanf_function* s,
                        ba0_printf_function* p, ba0_garbage1_function* g1,
                        ba0_garbage2_function* g2, ba0_copy_function* c)
{
    ba0_define_format_with_sizelt (id, -1, s, p, g1, g2, c);
}

BA0_DLL void ba0_define_format_with_sizelt (char* id, ba0_int_p sizelt, 
			ba0_scanf_function* s,
                        ba0_printf_function* p, ba0_garbage1_function* g1,
                        ba0_garbage2_function* g2, ba0_copy_function* c)

{   struct ba0_pair* pair;
    ba0_subformat sf;
    ba0_int_p i;

    for (i = 0; i < ba0_leaf_subformat.size; i++)
	if (strcmp (id, ba0_leaf_subformat.tab [i]->identificateur) == 0)
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
    
    ba0_push_stack (&ba0_format_stack);

    if (ba0_leaf_subformat.size == ba0_leaf_subformat.alloc)
	ba0_realloc_table ((ba0_table)&ba0_leaf_subformat, 
		   ba0_leaf_subformat.alloc + BA0_SUBFORMAT_REALLOC_VALUE);

    pair = (struct ba0_pair*)ba0_alloc (sizeof (struct ba0_pair));
    pair->identificateur = id;
    pair->value = sf = 
		(ba0_subformat)ba0_alloc (sizeof (struct ba0_subformat));
    
    sf->code = ba0_leaf_format;
    sf->u.leaf.sizelt = sizelt;

    if (s == (ba0_scanf_function*)-1)
	sf->u.leaf.scanf = (ba0_scanf_function*)&ba0_not_yet_implemented;
    else
	sf->u.leaf.scanf = s;

    if (p == (ba0_printf_function*)-1)
	sf->u.leaf.printf = (ba0_printf_function*)&ba0_not_yet_implemented;
    else
	sf->u.leaf.printf = p;

    if (g1 == (ba0_garbage1_function*)-1)
	sf->u.leaf.garbage1 = (ba0_garbage1_function*)&ba0_not_yet_implemented;
    else if (g1 == (ba0_garbage1_function*)0)
	sf->u.leaf.garbage1 = &ba0_empty_garbage1;
    else
	sf->u.leaf.garbage1 = g1;

    if (g2 == (ba0_garbage2_function*)-1)
        sf->u.leaf.garbage2 = (ba0_garbage2_function*)&ba0_not_yet_implemented;
    else if (g2 == (ba0_garbage2_function*)0)
        sf->u.leaf.garbage2 = &ba0_empty_garbage2;
    else
        sf->u.leaf.garbage2 = g2;

    if (c == (ba0_copy_function*)-1)
	sf->u.leaf.copy = (ba0_copy_function*)&ba0_not_yet_implemented;
    else if (c == (ba0_copy_function*)0)
	sf->u.leaf.copy = &identity;
    else
	sf->u.leaf.copy = c;

    ba0_leaf_subformat.tab [ba0_leaf_subformat.size++] = pair;
 
    ba0_pull_stack ();
}

/*****************************************************************************
 CONVERSION STRING -> FORMAT
 *****************************************************************************/

static bool is_open (char c)
{
    return c == '(' || c == '[' || c == '{' || c == '<';
}

static bool is_close (char c)
{
    return c == ')' || c == ']' || c == '}' || c == '>';
}

static char matching_char (char c)
{
    switch (c)
    {   case '(':
            return ')';
        case '[':
            return ']';
        case '{':
            return '}';
        case '<':
            return '>';
        default:
	    return '\0';
    }
}

/* 
 * Extract the identifier which follows some '%' in ba0_input_string.
 * After the call, ba0_index points to the last char of the identifier.
 * *b points to the identifier, stored in the ba0_buffer variable.
 */

static void get_code (char** b)
{   ba0_int_p i;

    if (ba0_input_string [ba0_index] == '%')
	ba0_index++;
    i = 0;
    while (isalnum((int)ba0_input_string [ba0_index]) || 
					ba0_input_string [ba0_index] == '_')
	ba0_buffer [i++] = ba0_input_string [ba0_index++];
    ba0_index--;
    ba0_buffer [i] = '\0';
    *b = ba0_buffer;
}

/* 
 * Extracts an opening parenthesis.
 * After the call, ba0_index points to this parenthesis.
 */

static char get_opening ()
{
    while (ba0_input_string [ba0_index] != '(' && 
		ba0_input_string [ba0_index] != '[' &&
		ba0_input_string [ba0_index] != '{' && 
		ba0_input_string [ba0_index] != '<' &&
		ba0_input_string [ba0_index] != '\0')
	ba0_index++;
    if (ba0_input_string [ba0_index] == '\0')
    	BA0_RAISE_EXCEPTION (BA0_ERRSYN);
    return ba0_input_string [ba0_index];
}

/* 
 * Analyzes ba0_input_string and computes the length of the text field,
 * as well as the number of subformats. Does not modify ba0_index.
 *
 * The stack (pile) contains '%' and opening parentheses.
 * nombre_de_pourcents gives the number of '%' in pile.
 *
 * The only stacked '%' are those which correspond to sub-ojects (%t, %l).
 *
 * We are at top_level if nombre_de_pourcents is zero.
 *
 * All opening parentheses are stacked.
 *
 * In the case of a string staring by an opening parenthesis, this one
 * is not taken into account.
 */

static void first_analysis (ba0_int_p* stringlength, ba0_int_p* linknumber)
{   static char pile [256];
    char   chrin, *code;
    ba0_int_p nombre_de_pourcents = 0;
    ba0_int_p sp = -1;
    ba0_int_p ba0_index_save = ba0_index;

    chrin = ba0_input_string [ba0_index];
    *stringlength = 0;
    *linknumber = 0;

    if (chrin == '\0')
        BA0_RAISE_EXCEPTION (BA0_ERRSYN);

    for (;;)
    {	

/*
 * If chrin is a closing parenthesis, the opening one is pulled.
 * If the former stacked char is a '%', it is pulled and nombre_de_pourcents
 * is decremented.
 * *stringlenth is never incremented.
 */
	if (chrin == '\0')
	{   if (sp < 0)
		break;
	    BA0_RAISE_EXCEPTION (BA0_ERRSYN);
	} else if (is_close (chrin))
	{   if (sp < 0)
		break;
	    if (chrin != matching_char (pile [sp]))
	    	BA0_RAISE_EXCEPTION (BA0_ERRSYN);
	    sp--;
	    if (sp >= 0 && pile [sp] == '%')
	    {	sp--;
		nombre_de_pourcents--;
	    }
	} else if (chrin == '%')
	{   
/*
 * A '%' accounts for one char since the possible opening and closing
 * parentheses which come with it are stored outside text.
 *
 * *linknumber is incremented if we are at the top level.
 *
 * If the code is l or t, a '%' and the opening parenthesis are stacked
 * and nombre_de_pourcents is incremented.
 */
	    if (nombre_de_pourcents == 0)
	    {	(*stringlength)++;
		(*linknumber)++;
	    }
	    get_code (&code);
	    if (strcmp (code, "t") == 0 || strcmp (code, "l") == 0 
		|| strcmp (code, "m") == 0 || strcmp (code, "a") == 0)
	    {	pile [++sp] = '%';
	    	pile [++sp] = get_opening ();
		nombre_de_pourcents++;
	    } 
	} else if (is_open (chrin))
	{   
/*
 * chrin is an opening parenthesis which does not come with any '%'.
 * It is stacked. If nombre_de_pourcents = 0, one counts also the 
 * matching closing parenthesis.
 */
	    pile [++sp] = chrin;
	    if (nombre_de_pourcents == 0)
		*stringlength += 2;
	} else if (nombre_de_pourcents == 0)

/* 
 * A tacky character is counted iff nombre_de_pourcents = 0.
 */

	    (*stringlength)++;
	chrin = ba0_input_string [++ba0_index];
    }
    ba0_index = ba0_index_save;
}

/*
 * The ba0_index variables points to the first character to analyze.
 * After the call, it points to the last analyzed one.
 *
 * The niveau variable provides the number of met opening parentheses.
 * One stops if niveau reaches -1.
 * It is pointless to do things more carefully.
 *
 * No need to use a stack because of recursive calls.
 */

static ba0_format _formate ()
{   ba0_int_p textlen, linknmb;
    ba0_int_p itext, ilink;
    ba0_format f;
    ba0_int_p niveau;
    char chrin;

    first_analysis (&textlen, &linknmb);

    f = (ba0_format) ba0_alloc (sizeof (struct ba0_format));
    f->text = (char*) ba0_alloc (textlen + 1);
    f->link = (ba0_subformat*) ba0_alloc (sizeof (ba0_subformat) * linknmb);
    f->linknmb = linknmb;
    itext = 0;
    ilink = 0;

    niveau = 0;
    chrin = ba0_input_string [ba0_index];
    for (;;)
    {	if (is_open (chrin))
	    niveau++;
	else if (chrin == '\0' || is_close (chrin))
	{   if (--niveau < 0)
		break;
	} else if (chrin == '%')
	{   char* code;
	    get_code (&code);
	    if (strcmp (code, "t") == 0 || strcmp (code, "l") == 0 || 
		strcmp (code, "a") == 0 || strcmp (code, "m") == 0)
	    {	ba0_subformat l = 
		    (ba0_subformat)ba0_alloc (sizeof (struct ba0_subformat));
		enum ba0_typeof_format c;
		if (strcmp (code, "t") == 0)
		    c = ba0_table_format;
		else if (strcmp (code, "a") == 0)
		    c = ba0_array_format;
		else if (strcmp (code, "l") == 0)
		    c = ba0_list_format;
		else 
		    c = ba0_matrix_format;
		f->link [ilink++] = l;
		l->u.node.po = get_opening ();
		l->u.node.pf = matching_char (l->u.node.po);
		ba0_index++;
		l->u.node.op = _formate ();
		l->code = c;
		if (l->u.node.op->linknmb != 1)
		    BA0_RAISE_EXCEPTION (BA0_ERRSYN);
		if (c == ba0_array_format &&
				l->u.node.op->link [0]->u.leaf.sizelt < 0)
		    BA0_RAISE_EXCEPTION (BA0_ERRSYN);
	    } else 
	    {	ba0_int_p  i;
		bool found = false;
		struct ba0_pair* pair;

		for (i = 0; i < ba0_leaf_subformat.size && !found; i++)
		{   pair = ba0_leaf_subformat.tab [i];
		    if (strcmp (pair->identificateur, code) == 0)
		    {	f->link [ilink++] = (ba0_subformat)pair->value;
			found = true;
		    }
		}
		if (!found)
		    BA0_RAISE_EXCEPTION (BA0_ERRSYN);
	    }
	}
	f->text [itext++] = chrin;
	chrin = ba0_input_string [++ba0_index];
    }
    f->text [itext] = '\0';
    return f;
}

/**********************************************************************
 HASH-TABLE MANAGEMENT
 **********************************************************************/

static ba0_int_p primary_key (char* s)
{   unsigned ba0_int_p t;

    t = (unsigned ba0_int_p)s;
    t /= BA0_ALIGN;
    t %= ba0_htable.size;
    return (ba0_int_p)t;
}

static ba0_int_p secondary_key (char* s)
{   unsigned ba0_int_p t;

    t = (unsigned ba0_int_p)s;
    t ^= -1;
    t /= BA0_ALIGN;
    t %= ba0_htable.size;
    if (t == 0) t = 1;		/* to avoid infinite loops */
    return (ba0_int_p)t;
}

/*
 * Transforms the string to a format. 
 * There are restrictions on the strings. 
 * In particular, parentheses must be paired. 
 * The format corresponding to a given string is only computed once. 
 * Error BA0_ERROOM is raised if the H-table gets filled.
 */

BA0_DLL ba0_format ba0_get_format (char* s)
{   ba0_int_p i, j;
    struct ba0_pair* pair;

    i = primary_key (s);
    j = secondary_key (s);
    pair = ba0_htable.tab [i];
    while (pair != (struct ba0_pair*)0 && pair->identificateur != s)
    {	i = (i + j) % ba0_htable.size;
	pair = ba0_htable.tab [i];
    }
	   
    if (pair == (struct ba0_pair*)0)
    {
	if (++ba0_nbelem_htable > ba0_htable.size / 3)
	    BA0_RAISE_EXCEPTION (BA0_ERROOM);
	ba0_push_stack (&ba0_format_stack);
	pair = (struct ba0_pair*)ba0_alloc (sizeof (struct ba0_pair));
	ba0_htable.tab [i] = pair;
	pair->identificateur = ba0_input_string = s;
	ba0_index = 0;
	pair->value = _formate ();
	ba0_pull_stack ();
    }

    return (ba0_format)pair->value;
}

