#ifndef BA0_MACROS_MPQ_H
#define BA0_MACROS_MPQ_H 1

#include "ba0_common.h"
#include "ba0_macros_mpz.h"

/* Macros for ba0_mpq */

#define ba0_mpq_t mpq_t

#define ba0_mpq_affect(rop,op)		rop [0] = op [0]
#define ba0_mpq_init(rop)		mpq_init(rop)
#define ba0_mpq_set(rop,op)		mpq_set(rop,op)
#define ba0_mpq_swap(rop,op)		mpq_swap(rop,op)

#if ! defined (BA0_USE_X64_GMP)
#    define ba0_mpq_set_si(rop,op)	mpq_set_si(rop,op,1)
#    define ba0_mpq_set_si_full(rop,opa,opb)				\
					mpq_set_si(rop,opa,opb)
#    define ba0_mpq_set_ui(rop,op)	mpq_set_ui(rop,op,1)
#else
#    define ba0_mpq_set_si(rop,op)	ba0_x64_mpq_set_si(rop,op,1)
#    define ba0_mpq_set_si_full(rop,opa,opb)				\
					ba0_x64_mpq_set_si(rop,opa,opb)
#    define ba0_mpq_set_ui(rop,op)	ba0_x64_mpq_set_ui(rop,op,1)
#endif

#define ba0_mpq_init_set(rop,op) 	do { 				\
					ba0_mpq_init(rop); 		\
					ba0_mpq_set(rop,op);		\
				    	} while (0)
#define ba0_mpq_init_set_si(rop,op) 	do { 				\
					ba0_mpq_init(rop);		\
					ba0_mpq_set_si(rop,op);		\
					} while (0)
#define ba0_mpq_init_set_ui(rop,op) 	do { 				\
					ba0_mpq_init(rop);	 	\
					ba0_mpq_set_ui(rop,op);	\
					} while (0)
#define ba0_mpq_sgn(op)			mpq_sgn(op)
#define ba0_mpq_cmp_ui(opa,opb)		mpq_cmp_ui(opa,opb,1)
#define ba0_mpq_is_one(op)		mpq_cmp_ui(op,1,1) == 0
#define	ba0_mpq_cmp(opa,opb)		mpq_cmp(opa,opb)
#define ba0_mpq_neg(opa,opb)		mpq_neg(opa,opb)
#define ba0_mpq_add(rop,opa,opb)	mpq_add(rop,opa,opb)
#define ba0_mpq_sub(rop,opa,opb)	mpq_sub(rop,opa,opb)
#define ba0_mpq_mul(rop,opa,opb)	mpq_mul(rop,opa,opb)
#define ba0_mpq_mul_ui(rop,opa,opb) 					\
		do { 							\
		ba0_mpq_set(rop,opa); 					\
		ba0_mpz_mul_ui(mpq_numref(rop),mpq_numref(opa),opb);	\
		mpq_canonicalize(rop);					\
		} while (0) 
#define ba0_mpq_mul_si(rop,opa,opb) 					\
		do { 							\
		ba0_mpq_set(rop,opa);					\
		ba0_mpz_mul_si(mpq_numref(rop),mpq_numref(rop),opb);	\
		mpq_canonicalize(rop);					\
		} while (0)
#define ba0_mpq_pow_ui(rop,opa,opb) 					\
		do { 							\
		ba0_mpz_pow_ui(mpq_numref(rop),mpq_numref(opa),opb);	\
		ba0_mpz_pow_ui(mpq_denref(rop),mpq_denref(opa),opb);	\
		} while (0)
#define ba0_mpq_div(rop,opa,opb)	mpq_div(rop,opa,opb)
#define ba0_mpq_invert(rop,op)		mpq_inv(rop,op)

#endif /* !BA0_MACROS_MPQ_H */
