#ifndef BA0_MACROS_MPZ_H
#define BA0_MACROS_MPZ_H 1

#include "ba0_common.h"

/* Macros for mpz_t */

#define ba0_mpz_t mpz_t

#define ba0_mpz_affect(rop,op)		rop [0] = op [0]
#define ba0_mpz_init(rop)		mpz_init(rop)
#define ba0_mpz_set(rop,op)		mpz_set(rop,op)
#define ba0_mpz_swap(rop,op)		mpz_swap(rop,op)
#define ba0_mpz_init_set(rop,op)	mpz_init_set(rop,op)

/*
 * ba0_mpz_si_pow_ui always mapped to ba0_x64_mpz_si_pow_ui.
 * This particular function does not rely on any assumption.
 */

#if ! defined (BA0_USE_X64_GMP)
#    define ba0_mpz_set_si(rop,op)		mpz_set_si(rop,op)
#    define ba0_mpz_set_ui(rop,op)		mpz_set_ui(rop,op)
#    define ba0_mpz_init_set_si(rop,op)		mpz_init_set_si(rop,op)	
#    define ba0_mpz_init_set_ui(rop,op)		mpz_init_set_ui(rop,op)
#    define ba0_mpz_mul_si(rop,opa,opb)		mpz_mul_si(rop,opa,opb)
#    define ba0_mpz_mul_ui(rop,opa,opb)		mpz_mul_ui(rop,opa,opb)
#    define ba0_mpz_mul_2exp(rop,opa,opb)	mpz_mul_2exp(rop,opa,opb)
#    define ba0_mpz_pow_ui(rop,opa,opb)		mpz_pow_ui(rop,opa,opb)
#    define ba0_mpz_ui_pow_ui(rop,opa,opb)	mpz_ui_pow_ui(rop,opa,opb)
#    define ba0_mpz_si_pow_ui(rop,opa,opb)	\
					ba0_x64_mpz_si_pow_ui(rop,opa,opb)
#else
#    define ba0_mpz_set_si(rop,op)		ba0_x64_mpz_set_si(rop,op)
#    define ba0_mpz_set_ui(rop,op)		ba0_x64_mpz_set_ui(rop,op)
#    define ba0_mpz_init_set_si(rop,op)	do { 				\
					mpz_init(rop); 			\
					ba0_x64_mpz_set_si(rop,op); 	\
					} while (0)
#    define ba0_mpz_init_set_ui(rop,op)	do { 				\
					mpz_init(rop); 			\
					ba0_x64_mpz_set_ui(rop,op); 	\
					} while (0)
#    define ba0_mpz_mul_si(rop,opa,opb)		ba0_x64_mpz_mul_si(rop,opa,opb)
#    define ba0_mpz_mul_ui(rop,opa,opb)		ba0_x64_mpz_mul_ui(rop,opa,opb)
#    define ba0_mpz_mul_2exp(rop,opa,opb)	\
					ba0_x64_mpz_mul_2exp(rop,opa,opb)
#    define ba0_mpz_pow_ui(rop,opa,opb)		ba0_x64_mpz_pow_ui(rop,opa,opb)
#    define ba0_mpz_ui_pow_ui(rop,opa,opb)	\
					ba0_x64_mpz_ui_pow_ui(rop,opa,opb)
#    define ba0_mpz_si_pow_ui(rop,opa,opb)	\
					ba0_x64_mpz_si_pow_ui(rop,opa,opb)
#endif

#define ba0_mpz_sgn(op)			mpz_sgn(op)
#define ba0_mpz_cmp_ui(opa,opb)		mpz_cmp_ui(opa,opb)
#define ba0_mpz_is_one(op)		mpz_cmp_ui(op,1) == 0
#define ba0_mpz_cmp(opa,opb)		mpz_cmp(opa,opb)
#define ba0_mpz_neg(opa,opb)		mpz_neg(opa,opb)
#define ba0_mpz_add(rop,opa,opb)	mpz_add(rop,opa,opb)
#define ba0_mpz_sub(rop,opa,opb)	mpz_sub(rop,opa,opb)
#define ba0_mpz_mul(rop,opa,opb)	mpz_mul(rop,opa,opb)
#define ba0_mpz_div(rop,opa,opb)	mpz_fdiv_q(rop,opa,opb)

#endif /* !BA0_MACROS_MPZ_H */
