#ifndef BA0_MACROS_MPZM_H
#define BA0_MACROS_MPZM_H 1

#include "ba0_common.h"
#include "ba0_macros_mpz.h"

/* Macros for ba0_mpzm */

#define ba0_mpzm_t mpz_t

#define ba0_mpzm_module			ba0_global.mpzm.module

#define ba0_mpzm_init(rop)		mpz_init(rop)

#define ba0_mpzm_set(rop,op)		mpz_set(rop,op)

#define ba0_mpzm_affect(rop,op)		rop [0] = op [0]

#define ba0_mpzm_swap(opa,opb)		mpz_swap(opa,opb)

#define ba0_mpzm_set_si(rop,op)						\
		do { 							\
		ba0_mpz_set_si(rop,op); 				\
		mpz_mod(rop,rop,ba0_mpzm_module); 			\
		} while (0)

#define ba0_mpzm_init_set(rop,op)	mpz_init_set(rop,op)

#define ba0_mpzm_init_set_si(rop,op) 					\
		do {							\
		ba0_mpz_init_set_si(rop,op); 				\
		mpz_mod(rop,rop,ba0_mpzm_module); 			\
		} while (0)

#define ba0_mpzm_init_set_ui(rop,op) 					\
		do { 							\
		mpz_init_set_ui(rop,op); 				\
		mpz_mod(rop,rop,ba0_mpzm_module); 			\
		} while (0)

#define ba0_mpzm_sgn(op)		mpz_sgn(op)

#define ba0_mpzm_cmp_ui(opa,opb)	mpz_cmp_ui(opa,opb)

#define ba0_mpzm_is_one(op)		mpz_cmp_ui(op,1) == 0

#define ba0_mpzm_cmp(opa,opb)		mpz_cmp(opa,opb)

#define ba0_mpzm_neg(rop,op)						\
		do { 							\
		mpz_neg(rop,op); 					\
		mpz_mod(rop,rop,ba0_mpzm_module); 			\
		} while (0)

#define ba0_mpzm_add(rop,opa,opb) 					\
		do { 							\
		mpz_add(rop,opa,opb); 					\
		mpz_mod(rop,rop,ba0_mpzm_module); 			\
		} while (0)

#define ba0_mpzm_sub(rop,opa,opb) 					\
		do { 							\
		mpz_sub(rop,opa,opb); 					\
		mpz_mod(rop,rop,ba0_mpzm_module); 			\
		} while (0)

#define ba0_mpzm_mul(rop,opa,opb) 					\
		do { 							\
		ba0_push_stack (&ba0_quiet_stack); 			\
		mpz_mul(ba0_mpzm_accum,opa,opb); 			\
		ba0_pull_stack (); 					\
		mpz_mod(rop,ba0_mpzm_accum,ba0_mpzm_module); 		\
		} while (0)

#define ba0_mpzm_mul_ui(rop,opa,opb) 					\
		do {							\
		ba0_push_stack (&ba0_quiet_stack); 			\
		ba0_mpz_mul_ui(ba0_mpzm_accum,opa,opb); 		\
		ba0_pull_stack (); 					\
		mpz_mod(rop,ba0_mpzm_accum,ba0_mpzm_module); 		\
		} while (0)

#define ba0_mpzm_mul_si(rop,opa,opb) 					\
		do { 							\
		ba0_mpz_mul_si(rop,opa,opb); 				\
		mpz_mod(rop,rop,ba0_mpzm_module); 			\
		} while (0)

#if ! defined (BA0_USE_X64_GMP)
#    define ba0_mpzm_pow_ui(rop,opa,opb)   				\
		mpz_powm_ui(rop,opa,opb,ba0_mpzm_module)
#else
#    define ba0_mpzm_pow_ui(rop,opa,opb)				\
		do {							\
		mpz_t bunk;						\
		ba0_push_another_stack ();				\
		ba0_x64_mpz_set_ui(bunk,opb);				\
		ba0_pull_stack ();					\
		mpz_powm(rop,opa,bunk,ba0_mpzm_module);			\
		} while (0)
#endif

#define ba0_mpzm_div(rop,opa,opb) 					\
		do { 							\
		ba0_push_stack (&ba0_quiet_stack); 			\
		mpz_invert(ba0_mpzm_accum,opb,ba0_mpzm_module);		\
		mpz_mul(ba0_mpzm_accum,ba0_mpzm_accum,opa); 		\
		ba0_pull_stack (); 					\
		mpz_mod(rop,ba0_mpzm_accum,ba0_mpzm_module); 		\
		} while (0)

#define ba0_mpzm_invert(rop,op)		mpz_invert(rop,op,ba0_mpzm_module)

#endif /* !BA0_MACROS_MPZM_H */
