#include "ba0_exception.h"
#include "ba0_stack.h"
#include "ba0_matrix.h"

BA0_DLL void ba0_init_matrix (ba0_matrix M)
{
    M->alloc = 0;
    M->nrow = 0;
    M->ncol = 0;
    M->entry = (void*)0;
}

BA0_DLL void ba0_reset_matrix (ba0_matrix M)
{
    M->nrow = 0;
    M->ncol = 0;
}

BA0_DLL void ba0_realloc_matrix (ba0_matrix M, ba0_int_p n, ba0_int_p m)
{   ba0_int_p new_size, old_size;
    void** new_entry;

    new_size = n * m;
    if (new_size > M->alloc)
    {	new_entry = (void**)ba0_alloc (sizeof (void*) * new_size);
	old_size = M->nrow * M->ncol;
	memcpy (new_entry, M->entry, old_size * sizeof (void*));
	M->entry = new_entry;
	M->alloc = new_size;
	M->nrow = n;
	M->ncol = m;
    }
}

BA0_DLL void ba0_realloc2_matrix 
	(ba0_matrix M, ba0_int_p n, ba0_int_p m, ba0_new_function* new_objet)
{   ba0_int_p new_size, old_size, i;
    void** new_entry;

    new_size = n * m;
    if (new_size > M->alloc)
    {   new_entry = (void**)ba0_alloc (sizeof (void*) * new_size);
        old_size = M->nrow * M->ncol;
	memcpy (new_entry, M->entry, old_size * sizeof (void*));
	for (i = old_size; i < new_size; i++)
	    new_entry [i] = (*new_objet) ();
        M->entry = new_entry;
        M->alloc = new_size;
        M->nrow = n;
        M->ncol = m;
    }
}

BA0_DLL ba0_matrix ba0_new_matrix (void)
{   ba0_matrix M;

    M = (ba0_matrix)ba0_alloc (sizeof (struct ba0_matrix));
    ba0_init_matrix (M);
    return M;
}

BA0_DLL void ba0_set_matrix (ba0_matrix A, ba0_matrix B)
{   ba0_int_p size;

    if (A != B)
    {	ba0_realloc_matrix (A, B->nrow, B->ncol);
	size = B->nrow * B->ncol;
	memcpy (A->entry, B->entry, size * sizeof (void*));
	A->nrow = B->nrow;
	A->ncol = B->ncol;
    }
}

BA0_DLL void ba0_set_matrix2 (ba0_matrix A, ba0_matrix B, 
		ba0_new_function* new_objet, ba0_binary_operation* set_objet)
{   ba0_int_p i, size;

    if (A != B)
    {	ba0_realloc2_matrix (A, B->nrow, B->ncol, new_objet);
	size = B->nrow * B->ncol;
	for (i = 0; i < size; i++)
	    (*set_objet) (A->entry [i], B->entry [i]);
	A->nrow = B->nrow;
	A->ncol = B->ncol;
    }
}

BA0_DLL void ba0_set_matrix_unity (ba0_matrix A, ba0_int_p n,
			   ba0_new_function* new_objet, 
			   ba0_unary_operation* set_objet_zero,
			   ba0_unary_operation* set_objet_one)
{   ba0_int_p i, j;

    ba0_reset_matrix (A);
    ba0_realloc2_matrix (A, n, n, new_objet);
    A->nrow = n;
    A->ncol = n;
    for (i = 0; i < n; i++)
    {	for (j = 0; j < i; j++)
	    (*set_objet_zero) (BA0_MAT(A, i, j));
	(*set_objet_one) (BA0_MAT(A, i, i));
	for (j = i + 1; j < n; j++)
	    (*set_objet_zero) (BA0_MAT(A, i, j));
    }
}

BA0_DLL bool 
	ba0_is_zero_matrix (ba0_matrix A, ba0_unary_predicate* is_zero_objet)
{   ba0_int_p i, size;
    bool zero;

    size = A->nrow * A->ncol;
    zero = true;

    for (i = 0; i < size && zero; i++)
	zero = (*is_zero_objet) (A->entry [i]);

    return zero;
}

BA0_DLL bool ba0_is_unity_matrix (ba0_matrix A, 
				ba0_unary_predicate* is_zero_objet, 
				ba0_unary_predicate* is_one_objet)
{
    ba0_int_p i, j;
    bool unity;

    unity = true;
    for (i = 0; i < A->nrow && unity; i++)
	for (j = 0; j < A->ncol && unity; j++)
	    if (i == j)
		unity = (*is_one_objet) (BA0_MAT(A, i, j));
	    else
		unity = (*is_zero_objet) (BA0_MAT(A, i, j));

    return unity;
}

BA0_DLL void ba0_swap_rows_matrix (ba0_matrix A, ba0_int_p i, ba0_int_p j)
{   ba0_int_p k;

    if (i < 0 || i >= A->nrow || j < 0 || j >= A->nrow)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    if (i != j)
    {	for (k = 0; k < A->ncol; k++)
    	{   BA0_SWAP (void*, BA0_MAT(A, i, k), BA0_MAT(A, j, k));
	}
    }
}

BA0_DLL void ba0_swap_columns_matrix (ba0_matrix A, ba0_int_p i, ba0_int_p j)
{   ba0_int_p k;

    if (i < 0 || i >= A->ncol || j < 0 || j >= A->ncol)
        BA0_RAISE_EXCEPTION (BA0_ERRALG);
    if (i != j)
    {   for (k = 0; k < A->nrow; k++)
        {   BA0_SWAP (void*, BA0_MAT(A, k, i), BA0_MAT(A, k, j));
        }
    }
}

BA0_DLL void ba0_add_matrix 
		(ba0_matrix R, ba0_matrix A, ba0_matrix B, 
		 ba0_new_function* new_objet, ba0_ternary_operation* add_objet)
{   ba0_int_p n, m;
    ba0_int_p i, j;

    if (A->nrow != B->nrow || A->ncol != B->ncol)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    n = A->nrow;
    m = A->ncol;
    ba0_reset_matrix (R);
    ba0_realloc2_matrix (R, n, m, new_objet);
    R->nrow = n;
    R->ncol = m;
    for (i = 0; i < n; i++)
	for (j = 0; j < m; j++)
	    (*add_objet) 
		(BA0_MAT(R, i, j), BA0_MAT(A, i, j), BA0_MAT(B, i, j));
}

BA0_DLL void ba0_mul_matrix 
		(ba0_matrix R, ba0_matrix A, ba0_matrix B,
		 ba0_new_function* new_objet, 
		 ba0_unary_operation* set_objet_zero,
		 ba0_binary_operation* set_objet,
		 ba0_ternary_operation* add_objet,
		 ba0_ternary_operation* mul_objet)
{   ba0_int_p i, j, k;
    struct ba0_mark M;
    void* z;

    if (A->ncol != B->nrow)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (A == R || B == R)
    {	ba0_matrix S;

	ba0_push_another_stack ();
	ba0_record (&M);
	S = ba0_new_matrix ();
	ba0_mul_matrix (S, A, B, new_objet, set_objet_zero, set_objet, 
							add_objet, mul_objet);
	ba0_pull_stack ();
	ba0_set_matrix2 (R, S, new_objet, set_objet);

	ba0_restore (&M);
	return;
    }

    ba0_reset_matrix (R);
    ba0_realloc2_matrix (R, A->nrow, B->ncol, new_objet);
    R->nrow = A->nrow;
    R->ncol = B->ncol;

    ba0_push_another_stack ();
    ba0_record (&M);
    z = (*new_objet) ();
    
    for (i = 0; i < R->nrow; i++)
    {	for (j = 0; j < R->ncol; j++)
	{   
	    ba0_pull_stack ();
	    (*set_objet_zero) (BA0_MAT(R, i, j));
	    ba0_push_another_stack ();
	    for (k = 0; k < A->ncol; k++)
	    {   (*mul_objet) (z, BA0_MAT(A, i, k), BA0_MAT(B, k, j));
		ba0_pull_stack ();
		(*add_objet) (BA0_MAT(R, i, j), BA0_MAT(R, i, j), z);
		ba0_push_another_stack ();
	    }
	}
    }

    ba0_pull_stack ();
    ba0_restore (&M);
}
