#include "ba0_basic_io.h"
#include "ba0_exception.h"
#include "ba0_stack.h"
#include "ba0_garbage.h"
#include "ba0_analex.h"
#include "ba0_mpfr.h"
#include "ba0_printf.h"
#include "ba0_string.h"
#include "ba0_global.h"

#if HAVE_MPFR_H

BA0_DLL bool ba0_domain_mpfr (void)
{
    return true;
}

BA0_DLL __mpfr_struct* ba0_new_mpfr (void)
{   __mpfr_struct* z;

    z = (__mpfr_struct*) ba0_alloc (sizeof (__mpfr_struct));
    ba0_mpfr_init (z);
    return z;
}

BA0_DLL void* ba0_scanf_mpfr (void* z)
{   struct ba0_mark M;
    ba0_int_p token_lmax;
    char *buf1;
    char *buf2;
    __mpfr_struct* e;
    bool minus_sign;
    bool mantissa;

    ba0_get_settings_analex (0, &token_lmax);
    ba0_push_another_stack ();
    ba0_record (&M);
    buf1 = (char*)ba0_alloc (token_lmax);
    buf2 = (char*)ba0_alloc (token_lmax);
    ba0_pull_stack ();

    if (z == (void*)0)
        e = ba0_new_mpfr ();
     else
        e = (__mpfr_struct*)z;
/*
   Possibly start with a minus sign.
*/
    if (ba0_sign_token_analex ("-"))
    {	ba0_get_token_analex ();
	minus_sign = true;
    } else
    {	if (ba0_sign_token_analex ("+"))
	    ba0_get_token_analex ();
	minus_sign = false;
    }
/*
   Possibly a non empty integer part
   Spaces are authorized between the minus sign and what follows.
*/
    mantissa = false;
    buf1 [0] = '\0';
    if (ba0_type_token_analex () == ba0_integer_token)
    {	strcat (buf1, ba0_value_token_analex ());
	ba0_get_token_analex ();
	mantissa = true;
    }
/*
   Possibly a dot plus a possible fractional part
*/
    if ((! mantissa || ! ba0_spaces_before_token_analex ()) && 
						ba0_sign_token_analex ("."))
    {	strcat (buf1, ".");
	ba0_get_token_analex ();
	if (! ba0_spaces_before_token_analex () && 
			ba0_type_token_analex () == ba0_integer_token)
	{   strcat (buf1, ba0_value_token_analex ());
	    ba0_get_token_analex ();
	}
	mantissa = true;
    }
/*
   There must be a mantissa
*/
    if (! mantissa)
	BA0_RAISE_EXCEPTION (BA0_ERRFLT);
/*
   Possibly an exponent.
*/
    strcpy (buf2, ba0_value_token_analex ());
    if (! ba0_spaces_before_token_analex () &&
	   (buf2 [0] == 'e' || buf2 [0] == 'E' || buf2 [0] == '@') &&
	   (buf2 [1] == '\0' || isdigit ((int)buf2 [1])))
    {   strcat (buf1, "e");
/*
   To simplify the coming analysis, the current token may be split
   into two different tokens
*/
	if (isdigit ((int)buf2 [1]))
	{   ba0_int_p i;
	    i = 2;
	    while (isdigit ((int)buf2 [i]))
		i++;
	    if (buf2 [i] == '\0')
		ba0_unget_given_token_analex 
					(buf2+1, ba0_integer_token, false);
	    else
	    {	ba0_unget_given_token_analex 
					(buf2+i, ba0_string_token, false);
		buf2 [i] = '\0';
		ba0_unget_given_token_analex 
					(buf2+1, ba0_integer_token, false);
	    }
	} 
/*
   The analysis of the exponent starts here
*/
	ba0_get_token_analex ();
	if (! ba0_spaces_before_token_analex () && 
			ba0_type_token_analex () == ba0_integer_token)
	{   strcat (buf1, ba0_value_token_analex ());
	} else if (! ba0_spaces_before_token_analex () && 
					(ba0_sign_token_analex ("+") || 
					 ba0_sign_token_analex ("-")))
	{   strcat (buf1, ba0_value_token_analex ());
	    ba0_get_token_analex ();
	    if (ba0_spaces_before_token_analex () ||
				ba0_type_token_analex () != ba0_integer_token)
		BA0_RAISE_EXCEPTION (BA0_ERRFLT);
	    strcat (buf1, ba0_value_token_analex ());
	} else
	    BA0_RAISE_EXCEPTION (BA0_ERRFLT);
    } else
	ba0_unget_token_analex ();

    if (mpfr_set_str (e, buf1, (int)10, ba0_mpfr_rnd) == -1)
        BA0_RAISE_EXCEPTION (BA0_ERRFLT);
    if (minus_sign)
        ba0_mpfr_neg (e, e);

    ba0_restore (&M);
    return e;
}

static void remove_trailing_zeros (char *s)
{   int l;

    l = strlen (s) - 1;
    while (s [l] == '0')
	s [l--] = '\0';
}

BA0_DLL void ba0_printf_mpfr (void* z)
{   __mpfr_struct* c = (__mpfr_struct*)z;
    mp_exp_t expt;
    char* s;
    int l;
    struct ba0_mark M;

    if (ba0_mpfr_cmp_ui (c, 0) == 0)
	ba0_put_string ("0");
    else
    {	ba0_record (&M);
	s = mpfr_get_str ((char*)0, &expt, 10, 0, c, ba0_mpfr_rnd);
	if (s [0] == '-')
	{   ba0_put_char (s [0]);
	    s += 1;
	}
	remove_trailing_zeros (s);
	l = strlen (s);
	if (expt < 0)
	{   ba0_put_char (s [0]);
	    expt -= 1;
	    if (s [1] != '\0')
	        ba0_printf (".%s", s+1);
	    ba0_printf ("e%d", expt);
	} else if (expt < l)
	{   if (expt == 0)
		ba0_put_char ('0');
	    else while (expt > 0)
	    {	ba0_put_char (*s);
		s += 1;
		expt -= 1;
	    }
	    ba0_printf (".%s", s);
	} else if (expt <= 3)
	{   ba0_put_string (s);
	    while (l < expt)
	    {	ba0_put_char ('0');
		l++;
	    }
	} else
	{   ba0_put_char (s [0]);
	    expt -= 1;
	    if (s [1] != '\0')
	    	ba0_printf (".%s", s+1);
	    ba0_printf ("e%d", expt);
	}
	ba0_restore (&M);
    }
}

/*
   Borrowed from mpfr-gmp.h
*/

/* Define BITS_PER_MP_LIMB
   Can't use sizeof(mp_limb_t) since it should be a preprocessor constant */
#if defined(GMP_NUMB_BITS) /* GMP 4.1.2 or above */
# define BITS_PER_MP_LIMB  (GMP_NUMB_BITS+GMP_NAIL_BITS)
#elif defined (__GMP_BITS_PER_MP_LIMB) /* Older versions 4.x.x */
# define BITS_PER_MP_LIMB  __GMP_BITS_PER_MP_LIMB
# define GMP_NUMB_BITS BITS_PER_MP_LIMB
# ifndef GMP_NAIL_BITS
#  define GMP_NAIL_BITS 0
# endif
#else
# error "Could not detect BITS_PER_MP_LIMB. Try with gmp internal files."
#endif

/*
   Borrowed from mpfr-impl.h
*/

#define MPFR_PREC(x)      ((x)->_mpfr_prec)
#define MPFR_EXP(x)       ((x)->_mpfr_exp)
#define MPFR_MANT(x)      ((x)->_mpfr_d)
#define MPFR_LIMB_SIZE(x) ((MPFR_PREC((x))-1)/BITS_PER_MP_LIMB+1)

/*
 * Readonly static data
 */

static char mpfr_struct_ [] = "__mpfr_struct";
static char mpfr_struct_mpfr_d_ [] = "__mpfr_struct._mpfr_d";

BA0_DLL ba0_int_p ba0_garbage1_mpfr (void* _c, enum ba0_garbage_code code)
{   __mpfr_struct* c = (__mpfr_struct*)_c;
    ba0_int_p n = 0;

    if (code == ba0_isolated)
	n += ba0_new_gc_info (c, sizeof (__mpfr_struct), mpfr_struct_);
    n += ba0_new_gc_info (c->_mpfr_d, MPFR_LIMB_SIZE(c)*sizeof(mp_limb_t), 
						mpfr_struct_mpfr_d_);
    return n;
}
 
BA0_DLL void* ba0_garbage2_mpfr (void* _c, enum ba0_garbage_code code)
{   __mpfr_struct* c;

    if (code == ba0_isolated)
	c = (__mpfr_struct*)ba0_new_addr_gc_info (_c, mpfr_struct_);
    else
	c = (__mpfr_struct*)_c;

    c->_mpfr_d = ba0_new_addr_gc_info (c->_mpfr_d, mpfr_struct_mpfr_d_);
    return c;
}

BA0_DLL void* ba0_copy_mpfr (void* z)
{   __mpfr_struct* c;

    c = ba0_new_mpfr ();
    ba0_mpfr_init_set (c, (__mpfr_struct*)z);
    return c;
}
 
#endif /* !HAVE_MPFR_H */

