#ifndef BA0_STACK_H
#define BA0_STACK_H

#include "ba0_common.h"
#include "ba0_table.h"

BEGIN_C_DECLS

struct ba0_mark {
    struct ba0_stack* stack;
    ba0_int_p index_in_cells;
    void* address;
    unsigned ba0_int_p ce_qui_reste;
};

struct ba0_stack {
    char* ident;
    struct ba0_table cells;
    unsigned ba0_int_p cell_size;
    struct ba0_mark free;
    struct ba0_mark max_alloc;
    ba0_int_p  nb_calls_to_alloc;
    ba0_int_p* bound;
};

#define ba0_main_stack		ba0_global.stack.main
#define ba0_second_stack	ba0_global.stack.second
#define ba0_analex_stack	ba0_global.stack.analex
#define ba0_quiet_stack		ba0_global.stack.quiet
#define ba0_format_stack	ba0_global.stack.format

#define ba0_stack_of_stacks	ba0_global.stack.stack_of_stacks
#define ba0_stack_sp		ba0_global.stack.stack_sp

/*
 * Some default values
 */

#define BA0_SIZE_CELL_MAIN_STACK  0xc00000
#define BA0_SIZE_CELL_QUIET_STACK  0x10000
#define BA0_SIZE_CELL_ANALEX_STACK 0x10000
#define BA0_NB_CELLS_PER_STACK       0x100
#define BA0_SIZE_STACK_OF_STACK      0x100

extern BA0_DLL void ba0_set_settings_no_oom (bool);
extern BA0_DLL void ba0_get_settings_no_oom (bool*);

extern BA0_DLL void ba0_set_settings_stack
		(ba0_int_p, ba0_int_p, ba0_int_p, ba0_int_p, ba0_int_p);
extern BA0_DLL void ba0_get_settings_stack
		(ba0_int_p*, ba0_int_p*, ba0_int_p*, ba0_int_p*, ba0_int_p*);

extern BA0_DLL void ba0_init_stack_of_stacks (void);
extern BA0_DLL void ba0_reset_stack_of_stacks (void);
extern BA0_DLL void ba0_clear_stack_of_stacks (void);

#define ba0_alloc_counter	ba0_global.stack.alloc_counter
#define ba0_malloc_counter	ba0_global.stack.malloc_counter
#define ba0_malloc_nbcalls	ba0_global.stack.malloc_nbcalls

extern BA0_DLL void ba0_set_settings_memory_functions 
		(void* (*) (size_t), void (*) (void*));
extern BA0_DLL void ba0_get_settings_memory_functions 
		(void* (**) (size_t), void (**) (void*));

extern BA0_DLL unsigned ba0_int_p ba0_ceil_align (unsigned ba0_int_p);
extern BA0_DLL unsigned ba0_int_p ba0_allocated_size (unsigned ba0_int_p);
extern BA0_DLL void* ba0_malloc (ba0_int_p);
extern BA0_DLL void* ba0_persistent_malloc (ba0_int_p);
extern BA0_DLL void ba0_free (void*);

extern BA0_DLL ba0_int_p ba0_cell_index_mark (void*, struct ba0_mark*);
extern BA0_DLL bool ba0_in_stack (void*, struct ba0_stack*);
extern BA0_DLL struct ba0_stack* ba0_which_stack (void*);
extern BA0_DLL struct ba0_stack* ba0_current_stack (void);
extern BA0_DLL unsigned ba0_int_p ba0_max_alloc_stack (struct ba0_stack*);

extern BA0_DLL void ba0_push_stack (struct ba0_stack*);
extern BA0_DLL void ba0_pull_stack (void);
extern BA0_DLL void ba0_push_another_stack (void);

extern BA0_DLL void ba0_init_stack (struct ba0_stack*);
extern BA0_DLL void ba0_init_one_cell_stack 
				(struct ba0_stack*, char*, void*, ba0_int_p);
extern BA0_DLL void ba0_reset_stack (struct ba0_stack*);
extern BA0_DLL void ba0_clear_cells_stack (struct ba0_stack*);
extern BA0_DLL void ba0_clear_one_cell_stack (struct ba0_stack*);
extern BA0_DLL void ba0_clear_stack (struct ba0_stack*);
extern BA0_DLL void* ba0_alloc_but_do_not_set_magic (unsigned ba0_int_p);
extern BA0_DLL void ba0_alloc_set_magic (void);
extern BA0_DLL void* ba0_alloc (unsigned ba0_int_p);
extern BA0_DLL void ba0_t1_alloc 
	(unsigned ba0_int_p, unsigned ba0_int_p, void**, unsigned ba0_int_p*);
extern BA0_DLL void ba0_t2_alloc 
	(unsigned ba0_int_p, unsigned ba0_int_p, unsigned ba0_int_p, 
	 void**, void**, unsigned ba0_int_p*);
extern BA0_DLL bool ba0_alloc_would_malloc_a_cell (unsigned ba0_int_p);
extern BA0_DLL void ba0_rotate_cells (ba0_int_p);

extern BA0_DLL void* ba0_alloc_mark (struct ba0_mark*, unsigned ba0_int_p);
extern BA0_DLL void ba0_record (struct ba0_mark*);
extern BA0_DLL void ba0_restore (struct ba0_mark*);
extern BA0_DLL unsigned ba0_int_p 
			ba0_range_mark (struct ba0_mark*, struct ba0_mark*);

END_C_DECLS

#endif /* !BA0_STACK_H */
