#include "ba0_stack.h"
#include "ba0_exception.h"
#include "ba0_list.h"
#include "ba0_table.h"

BA0_DLL unsigned ba0_int_p ba0_sizeof_table (ba0_table T)
{   unsigned ba0_int_p size;

    size = ba0_allocated_size (sizeof (struct ba0_table));
    size += ba0_allocated_size (T->size * sizeof (void*));
    return size;
}

BA0_DLL void ba0_realloc_table (ba0_table T, ba0_int_p n)
{   void** tab;
 
    if (n > T->alloc)
    {	tab = (void**)ba0_alloc (sizeof (void*) * n);
	memcpy (tab, T->tab, T->size * sizeof (void*));
	T->tab = tab;
	T->alloc = n;
    }
}

BA0_DLL void ba0_realloc2_table 
			(ba0_table T, ba0_int_p n, ba0_new_function* new_objet) {   void** tab;
    ba0_int_p i;

    if (n > T->alloc)
    {   tab = (void**)ba0_alloc (sizeof (void*) * n);
	memcpy (tab, T->tab, T->size * sizeof (void*));
	for (i = T->size; i < n; i++)
	    tab [i] = (*new_objet) ();
        T->tab = tab;
        T->alloc = n;
    }
}

BA0_DLL void ba0_init_table (ba0_table T)
{
    T->alloc = 0;
    T->size = 0;
    T->tab = (void**)0;
}
 
BA0_DLL void ba0_reset_table (ba0_table T)
{
    T->size = 0;
}

BA0_DLL ba0_table ba0_new_table (void)
{   ba0_table T;

    T = (ba0_table)ba0_alloc (sizeof (struct ba0_table));
    ba0_init_table (T);
    return T;
}

BA0_DLL void ba0_delete_table (ba0_table T, ba0_int_p i)
{
    if (i == T->size)
	T->size--;
    else if (i < T->size)
    {	void* x = T->tab [i];
	memmove 
	    (T->tab + i, T->tab + i + 1, (T->size - i - 1) * sizeof (void*));
	T->tab [--T->size] = x;
    }
}

BA0_DLL void ba0_insert_table (ba0_table T, ba0_int_p i, void* p)
{
    if (i < 0 || i > T->size)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    ba0_realloc_table (T, T->size + 1);
    memmove (T->tab + i + 1, T->tab + i, (T->size - i) * sizeof (void*));
    T->tab [i] = p;
    T->size++;
}

BA0_DLL bool ba0_member_table (void* p, ba0_table T)
{   ba0_int_p i;

    for (i = 0; i < T->size; i++)
	if (p == T->tab [i])
	    return true;
    return false;
}

BA0_DLL bool ba0_member2_table (void* p, ba0_table T, ba0_int_p* index)
{   ba0_int_p i;

    for (i = 0; i < T->size; i++)
        if (p == T->tab [i])
        {   if (index) *index = i;
            return true;
        }
    return false;
}

BA0_DLL bool ba0_equal_table (ba0_table T, ba0_table U)
{   ba0_int_p i;

    if (T->size != U->size)
	return false;
    for (i = 0; i < T->size; i++)
	if (T->tab [i] != U->tab [i])
	    return false;
    return true;
}

BA0_DLL bool ba0_unique_table (ba0_table T)
{   ba0_int_p i, j;

    for (i = 0; i < T->size-1; i++)
    {	for (j = i+1; j < T->size; j++)
	    if (T->tab [i] == T->tab [j])
		return false;
    }
    return true;
}

BA0_DLL void ba0_reverse_table (ba0_table T, ba0_table U)
{   ba0_int_p i, j;

    if (T == U)
    {	for (i = 0, j = T->size-1; i < j; i++, j--)
	    BA0_SWAP (void*, T->tab [i], T->tab [j]);
    } else
    {	ba0_reset_table (T);
	ba0_realloc_table (T, U->size);
	T->size = U->size;
	for (i = 0, j = U->size-1; j >= 0; i++, j--)
	    T->tab [i] = U->tab [j];
    }
}

BA0_DLL void ba0_concat_table (ba0_table T, ba0_table U, ba0_table V)
{
    ba0_realloc_table (T, U->size + V->size);
    memmove (T->tab + U->size, V->tab, sizeof (void*) * V->size);
    if (T != U)
	memmove (T->tab, U->tab, sizeof (void*) * U->size);
    T->size = U->size + V->size;
}

BA0_DLL void ba0_set_table (ba0_table T, ba0_table U)
{
    if (T != U)
    {	T->size = 0;
	ba0_realloc_table (T, U->size);
	memcpy (T->tab, U->tab, U->size * sizeof (void*));
	T->size = U->size;
    }
}

BA0_DLL void ba0_set_table_list (ba0_table T, ba0_list L)
{   ba0_int_p i, j;

    j = ba0_length_list (L);
    ba0_realloc_table (T, j);
    for (i = 0; i < j; i++)
    {   T->tab [i] = L->value;
        L = L->next;
    }
    T->size = j;
}

/*
 * Moves U.tab [g] at U.tab [U.size-1]. result in T.
 * The elements in U.tab [g+1, ... U.size-1] are shifted to the left.
 */

BA0_DLL void ba0_move_to_tail_table (ba0_table T, ba0_table U, ba0_int_p g)
{   void* elt;

    if (g < 0 || g >= U->size)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (T == U)
    {	if (g != U->size-1)
	{   elt = U->tab [g];
	    memmove (U->tab + g, U->tab + g + 1, 
					(U->size - 1 - g) * sizeof (void*));
	    U->tab [U->size - 1] = elt;
	}
    } else
    {	ba0_set_table (T, U);
	ba0_move_to_tail_table (T, T, g);
    }
}

/*
 * Reciprocal operation. Result in T.
 * Moves U.tab [U.size-1] to U.tab [g].
 * The elements in U.tab [g, ..., U.size-2] are shifted to the right.
 */

BA0_DLL void ba0_move_from_tail_table (ba0_table T, ba0_table U, ba0_int_p g)
{   void* elt;

    if (g < 0 || g >= U->size)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (T == U)
    {	if (g != U->size-1)
	{   elt = U->tab [U->size-1];
	    memmove (U->tab + g + 1, U->tab + g, 
					(U->size - 1 - g) * sizeof (void*));
	    U->tab [g] = elt;
	}
    } else
    {	ba0_set_table (T, U);
	ba0_move_from_tail_table (T, T, g);
    }
}

