#include "ba0_common.h"
#include "ba0_x64_gmp.h"
#include "ba0_macros_mpz.h"
#include "ba0_macros_mpq.h"

/*
 * This code is rewritten from the source code of gmp-4.2.4
 * It is intended to be used on Windows, 64 bits, under Visual Studio 2008.
 *
 * in this setting, we have:
 * SIZEOF_MP_LIMB_T	= 8
 * GMP_LIMB_BITS	= 64
 * GMP_NUMB_BITS	= 64
 * GMP_NAIL_BITS	= 0
 * GMP_NUMB_MASK	= -1 on 64 bits
 * GMP_NUMB_MAX		= unsigned -1 on 64 bits
 */

/*
 * Borrowed from gmp-impl.h line 471.
 */

/* GMP comment: From gmp.h, nicer names for internal use. */
#define ABS(x) ((x) >= 0 ? (x) : -(x))

/* GMP comment: Field access macros.  */
#define SIZ(x) ((x)->_mp_size)
#define ABSIZ(x) ABS (SIZ (x))
#define PTR(x) ((x)->_mp_d)
#define LIMBS(x) ((x)->_mp_d)
#define EXP(x) ((x)->_mp_exp)
#define PREC(x) ((x)->_mp_prec)
#define ALLOC(x) ((x)->_mp_alloc)

/* 
 * Borrowed from gmp-impl.h line 1311.
 * Original version:
 * #define MPZ_REALLOC(z,n) (UNLIKELY ((n) > ALLOC(z))     \
 *                           ? (mp_ptr) _mpz_realloc(z,n)  \
 *                           : PTR(z))
 */

#define MPZ_REALLOC(z,n) (((n) > ALLOC(z))		\
			  ? (mp_ptr) _mpz_realloc(z,n)	\
			  : PTR(z))

/*
 * Sort of memcpy. Generic code taken from gmp-impl.h line 1140.
 * Copy N limbs from SRC to DST decrementing, N==0 allowed.  
 *
 * Removed: two ASSERT statements.
 *
 * Used by mul_2exp
 */

#define MPN_COPY_DECR(dst, src, n)                      \
  do {                                                  \
    if ((n) != 0)                                       \
      {                                                 \
        mp_size_t __n = (n) - 1;                        \
        mp_ptr __dst = (dst) + __n;                     \
        mp_srcptr __src = (src) + __n;                  \
        mp_limb_t __x;                                  \
        __x = *__src--;                                 \
        if (__n != 0)                                   \
          {                                             \
            do                                          \
              {                                         \
                *__dst-- = __x;                         \
                __x = *__src--;                         \
              }                                         \
            while (--__n);                              \
          }                                             \
        *__dst-- = __x;                                 \
      }                                                 \
  } while (0)

/*
 * Sort of memset (dst, 0). Generic code taken from gmp-impl.h line 1229.
 *
 * Removed: one ASSERT statement.
 *
 * Used by mul_2exp
 */

#define MPN_ZERO(dst, n)                        \
  do {                                          \
    if ((n) != 0)                               \
      {                                         \
        mp_ptr __dst = (dst);                   \
        mp_size_t __n = (n);                    \
        do                                      \
          *__dst++ = 0;                         \
        while (--__n);                          \
      }                                         \
  } while (0)

/***********************************************************************
 * Functions for big integers (mpz_t).
 ***********************************************************************/

/*
 * From mpz/set_si.c
 */

BA0_DLL void ba0_x64_mpz_set_si (__mpz_struct* dest, ba0_int_p val)
{ mp_size_t size;
  mp_limb_t vl;

  vl = (mp_limb_t) (val >= 0 ? val : -val);	/* Removed cast */
  dest->_mp_d[0] = vl & GMP_NUMB_MASK;
  size = vl != 0;
			/* Removed #if GMP_NAIL_BITS != 0 */
  dest->_mp_size = val >= 0 ? size : -size;
}

/*
 * From mpz/set_ui.c
 */

BA0_DLL void ba0_x64_mpz_set_ui (__mpz_struct* dest, unsigned ba0_int_p val)
{ mp_size_t size;

  dest->_mp_d[0] = val & GMP_NUMB_MASK;
  size = val != 0;
			/* Removed #if BITS_PER_ULONG > GMP_NUMB_BITS */
  dest->_mp_size = size;
}

/*
 * From mpz/mul_i.h
 */

BA0_DLL void ba0_x64_mpz_mul_si 
		(__mpz_struct* prod, __mpz_struct* mult, ba0_int_p small_mult)
{
  mp_size_t size = SIZ(mult);
  mp_size_t sign_product = size;
  mp_limb_t sml;
  mp_limb_t cy;
  mp_ptr pp;

  if (size == 0 || small_mult == 0)
    {
      SIZ(prod) = 0;
      return;
    }

  size = ABS (size);

  sml = ABS (small_mult);			/* Macro MULTIPLICAND_ABS */
			/* Useless test, kept */
  if (small_mult <= GMP_NUMB_MAX)
    {
      MPZ_REALLOC (prod, size + 1);
      pp = PTR(prod);
      cy = mpn_mul_1 (pp, PTR(mult), size, sml & GMP_NUMB_MASK);
      pp[size] = cy;
      size += cy != 0;
    }
			/* Removed #if GMP_NAIL_BITS != 0 */
  SIZ(prod) = ((sign_product < 0) ^ (small_mult < 0)) ? -size : size;
}

/*
 * From mpz/mul_i.h
 */

BA0_DLL void ba0_x64_mpz_mul_ui 
		(__mpz_struct* prod, __mpz_struct* mult, 
		 unsigned ba0_int_p small_mult)
{
  mp_size_t size = SIZ(mult);
  mp_size_t sign_product = size;
  mp_limb_t sml;
  mp_limb_t cy;
  mp_ptr pp;

  if (size == 0 || small_mult == 0)
    {
      SIZ(prod) = 0;
      return;
    }

  size = ABS (size);

  sml = small_mult;				/* Macro MULTIPLICAND_ABS */
			/* Useless test, kept */
  if (small_mult <= GMP_NUMB_MAX)
    {
      MPZ_REALLOC (prod, size + 1);
      pp = PTR(prod);
      cy = mpn_mul_1 (pp, PTR(mult), size, sml & GMP_NUMB_MASK);
      pp[size] = cy;
      size += cy != 0;
    }
			/* Removed #if GMP_NAIL_BITS != 0 */
  SIZ(prod) = ((sign_product < 0) ^ (small_mult < 0)) ? -size : size;
}

/*
 * mpz/mul_2exp.c
 *
 * Actually, it is unlikely that cnt > LONG_MAX :-)
 */

BA0_DLL void ba0_x64_mpz_mul_2exp 
			(__mpz_struct* w, __mpz_struct* u, ba0_int_p cnt)
{
  mp_size_t usize = u->_mp_size;
  mp_size_t abs_usize = ABS (usize);
  mp_size_t wsize;
  mp_size_t limb_cnt;
  mp_ptr wp;
  mp_limb_t wlimb;

  if (usize == 0)
    {
      w->_mp_size = 0;
      return;
    }

  limb_cnt = (mp_size_t)cnt / GMP_NUMB_BITS;	/* Added cast */
  wsize = abs_usize + limb_cnt + 1;
  if (w->_mp_alloc < wsize)
    _mpz_realloc (w, wsize);

  wp = w->_mp_d;
  wsize = abs_usize + limb_cnt;

  cnt %= GMP_NUMB_BITS;
  if (cnt != 0)
    {
      wlimb = mpn_lshift (wp + limb_cnt, u->_mp_d, abs_usize, (mp_size_t)cnt);
						/* Added cast */
      if (wlimb != 0)
        {
          wp[wsize] = wlimb;
          wsize++;
        }
    }
  else
    {
      MPN_COPY_DECR (wp + limb_cnt, u->_mp_d, abs_usize);
    }

  /* Zero all whole limbs at low end.  Do it here and not before calling
     mpn_lshift, not to lose for U == W.  */
  MPN_ZERO (wp, limb_cnt);

  w->_mp_size = usize >= 0 ? wsize : -wsize;
}

/*
 * mpz/pow_ui.c is too complicated for me.
 * This one will be sufficient.
 */

BA0_DLL void ba0_x64_mpz_pow_ui
		(__mpz_struct* w, __mpz_struct* u, unsigned ba0_int_p n)
{   ba0_mpz_t a;
    unsigned ba0_int_p p;

    if (n == 0)
    {	ba0_mpz_set_ui (w, 1);
	return;
    }
    ba0_mpz_init_set (a, u);
    ba0_mpz_set_ui (w, 1);
    for (p = n; p != 1; p /= 2)
    {	if (p % 2 == 1)
	    ba0_mpz_mul (w, w, a);
	ba0_mpz_mul (a, a, a);
    }
    ba0_mpz_mul (w, w, a);
    mpz_clear (a);
}

BA0_DLL void ba0_x64_mpz_ui_pow_ui
		(__mpz_struct* w, unsigned ba0_int_p u, unsigned ba0_int_p n)
{   ba0_mpz_t a;

    ba0_mpz_init_set_ui (a, u);
    ba0_mpz_pow_ui (w, a, n);
    mpz_clear (a);
}

BA0_DLL void ba0_x64_mpz_si_pow_ui
                        (__mpz_struct* w, ba0_int_p u, unsigned ba0_int_p n)
{
    if (u >= 0)
        ba0_mpz_ui_pow_ui (w, (unsigned ba0_int_p)u, n);
    else
    {   ba0_mpz_ui_pow_ui (w, (unsigned ba0_int_p)(-u), n);
        if (n % 2 == 1)
            ba0_mpz_neg (w, w);
    }
}

/***********************************************************************
 * Functions for big rational numbers (mpq_t).
 ***********************************************************************/

/*
 * mpq/set_si.c
 */

BA0_DLL void ba0_x64_mpq_set_si 
			(__mpq_struct* dest, ba0_int_p num, ba0_int_p den)
{
  unsigned ba0_int_p abs_num;
			/* Removed if (GMP_NUMB_BITS < BITS_PER_ULONG) */
  abs_num = BA0_ABS(num);
  if (num == 0)
    {
      /* Canonicalize 0/d to 0/1.  */
      den = 1;
      dest->_mp_num._mp_size = 0;
    }
  else
    {
      dest->_mp_num._mp_d[0] = abs_num;
      dest->_mp_num._mp_size = num > 0 ? 1 : -1;
    }

  dest->_mp_den._mp_d[0] = den;
  dest->_mp_den._mp_size = (den != 0);
}

/*
 * mpq/set_ui.c
 */

BA0_DLL void ba0_x64_mpq_set_ui (__mpq_struct* dest, unsigned ba0_int_p num, 
						     unsigned ba0_int_p den)
{
			/* Removed if (GMP_NUMB_BITS < BITS_PER_ULONG) */
  if (num == 0)
    {
      /* Canonicalize 0/n to 0/1.  */
      den = 1;
      dest->_mp_num._mp_size = 0;
    }
  else
    {
      dest->_mp_num._mp_d[0] = num;
      dest->_mp_num._mp_size = 1;
    }

  dest->_mp_den._mp_d[0] = den;
  dest->_mp_den._mp_size = (den != 0);
}

