#include "ba0.h"

typedef struct my_complex {
    ba0_double realpart;   /* This a pointer to a double */
    double imagpart;       /* This is a double */
} *my_complex;

/* The default constructor */

static void init_my_complex (my_complex z)
{
    z->realpart = ba0_new_double ();
}

static my_complex new_my_complex (void)
{   my_complex z;

    z = (my_complex)ba0_alloc (sizeof (struct my_complex));
    init_my_complex (z);
    return z;
}

static void set_my_complex (my_complex z1, my_complex z2)
{
    *z1->realpart = *z2->realpart;
    z1->imagpart = z2->imagpart;
}

/* Very basic parser: my_complex = double [+/- i * double] */

static void* scanf_my_complex (void* x)
{   my_complex z;

/*
   Every scanf function receives either zero or a pointer to an initialized 
   data structure. In the first case, an area must be allocated.
*/
    if (x == (void*)0)
	z = new_my_complex ();
    else
        z = (my_complex)x;
/*
   The lexical analyzer is used for parsing my_complex
*/
    ba0_scanf ("%le", z->realpart);
    ba0_get_token_analex ();
    if (ba0_sign_token_analex ("+") || ba0_sign_token_analex ("-"))
    {   double posneg = ba0_sign_token_analex ("+") ? 1. : 0.;
        ba0_get_token_analex ();
        ba0_scanf ("i*%le", &z->imagpart);
        z->imagpart *= posneg;
    } else
    {   ba0_unget_token_analex ();
        z->imagpart = 0.;
    }
/*
   The pointer to the data structure must be returned
*/
    return z;
}

static void printf_my_complex (void* x)
{   my_complex z = (my_complex)x;

    if (z->imagpart == 0.)
        ba0_printf ("%le", z->realpart);
    else if (z->imagpart < 0.)
    {   double mz = - z->imagpart;
        ba0_printf ("%le - i * %le", z->realpart, &mz);
    } else
        ba0_printf ("%le + i * %le", z->realpart, &z->imagpart);
}

static char struct_my_complex [] = "struct my_complex";

static ba0_int_p garbage1_my_complex (void* x, enum ba0_garbage_code code)
{   my_complex z = (my_complex)x;
    ba0_int_p n;
/*
   Isolated structures must be moved separately by the garbage collector.
   Embedded structures are moved together with their embedding structure.
*/
    n = 0;
    if (code == ba0_isolated)
    {   ba0_new_gc_info (z, sizeof (struct my_complex), struct_my_complex);
        n += 1;
    }
/*
   The second call below is actually pointless.
*/
    n += ba0_garbage1 ("%le", z->realpart, ba0_isolated);
    n += ba0_garbage1 ("%le", &z->imagpart, ba0_embedded);         
/*
   The counter n of isolated data structures is returned
*/
    return n;
}

static void* garbage2_my_complex (void* x, enum ba0_garbage_code code)
{   my_complex z;
/*
   The string provided as second parameter to ba0_new_addr_gc_info
   is only useful for debugging. It must be exactly the same as that
   provided to ba0_new_gc_info by the garbage1 function.

   It must even be stored in a variable (Windows compiler).
*/
    if (code == ba0_isolated)
        z = (my_complex)ba0_new_addr_gc_info (x, struct_my_complex);
    else
        z = (my_complex)x;
/*
   The second call below is pointless.
*/
    z->realpart = (ba0_double)ba0_garbage2 ("%le", z->realpart, ba0_isolated);
    ba0_garbage2 ("%le", &z->imagpart, ba0_embedded);
/*
   The new address of the data structure is returned
*/
    return z;
}

static void* copy_my_complex (void* x)
{   my_complex z;

    z = new_my_complex ();
    set_my_complex (z, (my_complex)x);
    return z;
}

int main ()
{   ba0_table A;
    struct my_complex pi, **tab;
    struct ba0_mark M;
    bool b;

    ba0_restart (0, 0);
/*
   The new format %my_complex is defined
*/
    ba0_define_format_with_sizelt 
                ("my_complex", sizeof (struct my_complex),
                 &scanf_my_complex,
                 &printf_my_complex,
                 &garbage1_my_complex,
                 &garbage2_my_complex,
                 &copy_my_complex);
/*
   Reads one my_complex in pi. The variable must be initialized before parsing.
*/
    init_my_complex (&pi);
    ba0_sscanf2 ("3.14", "%my_complex", &pi);
/*
   Just wastes 10 bytes
*/
    ba0_record (&M);
    ba0_alloc (10);
/*
   Reads a table of my_complex. Then applies the garbage collector over it.
   The wasted 10 bytes are recovered.
*/
    A = ba0_new_table ();
    ba0_sscanf2 ("[3. + i*5e-1, -7.e0]", "%t[%my_complex]", A);
    ba0_garbage ("%t[%my_complex]", &M, &A);
/*
   Prints the table.
    ba0_printf ("pi = %my_complex, A = %t[%my_complex]\n", &pi, A);
*/

    tab = (my_complex*)A->tab;
    b = A->size == 2 && 
        *(tab [0]->realpart) == 3. && tab [0]->imagpart == 0.5 && 
        *(tab [1]->realpart) == -7. && tab [1]->imagpart == 0.;

    ba0_terminate (ba0_init_level);
    return b ? 0 : 1;
}
