#include "bad_nonzero.h"
#include "bad_reduction.h"
#include "bad_regularize.h"
#include "bad_quadruple.h"
#include "bad_Rosenfeld_Groebner.h"
#include "bad_global.h"
#include "bad_base_field.h"
#include "bad_stats.h"

static void bad_first_quadruples 
			(bad_tableof_quadruple, 
			 bad_attchain,
			 bap_tableof_polynom_mpz, 
			 bap_tableof_polynom_mpz,
			 enum bad_typeof_reduction,
			 bad_base_field);

static void bad_split_on_initial_and_separant
                        (bad_tableof_quadruple, bap_polynom_mpz);

static void bad_Euclid
			(bap_tableof_polynom_mpz,
			 bad_tableof_quadruple,
			 bap_polynom_mpz, bap_polynom_mpz, bav_variable);

#define BAD_RG_DEBUG 1
#undef BAD_RG_DEBUG

BAD_DLL void bad_Rosenfeld_Groebner 
			(bad_intersectof_regchain tabC, 
			 bap_tableof_polynom_mpz P, 
			 bap_tableof_polynom_mpz S,
			 bad_base_field K,
			 bad_splitting_control C)
{
/*
 * Initializations are just there to avoid stupid warnings
 */
    bad_nonzero_function* old_nonzero;
    void* old_nonzero_data;

    volatile bad_splitting_control control;
    ba0_int_p numberof_input_equations = 0;
    bool use_dimlb = false, reject;

    struct bad_tableof_quadruple tabG;
    bad_quadruple G;

    enum bad_typeof_reduction type_red = bad_full_reduction;

    struct bap_tableof_polynom_mpz ineqs;
    struct bap_tableof_product_mpz factored_ineqs;

    struct bap_product_mpz prod;

    struct bap_tableof_polynom_mpz tabP;

    bap_polynom_mpz p = (bap_polynom_mpz)0;

    bav_variable v;
    ba0_int_p i, g, k, counter, choice;
    struct ba0_mark M;
    struct ba0_exception_code code;
/*
 * Initializes stats
 */
    bad_init_stats (0);
    bad_stats_data.begin = time (0);
/*
 * Automatic properties will thus be inherited by the resulting regular chains.
 */
    tabC->inter.size = 0;
    bad_set_automatic_properties_attchain (&tabC->attrib);

    ba0_push_another_stack ();
    ba0_record (&M);
/*
 * From now on, use control.
 */
    if (C == (bad_splitting_control)0)
	control = bad_new_splitting_control ();
    else
	control = C;
/*
 * Should be stored within the "if". Placed here to avoid a stupid warning
 */
    bad_get_settings_nonzero (&old_nonzero, &old_nonzero_data);
    if (K != (bad_base_field)0)
    {	if (! bad_is_a_compatible_base_field (K, &tabC->attrib))
	    BA0_RAISE_EXCEPTION (BAD_ERRIBF);
	if (bad_has_desired_property_attchain 
			(&tabC->attrib, bad_autoreduced_desired_property))
	    bad_set_settings_nonzero 
				(&bad_nonzero_reduced_polynom_base_field, K);
	else
	    bad_set_settings_nonzero (&bad_nonzero_polynom_base_field, K);

	BA0_PUSH_EXCEPTION (code);
/*
 * The exception point is only designed for restoring the values of some
 * global variables that are modified when alg_indep is nonzero. The exception
 * is then raised again for further processing: volatile is not needed.
 */
	if (ba0_exception_is_raised (code))
	{   bad_set_settings_nonzero (old_nonzero, old_nonzero_data);
	    BA0_RAISE_EXCEPTION (ba0_mesgerr);
	}
    }

    if (bad_defines_a_differential_ideal_attchain (&tabC->attrib))
	type_red = bad_full_reduction;
    else
	type_red = bad_algebraic_reduction;
/*
 * The bad_first_quadruples function stores P and S in the components P and S 
 * of quadruples and stores them in tabG. 
 *
 * The base field defining equations are removed from P.
 */
    ba0_init_table ((ba0_table)&tabG);
    bad_first_quadruples (&tabG, &tabC->attrib, P, S, type_red, K);
/*
 * tabP is used by the Euclidean algorithm.
 */
    ba0_init_table ((ba0_table)&tabP);
    ba0_realloc2_table ((ba0_table)&tabP, 1, 
			(ba0_new_function*)&bap_new_polynom_mpz);
    tabP.size = 1;
/*
 * ineqs and factored_ineqs are used when some simplification
 * (factorization) of some inequation is discovered.
 */
    ba0_init_table ((ba0_table)&ineqs);
    ba0_init_table ((ba0_table)&factored_ineqs);
/*
 * p and prod are local variables.
 * counter is only used for debugging.
 */
    p = bap_new_polynom_mpz ();
    bap_init_product_mpz (&prod);
    counter = 0;
/*
 * bad_easy_criterion_penalty is set to 2 and not modified anymore.
 * Heuristic in the critical pair management.
 */
    choice = 1;
    bad_easy_criterion_penalty = 2;
/*
 * Main loop.
 * It stops if there is not any quadruple to process anymore or
 * if the splitting control variable tells us to do so.
 */
    while (tabG.size > 0 &&
			(!control->first_branch_only || tabC->inter.size != 1))
    {
	counter++;
#if defined (BAD_RG_DEBUG)
	ba0_printf ("------------------------------------------------\n");
	ba0_printf ("Round %d, |tabG| = %d\n", counter, tabG.size);
#endif

/*
 * Observe G is not pulled out of the stack. G is the top element of the stack
 */
	G = tabG.tab [tabG.size - 1];
#if defined (BAD_RG_DEBUG)
	ba0_printf ("%quadruple\n", G);
#endif
	if (bad_first_case_quadruple (G))
	{   
/*
 * The dimension lower bound. 
 * numberof_input_equations does not count the base field defining equations.
 */
	    use_dimlb = bad_apply_dimension_lower_bound_splitting_control 
						(control, G->P, &tabC->attrib);
	    numberof_input_equations = ba0_length_list ((ba0_list)G->P);
	}
/*
 * The dimension lower bound can be applied to reject quadruples
 */
	reject = false;
	if (use_dimlb)
	{   reject = 
		bad_codimension_regchain (&G->A, K) > numberof_input_equations;
	}

	if (reject)
	{
#if defined (BAD_RG_DEBUG)
	    ba0_printf ("Rejecting quadruple\n");
#endif
	    tabG.size -= 1;
	} else if (bad_is_a_listof_rejected_critical_pair (G->D) &&
					G->P == (bap_listof_polynom_mpz)0)
	{
/*
 * The purely algebraic treatment.
 * G is pulled out of the stack.
 * The reg_characteristic algorithm transforms it as finitely many regchains
 *     which are stacked over tabC.
 */
#if defined (BAD_RG_DEBUG)
	    ba0_printf ("Calling reg_characteristic\n");
#endif
	    tabG.size -= 1;
	    ba0_pull_stack ();
	    bad_reg_characteristic_quadruple (tabC, G, (bad_regchain)0);
#if defined (BAD_RG_DEBUG)
	    ba0_printf ("|tabC| = %d\n", tabC->inter.size);
#endif
	    ba0_push_another_stack ();
	} else 
	{   bad_critical_pair pair;
/*
	    ba0_printf ("loop %d. %t[\n\n%quadruple]\n", counter, &tabG);
*/
	    bad_pick_and_remove_quadruple (p, G, &pair, choice);
/*
 * If the reduction to zero test is deterministic, we may as well
 * skip it since we are going to reduce p anyway.
 *
 * If it is probabilistic, it is better to perform it before
 * the reduction.
 */
	    if (bad_redzero_strategy == bad_probabilistic_redzero_strategy)
	    {	if (bad_is_a_reduced_to_zero_polynom_by_regchain
							(p, &G->A, type_red))
		{   bad_stats_data.reductions_to_zero += 1;
		    continue;
		}
	    }

	    bad_reduce_polynom_by_regchain 
				(&prod, (bap_product_mpz)0, p, &G->A,
				 type_red, bad_all_derivatives_to_reduce);
/*
 * If the picked polynomial reduces to zero, pick the next one.
 */
	    if (bap_is_zero_product_mpz (&prod))
	    {	bad_stats_data.reductions_to_zero += 1;
		continue;
	    }
/*
 * The picked polynomial, which is now a product gets simplified.
 * The simplification preprocess may exhibit factorizations of
 * inequations, which are propagated in tabG.
 */
	    bad_preprocess_equation (&prod, &ineqs, &factored_ineqs, G);
	    bad_report_simplification_of_inequations 
					(&tabG, &ineqs, &factored_ineqs);

#if defined (BAD_RG_DEBUG)
	    ba0_printf ("new equation : %Pz\n", &prod);
#endif
	    if (prod.size == 0)
	    {	tabG.size -= 1;
		continue;
	    }
/*
 * Assume prod is f1 ... fn = 0
 * The following function performs the following :
 * 	for i = 1 to n-1 do
 * 	    H := G, "fi = 0"
 * 	    G := G, "fi != 0"
 * 	done
 * So that, afterwards, G involves f1 <> 0, ..., f{n-1} <> 0 and we 
 * only have to deal with the case fn = 0. 
 */
	    g = tabG.size - 1;
	    bad_realloc_tableof_position (&G->pos, G->pos.size + 1);

	    prod.size -= 1;
	    bad_split_on_factors_of_equations 
					(&tabG, &prod, (bap_product_mpz)0);

	    for (i = g; i < tabG.size; i++)
		bad_append_tableof_position 
			(&tabG.tab [i]->pos, &tabG.tab [i]->pos, 
			 bad_factor_case, i - g, tabG.size - 1 - g);
/*
 *  with the same notations as above, p = fn
 */
	    BA0_SWAP (struct bap_polynom_mpz, *p, prod.tab [prod.size].factor);
/*
 * The following function splits on the initial and separant of p.
 * This function calls bad_preprocess_equation and may simplify
 * the inequations of many quadruples.
 */
	    bad_split_on_initial_and_separant (&tabG, p);

	    v = bap_leader_polynom_mpz (p);
	    if (! bad_is_leader_of_regchain (v, &G->A, &k))
	    {	g = tabG.size - 1;
/*
 * The complete function transforms G as finitely many quadruples which 
 * overwrite G at the top of the stack.
 */
		bad_complete_quadruple (&tabG, p, (bad_regchain)0);
#if defined BAD_RG_DEBUG
		if (tabG.size - 1 < g)
		    ba0_printf 
			("component removed by bad_complete_quadruple\n");
#endif
	    } else
	    {   tabP.size = 1;
		g = tabG.size - 1;
/*
 * The Euclid function transforms G as finitely many quadruples G1, ..., Gn
 * which overwrite G at the top of the stack. It computes also the 
 * corresponding gcds g1, ..., gn of G->A.decision_system.tab [k] 
 * and p which overwrite the top element of the tabV stack.
 *
 * complete (Gi, gi) is performed afterwards for i = 1 ... n.
 */
#if defined BAD_RG_DEBUG
		ba0_printf ("\nEuclid (%Az, %Az)\n%quadruple\n", 
					G->A.decision_system.tab [k], p, G);
#endif
		bad_Euclid (&tabP, &tabG, G->A.decision_system.tab [k], p, v);

		for (i = g; i < tabG.size; i++)
		    bad_append_tableof_position
			(&tabG.tab [i]->pos, &tabG.tab [i]->pos,
			 bad_euclid_case, i - g, tabG.size - 1 - g);
#if defined BAD_RG_DEBUG
		for (i = 0; i < tabP.size; i++)
		{   ba0_printf ("\ngcd = %Az\n%quadruple\n", 
					tabP.tab [i], tabG.tab [i+g]);
		}
#endif
		for (i = tabP.size - 1; i >= 0; i--)
		{   ba0_move_to_tail_table 
				((ba0_table)&tabG, (ba0_table)&tabG, i + g);
		    bad_complete_quadruple 
				(&tabG, tabP.tab [i], (bad_regchain)0);
		}
	    }
	}
    }
    ba0_pull_stack ();
/*
 * All explicit regchains get the prime attribute
 */
    bad_fast_primality_test_intersectof_regchain (tabC);

    if (K != (bad_base_field)0)
    {	bad_set_settings_nonzero (old_nonzero, old_nonzero_data);
	ba0_pull_exception (code);
    }
    ba0_restore (&M);

    bad_stats_data.end = time (0);
}

/*
 * Generates the splittings corresponding to the vanishing of
 * the initial and the separant of p.
 *
 * G = the quadruple on the top of tabG
 * p = some new equation
 * 
 * init, reductum = initial and reductum of p
 * apply bad_preprocess_equation over these two equations
 * apply bad_split_on_factors_of_equations over the results.
 *
 * Perform similar computations with the separants.
 */

static void bad_split_on_initial_and_separant
			(bad_tableof_quadruple tabG, bap_polynom_mpz p)
{   bad_quadruple G;
    struct bap_tableof_product_mpz factored_ineqs;
    struct bap_tableof_polynom_mpz ineqs;
    struct bap_product_mpz prod_init, prod_reductum, prod_sep, prod_sepuctum;
    struct bap_polynom_mpz init, reductum, sep, sepuctum;
    struct ba0_mark M;
    ba0_int_p i, g;

    g = tabG->size - 1;
    G = tabG->tab [g];
    bad_realloc_tableof_position (&G->pos, G->pos.size + 1);

    ba0_push_another_stack ();
    ba0_record (&M);
/*
 * To manage exhibited factorizations of inequations
 */
    ba0_init_table ((ba0_table)&ineqs);
    ba0_init_table ((ba0_table)&factored_ineqs);
/*
 * First the initial
 */
    bap_init_readonly_polynom_mpz (&init);
    bap_init_readonly_polynom_mpz (&reductum);
    bap_initial_and_reductum_polynom_mpz (&init, &reductum, p);

    bap_init_product_mpz (&prod_init);
    bap_init_product_mpz (&prod_reductum);
/*
 * Discard obviously useless splittings
 */
    if (! (*bad_nonzero) (&init) && ! (*bad_nonzero) (&reductum))
    {	bap_factor_easy_polynom_mpz (&prod_init, &init, G->S);
	bad_preprocess_equation (&prod_init, &ineqs, &factored_ineqs, G);
	if (ineqs.size > 0)
	{   ba0_pull_stack ();
	    bad_report_simplification_of_inequations 
					(tabG, &ineqs, &factored_ineqs);
	    ba0_push_another_stack ();
	}
	if (prod_init.size > 0)
	{   
/*
 * The initial is never zero but the reductum can be zero.
 */
	    bap_factor_easy_polynom_mpz (&prod_reductum, &reductum, G->S);
	    bad_preprocess_equation
				(&prod_reductum, &ineqs, &factored_ineqs, G);
	    if (ineqs.size > 0)
            {   ba0_pull_stack ();
                bad_report_simplification_of_inequations
                                        (tabG, &ineqs, &factored_ineqs);
                ba0_push_another_stack ();
            }
	    if (prod_reductum.size > 0 || 
				bap_is_zero_product_mpz (&prod_reductum))
	    {   ba0_pull_stack ();
		bad_split_on_factors_of_equations
					(tabG, &prod_init, &prod_reductum);
		ba0_push_another_stack ();
	    }
	}
    }
/*
 * Then the separant, if needed.
 */    
    if (bap_leading_degree_polynom_mpz (p) > 1 &&
		 bad_has_desired_property_attchain 
			(&G->A.attrib, bad_squarefree_desired_property))
    {	bap_init_polynom_mpz (&sep);
	bap_init_polynom_mpz (&sepuctum);
	bap_separant_polynom_mpz (&sep, p);
	bap_prem_polynom_mpz 
		(&sepuctum, (bav_Idegree*)0, p, &sep, BAV_NOT_A_VARIABLE);

	bap_init_product_mpz (&prod_sep);
	bap_init_product_mpz (&prod_sepuctum);
/*
 * Discard obviously useless splittings. 
 * Observe that the separant involves the leading derivative of p.
 */
	if (! (*bad_nonzero) (&sepuctum))
	{   bap_factor_easy_polynom_mpz (&prod_sep, &sep, G->S);
	    bad_preprocess_equation (&prod_sep, &ineqs, &factored_ineqs, G);
            if (ineqs.size > 0)
            {   ba0_pull_stack ();
                bad_report_simplification_of_inequations
                                        (tabG, &ineqs, &factored_ineqs);
                ba0_push_another_stack ();
            }
	    if (prod_sep.size > 0)
	    {   
		bap_factor_easy_polynom_mpz (&prod_sepuctum, &sepuctum, G->S);
		bad_preprocess_equation 
				(&prod_sepuctum, &ineqs, &factored_ineqs, G);
		if (ineqs.size > 0)
		{   ba0_pull_stack (); 
		    bad_report_simplification_of_inequations
					(tabG, &ineqs, &factored_ineqs);
		    ba0_push_another_stack (); 
		}
		if (prod_sepuctum.size > 0 || 
				bap_is_zero_product_mpz (&prod_sepuctum))
		{   ba0_pull_stack ();
		    bad_split_on_factors_of_equations
					(tabG, &prod_sep, &prod_sepuctum);
		    ba0_push_another_stack ();
		}
	    }
	}
    }

    ba0_pull_stack ();

    for (i = g; i < tabG->size; i++)
    {	G = tabG->tab [i];
	bad_append_tableof_position 
	    (&G->pos, &G->pos, bad_splitting_case, i - g, tabG->size - 1 - g);
    }

    ba0_restore (&M);
}

/*
 * Initializes the first quadruples with P and S.
 * The process may lead to splittings.
 *
 * All generated quadruples hold the same number of input equations.
 *
 * The base field equations, if any, are inserted in the regchains.
 */

static void bad_first_quadruples 
	(bad_tableof_quadruple tabG, bad_attchain attrib,
	 bap_tableof_polynom_mpz P, bap_tableof_polynom_mpz S, 
	 enum bad_typeof_reduction type_red, bad_base_field K)
{   bad_quadruple G;
    struct bap_tableof_polynom_mpz ineqs;
    struct bap_tableof_product_mpz factored_ineqs;
    struct bap_product_mpz prod;
    struct ba0_mark M;
    ba0_int_p i, j, k;
    bool base_field;

    if (tabG->size != 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    ba0_realloc2_table
                ((ba0_table)tabG, 1, (ba0_new_function*)bad_new_quadruple);
    G = tabG->tab [0];
    tabG->size = 1;
/*
 * The relations of K are directly stored in G->A.
 * Overwrites the attributes of K->relations.attrib with attrib.
 */
    base_field = K != (bad_base_field)0;
    if (base_field)
        bad_set_regchain (&G->A, &K->relations);

    bad_set_attchain (&G->A.attrib, attrib);

    ba0_push_another_stack ();
    ba0_record (&M);

    ba0_init_table ((ba0_table)&ineqs);
    ba0_init_table ((ba0_table)&factored_ineqs);
    bap_init_product_mpz (&prod);
/*
 * First process the inequations.
 * No splitting occurs since "p*q != 0" is just "p != 0", "q != 0".
 * However, the quadruple G may be proved inconsistent and discarded.
 */
    if (S)
    {   for (i = 0; i < S->size && tabG->size > 0; i++)
        {   
/*
 * An inequation might be reduced to zero by a base field equation
 * factor_easy does not expect zero polynomials.
 */
	    if (base_field)
                bad_reduce_polynom_by_regchain
                    (&prod, (bap_product_mpz)0, S->tab [i], &G->A,
                     type_red, bad_all_derivatives_to_reduce);
            else if (bap_is_numeric_polynom_mpz (S->tab [i]))
		bap_set_product_polynom_mpz (&prod, S->tab [i], 1);
	    else
		bap_factor_easy_polynom_mpz (&prod, S->tab [i], G->S);
/*
 * In the case "0 != 0", remove the quadruple G from the stack.
 */
	    if (bap_is_zero_product_mpz (&prod))
		tabG->size -= 1;
	    else
	    {	bad_preprocess_equation (&prod, &ineqs, &factored_ineqs, G);
		ba0_pull_stack ();
		bad_report_simplification_of_inequations 
					(tabG, &ineqs, &factored_ineqs);
		for (j = 0; j < prod.size; j++)
		    G->S = bad_insert_in_listof_polynom_mpz
						(&prod.tab [j].factor, G->S);
		ba0_push_another_stack ();
	    }
        }
    }
/*
 * Then process the equations.
 * Splittings can occur, since "p*q = 0" is "p = 0" or "q = 0" (two cases).
 */
    for (i = 0; i < P->size; i++)
    {   for (k = tabG->size - 1; k >= 0; k--)
	{   
/*
 * Fix some new equation P->tab [i].
 * The quadruples to be considered are in tabG->tab [0 .. tabG->size-1].
 * At each loop, put the current one on the top of the stack and process
 * the new equation w.r.t. to it. This leads to finitely many cases.
 */
	    ba0_move_to_tail_table ((ba0_table)tabG, (ba0_table)tabG, k);
	    G = tabG->tab [tabG->size - 1];
	    if (base_field)
                bad_reduce_polynom_by_regchain
                    (&prod, (bap_product_mpz)0, P->tab [i], &G->A,
                     type_red, bad_all_derivatives_to_reduce);
	    else if (bap_is_numeric_polynom_mpz (P->tab [i]))
	        bap_set_product_polynom_mpz (&prod, P->tab [i], 1);
	    else
	        bap_factor_easy_polynom_mpz (&prod, P->tab [i], G->S);

	    if (bap_is_numeric_product_mpz (&prod))
	    {	if (! bap_is_zero_product_mpz (&prod))
		    tabG->size -= 1;
	    } else
	    {   bad_preprocess_equation (&prod, &ineqs, &factored_ineqs, G);

		ba0_pull_stack ();
		bad_report_simplification_of_inequations
					(tabG, &ineqs, &factored_ineqs);
	        if (prod.size == 0)
		    tabG->size -= 1;
	        else
		{   
/*
 * Apply bad_split_on_factors_of_equations for all but the last factor.
 * Store the last equation in G->P manually.
 */
		    prod.size -= 1;
		    bad_split_on_factors_of_equations 
					(tabG, &prod, (bap_product_mpz)0);
		    G->P = bad_insert_in_listof_polynom_mpz
					(&prod.tab [prod.size].factor, G->P);
		}
		ba0_push_another_stack ();
	    }
        } 
    }
/*
 * Initializes the positions
 */
    ba0_pull_stack ();

    for (i = 0; i < tabG->size; i++)
    {	G = tabG->tab [i];
	bad_append_tableof_position
		(&G->pos, &G->pos, bad_first_case, i, tabG->size-1);
    }

    ba0_restore (&M);
}

/*
 * Let G = tabG->tab [tabG->size-1]
 * 
 * Computes the gcd of A and B, viewed as polynomial in v,
 * with coefficients taken modulo the ideal defined by G.
 *
 * Some splittings may be generated. 
 * The different cases are stored in tabG.
 * The corresponding gcds are stored in tabP.
 *
 * A depends on v.
 * B not necessarily.
 * Both polynomials are supposed to be preprocessed.
 */

static void bad_Euclid
	(bap_tableof_polynom_mpz tabP,
	 bad_tableof_quadruple tabG,
	 bap_polynom_mpz A, bap_polynom_mpz B, bav_variable v)
{   bad_quadruple G, H, I, J;
    struct bap_product_mpz prod, prod_init, prod_reductum;
    struct bap_polynom_mpz R, init, reductum;
    struct bap_tableof_product_mpz factored_ineqs;
    struct bap_tableof_polynom_mpz ineqs;
    struct ba0_mark M;
    ba0_int_p i, j, k, l, g, h, p, q, delta;
    bool init_may_vanish;

/*
 * g = the index of G in tabG.
 * p = the index in tabP corresponding to G.
 * delta is used for debugging.
 */
    g = tabG->size - 1;
    p = tabP->size - 1;
    delta = g - p;
    G = tabG->tab [g];
/*
 * If B is zero, the gcd is A.
 * if B does not depend on v, then it is stored in P and thus
 * implicitly becomes equal to zero, whence the gcd is A.
 */
    if (bap_is_zero_polynom_mpz (B))
    {	
	bap_set_polynom_mpz (tabP->tab [p], A);
    } else if (! bap_depend_polynom_mpz (B, v))
    {	
	G->P = bad_insert_in_listof_polynom_mpz (B, G->P);
	bap_set_polynom_mpz (tabP->tab [p], A);
    } else
    {   ba0_push_another_stack ();
	ba0_record (&M);

	bap_init_polynom_mpz (&R);
	bap_init_product_mpz (&prod);
	ba0_init_table ((ba0_table)&ineqs);
	ba0_init_table ((ba0_table)&factored_ineqs);
        bap_init_readonly_polynom_mpz (&init);
        bap_init_readonly_polynom_mpz (&reductum);
        bap_init_product_mpz (&prod_init);
        bap_init_product_mpz (&prod_reductum);
/*
 * R = remainder (A, B, v)
 */
	bap_gcd_prem_polynom_mpz (&R, (bap_product_mpz)0, A, B, v);
/*
 * Get rid of the two special cases: R = 0 and R cannot vanish.
 */
	if (bap_is_zero_polynom_mpz (&R))
	{   ba0_pull_stack ();
	    bad_Euclid (tabP, tabG, B, &R, v);
	    ba0_push_another_stack ();
	} else if ((*bad_nonzero) (&R))
	{   tabP->size -= 1;
	    tabG->size -= 1;
	} else 
	{   
/*
 * R can vanish. Preprocess it.
 */
	    bap_factor_easy_polynom_mpz (&prod, &R, G->S);
	    bad_preprocess_equation (&prod, &ineqs, &factored_ineqs, G);
	    if (ineqs.size > 0)
	    {   ba0_pull_stack ();
		bad_report_simplification_of_inequations
					(tabG, &ineqs, &factored_ineqs);
		ba0_push_another_stack ();
	    }
	    if (prod.size == 0)
	    {	
/*
 * This case may happen if prod only involve factors which occur in G->S
 */
		tabP->size -= 1;
		tabG->size -= 1;
	    } else
	    {
                for (i = 0; i < prod.size; i++)
                {        
/*
 * For each factor of R (index is i)
 */
                    ba0_pull_stack ();
/*
 * H = G
 * G = G + "ith factor of R != 0"
 * except for the last factor, in which case H = G.
 *
 * h = the index of H in tabG
 * q = the index in tabP corresponding to H
 */
                    if (i < prod.size - 1)
                    {   ba0_realloc2_table
                            ((ba0_table)tabG, tabG->size + 1,
                             (ba0_new_function*)&bad_new_quadruple);
                        ba0_realloc2_table
                            ((ba0_table)tabP, tabP->size + 1,
                             (ba0_new_function*)&bap_new_polynom_mpz);
                        tabG->size += 1;
                        tabP->size += 1;
                        h = tabG->size - 1;
                        q = tabP->size - 1;
                        H = tabG->tab [h];
                        bad_set_quadruple (H, G);
                        G->S = bad_insert_in_listof_polynom_mpz 
                                                (&prod.tab [i].factor, G->S);
                    } else
                    {   ba0_move_to_tail_table 
					((ba0_table)tabG, (ba0_table)tabG, g);
                        ba0_move_to_tail_table 
					((ba0_table)tabP, (ba0_table)tabP, p);
                        g = tabG->size - 1;
                        p = tabP->size - 1;
                        h = tabG->size - 1;
                        q = tabP->size - 1;
                        H = G;
                    }
                    ba0_push_another_stack ();
                    if (! bap_depend_polynom_mpz (&prod.tab [i].factor, v))
                    {   
/*
 * If the ith factor of R does not depend on v, suppose it equal to
 * zero and take B for gcd.
 */
                        ba0_pull_stack ();
                        bad_Euclid (tabP, tabG, B, &prod.tab [i].factor, v);
                        ba0_push_another_stack ();
                    } else
                    {
/*
 * Take the initial and reductum of the ith factor of R
 * The ith factor of prod should not involve trivial factors so
 * that the reductum should not be zero.
 */
                        bap_initial_and_reductum2_polynom_mpz 
                                   (&init, &reductum, &prod.tab [i].factor, v);
                        if (bap_is_zero_polynom_mpz (&reductum))
                            BA0_RAISE_EXCEPTION (BA0_ERRALG);
                        init_may_vanish = ! (*bad_nonzero) (&init);
                        if (init_may_vanish && ! (*bad_nonzero) (&reductum))
                        {   
/*
 * If both the initial and the reductum can vanish, generate splittings.
 */
                            bap_factor_easy_polynom_mpz 
				    (&prod_init, &init, H->S);
                            bad_preprocess_equation 
                                    (&prod_init, &ineqs, &factored_ineqs, H);
                            if (ineqs.size > 0)
                            {   ba0_pull_stack ();
                                bad_report_simplification_of_inequations
                                               (tabG, &ineqs, &factored_ineqs);
                                ba0_push_another_stack ();
                            }
                            if (prod_init.size > 0)
                            {   bap_factor_easy_polynom_mpz 
                                  (&prod_reductum, &reductum, H->S);
                                bad_preprocess_equation
                                  (&prod_reductum, &ineqs, &factored_ineqs, H);
                                if (ineqs.size > 0)
                                {   ba0_pull_stack ();
                                    bad_report_simplification_of_inequations
                                               (tabG, &ineqs, &factored_ineqs);
                                    ba0_push_another_stack ();
                                }
                                if (prod_reductum.size > 0)
                                {   
                                    ba0_pull_stack ();
/*
 * If, really, both initial and reductum can vanish, split.
 * Observe the (bap_product_mpz)0, which means that the reductum
 * is temporarily assumed identically zero.
 */
                                    bad_split_on_factors_of_equations
                                        (tabG, &prod_init, (bap_product_mpz)0);
/*
 * H was moved on the top of tabG so that h is no more the index of H !
 */
                                    if (H != tabG->tab [tabG->size - 1])
                                        BA0_RAISE_EXCEPTION (BA0_ERRALG);
                                    ba0_realloc2_table 
                                        ((ba0_table)tabP, 
                                         tabP->size + tabG->size - h - 1,
                                         (ba0_new_function*)
						&bap_new_polynom_mpz);
                                    tabP->size += tabG->size - h - 1;
/*
 * For each new quadruple generated by the splittings, 
 *     for each factor of the reductum, recursive call:
 *          Euclid (B, the factor of the reductum) over the quadruple
 */
                                    for (j = tabG->size - 2, 
                                         k = tabP->size - 2; j >= h; j--, k--)
                                    {   for (l = 0; l < prod_reductum.size; l++)
                                        {   
/*
 * I = the new quadruple
 * J = I 
 * I = I + "the lth factor of the reductum != 0"
 * except for the last factor, in which case I = J
 */
                                            if (l < prod_reductum.size - 1)
                                            {   ba0_realloc2_table
                                                    ((ba0_table)tabG, 
                                                     tabG->size + 1,
                                                     (ba0_new_function*)
                                                           &bad_new_quadruple);
                                                ba0_realloc2_table
                                                    ((ba0_table)tabP, 
                                                     tabP->size + 1,
                                                     (ba0_new_function*)
                                                         &bap_new_polynom_mpz);
                                                tabG->size += 1;
                                                tabP->size += 1;
                                                I = tabG->tab [j];
                                                J = tabG->tab [tabG->size - 1];
                                                bad_set_quadruple (J, I);
                                                I->S = bad_insert_in_listof_polynom_mpz
							(&prod_reductum.tab [l].factor, I->S);
                                            } else
                                            {   ba0_move_to_tail_table 
                                                    ((ba0_table)tabG, 
                                                     (ba0_table)tabG, j);
                                                ba0_move_to_tail_table
                                                    ((ba0_table)tabP, 
                                                     (ba0_table)tabP, k);
                                            }
                                            bad_Euclid 
                                                (tabP, tabG, B, 
                                                 &prod_reductum.tab [l].factor,
						 v);
                                        }
                                    }
                                    ba0_push_another_stack ();
                                }
                            }
                        } else if (init_may_vanish)
                        {   
/*
 * The reductum cannot vanish but the initial, isolated, could.
 * To make sure that the gcd is correct, force it not to vanish.
 */
                            bap_factor_easy_polynom_mpz 
				      (&prod_init, &init, H->S);
                            bad_preprocess_equation 
                                      (&prod_init, &ineqs, &factored_ineqs, H);
                            if (ineqs.size > 0)
                            {   ba0_pull_stack ();
                                bad_report_simplification_of_inequations
                                               (tabG, &ineqs, &factored_ineqs);
                                ba0_push_another_stack ();
                            }
                            ba0_pull_stack ();
                            for (j = 0; j < prod_init.size; j++)
                                H->S = bad_insert_in_listof_polynom_mpz 
                                             (&prod_init.tab [j].factor, H->S);
                            ba0_push_another_stack ();
                        }
                        if (H != tabG->tab [h])
                            BA0_RAISE_EXCEPTION (BA0_ERRALG);
/*
 * Over H, the initial of the ith factor of R cannot vanish.
 * Recursive call:
 *    Euclid (B, ith factor of R) over H
 */
                        ba0_pull_stack ();
                        ba0_move_to_tail_table 
					((ba0_table)tabG, (ba0_table)tabG, h);
                        ba0_move_to_tail_table 
					((ba0_table)tabP, (ba0_table)tabP, q);
                        bad_Euclid (tabP, tabG, B, &prod.tab [i].factor, v);
                        ba0_push_another_stack ();
                    }
                }
	    }
	}
	ba0_pull_stack ();
	ba0_restore (&M);
    }
    if (delta != tabG->size - tabP->size)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
}

