#include "bad_attchain.h"

/*
 * The integer fields {\tt structural_property} and {\tt desired_property}
 * have the $i$--th bit set to one if and only if the embedding chain
 * holds the $i$--th property of the corresponding enum type.
 * The field {\tt desired_property} gives properties that one wants
 * to achieve: coherence, squarefreeness, autoreduction, primitivity
 * and strong normalization.
 * 
 * The field {\tt ordering} gives the ordering w.r.t. the chain is defined.
 */

/*
 * Sets the field {\tt ordering} to the current ordering.
 * Clears the fields {\tt structural_property} and {\tt desired_property}
 * of~$A$.
 */

BAD_DLL void bad_init_attchain (bad_attchain A)
{
    bad_reset_attchain (A);
}

/*
 * Sets the field {\tt ordering} to the current ordering.
 * Clears the fields {\tt structural_property} and {\tt desired_property}
 * of~$A$.
 */

BAD_DLL void bad_reset_attchain (bad_attchain A)
{
    A->ordering = bav_R_Iordering ();
    A->structural_property = 0;
    A->desired_property = 0;
}

BAD_DLL bool bad_equal_attchain (bad_attchain A, bad_attchain B)
{
    return A->ordering == B->ordering &&
		A->structural_property == B->structural_property &&
		A->desired_property == B->desired_property;
}

/*
 * Sets~$A$ to~$B$.
 */

BAD_DLL void bad_set_attchain (bad_attchain A, bad_attchain B)
{
    *A = *B;
}

BAD_DLL void bad_intersect_attchain (bad_attchain A, bad_attchain B)
{
    if (A->ordering != B->ordering)
	BA0_RAISE_EXCEPTION (BAD_ERRIAC);
    if (bad_has_structural_property_attchain 
			(A, bad_differential_structural_property) != 
		bad_has_structural_property_attchain 
			(B, bad_differential_structural_property))
	BA0_RAISE_EXCEPTION (BAD_ERRIAC);
    bad_clear_structural_property_attchain (A, bad_prime_structural_property);
    A->desired_property &= B->desired_property;
}

/*
 * The list~$P$ contains a list of identifiers for structural or desired
 * properties. Sets~$A$ to the corresponding properties.
 * Exception BAD_ERRNAC is raised in the case of a mistake.
 */

BAD_DLL void bad_set_properties_attchain (bad_attchain A, ba0_tableof_string P)
{   enum bad_structural_property_attchain structural_prop;
    enum bad_desired_property_attchain desired_prop;
    ba0_int_p i;

    A->structural_property = 0;
    A->desired_property = 0;

    if (P != (ba0_tableof_string)0)
    {	for (i = 0; i < P->size; i++)
        {   if (bad_is_a_structural_property_attchain 
					(P->tab [i], &structural_prop))
                bad_set_structural_property_attchain (A, structural_prop);
            else if (bad_is_a_desired_property_attchain 
					(P->tab [i], &desired_prop))
                bad_set_desired_property_attchain (A, desired_prop);
            else
                BA0_RAISE_EXCEPTION (BAD_ERRNAC);
	}
    }
}

/*
 * no derivation => not differential
 * not differential or ODE => not coherent
 * differential + PDE => coherent
 * differential => squarefree
 * prime => squarefree
 * normalized => primitive + autoreduced
 */

BAD_DLL void bad_set_automatic_properties_attchain (bad_attchain A)
{
    if (bav_R.ders.size == 0)
	bad_clear_structural_property_attchain
				(A, bad_differential_structural_property);
    if (bad_has_structural_property_attchain 
				(A, bad_differential_structural_property))
    {	bad_set_desired_property_attchain (A, bad_squarefree_desired_property);
	if (bav_R.ders.size >= 2)
	    bad_set_desired_property_attchain
					(A, bad_coherence_desired_property);
	else
	    bad_clear_desired_property_attchain
					(A, bad_coherence_desired_property);
    } else
	bad_clear_desired_property_attchain 
					(A, bad_coherence_desired_property);
					
    if (bad_has_structural_property_attchain
					(A, bad_prime_structural_property))
	bad_set_desired_property_attchain (A, bad_squarefree_desired_property);
    if (bad_has_desired_property_attchain (A, bad_normalized_desired_property))
    {	bad_set_desired_property_attchain 
				(A, bad_autoreduced_desired_property);
	bad_set_desired_property_attchain (A, bad_primitive_desired_property);
    }
}

/*
 * The list~$P$ contains a list of identifiers for structural properties. 
 * Sets~$A$ to the corresponding properties.
 * Exception BAD_ERRNAC is raised in the case of a mistake.
 */

BAD_DLL void bad_set_structural_properties_attchain 
				(bad_attchain A, ba0_tableof_string P)
{   enum bad_structural_property_attchain structural_prop;
    ba0_int_p i;

    A->structural_property = 0;

    for (i = 0; i < P->size; i++)
    {	if (bad_is_a_structural_property_attchain 
						(P->tab [i], &structural_prop))
	    bad_set_structural_property_attchain (A, structural_prop);
	else
	    BA0_RAISE_EXCEPTION (BAD_ERRNAC);
    }
}

/*
 * The list~$P$ contains a list of identifiers for desired properties. 
 * Sets~$A$ to the corresponding properties.
 * Exception BAD_ERRNAC is raised in the case of a mistake.
 */

BAD_DLL void bad_set_desired_properties_attchain (bad_attchain A, ba0_tableof_string P)
{   enum bad_desired_property_attchain desired_prop;
    ba0_int_p i;

    A->desired_property = 0;

    for (i = 0; i < P->size; i++)
    {	if (bad_is_a_desired_property_attchain (P->tab [i], &desired_prop))
	    bad_set_desired_property_attchain (A, desired_prop);
	else
	    BA0_RAISE_EXCEPTION (BAD_ERRNAC);
    }
}

/*
 * Sets the structural property~$p$ of~$A$.
 */

BAD_DLL void bad_set_structural_property_attchain 
                       (bad_attchain A, enum bad_structural_property_attchain p)
{
    A->structural_property |= (ba0_int_p)(1 << (int)p);
}

/*
 * Sets the desired property~$p$ of~$A$.
 */

BAD_DLL void bad_set_desired_property_attchain
		(bad_attchain A, enum bad_desired_property_attchain p)
{
    A->desired_property |= (ba0_int_p)(1 << (int)p);
}

/*
 * Clears the structural property~$p$ of~$A$.
 */

BAD_DLL void bad_clear_structural_property_attchain
		(bad_attchain A, enum bad_structural_property_attchain p)
{
    A->structural_property &= - 1 - (1 << (int)p);
}

/*
 * Clears the desired property~$p$ of~$A$.
 */

BAD_DLL void bad_clear_desired_property_attchain
			(bad_attchain A, enum bad_desired_property_attchain p)
{
    A->desired_property &= - 1 - (1 << (int)p);
}

/*
 * Readonly static structures
 */

static struct {
	char* ident;
	enum bad_structural_property_attchain prop;
    } structural_properties [] =  
	{ { "differential", bad_differential_structural_property },
	  { "prime", bad_prime_structural_property }
	};

static struct {
	char* ident;
	enum bad_desired_property_attchain prop;
    } desired_properties [] =  
	{ { "autoreduced", bad_autoreduced_desired_property },
	  { "primitive", bad_primitive_desired_property },
	  { "squarefree", bad_squarefree_desired_property },
	  { "coherent", bad_coherence_desired_property },
	  { "normalized", bad_normalized_desired_property }
	};

/*
 * Returns true if $s$ is a structural property.
 * Legal chains for~$s$ are ``differential'' and ``prime''.
 * If so and $p$ is non zero then $\ast p$ receives the corresponding
 * element of the following type.
 */

BAD_DLL bool bad_is_a_structural_property_attchain 
			(char* s, enum bad_structural_property_attchain* p)
{   ba0_int_p i, n;

    n = sizeof (structural_properties) / sizeof (*structural_properties);
    for (i = 0; i < n; i++)
    {	if (ba0_strcasecmp (s, structural_properties [i].ident) == 0)
	{   if (p != (enum bad_structural_property_attchain*)0)
		*p = structural_properties [i].prop;
	    return true;
	}
    }
    return false;
}

/*
 * Appends to T the list of the structural properties of C.
 * The identifiers are the same as the ones used for parsing.
 */

BAD_DLL void bad_structural_properties_attchain 
					(ba0_tableof_string T, bad_attchain C)
{
    ba0_int_p i, n;

    n = sizeof (structural_properties) / sizeof (*structural_properties);
    ba0_realloc_table ((ba0_table)T, T->size + n);
    for (i = 0; i < n; i++)
    {	if (bad_has_structural_property_attchain 
					(C, structural_properties [i].prop))
	    T->tab [T->size++] = structural_properties [i].ident;
    }
}

/*
 * Returns true if $s$ is a desired property.
 * Legal chains for~$s$ are ``autoreduced'', ``primitive'', ``squarefree'',
 * ``coherent'' and ``normalized''.
 * If so and $p$ is non zero then $\ast p$ receives the corresponding 
 * element of the following type.
 */

BAD_DLL bool bad_is_a_desired_property_attchain 
			(char* s, enum bad_desired_property_attchain* p)
{   ba0_int_p i, n;

    n = sizeof (desired_properties) / sizeof (*desired_properties);
    for (i = 0; i < n; i++)
    {   if (ba0_strcasecmp (s, desired_properties [i].ident) == 0)
	{   if (p != (enum bad_desired_property_attchain*)0)
		*p = desired_properties [i].prop;
	    return true;
	}
    }
    return false;
}

/*
 * Appends to T the list of the desired properties of C.
 * The identifiers are the same as the ones used for parsing.
 */

BAD_DLL void bad_desired_properties_attchain 
					(ba0_tableof_string T, bad_attchain C)
{
    ba0_int_p i, n;

    n = sizeof (desired_properties) / sizeof (*desired_properties);
    ba0_realloc_table ((ba0_table)T, T->size + n);
    for (i = 0; i < n; i++)
    {	if (bad_has_desired_property_attchain 
					(C, desired_properties [i].prop))
	    T->tab [T->size++] = desired_properties [i].ident;
    }
}

/*
 * Returns true if $A$ holds the structural property~$p$.
 */

BAD_DLL bool bad_has_structural_property_attchain
		(bad_attchain A, enum bad_structural_property_attchain p)
{
    return (A->structural_property & (ba0_int_p)(1 << (int)p)) != 0;
}

/*
 * Returns true if $A$ holds the desired property~$p$.
 */

BAD_DLL bool bad_has_desired_property_attchain
			(bad_attchain A, enum bad_desired_property_attchain p)
{
    return (A->desired_property & (ba0_int_p)(1 << (int)p)) != 0;
}

/*
 * Returns true if the ideal defined by~$C$ is prime.
 */

BAD_DLL bool bad_defines_a_prime_ideal_attchain (bad_attchain C)
{
    return bad_has_structural_property_attchain 
					(C, bad_prime_structural_property);
}

/*
 * Returns true if the ideal defined by~$C$ is differential.
 */

BAD_DLL bool bad_defines_a_differential_ideal_attchain (bad_attchain C)
{
    return bad_has_structural_property_attchain 
				(C, bad_differential_structural_property);
}

