#ifndef BAD_BASE_FIELD_H
#define BAD_BASE_FIELD_H 1

#include "bad_common.h"
#include "bad_regchain.h"

BEGIN_C_DECLS

/*
 * Base fields are presented by generators and relations.
 * They may be either differential or non differential fields.
 *
 * All derivatives less than or equal to varmax are in the base field.
 * In the differential case, the derivatives of varmax are also in the 
 * base field.
 *
 * The relations are defined by a regchain, which is assumed to define
 * a prime ideal.
 *
 * Observe that the behaviour of the functions depends on the ranking.
 */

typedef struct bad_base_field {
    bool differential;
    bav_variable varmax;
    struct bad_regchain relations;
} *bad_base_field;

extern BAD_DLL void bad_init_base_field (bad_base_field);
extern BAD_DLL bad_base_field bad_new_base_field (void);
extern BAD_DLL void bad_force_algebraic_base_field (bad_base_field);

extern BAD_DLL void bad_base_field_generators 
				(bav_tableof_variable, bad_base_field);

extern BAD_DLL void bad_move_base_field_generator
				(bad_base_field, bav_variable);

extern BAD_DLL void bad_set_base_field_generators_and_relations
				(bad_base_field, bav_tableof_variable,
				 bad_regchain, bav_tableof_parameter, 
				 bool, bool);

extern BAD_DLL void bad_base_field_implicit_generators
				(bav_tableof_variable, bad_base_field,
				 bav_tableof_variable, bad_regchain);

extern BAD_DLL bool bad_is_a_compatible_base_field 
				(bad_base_field, bad_attchain);

extern BAD_DLL bool bad_member_reduced_nonzero_polynom_base_field 
				(bap_polynom_mpz, bad_base_field);
extern BAD_DLL bool bad_member_nonzero_polynom_base_field 
				(bap_polynom_mpz, bad_base_field);
extern BAD_DLL bool bad_member_polynom_base_field
				(bap_polynom_mpz, bad_base_field);

extern BAD_DLL ba0_scanf_function bad_scanf_base_field;
extern BAD_DLL ba0_printf_function bad_printf_base_field;

END_C_DECLS

#endif /*! BAD_BASE_FIELD_H */
