#include "bad_reduction.h"
#include "bad_regularize.h"
#include "bad_low_power_theorem.h"
#include "bad_global.h"

#define bad_zstring	bad_initialized_global.preparation.zstring

/*
 * SETTINGS
 */

BAD_DLL void bad_set_settings_preparation (char* zstring)
{
    bad_zstring = zstring ? zstring : BAD_ZSTRING;
}

BAD_DLL void bad_get_settings_preparation (char** zstring)
{
    if (zstring) *zstring = bad_zstring;
}

/*
 * PREPARATION TERMS
 */

/*
 * Initialized to the term 1
 */

static void bad_init_preparation_term (bad_preparation_term T)
{
    ba0_init_table ((ba0_table)&T->z);
    ba0_init_table ((ba0_table)&T->theta);
    ba0_init_table ((ba0_table)&T->deg);
}

static bad_preparation_term bad_new_preparation_term (void)
{   bad_preparation_term T;

    T = (bad_preparation_term)ba0_alloc (sizeof (struct bad_preparation_term));
    bad_init_preparation_term (T);
    return T;
}

static void bad_realloc_preparation_term (bad_preparation_term T, ba0_int_p n)
{
    if (T->theta.alloc < n)
    {	ba0_realloc_table ((ba0_table)&T->z, n);
	ba0_realloc2_table 
		((ba0_table)&T->theta, n, (ba0_new_function*)&bav_new_term);
	ba0_realloc_table ((ba0_table)&T->deg, n);
    }
}

static void bad_set_preparation_term_one (bad_preparation_term T)
{
    ba0_reset_table ((ba0_table)&T->z);
    ba0_reset_table ((ba0_table)&T->theta);
    ba0_reset_table ((ba0_table)&T->deg);
}

static void bad_set_preparation_term 
			(bad_preparation_term T, bad_preparation_term U)
{   ba0_int_p i;

    if (T != U)
    {	bad_realloc_preparation_term (T, U->z.size);
	ba0_set_table ((ba0_table)&T->z, (ba0_table)&U->z);
	ba0_reset_table ((ba0_table)&T->theta);
	for (i = 0; i < U->theta.size; i++)
	{   bav_set_term (T->theta.tab [T->theta.size], U->theta.tab [i]);
	    T->theta.size += 1;
	}
	ba0_set_table ((ba0_table)&T->deg, (ba0_table)&U->deg);
    }
}

/*
 * If a term involves a zi corresponding to a base field equation then
 * its degree is +infinity, so that it gets moved to the end of the
 * preparation equation.
 */

static bav_Idegree bad_degree_preparation_term 
				(bad_preparation_term T, bad_regchain K)
{   bav_Idegree d;
    ba0_int_p i, n;

    n = K->decision_system.size;
    d = 0;
    for (i = 0; i < T->deg.size && d != BA0_MAX_INT_P; i++)
    {	if (T->z.tab [i] >= n)
	    d += T->deg.tab [i];
	else
	    d = BA0_MAX_INT_P;
    }
    return d;
}

static void bad_printf_preparation_term (bad_preparation_term T)
{   struct bav_rank rg;
    bav_variable v, fakevar;
    bav_symbol fakesym;
    bool first;
    struct ba0_mark M;
    ba0_int_p i;
    ba0_scanf_function *scanfsym;
    ba0_printf_function *printsym;
    static char buffer [BA0_BUFSIZE];

    ba0_record (&M);

    if (T->z.size == 0)
    {	ba0_printf ("1");
	return;
    }
/*
 * v = any dependent variable
 */
    if (bav_R.deps.size == 0)
	BA0_RAISE_EXCEPTION (BA0_ERRNYP);
    v = bav_R.vars.tab [bav_R.deps.tab [0]];
/*
 * The idea consists in building a fake variable (from v) by replacing
 * the identifier of v by some zi. In order to benefit from the
 * variable printing functions (which may be customized by the user).
 */
    fakesym = ba0_alloc (sizeof (struct bav_symbol));
    memcpy (fakesym, v->root, sizeof (struct bav_symbol));
    fakesym->ident = buffer;
    fakevar = ba0_alloc (sizeof (struct bav_variable));
/*
 * The bav_printf_numbered_symbol makes no sense in this context.
 * If it is the symbol printing function, switch to the default one.
 */
    bav_get_settings_symbol (&scanfsym, &printsym);
    if (printsym == &bav_printf_numbered_symbol)
	bav_set_settings_symbol (scanfsym, &bav_printf_default_symbol);
/*
 * Prints a term
 */ 
    first = true;
    for (i = 0; i < T->z.size; i++)
    {	
/*
 * The identifier becomes zi
 */
	ba0_sprintf (buffer, bad_zstring, T->z.tab [i] + 1);

	if (! first)
            ba0_printf ("*");
	first = false;
/*
 * Recover the derivation operator by differentiating v
 */
	v = bav_order_zero_variable (v);
	v = bav_diff2_variable (v, T->theta.tab [i]);
	memcpy (fakevar, v, sizeof (struct bav_variable));
	fakevar->root = fakesym;
/*
 * Build the rank
 */
	rg.var = fakevar;
	rg.deg = T->deg.tab [i];
/*
 * Print it. 
 */
	ba0_printf ("%rank", &rg);
    }
/*
 * Restore
 */
    bav_set_settings_symbol (scanfsym, printsym);
    ba0_restore (&M);
}

static void bad_eval_preparation_term 
		(bap_polynom_mpz R, bad_preparation_term T, bad_regchain A)
{   struct bap_polynom_mpz S1, S2;
    struct bav_tableof_variable nulles;
    struct ba0_mark M;
    ba0_int_p i;

    ba0_push_another_stack ();
    ba0_record (&M);

    ba0_init_table ((ba0_table)&nulles);

    bap_init_polynom_mpz (&S1);
    bap_init_polynom_mpz (&S2);

    bap_set_polynom_one_mpz (&S1);
    for (i = 0; i < T->z.size; i++)
    {	bap_diff2_polynom_mpz (&S2, A->decision_system.tab [T->z.tab [i]], 
					 T->theta.tab [i], &nulles);
	bap_pow_polynom_mpz (&S2, &S2, T->deg.tab [i]);
	bap_mul_polynom_mpz (&S1, &S1, &S2);
    }
    ba0_pull_stack ();
    bap_set_polynom_mpz (R, &S1);
    ba0_restore (&M);
}

/*
 * Multiplication by (theta z)^d
 */

static void bad_mul_preparation_term_theta_z
		(bad_preparation_term T, bad_preparation_term U,
		 bav_term theta, ba0_int_p z, bav_Idegree d)
{   ba0_int_p i;
    bool found;

    if (T != U)
	bad_set_preparation_term (T, U);

    i = 0;
    found = false;
    while (i < T->theta.size && !found)
    {	if (T->z.tab [i] == z && bav_equal_term (T->theta.tab [i], theta))
	    found = true;
	else
	    i += 1;
    }
    if (found)
	T->deg.tab [i] += d;
    else
    {	bad_realloc_preparation_term (T, T->theta.size + 1);

	T->z.tab [T->z.size] = z;
	bav_set_term (T->theta.tab [T->theta.size], theta);
	T->deg.tab [T->deg.size] = d;

	T->z.size += 1;
	T->theta.size += 1;
	T->deg.size += 1;
    }
}

/*
 * T = U * V
 */

static void bad_mul_preparation_term 
    (bad_preparation_term T, bad_preparation_term U, bad_preparation_term V)
{   struct bad_preparation_term R;
    ba0_int_p i;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);
    bad_init_preparation_term (&R);
    bad_set_preparation_term (&R, U);
    for (i = 0; i < V->z.size; i++)
	bad_mul_preparation_term_theta_z 
		(&R, &R, V->theta.tab [i], V->z.tab [i], V->deg.tab [i]);
    ba0_pull_stack ();
    bad_set_preparation_term (T, &R);
    ba0_restore (&M);
}

/*
 * PREPARATION EQUATIONS
 */

BAD_DLL void bad_init_preparation_equation (bad_preparation_equation E)
{
    bap_init_product_mpz (&E->H);
    ba0_init_table ((ba0_table)&E->coeffs);
    ba0_init_table ((ba0_table)&E->terms);

    E->F = (bap_polynom_mpz)0;
    E->A = (bad_regchain)0;
    E->K = (bad_regchain)0;
}

BAD_DLL bad_preparation_equation bad_new_preparation_equation (void)
{   bad_preparation_equation E;

    E = (bad_preparation_equation)ba0_alloc 
				(sizeof (struct bad_preparation_equation));
    bad_init_preparation_equation (E);
    return E;
}

static void bad_realloc_preparation_equation 
				(bad_preparation_equation P, ba0_int_p n)
{
    ba0_realloc2_table ((ba0_table)&P->coeffs, n,
				(ba0_new_function*)&bap_new_polynom_mpz);
    ba0_realloc2_table ((ba0_table)&P->terms, n,
				(ba0_new_function*)&bad_new_preparation_term);
}

BAD_DLL void bad_printf_preparation_equation (void* A)
{   bad_preparation_equation E = (bad_preparation_equation)A;
    ba0_int_p i, n;
    bool first;

    if (E->F == (bap_polynom_mpz)0)
	return;

    n = E->K->decision_system.size;

    if (E->denom)
	ba0_printf ("%Pz * (1/%z)*(%Az) = ", &E->H, E->denom, E->F);
    else
        ba0_printf ("%Pz * (%Az) = ", &E->H, E->F);
    if (E->coeffs.size == 0)
	ba0_printf ("0");
    else
    {	first = true;
	for (i = 0; i < E->coeffs.size; i++)
	{   if (! first)
		ba0_printf (" + ");
	    first = false;
	    if (E->denom)
		ba0_printf ("(1/%z)*(%Az)*", E->denom, E->coeffs.tab [i]);
	    else
	        ba0_printf ("(%Az)*", E->coeffs.tab [i]);
	    bad_printf_preparation_term (E->terms.tab [i]);
	}
    }
}

BAD_DLL void bad_check_preparation_equation (bad_preparation_equation E)
{   struct bap_polynom_mpz lhs, rhs, R;
    struct ba0_mark M;
    ba0_int_p i;

    ba0_record (&M);

    bap_init_polynom_mpz (&lhs);
    bap_init_polynom_mpz (&rhs);

    bap_expand_product_mpz (&lhs, &E->H);
    bap_mul_polynom_mpz (&lhs, &lhs, E->F);

    bap_init_polynom_mpz (&R);
    for (i = 0; i < E->coeffs.size; i++)
    {	bad_eval_preparation_term (&R, E->terms.tab [i], E->A);
	bap_mul_polynom_mpz (&R, &R, E->coeffs.tab [i]);
	bap_add_polynom_mpz (&rhs, &rhs, &R);
    }

    if (! bap_equal_polynom_mpz (&lhs, &rhs))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    ba0_restore (&M);
}

static void bad_set_preparation_equation_polynom2
	(bad_preparation_equation E, 
	 bap_polynom_mpz F, ba0_mpz_t denom,
	 bad_regchain A, bad_regchain K);

/*
 * We have
 *
 * (E)    H * F = sum coeffs [i] * terms [i]
 *
 * moreover
 *
 * coeffs [k] is reducible by A [z] at rank rg
 *
 * One wants to reduce coeffs [k]. One computes:
 *
 * Hk * coeffs [k] = Q * theta A [z] + R
 *
 * (Eq)   Hq * Q = sum coeffs_q [l] * terms_q [l]
 *
 * Multiply (E) by Hk * Hq. One gets:
 *
 * Hk * Hq * F = sum {i != k} Hk * Hq * coeffs [i] * terms [i]
 *             + Hq * Hk * coeffs [k] * terms [k]
 *
 *             = sum {i != k} Hk * Hq * coeffs [i] * terms [i]
 *             + Hq * Q * theta A [z] * terms [k]
 *             + Hq * R * terms [k]
 *
 *             = sum {i != k} Hk * Hq * coeffs [i] * terms [i]
 *             + sum coeffs_q [l] * (terms_q [l] * theta A [z] * terms [k])
 *             + Hq * R * terms [k]
 */

static void bad_reduce_preparation_equation 
	(bad_preparation_equation E, ba0_int_p k, ba0_int_p z, bav_rank rg)
{   bap_tableof_polynom_mpz A;
    struct bad_preparation_equation Eq;
    struct bap_polynom_mpz coeffk, B, Q, R, init, Hqexp;
    bap_polynom_mpz Hk;
    struct bad_preparation_term termk;
    struct bav_tableof_variable nulles;
    struct bav_term theta;
    struct bav_rank rg_z;
    bav_Idegree e;
    ba0_int_p i;
    struct ba0_mark M;
/*
ba0_printf ("bad_reduce_preparation_equation\ninput: %preparation_equation, k = %d, z = %d, rg = %rank\n\n", E, k, z, rg);
*/
    A = &E->A->decision_system;

    ba0_push_another_stack ();
    ba0_record (&M);

    bad_init_preparation_equation (&Eq);
    bap_init_polynom_mpz (&coeffk);
    bad_init_preparation_term (&termk);
    bap_init_polynom_mpz (&B);
    bap_init_polynom_mpz (&Q);
    bap_init_polynom_mpz (&R);
    bap_init_polynom_mpz (&Hqexp);
    ba0_init_table ((ba0_table)&nulles);
    bav_init_term (&theta);

    bap_set_polynom_mpz (&coeffk, E->coeffs.tab [k]);
    bad_set_preparation_term (&termk, E->terms.tab [k]);
/*
 * Perform the pseudo division of coeffk by theta A[z]
 */
    rg_z = bap_rank_polynom_mpz (A->tab [z]);
    bav_operator_between_derivatives (&theta, rg->var, rg_z.var);

    bad_reduced_to_zero_derivatives_by_regchain 
					(&nulles, E->K, bad_full_reduction);

    bap_diff2_polynom_mpz (&B, A->tab [z], &theta, &nulles);
    bap_pseudo_division_polynom_mpz (&Q, &R, &e, &coeffk, &B, rg->var);

    if (bap_is_zero_polynom_mpz (&Q))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (bav_is_one_term (&theta))
    {	bap_init_readonly_polynom_mpz (&init);
	bap_initial_polynom_mpz (&init, A->tab [z]);
	Hk = &init;
    } else
    {   Hk = bap_new_polynom_mpz ();
	bap_separant_polynom_mpz (Hk, A->tab [z]);
    }
/*
 * Compute the preparation equation for the quotient Q
 */
    bad_set_preparation_equation_polynom2 
				(&Eq, &Q, (__mpz_struct*)0, E->A, E->K);
    bap_expand_product_mpz (&Hqexp, &Eq.H);
/*
 * Incorporation in (E)
 */
    ba0_pull_stack ();
/*
 * step 1: delete the kth entry of E.
 */
    ba0_delete_table ((ba0_table)&E->coeffs, k);
    ba0_delete_table ((ba0_table)&E->terms, k);
    bad_realloc_preparation_equation (E, E->coeffs.size + Eq.coeffs.size + 1);
/* 
 * step 2: H and the coeffs [i] (i != k) are multiplied by Hk * Hq
 */
    bap_mul_product_mpz (&E->H, &E->H, &Eq.H);
    bap_mul_product_polynom_mpz (&E->H, &E->H, Hk, 1);
    for (i = 0; i < E->coeffs.size; i++)
    {	bap_mul_polynom_mpz (E->coeffs.tab [i], E->coeffs.tab [i], &Hqexp);
	bap_mul_polynom_mpz (E->coeffs.tab [i], E->coeffs.tab [i], Hk);
    }
/*
 * step 3: add (sum coeffs_q [l]* terms_q [l]) * termk * theta z
 */
    for (i = 0; i < Eq.coeffs.size; i++)
    {	bap_set_polynom_mpz (E->coeffs.tab [E->coeffs.size], Eq.coeffs.tab [i]);
	bad_mul_preparation_term 
		(E->terms.tab [E->terms.size], Eq.terms.tab [i], &termk);
	bad_mul_preparation_term_theta_z 
		(E->terms.tab [E->terms.size], E->terms.tab [E->terms.size], 
		 &theta, z, 1);
	E->coeffs.size += 1;
	E->terms.size += 1;
    }
/*
 * step 4: add Hq * R * termk to E provided that R is non zero
 */
    if (! bap_is_zero_polynom_mpz (&R))
    {	bap_mul_polynom_mpz (E->coeffs.tab [E->coeffs.size], &R, &Hqexp);
    	bad_set_preparation_term (E->terms.tab [E->terms.size], &termk);
    	E->coeffs.size += 1;
    	E->terms.size += 1;
    }
/*
ba0_printf ("bad_reduce_preparation_equation\noutput: %preparation_equation, k = %d, z = %d, rg = %rank\n", E, k, z, rg);
*/
    ba0_restore (&M);
}

/*
 * Computes most of a preparation equation for F and stores it in E since
 * the coeffs [i] are not guaranteed to be regular w.r.t. A.
 */

static void bad_set_preparation_equation_polynom2 
	(bad_preparation_equation E, 
	 bap_polynom_mpz F, mpz_t denom, bad_regchain A, bad_regchain K)
{   struct bav_rank rg;
    ba0_int_p i, z;
    bool modified;
/*
 * Initialize (E) to 1 * F = F * 1
 */
    E->F = F;
    E->denom = denom;
    E->A = A;
    E->K = K;

    bap_set_product_one_mpz (&E->H);
    bad_realloc_preparation_equation (E, 1);
    bap_set_polynom_mpz (E->coeffs.tab [0], F);
    bad_set_preparation_term_one (E->terms.tab [0]);
    E->coeffs.size = 1;
    E->terms.size = 1;
/*
 * Reduce
 */
    do
    {	modified = false;
	i = 0; 
	while (i < E->coeffs.size && ! modified)
	{   
/*
 * First perform the reductions by the base field equations
 */
	    if (bad_is_a_reducible_polynom_by_regchain 
				(E->coeffs.tab [i], K, bad_full_reduction,
				 bad_all_derivatives_to_reduce, &rg, &z))
	    {   bad_reduce_preparation_equation (E, i, z, &rg);
		modified = true;
/* 
 * Then perform the reductions by the other equations
 */
	    } else if (bad_is_a_reducible_polynom_by_regchain
				(E->coeffs.tab [i], A, bad_full_reduction,
                                 bad_all_derivatives_to_reduce, &rg, &z))
	    {	bad_reduce_preparation_equation (E, i, z, &rg);
		modified = true;
	    } else
		i += 1;
	}
    } while (modified);
}

/*
 * Assigns to (E) a preparation equation for F and A.
 * The result is sorted by increasing total degree of the terms.
 *
 * The regularity test on the coefficients of the equation may raise
 * an exception (the zero divisor is then recovered in *ddz).
 * 
 * The n lowest equations of A are assumed to be base field equations.
 * The reductions by these equations will not be taken into account in
 * the preparation equations.
 */

BAD_DLL void bad_set_preparation_equation_polynom 
		(bad_preparation_equation E, bap_polynom_mpz F, 
		 ba0_mpz_t denom, bad_regchain A, ba0_int_p n, 
		 bap_polynom_mpz* ddz)
{   bad_regchain K;
    ba0_int_p i, j;
    bav_Idegree di, dj;
    bool swap;

    if (n > A->decision_system.size)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    K = bad_new_regchain ();
    bad_set_regchain (K, A);
    K->decision_system.size = n;
    
    bad_set_preparation_equation_polynom2 (E, F, denom, A, K);
/*
 * Special case
 */
    if (bap_is_zero_polynom_mpz (F))
    {	E->coeffs.size = 0;
	E->terms.size = 0;
	return;
    }
/*
 * Terms are sorted by increasing degree
 */
    do
    {	swap = false;
	for (i = 0; i < E->terms.size - 1; i++)
	{   di = bad_degree_preparation_term (E->terms.tab [i], K);
	    j = i+1;
	    dj = bad_degree_preparation_term (E->terms.tab [j], K);
	    if (dj < di)
	    {	BA0_SWAP (bad_preparation_term, 
					E->terms.tab [i], E->terms.tab [j]);
		BA0_SWAP (bap_polynom_mpz, 
					E->coeffs.tab [i], E->coeffs.tab [j]);
		swap = true;
	    }
	}
    } while (swap);
/*
 * Regularity testing. May raise an exception.
 */
    for (i = 0; i < E->coeffs.size; i++)
	bad_check_regularity_polynom_mod_regchain (E->coeffs.tab [i], A, ddz);
}

/*
 * Computes q = the minimum degree of the terms
 *          l = the number of terms of degree q
 * The l terms of degree q stand at the beginning of E->terms.
 */

BAD_DLL void bad_preparation_congruence 
		(ba0_int_p* l, bav_Idegree* q, bad_preparation_equation E)
{   bav_Idegree d;
    ba0_int_p i;

    if (E->terms.size == 0)
    {	*l = 0;
	*q = 0;
    } else
    {	d = bad_degree_preparation_term (E->terms.tab [0], E->K);
	if (d == BA0_MAX_INT_P)
	{   *l = 0;
	    *q = 0;
	} else
	{   i = 1;
	    while (i < E->terms.size && 
		bad_degree_preparation_term (E->terms.tab [i], E->K) == d)
	        i += 1;
	    *q = d;
	    *l = i;
	}
    }
}

/*
 * Kolchin IV, 15, Theorem 6
 * Low Power Theorem
 */

BAD_DLL bool bad_low_power_theorem_condition_to_be_a_component 
				(bad_preparation_equation E)
{   bad_preparation_term T;
    bav_Idegree q;
    ba0_int_p l;

    bad_preparation_congruence (&l, &q, E);
    if (l != 1 || q == 0)
	return false;
    T = E->terms.tab [0];
    if (T->z.size != 1 || ! bav_is_one_term (T->theta.tab [0]))
	return false;
    return true;
}

/*
 * T is assumed to be produced by bad_Rosenfeld_Groebner over a single
 * polynomial F, with coefficients in the base field K.
 *
 * Removes from T the redundant components of the differential ideal
 * generated by F. The result is stored in U. The resulting intersection
 * may involve components, not present in T. It may even be larger than T.
 */

BAD_DLL void bad_low_power_theorem_simplify_intersectof_regchain
	(bad_intersectof_regchain S0,
	 bad_intersectof_regchain T0, bad_base_field K)
{   struct bad_preparation_equation E;
    struct bad_intersectof_regchain S, T, aux;
    bad_regchain general_component, A;
    bap_polynom_mpz F, ddz;
    ba0_mpz_t denom;
    struct ba0_exception_code code;
    struct ba0_mark M;
    volatile ba0_int_p i, n;

    if (T0->inter.size == 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    n = K->relations.decision_system.size;
    general_component = T0->inter.tab [0];
    if (general_component->decision_system.size != n+1)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    F = general_component->decision_system.tab [n];

    ba0_push_another_stack ();
    ba0_record (&M);

    bad_init_intersectof_regchain (&T);
    bad_set_intersectof_regchain (&T, T0);

    bad_init_intersectof_regchain (&S);
    bad_set_intersectof_regchain_regchain (&S, general_component);
    bad_init_intersectof_regchain (&aux);

    ba0_mpz_init_set_ui (denom, 1);
    for (i = 1; i < T.inter.size; i++)
    {	A = T.inter.tab [i];
	if (A->decision_system.size == n+1)
	{   
	    BA0_PUSH_EXCEPTION (code);
	    if (ba0_exception_is_set (code))
	    {	bad_init_preparation_equation (&E);
		bad_set_preparation_equation_polynom (&E, F, denom, A, n, &ddz);
		ba0_pull_exception (code);

		if (bad_low_power_theorem_condition_to_be_a_component (&E))
		    bad_append_intersectof_regchain_regchain (&S, A);
	    } else
	    {	if (ba0_mesgerr != BAD_EXRDDZ)
		    BA0_RAISE_EXCEPTION (ba0_mesgerr);
		bad_set_intersectof_regchain_regchain (&aux, A);
		bad_handle_splitting_exceptions_regchain 
				(&aux, ddz, (bad_regchain)0, BAD_EXRDDZ);
		bad_append_intersectof_regchain (&T, &aux);
	    }
	}
    }

    ba0_pull_stack ();
    bad_set_intersectof_regchain (S0, &S);
    ba0_restore (&M);
}

