#ifndef BAD_LOW_POWER_THEOREM_H
#define BAD_LOW_POWER_THEOREM_H 1

#include "bad_common.h"
#include "bad_regchain.h"
#include "bad_intersectof_regchain.h"
#include "bad_base_field.h"

BEGIN_C_DECLS

#define BAD_ZSTRING	"z%d"

extern BAD_DLL void bad_set_settings_preparation (char*);
extern BAD_DLL void bad_get_settings_preparation (char**);

/*
 * Substructure of bad_preparation_equation
 *
 * Denote A = A1, ..., Ar a differential regular chain
 *
 * Introduce r differential indeterminates z1, ..., zr.
 * A bad_preparation_term represents a term on the derivatives of the zi.
 * All tables have the same size. For each index i
 *
 * z[i]     = an index in [1, ..., r] (actually in 0 ... r-1)
 * theta[i] = a derivation operator (power product of derivations)
 * deg[i]   = a degree
 */

typedef struct bad_preparation_term {
    struct ba0_tableof_int_p z;
    struct bav_tableof_term theta;
    struct bav_tableof_Idegree deg;
} *bad_preparation_term;

typedef struct bad_tableof_preparation_term {
    ba0_int_p alloc;
    ba0_int_p size;
    bad_preparation_term* tab;
} *bad_tableof_preparation_term;

/*
 * Represents a preparation equation (Kolchin, IV, 13).
 *
 * Given F and A = A1, ..., Ar, compute:
 *
 * H*F = sum coeffs [i] * terms [i] mod (zi = Ai)
 *
 * The coeffs [i] must be reduced and regular w.r.t. A.
 *
 * K = K1, ..., Ks provides the base field equations.
 * We have K1 = A1, ..., Ks = As.
 */

typedef struct bad_preparation_equation {
    struct bap_product_mpz H;
    struct bap_tableof_polynom_mpz coeffs;
    struct bad_tableof_preparation_term terms;
    bap_polynom_mpz F;
    __mpz_struct* denom;
    bad_regchain A;
    bad_regchain K;
} *bad_preparation_equation;

extern BAD_DLL void bad_init_preparation_equation (bad_preparation_equation);
extern BAD_DLL bad_preparation_equation bad_new_preparation_equation (void);
extern BAD_DLL void bad_set_preparation_equation_polynom 
			(bad_preparation_equation, bap_polynom_mpz, ba0_mpz_t,
			 bad_regchain, ba0_int_p, bap_polynom_mpz*);

extern BAD_DLL void bad_check_preparation_equation (bad_preparation_equation);
extern BAD_DLL ba0_printf_function bad_printf_preparation_equation;

extern BAD_DLL void bad_preparation_congruence 
			(ba0_int_p*, bav_Idegree*, bad_preparation_equation);

extern BAD_DLL bool bad_low_power_theorem_condition_to_be_a_component 
			(bad_preparation_equation E);

extern BAD_DLL void bad_low_power_theorem_simplify_intersectof_regchain
			(bad_intersectof_regchain, bad_intersectof_regchain,
			 bad_base_field);

END_C_DECLS

#endif /* !BAD_LOW_POWER_THEOREM_H */
