#include "bad_reduced_form.h"
#include "bad_regularize.h"

/*
 * If F = 0 mod C then 0.
 * If NF (F, C) is a ratfrac whose denominator has rank less than v,
 *    then NF (F, C). 
 * Else
 *    RF (initial) * RF (rank) + RF (tail)
 *
 * Observe v can be zero.
 * The ddz parameter is dummy since the normal form of a polynomial
 *    cannot fail.
 */

static void bad_reduced_form_polynom_mod_regchain2 
	(bap_ratfrac_mpz R, bap_polynom_mpz F, bav_variable v, 
	 bad_regchain C, bool stop, bap_polynom_mpz* ddz)
{   struct bap_ratfrac_mpz RNF, Rinit, Rrg, Rred;
    struct bap_polynom_mpz Frg, init, red;
    struct bav_rank rgv, rgF, rgNF;
    struct ba0_mark M;

    rgv.var = v;
    rgv.deg = 1;

    rgF = bap_rank_polynom_mpz (F);

    if (bap_is_numeric_polynom_mpz (F))
    {	bap_set_ratfrac_polynom_mpz (R, F);
	return;
    }

    ba0_push_another_stack ();
    ba0_record (&M);

    bap_init_ratfrac_mpz (&RNF);
    bad_normal_form_polynom_mod_regchain (&RNF, F, C, ddz);
    rgNF = bap_rank_polynom_mpz (&RNF.denom);
    if ((v == BAV_NOT_A_VARIABLE && bap_is_numeric_polynom_mpz (&RNF.denom)) ||
		    (v != BAV_NOT_A_VARIABLE && bav_lt_rank (&rgNF, &rgv)))
    {   ba0_pull_stack ();
        bap_set_ratfrac_mpz (R, &RNF);
    } else if (stop)
/*
 * stop = true iff we are processing RF (rank).
 * To avoid infinite recursion.
 */
    {   ba0_pull_stack ();
        bap_set_ratfrac_polynom_mpz (R, F);
    } else
    {   bap_init_readonly_polynom_mpz (&init);
        bap_init_readonly_polynom_mpz (&red);
        bap_initial_and_reductum_polynom_mpz (&init, &red, F);
        bap_init_ratfrac_mpz (&Rinit);
        bad_reduced_form_polynom_mod_regchain2 
					(&Rinit, &init, v, C, false, ddz);
        if (! bap_is_zero_ratfrac_mpz (&Rinit))
        {   bap_init_polynom_mpz (&Frg);
	    bap_set_polynom_variable_mpz (&Frg, rgF.var, rgF.deg);
    	    bap_init_ratfrac_mpz (&Rrg);
            bad_reduced_form_polynom_mod_regchain2 
					(&Rrg, &Frg, v, C, true, ddz);
            bap_mul_ratfrac_mpz (&Rinit, &Rinit, &Rrg);
        }
	bap_init_ratfrac_mpz (&Rred);
        bad_reduced_form_polynom_mod_regchain2 
					(&Rred, &red, v, C, false, ddz);

        ba0_pull_stack ();
        bap_add_ratfrac_mpz (R, &Rinit, &Rred);
    }

    ba0_restore (&M);
}

BAD_DLL void bad_reduced_form_polynom_mod_regchain 
    (bap_ratfrac_mpz R, bap_polynom_mpz F, bav_variable v, 
     bad_regchain C)
{
    bad_reduced_form_polynom_mod_regchain2 
				(R, F, v, C, false, (bap_polynom_mpz*)0);
    bap_reduce_ratfrac_mpz (R, R);
}

BAD_DLL void bad_reduced_form_polynom_mod_intersectof_regchain
    (bap_tableof_ratfrac_mpz tabR, bap_polynom_mpz F, bav_variable v,
     bad_intersectof_regchain tabC)
{   ba0_int_p i;

    ba0_reset_table ((ba0_table)tabR);
    ba0_realloc2_table ((ba0_table)tabR, tabC->inter.size,
			(ba0_new_function*)&bap_new_ratfrac_mpz);
    for (i = 0; i < tabC->inter.size; i++)
    {	bad_reduced_form_polynom_mod_regchain2 
				(tabR->tab [i], F, v, tabC->inter.tab [i], 
				 false, (bap_polynom_mpz*)0);
	bap_reduce_ratfrac_mpz (tabR->tab [i], tabR->tab [i]);
	tabR->size = i+1;
    }
}

