#include "bad_regchain.h"
#include "bad_regularize.h"
#include "bad_intersectof_regchain.h"
#include "bad_reduction.h"
#include "bad_global.h"
#include "bad_nonzero.h"

/****************************************************************************
 CONSTRUCTORS AND RELATED FUNCTIONS
 ****************************************************************************/

BAD_DLL void bad_init_regchain (bad_regchain C)
{
    bad_init_attchain (&C->attrib);
    ba0_init_table ((ba0_table)&C->decision_system);
/*
    ba0_init_table ((ba0_table)&C->simplification_system);
*/
}

BAD_DLL void bad_reset_regchain (bad_regchain C)
{
    bad_reset_attchain (&C->attrib);
    ba0_reset_table ((ba0_table)&C->decision_system);
/*
    ba0_reset_table ((ba0_table)&C->simplification_system);
*/
}

BAD_DLL bad_regchain bad_new_regchain (void)
{   bad_regchain C;

    C = (bad_regchain)ba0_alloc (sizeof(struct bad_regchain));
    bad_init_regchain (C);
    return C;
}

BAD_DLL void bad_realloc_regchain (bad_regchain C, ba0_int_p n)
{
    ba0_realloc2_table 
	((ba0_table)&C->decision_system, n, 
				(ba0_new_function*)&bap_new_polynom_mpz);
/*
    ba0_realloc2_table
	((ba0_table)&C->simplification_system, m,
				(ba0_new_function*)&bap_new_polynom_mpz);
*/
}

BAD_DLL void bad_set_regchain (bad_regchain A, bad_regchain B)
{   ba0_int_p i;

    if (A == B)
	return;

    bav_R_push_ordering (B->attrib.ordering);

    bad_realloc_regchain (A, B->decision_system.size);
    for (i = 0; i < B->decision_system.size; i++)
	bap_set_polynom_mpz (A->decision_system.tab [i],
					B->decision_system.tab [i]);
    A->decision_system.size = B->decision_system.size;
/*
    for (i = 0; i < B->simplification_system.size; i++)
	bap_set_polynom_mpz (A->simplification_system.tab [i],
					B->simplification_system.tab [i]);
    A->simplification_system.size = B->simplification_system.size;
*/
    bad_set_attchain (&A->attrib, &B->attrib);

    bav_R_pull_ordering ();
}

/*
 * Given to qsort, this function sorts a table of polynomials, in 
 * increasing order.
 */

static int bad_comp_for_increasing (const void* PP, const void* QQ)
{   bap_polynom_mpz P = *(bap_polynom_mpz*)PP;
    bap_polynom_mpz Q = *(bap_polynom_mpz*)QQ;

    if (bap_lt_rank_polynom_mpz (P, Q))
	return -1;
    else if (bap_equal_rank_polynom_mpz (P, Q))
	return 0;
    else
	return 1;
}

/*
 * The zero ideal is not prime
 */

BAD_DLL void bad_fast_primality_test_regchain (bad_regchain C)
{
    if (C->decision_system.size > 0 && bad_is_explicit_regchain (C))
	bad_set_structural_property_attchain 
			(&C->attrib, bad_prime_structural_property);
    else if (C->decision_system.size == 0)
	bad_clear_structural_property_attchain
			(&C->attrib, bad_prime_structural_property);
}

/*
 * C must be [non empty,] free of independent polynomials, triangular
 * differential => differentially triangular
 */

static void bad_check_consistency_regchain (bad_regchain C)
{   bav_variable u, v;
    ba0_int_p i, j;
/*
    if (C->decision_system.size == 0)
	BA0_RAISE_EXCEPTION (BAD_ERRNRC);
*/
    for (i = 0; i < C->decision_system.size; i++)
	if (bap_is_independent_polynom_mpz (C->decision_system.tab [i]))
	    BA0_RAISE_EXCEPTION (BAD_ERRNRC);

    for (i = 0; i < C->decision_system.size; i++)
    {	u = bap_leader_polynom_mpz (C->decision_system.tab [i]);
	for (j = i + 1; j < C->decision_system.size; j++)
	{   v = bap_leader_polynom_mpz (C->decision_system.tab [j]);
	    if (u == v)
		BA0_RAISE_EXCEPTION (BAD_ERRNRC);
	    if (bad_defines_a_differential_ideal_regchain (C) &&
			(bav_is_derivative (u, v) || bav_is_derivative (v, u)))
		BA0_RAISE_EXCEPTION (BAD_ERRNRC);
	}
    }
}

BAD_DLL void bad_remove_zero_derivatives_of_tableof_parameter_from_regchain 
		(bad_regchain C, bad_regchain D, bav_tableof_parameter P)
{   struct bav_tableof_variable nulles;
    bav_variable v;
    ba0_int_p i;
    struct ba0_mark M;

    if (C != D)
	bad_set_regchain (C, D);

    ba0_record (&M);
    ba0_init_table ((ba0_table)&nulles);
    bav_zero_derivatives_of_tableof_parameter (&nulles, P);

    bav_R_mark_variables (false);
    for (i = 0; i < nulles.size; i++)
        nulles.tab [i]->hack = true;
    for (i = C->decision_system.size - 1; i >= 0; i--)
    {   v = bap_leader_polynom_mpz (C->decision_system.tab [i]);
        if (v->hack)
            ba0_delete_table ((ba0_table)&C->decision_system, i);
    }
    ba0_restore (&M);
}

static void bad_set_and_extend_regchain_tableof_polynom_mpz_all_parameters
   (bad_regchain C, bap_tableof_polynom_mpz T, bav_tableof_parameter P,
    ba0_tableof_string properties, bool pretend)
{   
    ba0_int_p i;
/*
 * Set. T may be C->decision_system. T may also be zero.
 */
    if (&C->decision_system != T)
    {	if (T == (bap_tableof_polynom_mpz)0)
	   C->decision_system.size = 0;
	else
	{   bad_realloc_regchain (C, T->size);
	    for (i = 0; i < T->size; i++)
	        bap_set_polynom_mpz (C->decision_system.tab [i], T->tab [i]);
	    C->decision_system.size = T->size;
	}
    }
/*
 * Extend. Then free Q.
 */
    if (P != (bav_tableof_parameter)0)
	bap_zero_derivatives_of_tableof_parameter_mpz (&C->decision_system, P);
/*
 * Sort decision_system before checking
 */
    qsort (C->decision_system.tab, C->decision_system.size,
		sizeof (bap_polynom_mpz), &bad_comp_for_increasing);
/*
 * Set the properties (+ automatic)
 */
    bad_set_properties_attchain (&C->attrib, properties);
    bad_set_automatic_properties_attchain (&C->attrib);

    if (! pretend)
    {	bad_check_consistency_regchain (C);
	bad_quench_regchain (C, C, 0, (bap_polynom_mpz*)0);
    }
}

static void 
	bad_set_and_extend_regchain_tableof_polynom_mpz_necessary_parameters
   (bad_regchain C, bap_tableof_polynom_mpz T, bav_tableof_parameter P,
    ba0_tableof_string properties, bool pretend)
{   
    struct bav_tableof_parameter Q;
    struct ba0_mark M;
    ba0_int_p i;
/*
 * Compute the parameters of P, actually involved in T
 */
    ba0_push_another_stack ();
    ba0_record (&M);
    ba0_init_table ((ba0_table)&Q);
    if (P != (bav_tableof_parameter)0)
    {	if (T != (bap_tableof_polynom_mpz)0)
        {   for (i = 0; i < T->size; i++)
	        bap_involved_parameters_polynom_mpz (&Q, T->tab [i], P);
	}
    }
    ba0_pull_stack ();
    bad_set_and_extend_regchain_tableof_polynom_mpz_all_parameters 
					(C, T, &Q, properties, pretend);
    ba0_restore (&M);
}

BAD_DLL void bad_set_and_extend_regchain_tableof_polynom_mpz 
   (bad_regchain C, bap_tableof_polynom_mpz T, bav_tableof_parameter P,
    ba0_tableof_string properties, bool allpars, bool pretend)
{
    if (allpars)
	bad_set_and_extend_regchain_tableof_polynom_mpz_all_parameters
						(C, T, P, properties, pretend);
    else
	bad_set_and_extend_regchain_tableof_polynom_mpz_necessary_parameters
						(C, T, P, properties, pretend);
}

BAD_DLL void bad_set_and_extend_regchain_tableof_ratfrac_mpz
    (bad_regchain C, bap_tableof_ratfrac_mpz T, bav_tableof_parameter P,
     ba0_tableof_string properties, bool allpars, bool pretend)
{   struct bap_ratfrac_mpz R;
    struct ba0_mark M;
    ba0_int_p i;

    ba0_push_another_stack ();
    ba0_record (&M);
    bap_init_ratfrac_mpz (&R);
    ba0_pull_stack ();

    bad_realloc_regchain (C, T->size);
    for (i = 0; i < T->size; i++)
    {	ba0_push_another_stack ();
	bap_reduce_ratfrac_mpz (&R, T->tab [i]);
	ba0_pull_stack ();
	bap_numer_ratfrac_mpz (C->decision_system.tab [i], &R);
    }
    C->decision_system.size = T->size;

    bad_set_and_extend_regchain_tableof_polynom_mpz 
		(C, &C->decision_system, P, properties, allpars, pretend);
}

BAD_DLL void bad_set_properties_regchain (bad_regchain C, ba0_tableof_string T)
{
    bad_set_properties_attchain (&C->attrib, T);
}

BAD_DLL void bad_set_structural_properties_regchain 
				(bad_regchain C, ba0_tableof_string T)
{
    bad_set_structural_properties_attchain (&C->attrib, T);
}

BAD_DLL void bad_set_desired_properties_regchain 
                                (bad_regchain C, ba0_tableof_string T)
{
    bad_set_desired_properties_attchain (&C->attrib, T);
}

BAD_DLL bool bad_has_desired_property_regchain
		(bad_regchain C, enum bad_desired_property_attchain P)
{
    return bad_has_desired_property_attchain (&C->attrib, P);
}

BAD_DLL bool bad_has_structural_property_regchain
		(bad_regchain C, enum bad_structural_property_attchain P)
{
    return bad_has_structural_property_attchain (&C->attrib, P);
}

BAD_DLL bool bad_defines_a_differential_ideal_regchain (bad_regchain C)
{
    return bad_defines_a_differential_ideal_attchain (&C->attrib);
}

BAD_DLL bool bad_defines_a_prime_ideal_regchain (bad_regchain C)
{
    return bad_defines_a_prime_ideal_attchain (&C->attrib);
}

BAD_DLL void bad_sort_regchain (bad_regchain A, bad_regchain B)
{   ba0_int_p i;

    if (A != B)
    {	bad_reset_regchain (A);
	bad_realloc_regchain (A, B->decision_system.size);
	for (i = 0; i < B->decision_system.size; i++)
	    bap_sort_polynom_mpz 
		(A->decision_system.tab [i], B->decision_system.tab [i]);
	A->decision_system.size = B->decision_system.size;
/*
	for (i = 0; i < B->simplification_system.size; i++)
	    bap_sort_polynom_mpz 
		(A->simplification_system.tab [i], 
					B->simplification_system.tab [i]);
	A->simplification_system.size = B->simplification_system.size;
*/
	A->attrib.ordering = bav_R_Iordering ();
	A->attrib.structural_property = B->attrib.structural_property;
	A->attrib.desired_property = B->attrib.desired_property;
    } else
    {	for (i = 0; i < A->decision_system.size; i++)
	    bap_sort_polynom_mpz 
		(A->decision_system.tab [i], A->decision_system.tab [i]);
/*
        for (i = 0; i < A->simplification_system.size; i++)
	    bap_sort_polynom_mpz 
		(A->simplification_system.tab [i], 
					A->simplification_system.tab [i]);
*/
	A->attrib.ordering = bav_R_Iordering ();
    }
}

BAD_DLL void* bad_scanf_regchain (void* A)
{   bad_regchain C;
    bap_tableof_ratfrac_mpz ratfrac;
    ba0_tableof_string properties;
    struct ba0_mark M;

    if (A == (void*)0)
	C = bad_new_regchain ();
    else
    {	C = (bad_regchain)A;
	bad_reset_attchain (&C->attrib);
    }

    ba0_push_another_stack ();
    ba0_record (&M);
    ratfrac = (bap_tableof_ratfrac_mpz)ba0_new_table ();
    properties = (ba0_tableof_string)ba0_new_table ();
    ba0_scanf ("regchain (%t[%Qz], %t[%s])", ratfrac, properties);
    ba0_pull_stack ();
    bad_set_and_extend_regchain_tableof_ratfrac_mpz 
	    (C, ratfrac, (bav_tableof_parameter)0, properties, true, false);
    ba0_restore (&M);
    return C;
}

BAD_DLL void* bad_scanf_pretend_regchain (void* A)
{   bad_regchain C;
    bap_tableof_ratfrac_mpz ratfrac;
    ba0_tableof_string properties;
    struct ba0_mark M;

    if (A == (void*)0)
	C = bad_new_regchain ();
    else
    {	C = (bad_regchain)A;
	bad_reset_attchain (&C->attrib);
    }

    ba0_push_another_stack ();
    ba0_record (&M);
    ratfrac = (bap_tableof_ratfrac_mpz)ba0_new_table ();
    properties = (ba0_tableof_string)ba0_new_table ();
    ba0_scanf ("regchain (%t[%Qz], %t[%s])", ratfrac, properties);
    ba0_pull_stack ();
    bad_set_and_extend_regchain_tableof_ratfrac_mpz 
	    (C, ratfrac, (bav_tableof_parameter)0, properties, true, true);
    ba0_restore (&M);
    return C;
}

BAD_DLL void bad_printf_regchain (void* A)
{   bad_regchain C = (bad_regchain)A;
    ba0_tableof_string P;
    struct ba0_mark M;

    ba0_record (&M);

    P = (ba0_tableof_string)ba0_new_table ();
    bad_structural_properties_attchain (P, &C->attrib);
    bad_desired_properties_attchain (P, &C->attrib);
    ba0_printf ("regchain (%t[%Az], %t[%s])", &C->decision_system, P);

    ba0_restore (&M);
}

BAD_DLL void bad_printf_regchain_equations (void* A)
{   bad_regchain C = (bad_regchain)A;

    ba0_printf ("%t[%Az]", &C->decision_system);
}

/*
 * Readonly static data
 */

static char _regchain [] = "struct bad_regchain";
static char _regchain_decision [] = "bad_regchain->decision_system.tab";

BAD_DLL ba0_int_p bad_garbage1_regchain (void* A, enum ba0_garbage_code code)
{   bad_regchain C = (bad_regchain)A;
    ba0_int_p i, n = 0;

    if (code == ba0_isolated)
	n += ba0_new_gc_info 
		(C, sizeof(struct bad_regchain), _regchain);

    if (C->decision_system.alloc > 0)
    {   n += ba0_new_gc_info 
		(C->decision_system.tab, 
		 C->decision_system.alloc * sizeof(bap_polynom_mpz), 
		 _regchain_decision);

        for (i = 0; i < C->decision_system.alloc; i++)
            n += bap_garbage1_polynom_mpz 
				(C->decision_system.tab [i], ba0_isolated);
    }
/*
    if (C->simplification_system.alloc > 0)
    {	n += ba0_new_gc_info 
		(C->simplification_system.tab,
		 C->simplification_system.alloc * sizeof(bap_polynom_mpz),
		 "bad_regchain->simplification_system.tab");

	for (i = 0; i < C->simplification_system.alloc; i++)
	    n += bap_garbage1_polynom_mpz 
			(C->simplification_system.tab [i], ba0_isolated);
    }
*/
    return n;
}

BAD_DLL void* bad_garbage2_regchain (void* A, enum ba0_garbage_code code)
{   bad_regchain C;
    ba0_int_p i;

    if (code == ba0_isolated)
	C = (bad_regchain)ba0_new_addr_gc_info (A, _regchain);
    else
	C = (bad_regchain)A;

    if (C->decision_system.alloc > 0)
    {   C->decision_system.tab = (bap_polynom_mpz*)ba0_new_addr_gc_info
	    (C->decision_system.tab, _regchain_decision);

        for (i = 0; i < C->decision_system.alloc; i++)
            C->decision_system.tab [i] = bap_garbage2_polynom_mpz 
				(C->decision_system.tab [i], ba0_isolated);
    }
/*
    if (C->simplification_system.alloc > 0)
    {	C->simplification_system.tab = (bap_polynom_mpz*)ba0_new_addr_gc_info
	    (C->simplification_system.tab, 
	     "bad_regchain->simplification_system.tab");

	for (i = 0; i < C->simplification_system.alloc; i++)
	    C->simplification_system.tab [i] =
		bap_garbage2_polynom_mpz 
			(C->simplification_system.tab [i], ba0_isolated);
    }
*/
    return C;
}

BAD_DLL void* bad_copy_regchain (void* A)
{   bad_regchain B, C = (bad_regchain)A;

    B = bad_new_regchain ();
    bad_set_regchain (B, C);
    return B;
}

/****************************************************************************
 PREDICATES
 ****************************************************************************/


BAD_DLL bool bad_is_rank_of_regchain 
				(bav_rank rg, bad_regchain C, ba0_int_p* i)
{   ba0_int_p k;
    bool b;

    bav_R_push_ordering (C->attrib.ordering);
    if (! bad_is_leader_of_regchain (rg->var, C, &k))
        b = false;
    else if (rg->deg != bap_leading_degree_polynom_mpz 
						(C->decision_system.tab [k]))
	b = false;
    else 
    {	if (i != (ba0_int_p*)0)
	    *i = k;
	b = true;
    }
    bav_R_pull_ordering ();
    return b;
}

BAD_DLL bool bad_is_leader_of_regchain 
				(bav_variable v, bad_regchain C, ba0_int_p* i)
{   bool b;
    ba0_int_p k;

    bav_R_push_ordering (C->attrib.ordering);
    b = false;
    for (k = 0; (! b) && k < C->decision_system.size; k++)
    {   if (v == bap_leader_polynom_mpz (C->decision_system.tab [k]))
        {   if (i != (ba0_int_p*)0)
                *i = k;
            b = true;
        }
    }
    bav_R_pull_ordering ();
    return b;
}

BAD_DLL bool bad_is_derivative_of_leader_of_regchain 
				(bav_variable v, bad_regchain C, ba0_int_p* i)
{   bav_variable w;
    bool b;
    ba0_int_p k;

    bav_R_push_ordering (C->attrib.ordering);
    b = false;
    for (k = 0; k < C->decision_system.size && (! b); k++)
    {   w = bap_leader_polynom_mpz (C->decision_system.tab [k]);
	if (bav_is_derivative (v, w))
        {   if (i != (ba0_int_p*)0)
                *i = k;
            b = true;
        }
    }
    bav_R_pull_ordering ();
    return b;
}

BAD_DLL bool bad_is_solved_regchain (bad_regchain C)
{   bav_variable x;
    bap_tableof_polynom_mpz T;
    ba0_int_p i;

    T = &C->decision_system;
    if (T->size == 0 || T->tab [0]->rang_total.size > 1)
        return false;
    x = bap_leader_polynom_mpz (T->tab [0]);
    for (i = 1; i < T->size; i++)
    {   if (! bap_is_solved_polynom_mpz (T->tab [i]))
            return false;
        if (T->tab [i]->rang_total.size > 2 ||
	                (T->tab [i]->rang_total.size == 2 &&
                                 ! bap_depend_polynom_mpz (T->tab [i], x)))
            return false;
    }
    return true;
}

BAD_DLL bool bad_is_orthonomic_regchain (bad_regchain C)
{   struct bap_polynom_mpz init;
    struct ba0_mark M;
    ba0_int_p i;
    bool b;

    ba0_record (&M);
    bap_init_readonly_polynom_mpz (&init);
    b = true;
    for (i = 0; b && i < C->decision_system.size; i++)
    {	if (bap_leading_degree_polynom_mpz (C->decision_system.tab [i]) != 1)
	    b = false;
	else
	{   bap_initial_polynom_mpz (&init, C->decision_system.tab [i]);
	    b = bap_is_independent_polynom_mpz (&init);
	}
    }
    ba0_restore (&M);
    return b;
}

BAD_DLL bool bad_is_explicit_regchain (bad_regchain C)
{   ba0_int_p i;

    for (i = 0; i < C->decision_system.size; i++)
        if (bap_leading_degree_polynom_mpz (C->decision_system.tab [i]) != 1)
            return false;
    return true;
}

BAD_DLL bool bad_is_zero_regchain (bad_regchain C)
{
    return C->decision_system.size == 0;
}

BAD_DLL ba0_int_p bad_product_of_leading_degrees_regchain (bad_regchain C)
{   ba0_int_p r, i;

    r = 1;
    for (i = 0; i < C->decision_system.size; i++)
        r *= bap_leading_degree_polynom_mpz (C->decision_system.tab [i]);
    return r;
}


/************************************************************************
 MISCELLANEOUS FUNCTIONS
 ************************************************************************/

BAD_DLL bav_Iordering bad_ordering_eliminating_leaders_of_regchain 
							(bad_regchain C)
{
    bav_Iordering r;
    ba0_int_p i;

    r = bav_R_copy_ordering (C->attrib.ordering);
    bav_R_push_ordering (r);
    for (i = 0; i < C->decision_system.size; i++)
	bav_R_set_maximal_variable 
			(bap_leader_polynom_mpz (C->decision_system.tab [i]));
    bav_R_pull_ordering ();
    return r;
}

static void bad_normalize_polynom_mpz
	(bap_polynom_mpz, bap_product_mpz, bap_product_mpz, bap_polynom_mpz);

BAD_DLL void bad_quench_regchain
        (bad_regchain Cbar, bad_regchain C, ba0_int_p k, bap_polynom_mpz* ddz)
{
    struct bad_regchain D;
    struct bap_product_mpz R, U, G;
    struct bap_polynom_mpz init, sep, Di;
    volatile bap_polynom_mpz zerodiv; 	
    struct bav_rank rank_before, rank_after;
    bav_variable v;
    ba0_mpz_t *lc;
    struct ba0_exception_code code;
    volatile ba0_int_p i, j;
    volatile bool the_separant_also;
    volatile bool exception_due_to_the_initial, exception_due_to_the_separant;
    struct ba0_mark M;

    bav_R_push_ordering (C->attrib.ordering);

    ba0_push_another_stack ();
    ba0_record (&M);

    bad_init_regchain (&D);
    bad_set_regchain (&D, C);

    bap_init_product_mpz (&R);
    bap_init_product_mpz (&U);
    bap_init_product_mpz (&G);

    bap_init_readonly_polynom_mpz (&init);
    bap_init_polynom_mpz (&sep);
    bap_init_polynom_mpz (&Di);

    i = k;
    while (i < C->decision_system.size)
    {
/*
 * If desired, the current equation is reduced using the former ones. 
 * It may happen that the rank of the equation decreases. 
 * In that case, the ideal is unit and the exception BAD_EXRCNC is raised.
 *
 * Expanding the product is necessary in general.
 */
	rank_before = bap_rank_polynom_mpz (D.decision_system.tab [i]);
	if (bad_defines_a_differential_ideal_regchain (&D))
	{
	    if (bad_has_desired_property_attchain 
				(&D.attrib, bad_autoreduced_desired_property))
	    {	bad_reduce_polynom_by_regchain 
			(&R, (bap_product_mpz)0, D.decision_system.tab [i], &D, 
			 bad_full_reduction, bad_all_but_leader_to_reduce);
		bap_expand_product_mpz (D.decision_system.tab [i], &R);
	    } else if (! bad_is_a_partially_reduced_polynom_wrt_regchain 
						(D.decision_system.tab [i], &D))
	    {	bad_reduce_polynom_by_regchain
			(&R, (bap_product_mpz)0, D.decision_system.tab [i], &D,
			 bad_partial_reduction, bad_all_but_leader_to_reduce);
		bap_expand_product_mpz (D.decision_system.tab [i], &R);
	    }
	} else if (bad_has_desired_property_attchain
				(&D.attrib, bad_autoreduced_desired_property))
	{   bad_reduce_polynom_by_regchain
		(&R, (bap_product_mpz)0, D.decision_system.tab [i], &D,
		 bad_algebraic_reduction, bad_all_but_leader_to_reduce);
	    bap_expand_product_mpz (D.decision_system.tab [i], &R);
	}
	rank_after = bap_rank_polynom_mpz (D.decision_system.tab [i]);
	if (! bav_equal_rank (&rank_before, &rank_after))
	    BA0_RAISE_EXCEPTION (BAD_EXRCNC);
/*
 * The reduction process may exhibit a nonzero relation.
 * Case of nonzero variables not at the bottom of the ordering.
 */
	if ((*bad_nonzero) (D.decision_system.tab [i]))
	    BA0_RAISE_EXCEPTION (BAD_EXRCNC);
/*
 * Extracting the initial and the separant
 */
	bap_initial_polynom_mpz (&init, D.decision_system.tab [i]);
	the_separant_also = 
	    ((bad_defines_a_differential_ideal_regchain (&D) ||
              bad_has_desired_property_attchain
                        (&D.attrib, bad_squarefree_desired_property)) &&
             bap_leading_degree_polynom_mpz (D.decision_system.tab [i]) > 1);
	if (the_separant_also)
	    bap_separant_polynom_mpz (&sep, D.decision_system.tab [i]);

/*
 * Regularization processes
 *
 *    The step consists in checking the regularity of the initial of the 
 * current equation modulo the ideal defined by the already processed equations 
 * (in the case of a strong normalization desire, one even tries to compute an 
 * algebraic inverse of this initial but computations are essentially the same).
 *
 *    If the initial is zero (i.e. if the regularity check raises the exception
 * BAD_EXRNUL) the ideal is proven to be the unit one and the exception 
 * BAD_EXRCNC is raised.
 *
 *    If the regularity check raises the exception BAD_EXRDDZ then two cases
 * are distinguished: if the initial and the zero divisor have the same leader 
 * then the zero divisor is a divisor of the initial thus the current equation 
 * is factored out by this divisor: the exception is handled locally, since 
 * one only keeps one branch. 
 *
 *    In the other case, the exception BAD_EXRDDZ is re-raised: one may have
 * to keep different branches. This is not handled locally.
 *
 *    If desired, the invertibility of the separant is checked too. The 
 * management of the exceptions is the same as for the initial.
 */
	exception_due_to_the_initial = false;
	if (! bap_is_numeric_polynom_mpz (&init))
	{
/*
 * Save D.decision_system.tab [i] to recover it if an exception is raised.
 * This is only needed in the case of a strong normalization.
 */
	    if (bad_has_desired_property_attchain
				(&D.attrib, bad_normalized_desired_property)) 
	        bap_set_polynom_mpz (&Di, D.decision_system.tab [i]);
	    BA0_PUSH_EXCEPTION (code);
	    if (ba0_exception_is_set (code))
	    { 
/*
 * U, G and D.decision_system.tab [i] are likely to be modified.
 */
	        if (bad_has_desired_property_attchain 
				(&D.attrib, bad_normalized_desired_property))
	        {   bad_invert_polynom_mod_regchain 
			(&U, &G, &init, &D, (bap_polynom_mpz*)&zerodiv);
	            bad_normalize_polynom_mpz (D.decision_system.tab [i], 
					   &U, &G, D.decision_system.tab [i]);
/*
 * The rank cannot decrease for the initial is now independent.
 */
		    if (bad_has_desired_property_attchain
				(&D.attrib, bad_autoreduced_desired_property))
		    {   bad_reduce_polynom_by_regchain
				(&R, (bap_product_mpz)0, 
				 D.decision_system.tab [i], &D,
				 bad_algebraic_reduction, 
				 bad_all_but_leader_to_reduce);
		        bap_expand_product_mpz (D.decision_system.tab [i], &R);
		    }
	        } else
		    bad_check_regularity_polynom_mod_regchain 
				(&init, &D, (bap_polynom_mpz*)&zerodiv);
		ba0_pull_exception (code);
	    } else
		exception_due_to_the_initial = true;
	}
/*
 * Primitive parts can be taken now since the initial is guaranteed.
 */
	if ((! exception_due_to_the_initial) && 
			bad_has_desired_property_attchain
				(&D.attrib, bad_primitive_desired_property))
	    bap_primpart_polynom_mpz (D.decision_system.tab [i], 
				D.decision_system.tab [i], BAV_NOT_A_VARIABLE);
/*
 * The new relation may be considered as non zero.
 */
	if ((*bad_nonzero) (D.decision_system.tab [i]))
	    BA0_RAISE_EXCEPTION (BAD_EXRCNC);
/*
 * The squarefreeness desire
 */
	exception_due_to_the_separant = false;
	if ((! exception_due_to_the_initial) && the_separant_also)
	{   BA0_PUSH_EXCEPTION (code);
	    if (ba0_exception_is_set (code))
	    {	bad_check_regularity_polynom_mod_regchain 
					(&sep, &D, (bap_polynom_mpz*)&zerodiv);
		ba0_pull_exception (code);
	    } else
		exception_due_to_the_separant = true;
	} 
/*
 * Normalize the numerical leading coefficient
 */
	if ((! exception_due_to_the_initial) && 
					(! exception_due_to_the_separant))
	{
	    lc = bap_numeric_initial_polynom_mpz (D.decision_system.tab [i]);
	    if (ba0_mpz_sgn (*lc) < 0)
		bap_neg_polynom_mpz 
			(D.decision_system.tab [i], D.decision_system.tab [i]);
	    i++;
	} else
	{   
	    if (ba0_mesgerr == BAD_EXRNUL)
		BA0_RAISE_EXCEPTION (BAD_EXRCNC);
	    else if (ba0_mesgerr != BAD_EXRDDZ)
		BA0_RAISE_EXCEPTION (ba0_mesgerr);
/*
 * If BAD_EXRDDZ was raised then the zero divisor is stored in zerodiv.
 * Let j = the index of the element of D with the same leader as zerodiv.
 */
	    v = bap_leader_polynom_mpz (zerodiv);
	    while (!  bad_is_leader_of_regchain (v, &D, (ba0_int_p*)&j))
	    {   
/*
 * tests/rg17.c
 */
		bap_initial_polynom_mpz (zerodiv, zerodiv);
		if (bap_is_independent_polynom_mpz (zerodiv))
		    BA0_RAISE_EXCEPTION (BA0_ERRALG);
		v = bap_leader_polynom_mpz (zerodiv);
	    }
/*
 * In these cases, the exception is handled non-locally (see above).
 */
	    if ((exception_due_to_the_initial &&
				v != bap_leader_polynom_mpz (&init)) ||
		(exception_due_to_the_separant &&
				v != bap_leader_polynom_mpz (&sep)))
	    {	if (ddz != (bap_polynom_mpz*)0)
		{   *ddz = zerodiv;
		    BA0_RAISE_EXCEPTION2 (BAD_EXRDDZ, "%Az", (void**)ddz);
		} else
		    BA0_RAISE_EXCEPTION (BAD_EXRDDZ);
	    }
/*
 * The exception is handled locally. Observe j < i.
 */
	    bap_pquo_polynom_mpz (D.decision_system.tab [j], (bav_Idegree*)0, 
				  D.decision_system.tab [j], zerodiv, v);
/*
 * Variables modified between the moment the exception catching point
 * was set and the moment the exception was raised must be reinitialized.
 */
	    if (exception_due_to_the_initial &&
			bad_has_desired_property_attchain 
				(&D.attrib, bad_normalized_desired_property))
	    {	bap_init_product_mpz (&R);
		bap_init_product_mpz (&U);
		bap_init_product_mpz (&G);
		bap_init_polynom_mpz (D.decision_system.tab [i]);
		bap_set_polynom_mpz (D.decision_system.tab [i], &Di);
	    }
/*
 * Quenching must now restart at index j.
 */
	    i = j;
	} 
    }
    ba0_pull_stack ();
    bad_set_regchain (Cbar, &D);
    ba0_restore (&M);

    bav_R_pull_ordering ();
}

/*
 * We have U * initial (A) = G.
 * Sets R to G * rank (A) + U * reductum (A)
 */

static void bad_normalize_polynom_mpz 
   (bap_polynom_mpz R, bap_product_mpz U, bap_product_mpz G, bap_polynom_mpz A)
{   struct bap_polynom_mpz init, tail, red;
    struct bav_term T;
    struct bav_rank rg;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);

    rg = bap_rank_polynom_mpz (A);

    bav_init_term (&T);
    bav_set_term_rank (&T, &rg);

    bap_init_polynom_mpz (&init);
    bap_expand_product_mpz (&init, G);
    bap_mul_polynom_term_mpz (&init, &init, &T);

    bap_init_readonly_polynom_mpz (&red);
    bap_reductum_polynom_mpz (&red, A);
    bap_init_polynom_mpz (&tail);
    bap_expand_product_mpz (&tail, U);
    bap_mul_polynom_mpz (&tail, &tail, &red);

    ba0_pull_stack ();
    bap_add_polynom_mpz (R, &init, &tail);
    ba0_restore (&M); 
}

BAD_DLL void bad_quench_and_handle_exceptions_regchain
	(bad_intersectof_regchain tabC, ba0_int_p k, bad_regchain ideal)
{   bad_regchain C;
    bap_polynom_mpz g;
    struct ba0_exception_code code;

    if (ideal != (bad_regchain)0 && 
				! bad_defines_a_prime_ideal_regchain (ideal))
	BA0_RAISE_EXCEPTION (BA0_ERRNYP);
    bav_R_push_ordering (tabC->attrib.ordering);
    C = tabC->inter.tab [tabC->inter.size - 1];

    BA0_PUSH_EXCEPTION (code);
    if (ba0_exception_is_set (code))
    {	bad_quench_regchain (C, C, k, &g);
/*
   The above quenching operation has not raised any unhandled exception
*/
	ba0_pull_exception (code);
	bav_R_pull_ordering ();
    } else 
    {   bav_R_pull_ordering ();
	bad_handle_splitting_exceptions_regchain (tabC, g, ideal, ba0_mesgerr);
    }
}

/*
 * An exception EXRCNC or EXRDDZ was raised, exhibiting an inconsistency 
 * (BAD_EXRCNC) or a factorization of the regchain on the top of tabC 
 * (the factor is in g). 
 * 
 * The following function is called in order to split the current 
 * regchain (eventually to kill it) and perform some quenching.
 *
 * Observe that EXRNUL is *not* handled in this function (see normal form
 * functions).
 */

BAD_DLL void bad_handle_splitting_exceptions_regchain 
	(bad_intersectof_regchain tabC, bap_polynom_mpz g, 
	 bad_regchain ideal, char* mesgerr)
{   bad_regchain C, D = (bad_regchain)0;
    enum bad_typeof_reduction type_red;
    bap_polynom_mpz h;
    bav_variable v;
    ba0_int_p l;
    bool consider_g, consider_h;
    struct ba0_mark M;

    if (ideal != (bad_regchain)0 && 
			! bad_defines_a_prime_ideal_regchain (ideal))
	BA0_RAISE_EXCEPTION (BA0_ERRNYP);
    bav_R_push_ordering (tabC->attrib.ordering);
    C = tabC->inter.tab [--tabC->inter.size];

    if (mesgerr == BAD_EXRCNC)
    {
/*
 * The current branch leads to the unit ideal. 
 */
	bav_R_pull_ordering ();
	return;
    } else if (mesgerr != BAD_EXRDDZ)
	BA0_RAISE_EXCEPTION (mesgerr);
/* 
 * A factorization C->decision_system [l] = g * h is exhibited. 
 * This factorization leads to a splitting (some branches may be 
 * discarded using the regular chain ideal) at index l. 
 * The quenching operation is restarted on all branches from index l.
 */
    v = bap_leader_polynom_mpz (g);
    if (! bad_is_leader_of_regchain (v, C, &l))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    type_red = bad_defines_a_differential_ideal_attchain (&tabC->attrib) ? 
				bad_full_reduction : bad_algebraic_reduction;

    consider_g = 
	ideal == (bad_regchain)0 ||
	bad_is_a_reduced_to_zero_polynom_by_regchain (g, ideal, type_red);

    ba0_push_another_stack ();
    ba0_record (&M);
    h = bap_new_polynom_mpz ();
    bap_pquo_polynom_mpz (h, (bav_Idegree*)0, C->decision_system.tab [l], g, v);
    ba0_pull_stack ();

    consider_h = ideal == (bad_regchain)0 ||
	 bad_is_a_reduced_to_zero_polynom_by_regchain (h, ideal, type_red);
    if (consider_h && ! consider_g)
    {	BA0_SWAP (bap_polynom_mpz, g, h);
	BA0_SWAP (bool, consider_g, consider_h);
    }

    if (consider_h)
	D = (bad_regchain)bad_copy_regchain (C);

    if (consider_g)
    {
/*
 * Sufficient to push C at the tail of tabC.
 */
	tabC->inter.size += 1;
	bap_set_polynom_mpz (C->decision_system.tab [l], g);
	bad_quench_and_handle_exceptions_regchain (tabC, l, ideal);
    }

    if (consider_h)
    {	ba0_realloc_table ((ba0_table)&tabC->inter, tabC->inter.size + 1);
	tabC->inter.tab [tabC->inter.size++] = D;
	bap_set_polynom_mpz (D->decision_system.tab [l], h);
	bad_quench_and_handle_exceptions_regchain (tabC, l, ideal);
    }

    ba0_restore (&M);
    bav_R_pull_ordering ();
}

BAD_DLL ba0_int_p bad_codimension_regchain (bad_regchain C, bad_base_field K)
{   
    struct bav_tableof_variable T;
    bav_variable v;
    ba0_int_p i, n, codim = 0;
    struct ba0_mark M;

/*
 * The defining equations of K (i.e. K->relations) are part of
 * the equations of C. They must not be taken inbto account while
 * counting the codimension.
 */
    if (bad_has_structural_property_regchain 
				(C, bad_differential_structural_property))
    {	
/*
 * ODE case. It is sufficient to count the equations.
 */
	if (bav_R.ders.size == 1)
	{   if (K != (bad_base_field)0)
	        n = K->relations.decision_system.size;
	    else
	        n = 0;
	    codim = C->decision_system.size - n;
	} else
/*
 * PDE case. One counts the dependent variables, for which, at
 * least one derivative is a leader.
 */
	{   ba0_record (&M);
	    ba0_init_table ((ba0_table)&T);
	    ba0_realloc_table ((ba0_table)&T, bav_R.vars.size);
	    if (K != (bad_base_field)0)
	    {	for (i = 0; i < K->relations.decision_system.size; i++)
		{   v = bap_leader_polynom_mpz 
				(K->relations.decision_system.tab [i]);
		    v = bav_order_zero_variable (v);
		    if (! ba0_member_table (v, (ba0_table)&T))
		    {	T.tab [T.size] = v;
			T.size += 1;
		    }
		}
		n = T.size;
	    } else
		n = 0;
/*
 * The defining equations of K are counted twice. Could be optimized.
 */
	    ba0_reset_table ((ba0_table)&T);
	    for (i = 0; i < C->decision_system.size; i++)
	    {	v = bap_leader_polynom_mpz (C->decision_system.tab [i]);
		v = bav_order_zero_variable (v);
		if (! ba0_member_table (v, (ba0_table)&T))
		{   T.tab [T.size] = v;
		    T.size += 1;
		}
	    }
	    ba0_restore (&M);
	    codim = T.size - n;
	}
    } else
/*
 * Algebraic case.  It is sufficient to count the equations.
 */
    {   if (K != (bad_base_field)0)
	    n = K->relations.decision_system.size;
	else
	    n = 0;
	codim = C->decision_system.size - n;
    }
    return codim;
}

BAD_DLL unsigned ba0_int_p bad_sizeof_regchain (bad_regchain C)
{   unsigned ba0_int_p size;
    ba0_int_p i;

    size = ba0_allocated_size (sizeof (struct bad_regchain));
    size += ba0_allocated_size 
		(C->decision_system.size * sizeof (bap_polynom_mpz));
    for (i = 0; i < C->decision_system.size; i++)
	size += bap_sizeof_polynom_mpz (C->decision_system.tab [i]);
    return size;
}

BAD_DLL void bad_switch_ring_regchain (bad_regchain C, bav_differential_ring R)
{   ba0_int_p i;

    for (i = 0; i < C->decision_system.size; i++)
	bap_switch_ring_polynom_mpz (C->decision_system.tab [i], R);
}

