#include "bad_nonzero.h"
#include "bad_reduction.h"
#include "bad_regularize.h"
#include "bad_invert.h"
#include "bad_global.h"

BAD_DLL void bad_set_settings_regularize 
			(enum bad_typeof_regularize_strategy strategy)
{
    bad_regularize_strategy = strategy ? strategy :
					bad_subresultant_regularize_strategy;
}

BAD_DLL void bad_get_settings_regularize 
			(enum bad_typeof_regularize_strategy* strategy)
{
    if (strategy) *strategy = bad_regularize_strategy;
}

/**********************************************************************
 CONTEXT FUNCTION FOR BAD_EUCLID
 **********************************************************************/

static enum bad_typeof_context bad_context 
				(bad_tableof_quadruple tabG, bad_regchain C)
{
    if (C != (bad_regchain)0)
	if (tabG != (bad_tableof_quadruple)0)
	    return bad_pardi_context;
	else
	    return bad_inverse_context;
    else
	return bad_rg_context;
}

typedef void bad_Euclid_function
		(bap_tableof_tableof_polynom_mpz, 
		 bad_tableof_quadruple,
		 enum bad_typeof_Euclid,
		 bap_polynom_mpz, bap_polynom_mpz, 
		 bav_variable,
		 bool, bool, bad_regchain, bap_polynom_mpz*);
	
/*****************************************************************************
 THE EUCLIDEAN ALGORITHM 

 It is performed with coefficients taken modulo an ideal.

 In the Rosenfeld-Grbner context, splitting cases may lead to
 recursive calls, through bad_ensure_regular_lcoeff.
 *****************************************************************************/

static bad_Euclid_function bad_Euclid_gcd_prem_and_factor_mod_regchain;
static bad_Euclid_function bad_Euclid_gcd_prem_mod_regchain;
static bad_Euclid_function bad_Euclid_subresultant_mod_regchain;

BAD_DLL void bad_Euclid_mod_regchain 
	       (bap_tableof_tableof_polynom_mpz tabV, 
		bad_tableof_quadruple tabG,
		enum bad_typeof_Euclid type,
		bap_polynom_mpz A, bap_polynom_mpz B, bav_variable v, 
		bool reginitA, bool reginitB,
		bad_regchain C, bap_polynom_mpz* ddz)
{
    switch (bad_regularize_strategy)
    {	case bad_subresultant_regularize_strategy:
	    bad_Euclid_subresultant_mod_regchain
		(tabV, tabG, type, A, B, v, reginitA, reginitB, C, ddz);
	    break;
	case bad_gcd_prem_regularize_strategy:
	    bad_Euclid_gcd_prem_mod_regchain
		(tabV, tabG, type, A, B, v, reginitA, reginitB, C, ddz);
	    break;
	case bad_gcd_prem_and_factor_regularize_strategy: 
	    bad_Euclid_gcd_prem_and_factor_mod_regchain 
		(tabV, tabG, type, A, B, v, reginitA, reginitB, C, ddz);
	    break;
    }
}

/*
   Subfunction of bad_Euclid_mod_regchain.
   Removes lcoeff (A, v) from A while this coefficient is zero mod C.
   At the end of the loop (i.e. when lcoeff (A, v) is non zero), checks
	that lcoeff (A, v) is regular mod C.
   If non regular, the exception BAD_EXRDDZ is raised and a divisor
	of some element of C is stored in *ddz (provided ddz is nonzero).
   The boolean *rankfall receives true if at least one of the lcoeff (A, v)
	is zero (so that the rank of A falls).

   Stores in P the coefficients which are zero mod C but not mod the
	component A of G.

   tabV and tabG lie in the current stack. They may be modified in rg context.
   B lies in the other stack. It may be modified in inverse and pardi contexts.
*/

static bool bad_ensure_regular_lcoeff 
	(bap_tableof_tableof_polynom_mpz tabV, bad_tableof_quadruple tabG, 
	 enum bad_typeof_Euclid type, 
	 bap_polynom_mpz A, bap_polynom_mpz B, bav_variable v, 
	 bool reginitA, 
	 bad_regchain C, bap_polynom_mpz* ddz)
{   bad_quadruple G = (bad_quadruple)0, H;
    enum bad_typeof_reduction type_red;
    enum bad_typeof_context context;
    bap_tableof_polynom_mpz V, W;
    struct bap_polynom_mpz init, reductum;
    ba0_int_p index_G, index_V;
    bool consistent;

/*
   init and reductum are going to receive parts of B.
   They thus lie in the other stack.
   Actually, the stack changing operations are not necessary since the
	initial and reductum inst. are implemented with no memory allocation.
*/
    ba0_push_another_stack ();
    bap_init_readonly_polynom_mpz (&init);
    bap_init_readonly_polynom_mpz (&reductum);
    ba0_pull_stack ();

    consistent = true;
    context = bad_context (tabG, C);
    switch (context)
    {	case bad_rg_context:
	    index_G = tabG->size - 1;
	    index_V = tabV->size - 1;
	    G = tabG->tab [index_G];
	    V = tabV->tab [index_V];
/*
   A splitting is generated :
	H = G + initial, reductum = 0
	G = G + initial <> 0.
   About H. The initial goes in the P component. 
	    For the reductum, it depends on its degree in v. 
   If this degree is > 0 then 
	B degenerates to the reductum and a recursive call to bad_Euclid 
	is performed to handle that degenerated gcd computation. 
   If this degree is <= 0 then 
	the reductum is stored in the P component of H and, modulo the 
	equations present in P, the polynomial B degenerates to zero and 
	the gcd, which is A, is stored in W.
*/
	    ba0_push_another_stack ();
	    bap_initial_and_reductum2_polynom_mpz (&init, &reductum, B, v);
	    ba0_pull_stack ();
	    ba0_realloc2_table ((ba0_table)tabG, tabG->size + 1,
				(ba0_new_function*)&bad_new_quadruple);
	    ba0_realloc2_table ((ba0_table)tabV, tabV->size + 1,
				(ba0_new_function*)&ba0_new_table);
	    H = tabG->tab [tabG->size++];
	    W = tabV->tab [tabV->size++];
	    bad_set_quadruple (H, G);
/*
   If G + initial = 0 (possibly + reductum = 0) is inconsistent then
        the quadruple H is discarded.
*/
	    if (! bad_simplify_and_store_in_P_quadruple (H, &init))
	    {   tabG->size--;
		tabV->size--;
	    } else
	    {   if (! bap_depend_polynom_mpz (&reductum, v))
		{   if (! bad_simplify_and_store_in_P_quadruple (H, &reductum))
		    {   tabG->size--;
			tabV->size--;
		    } else
		    {	ba0_realloc2_table ((ba0_table)W, 1,
				(ba0_new_function*)&bap_new_polynom_mpz);
			bap_set_polynom_mpz (W->tab [0], A);
			W->size = 1;
		    }
		} else
		    bad_Euclid_mod_regchain
			(tabV, tabG, type,
			 A, &reductum, v, reginitA, false, C, ddz);
/*
   The quadruple H and possibly many other ones, due to the recursive call
	to bad_Euclid are stored in tabG, above G (the corresponding gcds
	are stored in tabV, above V).
   I manage to put (G, V) at the top of the stacks tabG and tabV.
*/
		BA0_SWAP (bad_quadruple, 
				tabG->tab [tabG->size - 1], 
				tabG->tab [index_G]);
		BA0_SWAP (bap_tableof_polynom_mpz, 
				tabV->tab [tabV->size - 1],
				tabV->tab [index_V]);
	    }
	    consistent = bad_simplify_and_store_in_S_quadruple (G, &init);
	    break;
	case bad_pardi_context:
	    G = tabG->tab [tabG->size - 1];
	case bad_inverse_context:
	    type_red = 
		bad_defines_a_differential_ideal_regchain (C) ? 
				bad_full_reduction : bad_algebraic_reduction;
/*
   If lcoeff (B) = 0 mod C then 
      B = reductum (B)
      in the pardi case: if lcoeff (B) <> 0 mod the component A of G then 
	this lcoeff is stored it is stored in the component P of G.
   If deg (B, v) = 0 then, 
      in the pardi case: B is necessarily zero mod C.
*/
	    for (;;)
	    {	if (bap_is_zero_polynom_mpz (B))
		    break;
		ba0_push_another_stack ();
		bap_initial2_polynom_mpz (&init, B, v);
		ba0_pull_stack ();
		if (bap_depend_polynom_mpz (B, v) || 
						context == bad_inverse_context)
		{   if (! bad_is_a_reduced_to_zero_polynom_by_regchain 
							(&init, C, type_red))
			break;
		}
		if (context == bad_pardi_context &&
			! bad_is_a_reduced_to_zero_polynom_by_regchain
						(&init, &G->A, type_red))
		    bad_simplify_and_store_in_P_quadruple (G, &init);
		ba0_push_another_stack ();
		bap_lcoeff_and_reductum_polynom_mpz 
					(BAP_NOT_A_POLYNOM_mpz, B, B, v);
		ba0_pull_stack ();
	    }
/*
   In the inverse context, we only know that the lcoeff of B is non zero.
   We need to make sure it is regular.
   The problem does not arise with pardi since the ideal is prime.
*/
	    if (context == bad_inverse_context &&
			! bad_defines_a_prime_ideal_regchain (C))
	    {	ba0_push_another_stack ();
		bap_initial2_polynom_mpz (&init, B, v);
		ba0_pull_stack ();
		bad_check_regularity_polynom_mod_regchain (&init, C, ddz);
	    }
	    break;
    }
    return consistent;
}

/*
  Subfunction of bad_Euclid_mod_regchain
  Computations are performed using subresultant sequences
*/

static bad_Euclid_function bad_basic_Euclid_subresultant_mod_regchain;
static bad_Euclid_function bad_extended_Euclid_subresultant_mod_regchain;

static void bad_Euclid_subresultant_mod_regchain 
	       (bap_tableof_tableof_polynom_mpz tabV, 
		bad_tableof_quadruple tabG,
		enum bad_typeof_Euclid type,
		bap_polynom_mpz A, bap_polynom_mpz B, bav_variable v, 
		bool reginitA, bool reginitB,
		bad_regchain C, bap_polynom_mpz* ddz)
{   
    if (type == bad_basic_Euclid)
	bad_basic_Euclid_subresultant_mod_regchain 
		(tabV, tabG, type, A, B, v, reginitA, reginitB, C, ddz);
    else
	bad_extended_Euclid_subresultant_mod_regchain 
		(tabV, tabG, type, A, B, v, reginitA, reginitB, C, ddz);
}

/*
   Subfunction of bad_Euclid_subresultant_mod_regchain
   Only the last nonzero subresultant is required (not the Bezout identity)

   Stores in S the actual leading coefficients of the remainders (but not the
	leading coefficient of the first remainder).
   The function ensure_regular_lcoeff modifies P.
*/

static void bad_basic_Euclid_subresultant_mod_regchain 
	(bap_tableof_tableof_polynom_mpz tabV, bad_tableof_quadruple tabG, 
	 enum bad_typeof_Euclid type,
	 bap_polynom_mpz A, bap_polynom_mpz B, bav_variable v, 
	 bool reginitA, bool reginitB,
	 bad_regchain C, bap_polynom_mpz* ddz)
{
    bad_quadruple G = (bad_quadruple)0; /* to avoid a warning */
    enum bad_typeof_context context;
    bap_tableof_polynom_mpz V;
    bap_polynom_mpz P, Q, Z;
    struct bap_polynom_mpz coeff, s;
    bav_Idegree d, delta;
    bool consistent, rankfall;
    struct ba0_mark M;

    P = (bap_polynom_mpz)0;
    Q = (bap_polynom_mpz)0;
    Z = (bap_polynom_mpz)0;

    ba0_push_another_stack ();
    ba0_record (&M);
/*
   The result (the gcd of A and B) is stored in the V on the top of
	the stack tabV. 
   In the pardi and rg case, one also work with the G on the top of
	the stack tabG. In the rg case, it may happen that G gets
	inconsistent. It is then pulled out of the stack.
*/
    context = bad_context (tabG, C);
    switch (context)
    {	case bad_rg_context:
	case bad_pardi_context:
	    G = tabG->tab [tabG->size - 1];
	    break;
	case bad_inverse_context:
	    break;
    }
    V = tabV->tab [tabV->size - 1];
    if (bap_leading_degree_polynom_mpz (A) < bap_leading_degree_polynom_mpz (B))
    {
        BA0_SWAP (bap_polynom_mpz, A, B);
	BA0_SWAP (bool, reginitA, reginitB);
    }
/*
   We make sure that the leading coefficient of the lower degree polynomial
	is regular.
*/
    consistent = true;
    if (! reginitB)
    {	ba0_pull_stack ();
	switch (context)
	{   case bad_inverse_context:
	    case bad_pardi_context:
	 	bad_ensure_regular_lcoeff 
				(tabV, tabG, type, A, B, v, reginitA, C, ddz);
		break;
	    case bad_rg_context:
		consistent = bad_ensure_regular_lcoeff
				(tabV, tabG, type, A, B, v, reginitA, C, ddz);
		break;
	}
	ba0_push_another_stack ();
    }
    if (consistent)
    {	bap_init_readonly_polynom_mpz (&coeff);
	bap_init_polynom_mpz (&s);

	P = bap_new_polynom_mpz ();
	Q = bap_new_polynom_mpz ();
	Z = bap_new_polynom_mpz ();
	bap_set_polynom_mpz (P, A);
	bap_set_polynom_mpz (Q, B);
    }
/*
   Result in P at the end of the loop (unless inconsistent)
*/
    while (consistent && ! bap_is_zero_polynom_mpz (Q))
    {
/*
   (P, Q) = (Q, prem (P, Q)).
   The original code of Lionel Ducos takes prem (P, - Q).
*/
	if (bap_degree_polynom_mpz (Q, v) == 0)
	{   BA0_SWAP (bap_polynom_mpz, P, Q);
	    bap_set_polynom_zero_mpz (Q);
	} else
	{   bap_initial2_polynom_mpz (&coeff, Q, v);
	    delta = bap_leading_degree_polynom_mpz (P) - 
					bap_leading_degree_polynom_mpz (Q);
	    bap_pow_polynom_mpz (&s, &coeff, delta);
	    bap_prem_polynom_mpz (Z, (bav_Idegree*)0, P, Q, v);
	    BA0_SWAP (bap_polynom_mpz, Z, Q);
	    BA0_SWAP (bap_polynom_mpz, Z, P);
	}
	rankfall = false;
	while (consistent && ! bap_is_zero_polynom_mpz (Q) && ! rankfall)
	{   d = bap_degree_polynom_mpz (Q, v);
	    ba0_pull_stack ();
	    switch (context)
	    {	case bad_inverse_context:
		case bad_pardi_context:
		    bad_ensure_regular_lcoeff 
				(tabV, tabG, type, P, Q, v, true, C, ddz);
		    break;
		case bad_rg_context:
		    if (bap_depend_polynom_mpz (Q, v))
			consistent = bad_ensure_regular_lcoeff
				(tabV, tabG, type, P, Q, v, true, C, ddz);
		    else
		    {	consistent = 
				bad_simplify_and_store_in_P_quadruple (G, Q);
			bap_set_polynom_zero_mpz (Q);
		    }
		    break;
	    }
	    ba0_push_another_stack ();
	    if (! consistent || bap_is_zero_polynom_mpz (Q))
		continue;
	    rankfall = d != bap_degree_polynom_mpz (Q, v);
	    if (rankfall)
		continue;
	    bap_initial2_polynom_mpz (&coeff, Q, v);
	    delta = bap_leading_degree_polynom_mpz (P) -
					bap_degree_polynom_mpz (Q, v);
	    bap_muldiv2_Lazard_polynom_mpz (Z, Q, &coeff, &s, delta);
	    bap_nsr2_Ducos_polynom_mpz (P, P, Q, Z, &s, v);
	    BA0_SWAP (bap_polynom_mpz, P, Q);
	    bap_lcoeff_polynom_mpz (&s, Z, v);
	}
    }
    ba0_pull_stack ();
    if (! consistent)
    {	tabG->size--;
	tabV->size--;
    } else
    {	ba0_realloc2_table 
		((ba0_table)V, 1, (ba0_new_function*)&bap_new_polynom_mpz);    
	bap_set_polynom_mpz (V->tab [0], P);
	V->size = 1;
    }
    ba0_restore (&M);
}

/*
   Subfunction of bad_Euclid_subresultant_mod_regchain
   The table result is initialized.
   Its size gives the part of the Bezout identity which is desired.

   Stores in S the actual leading coefficients of the remainders (but not the
	first one).
   The function ensure_regular_lcoeff modifies P.
*/

static void bad_extended_Euclid_subresultant_mod_regchain
        (bap_tableof_tableof_polynom_mpz tabV, bad_tableof_quadruple tabG,
         enum bad_typeof_Euclid type,
         bap_polynom_mpz P, bap_polynom_mpz Q, bav_variable v,
         bool reginitP, bool reginitQ,
         bad_regchain C, bap_polynom_mpz* ddz)
{   bad_quadruple G;
    enum bad_typeof_context context;
    bap_tableof_polynom_mpz V, VZ, VP, VQ;
    struct bap_polynom_mpz s, coeff, bunk;
    bav_Idegree degP, degQ, d;
    bool perm, rankfall, consistent;
    ba0_int_p i, n = 0;
    struct ba0_mark M;

    VP = (bap_tableof_polynom_mpz)0;
    VQ = (bap_tableof_polynom_mpz)0;
    VZ = (bap_tableof_polynom_mpz)0;

    context = bad_context (tabG, C);

    if (context != bad_inverse_context)
	BA0_RAISE_EXCEPTION (BA0_ERRNYP);
    if (bap_degree_polynom_mpz (P, v) <= 0 || 
					bap_degree_polynom_mpz (Q, v) <= 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    ba0_push_another_stack ();
    ba0_record (&M);

    switch (context)
    {	case bad_rg_context:
	case bad_pardi_context:
	    G = tabG->tab [tabG->size - 1];
	    break;
	case bad_inverse_context:
	    break;
    }
    V = tabV->tab [tabV->size - 1];
    switch (type)
    {	case bad_basic_Euclid:
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
	    break;
	case bad_half_extended_Euclid:
	    n = 2;
	    break;
	case bad_extended_Euclid:
	    n = 3;
	    break;
    }
    degP = bap_degree_polynom_mpz (P, v);
    degQ = bap_degree_polynom_mpz (Q, v);
    if (degP < degQ)
    {	BA0_SWAP (bap_polynom_mpz, P, Q);
	BA0_SWAP (bav_Idegree, degP, degQ);
	BA0_SWAP (bool, reginitP, reginitQ);
	perm = true;
    } else
	perm = false;

    consistent = true;
    if (! reginitQ)
    {	ba0_pull_stack ();
	switch (context)
	{   case bad_inverse_context:
	    case bad_pardi_context:
		bad_ensure_regular_lcoeff 
				(tabV, tabG, type, P, Q, v, reginitP, C, ddz);
		break;
	    case bad_rg_context:
		consistent = bad_ensure_regular_lcoeff 
				(tabV, tabG, type, P, Q, v, reginitP, C, ddz);
		break;
	}
	ba0_push_another_stack ();
    }
    if (consistent)
    {
	bap_init_polynom_mpz (&bunk);
	bap_init_readonly_polynom_mpz (&coeff);
	bap_init_polynom_mpz (&s);
	VP = (bap_tableof_polynom_mpz)ba0_new_table ();
	VQ = (bap_tableof_polynom_mpz)ba0_new_table ();
	VZ = (bap_tableof_polynom_mpz)ba0_new_table ();
	ba0_realloc2_table 
		((ba0_table)VP, n, (ba0_new_function*)&bap_new_polynom_mpz);
	ba0_realloc2_table 
		((ba0_table)VQ, n, (ba0_new_function*)&bap_new_polynom_mpz);
	ba0_realloc2_table 
		((ba0_table)VZ, n, (ba0_new_function*)&bap_new_polynom_mpz);
	VP->size = n;
	VQ->size = n;
	VZ->size = n;

	bap_initial2_polynom_mpz (&coeff, Q, v);

	bap_pow_polynom_mpz (&s, &coeff, degP - degQ);
	bap_set_polynom_mpz (VP->tab [0], Q);
	bap_set_polynom_mpz (VZ->tab [0], Q);
	if (n == 2)
	{   if (perm)
	    {   bap_set_polynom_one_mpz (VP->tab [1]);
		bap_set_polynom_one_mpz (VZ->tab [1]);
	    }
	} else if (n == 3)
	{   if (perm)
	    {   bap_set_polynom_one_mpz (VP->tab [1]);
		bap_set_polynom_one_mpz (VZ->tab [1]);
	    } else
	    {   bap_set_polynom_one_mpz (VP->tab [2]);
		bap_set_polynom_one_mpz (VZ->tab [2]);
	    }
	}

	if (n == 1)
	    bap_pseudo_division_polynom_mpz 
		(BAP_NOT_A_POLYNOM_mpz, VQ->tab [0], (bav_Idegree*)0, P, Q, v);
	else if (n == 2)
	{   if (perm)
	    {   bap_pseudo_division_polynom_mpz 
			(VQ->tab [1], VQ->tab [0], (bav_Idegree*)0, P, Q, v);
		bap_neg_polynom_mpz (VQ->tab [1], VQ->tab [1]);
	    } else
	    {   bap_pseudo_division_polynom_mpz 
			(BAP_NOT_A_POLYNOM_mpz, VQ->tab [0], &d, P, Q, v);
	        bap_initial2_polynom_mpz (&coeff, Q, v);
	        bap_pow_polynom_mpz (VQ->tab [1], &coeff, d);
	    }
	} else
	{   if (perm)
	    {   bap_pseudo_division_polynom_mpz 
				(VQ->tab [1], VQ->tab [0], &d, P, Q, v);
		bap_neg_polynom_mpz (VQ->tab [1], VQ->tab [1]);
	        bap_initial2_polynom_mpz (&coeff, Q, v);
	        bap_pow_polynom_mpz (VQ->tab [2], &coeff, d);
	    } else
	    {   bap_pseudo_division_polynom_mpz 
				(VQ->tab [2], VQ->tab [0], &d, P, Q, v);
		bap_neg_polynom_mpz (VQ->tab [2], VQ->tab [2]);
		bap_initial2_polynom_mpz (&coeff, Q, v);
		bap_pow_polynom_mpz (VQ->tab [1], &coeff, d);
	    }
	}
    }
    while (consistent && ! bap_is_zero_polynom_mpz (VQ->tab [0]))
    {
	d = bap_degree_polynom_mpz (VQ->tab [0], v);
	ba0_pull_stack ();
	switch (context)
	{   case bad_inverse_context:
	    case bad_pardi_context:
		bad_ensure_regular_lcoeff 
			(tabV, tabG, type, VP->tab [0], VQ->tab [0], 
							v, true, C, ddz);
		break;
	    case bad_rg_context:
		consistent = bad_ensure_regular_lcoeff 
			(tabV, tabG, type, VP->tab [0], VQ->tab [0], 
							v, true, C, ddz);
		break;
	}
	ba0_push_another_stack ();
	if (! consistent)
	    continue;
	rankfall = d != bap_degree_polynom_mpz (VQ->tab [0], v);
	if (rankfall)
	{   degP = bap_degree_polynom_mpz (VZ->tab [0], v);
	    degQ = bap_degree_polynom_mpz (VQ->tab [0], v);
	    bap_initial2_polynom_mpz (&coeff, VQ->tab [0], v);
	    bap_pow_polynom_mpz (&s, &coeff, degP - degQ);
	    bap_set_polynom_mpz (&bunk, VZ->tab [0]);
	    for (i = 0; i < n; i++)
	    {	bap_set_polynom_mpz (VP->tab [i], VQ->tab [i]);
		bap_set_polynom_mpz (VZ->tab [i], VQ->tab [i]);
	    }
	    if (n == 1)
		bap_pseudo_division_polynom_mpz 
			(BAP_NOT_A_POLYNOM_mpz, VQ->tab [0], (bav_Idegree*)0, 
			 &bunk, VQ->tab [0], v);
	    else if (n == 2)
	    {	bap_pseudo_division_polynom_mpz 
			(BAP_NOT_A_POLYNOM_mpz, VQ->tab [0], &d, 
			 &bunk, VQ->tab [0], v);
		bap_initial2_polynom_mpz (&coeff, VQ->tab [0], v);
		bap_pow_polynom_mpz (VQ->tab [1], &coeff, d);
	    } else
	    {	bap_pseudo_division_polynom_mpz 
			(VQ->tab [2], VQ->tab [0], &d, &bunk, VQ->tab [0], v);
		bap_neg_polynom_mpz (VQ->tab [2], VQ->tab [2]);
		bap_initial2_polynom_mpz (&coeff, VQ->tab [0], v);
		bap_pow_polynom_mpz (VQ->tab [1], &coeff, d);
	    }
	} else
	{   bap_initial2_polynom_mpz (&coeff, VQ->tab [0], v);
	    degQ = bap_degree_polynom_mpz (VQ->tab [0], v);
	    degP = bap_degree_polynom_mpz (VZ->tab [0], v);
	    bap_muldiv3_Lazard_polynom_mpz (VZ, VQ, &coeff, &s, degP - degQ);
	    bap_initial2_polynom_mpz (&coeff, VZ->tab [0], v);
	    bap_nsr3_Ducos_polynom_mpz (VP, VP, VQ, &s, &coeff, v);
	    BA0_SWAP (bap_tableof_polynom_mpz, VP, VQ);
	    bap_set_polynom_mpz (&s, &coeff);
	}
    }
    ba0_pull_stack ();
    if (! consistent)
    {	tabG->size--;
	tabV->size--;
    } else
    {	ba0_realloc2_table
		((ba0_table)V, n, (ba0_new_function*)bap_new_polynom_mpz);
	for (i = 0; i < n; i++)
	    bap_set_polynom_mpz (V->tab [i], VZ->tab [i]);
	V->size = n;
    }
    ba0_restore (&M);
}

static void bad_Euclid_gcd_prem_and_factor_mod_regchain
	(bap_tableof_tableof_polynom_mpz tabV, bad_tableof_quadruple G,
	 enum bad_typeof_Euclid type,
         bap_polynom_mpz A, bap_polynom_mpz B, bav_variable v,
         bool reginitA, bool reginitB,
         bad_regchain C, bap_polynom_mpz* ddz)
{
    tabV = (bap_tableof_tableof_polynom_mpz)0;
    G = (bad_tableof_quadruple)0;
    type = bad_half_extended_Euclid;
    A = (bap_polynom_mpz)0;
    B = (bap_polynom_mpz)0;
    v = (bav_variable)0;
    reginitA = true;
    reginitB = true;
    C = (bad_regchain)0;
    ddz = (bap_polynom_mpz*)0;

    BA0_RAISE_EXCEPTION (BA0_ERRNYP);
}

static void bad_Euclid_gcd_prem_mod_regchain
        (bap_tableof_tableof_polynom_mpz tabV, bad_tableof_quadruple G,
         enum bad_typeof_Euclid type,
         bap_polynom_mpz A, bap_polynom_mpz B, bav_variable v,
         bool reginitA, bool reginitB,
         bad_regchain C, bap_polynom_mpz* ddz)
{
    tabV = (bap_tableof_tableof_polynom_mpz)0;
    G = (bad_tableof_quadruple)0;
    type = bad_half_extended_Euclid;
    A = (bap_polynom_mpz)0;
    B = (bap_polynom_mpz)0;
    v = (bav_variable)0;
    reginitA = true;
    reginitB = true;
    C = (bad_regchain)0;
    ddz = (bap_polynom_mpz*)0;

    BA0_RAISE_EXCEPTION (BA0_ERRNYP);
}

/****************************************************************************
 CHECK REGULARITY
 ****************************************************************************/

static void bad_check_algebraic_regularity_mod_regchain
		(bap_polynom_mpz, bad_regchain, 
		 bap_listof_polynom_mpz, bap_polynom_mpz*);

/*
 * Tests the regularity of A modulo the ideal defined by C.
 * This function either returns or raises BAD_EXRNUL (case of A = 0)
 * or BAD_EXRDDZ (case of A | 0). In the last case, a non-trivial
 * divisor of some element of C is returned in *ddz.
 */

BAD_DLL void bad_check_regularity_polynom_mod_regchain
                (bap_polynom_mpz A, bad_regchain C, bap_polynom_mpz* ddz)
{
    enum bad_typeof_reduction type_red;
    struct bap_product_mpz P;
    struct ba0_mark M;
    ba0_int_p i;

    if (bap_is_zero_polynom_mpz (A))
	BA0_RAISE_EXCEPTION (BAD_EXRNUL);
    else if (bad_is_zero_regchain (C))
	return;
    else if (bad_defines_a_prime_ideal_regchain (C))
    {	enum bad_typeof_reduction type_red;
/*
 * If the ideal is prime then the regularity test reduces to a non zero test
 */
	type_red = bad_defines_a_differential_ideal_regchain (C) ? 
				bad_full_reduction : bad_algebraic_reduction;
	if (bad_is_a_reduced_to_zero_polynom_by_regchain (A, C, type_red))
	    BA0_RAISE_EXCEPTION (BAD_EXRNUL);
	return;
    } else
    {
/*
 * The non prime ideal case.
 * One first reduces A, then applies the algebraic test.
 */
	ba0_record (&M);

	if (bad_defines_a_differential_ideal_regchain (C))
	    type_red = bad_full_reduction;
	else
	    type_red = bad_algebraic_reduction;

	bap_init_product_mpz (&P);
	bad_reduce_polynom_by_regchain (&P, (bap_product_mpz)0, A, C, 
				    type_red, bad_all_derivatives_to_reduce);
	if (bap_is_zero_product_mpz (&P))
	    BA0_RAISE_EXCEPTION (BAD_EXRNUL);
/*
 * From now on, BAD_EXRNUL cannot be raised anymore.
 */
	for (i = 0; i < P.size; i++)
	    bad_check_algebraic_regularity_mod_regchain 
			(&P.tab [i].factor, C, (bap_listof_polynom_mpz)0, ddz);

	ba0_restore (&M);
    }
}

/*
 * Subfunction of bad_check_regularity_polynom_mod_regchain
 * Tests the regularity of A mod the algebraic ideal defined by C.
 *
 * A is nonzero modulo C.
 * The initial of A is nonzero modulo C.
 * C is not the zero regchain and should define a prime ideal.
 *
 * The function either returns or raises the exception BAD_EXRDDZ.
 * In the second case, a non-trivial divisor of some element of C
 * is returned in *ddz.
 *
 * S is a list of polynomials that might help factoring polynomials.
 */

static void bad_Euclid_check_algebraic_regularity_mod_regchain
	(bap_polynom_mpz A, bap_polynom_mpz B, bav_variable v,
	 bad_regchain C, bap_listof_polynom_mpz S, bap_polynom_mpz* ddz);

static void bad_check_algebraic_regularity_mod_regchain
		(bap_polynom_mpz A, bad_regchain C, 
		 bap_listof_polynom_mpz S, bap_polynom_mpz* ddz)
{   struct bap_polynom_mpz init;
    bav_variable v;
    bav_Idegree dA, dk;
    ba0_int_p k;

    if (! (*bad_nonzero) (A))
    {	v = bap_leader_polynom_mpz (A);
	bap_init_readonly_polynom_mpz (&init);
	bap_initial_polynom_mpz (&init, A);
	bad_check_algebraic_regularity_mod_regchain (&init, C, S, ddz);
        if (bad_is_leader_of_regchain (v, C, &k))
	{   dA = bap_leading_degree_polynom_mpz (A);
	    dk = bap_leading_degree_polynom_mpz (C->decision_system.tab [k]);
	    if (dA < dk)
                bad_Euclid_check_algebraic_regularity_mod_regchain
				(C->decision_system.tab [k], A, v, C, S, ddz);
	    else
		bad_Euclid_check_algebraic_regularity_mod_regchain
				(A, C->decision_system.tab [k], v, C, S, ddz);
	}
    }
}

/* 
 * Subfunction of bad_check_algebraic_regularity_mod_regchain
 * A and B depend on v, degree (A, v) > degree (B, v), and the initials 
 * of A and B are regular modulo the algebraic ideal defined by C.
 *
 * Either returns or raises BAD_EXRDDZ.
 *
 * Performs a variant of the Euclidean algorithm.
 */

static void bad_Euclid_check_algebraic_regularity_mod_regchain
	(bap_polynom_mpz A, bap_polynom_mpz B, bav_variable v,
	 bad_regchain C, bap_listof_polynom_mpz S, bap_polynom_mpz* ddz)
{   struct bap_product_mpz prod1, prod2;
    struct bap_polynom_mpz R, init;
    struct ba0_mark M;
    ba0_int_p i, j;

    ba0_record (&M);

    bap_init_readonly_polynom_mpz (&init);
    bap_init_polynom_mpz (&R);
    bap_init_product_mpz (&prod1);
    bap_init_product_mpz (&prod2);

    bap_gcd_prem_polynom_mpz (&R, (bap_product_mpz)0, A, B, v);
    bad_reduce_easy_polynom_by_regchain (&R, &R, C, bad_algebraic_reduction);
    bap_factor_easy_polynom_mpz (&prod1, &R, S);

    if (bap_is_zero_product_mpz (&prod1))
    {	if (ddz != (bap_polynom_mpz*)0)
	{   *ddz = B;
	    BA0_RAISE_EXCEPTION2 (BAD_EXRDDZ, "%Az", (void**)ddz);
	} else
	    BA0_RAISE_EXCEPTION (BAD_EXRDDZ);
    } 
/*
 * The remainder (prod1) of A by B seems not to be zero.
 */
    for (i = 0; i < prod1.size; i++)
    {   
#define BAD_DO_REDUCE 1
#undef BAD_DO_REDUCE
#if defined BAD_DO_REDUCE
	bad_reduce_polynom_by_regchain 
		(&prod2, (bap_product_mpz)0, &prod1.tab [i].factor, C, 
		 bad_algebraic_reduction, bad_all_derivatives_to_reduce);
#else
	if (bad_is_a_reduced_to_zero_polynom_by_regchain 
			(&prod1.tab [i].factor, C, bad_algebraic_reduction))
	    bap_set_product_zero_mpz (&prod2);
	else
	{   bad_ensure_initial_mod_regchain 
			(&prod1.tab [i].factor, &prod1.tab [i].factor, 
			 C, bad_algebraic_reduction);
	    bap_set_product_polynom_mpz (&prod2, &prod1.tab [i].factor, 1);
	}
#endif
/*
 * The above reduction is necessary. It may prove that the remainder is zero.
 * prod2 is a factorization of reduced prod1.
 */
	if (bap_is_zero_product_mpz (&prod2))
	{   if (ddz != (bap_polynom_mpz*)0)
	    {   *ddz = B;
		BA0_RAISE_EXCEPTION2 (BAD_EXRDDZ, "%Az", (void**)ddz);
	    } else
		BA0_RAISE_EXCEPTION (BAD_EXRDDZ);
	} 
/*
 * At this stage, one makes sure that no factor of prod2 can be zero mod C.
 * Thus, one makes sure that B is not a gcd of C->decision_system.tab [k] and A.
 */
	for (j = 0; j < prod2.size; j++)
	{   if (! bap_depend_polynom_mpz (&prod2.tab [j].factor, v))
		bad_check_algebraic_regularity_mod_regchain
					    (&prod2.tab [j].factor, C, S, ddz);
	    else
	    {   bap_initial_polynom_mpz (&init, &prod2.tab [j].factor);
		bad_check_algebraic_regularity_mod_regchain (&init, C, S, ddz);
		bad_Euclid_check_algebraic_regularity_mod_regchain
				      (B, &prod2.tab [j].factor, v, C, S, ddz);
	    }
	}
    }
    ba0_restore (&M);
}

/**********************************************************************
 REG CHARACTERISTIC
 **********************************************************************/

BAD_DLL void bad_reg_characteristic_regchain 
   (bad_intersectof_regchain tabC, bap_listof_polynom_mpz S, bad_regchain ideal)
{   bad_intersectof_regchain tabaux;
    bad_regchain C, D = (bad_regchain)0;
    enum bad_typeof_reduction type_red;
    bap_polynom_mpz g, h, r;
    bav_variable vg, vh;
    bool b, consider_g, consider_h;
    ba0_int_p  i, k;
    struct ba0_exception_code code;
    struct ba0_mark M;

    if (S == (bap_listof_polynom_mpz)0)
	return;

    C = tabC->inter.tab [tabC->inter.size - 1];
/*
   Only g and h are generated in the working stack
*/
    ba0_push_another_stack ();
    ba0_record (&M);

    BA0_PUSH_EXCEPTION (code);
    if (ba0_exception_is_set (code))
    {	bad_check_regularity_polynom_mod_regchain (S->value, C, &g);
/*
   The current inequation is regular w.r.t. C. One handles the other ones.
*/
	ba0_pull_exception (code);
	ba0_pull_stack ();
	ba0_restore (&M);

	bad_reg_characteristic_regchain (tabC, S->next, ideal);
	return;
    } else if (ba0_mesgerr == BAD_EXRNUL)
    {	ba0_pull_stack ();
	ba0_restore (&M);
/*
   The current inequation is zero mod C. One discards C from tabC.
*/
	tabC->inter.size--;
	return;
    } else if (ba0_mesgerr != BAD_EXRDDZ)
        BA0_RAISE_EXCEPTION (ba0_mesgerr);
/*
   The current inequation is a zero divisor mod C.
   It exhibits a factorization C->decision_system [k] = g * h
*/
    vg = bap_leader_polynom_mpz (g);
    b = bad_is_leader_of_regchain (vg, C, &k);
    if (! b)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    type_red = bad_defines_a_differential_ideal_attchain (&tabC->attrib) ?
                                bad_full_reduction : bad_algebraic_reduction;

    h = bap_new_polynom_mpz ();
    r = bap_new_polynom_mpz ();
    bap_pquo_polynom_mpz 
		(h, (bav_Idegree*)0, C->decision_system.tab [k], g, vg);

    ba0_pull_stack ();
/*
   consider_g means that the branch g = 0 must be considered.
   consider_h means that the branch h = 0 must be considered.
*/
    if (vg == bap_leader_polynom_mpz (S->value))
    {	bap_prem_polynom_mpz (r, (bav_Idegree*)0, S->value, g, vg);
	consider_g = ! bap_is_zero_polynom_mpz (g);
    } else if (ideal != (bad_regchain)0)
	consider_g = bad_is_a_reduced_to_zero_polynom_by_regchain 
							(g, ideal, type_red);
    else
	consider_g = true;

    vh = bap_leader_polynom_mpz (h);
    if (vh == bap_leader_polynom_mpz (S->value))
    {	bap_prem_polynom_mpz (r, (bav_Idegree*)0, S->value, h, vh);
	consider_h = ! bap_is_zero_polynom_mpz (h);
    } else if (ideal != (bad_regchain)0)
	consider_h = bad_is_a_reduced_to_zero_polynom_by_regchain 
							(h, ideal, type_red);
    else
	consider_h = true;
/*
   Avoids the case h = 0 must be considered and g = 0 must not.
*/ 
    if (consider_h && ! consider_g)
    {	BA0_SWAP (bap_polynom_mpz, g, h);
	BA0_SWAP (bool, consider_g, consider_h);
    }
/*
   If both branches must be considered, D = a copy of C.
*/
    if (consider_h)
	D = (bad_regchain)bad_copy_regchain (C);
/*
   tabaux is a table of regchain generated by the composition of
	the splitting (g = 0 and/or h = 0) and the quenching operation.
*/
    tabaux = bad_new_intersectof_regchain ();
    bad_set_attchain (&tabaux->attrib, &tabC->attrib);

    if (consider_g)
    {	ba0_realloc_table ((ba0_table)&tabaux->inter, tabaux->inter.size + 1);
	tabaux->inter.tab [tabaux->inter.size++] = C;
	bap_set_polynom_mpz (C->decision_system.tab [k], g);
	bad_quench_and_handle_exceptions_regchain (tabaux, k, ideal);
    }

    if (consider_h)
    {	ba0_realloc_table ((ba0_table)&tabaux->inter, tabaux->inter.size + 1);
	tabaux->inter.tab [tabaux->inter.size++] = D;
	bap_set_polynom_mpz (D->decision_system.tab [k], h);
	bad_quench_and_handle_exceptions_regchain (tabaux, k, ideal);
    }
/*
   One does not need g and h anymore
*/
    ba0_restore (&M);
/*
   One removes G from tabG (for it is rewritten by splittings) and
	one resizes tabG heuristically, to avoid many resizing
*/
    tabC->inter.size--;
    bad_realloc_intersectof_regchain 
				(tabC, tabC->inter.size + tabaux->inter.size);
/*
   Recursive calls with split regchain and the same set of inequations
   Tacky assignment is permitted.
*/
    for (i = 0; i < tabaux->inter.size; i++)
    {
	bad_realloc_intersectof_regchain (tabC, tabC->inter.size + 1);
	tabC->inter.tab [tabC->inter.size++] = tabaux->inter.tab [i];
	bad_reg_characteristic_regchain (tabC, S, ideal);
    }
}

/*****************************************************************************
 NORMAL FORM
 *****************************************************************************/

BAD_DLL void bad_normal_form_polynom_mod_regchain 
		(bap_ratfrac_mpz NF, bap_polynom_mpz A, 
		 bad_regchain C, bap_polynom_mpz* ddz)
{   struct bap_product_mpz Pbar, Qbar, H, R, U, G;
    enum bad_typeof_reduction type_red;
    struct ba0_mark M;
    ba0_mpz_t *lc;

    ba0_push_another_stack ();
    ba0_record (&M);

    type_red = bad_defines_a_differential_ideal_regchain (C) ? 
				bad_full_reduction : bad_algebraic_reduction;

    bap_init_product_mpz (&Pbar);
    bap_init_product_mpz (&Qbar);
    bap_init_product_mpz (&R);
    bap_init_product_mpz (&H);
    bap_init_product_mpz (&U);
    bap_init_product_mpz (&G);

    bap_set_product_polynom_mpz (&Pbar, A, 1);

    while (bad_is_a_reducible_product_by_regchain 
      (&Pbar, C, type_red, bad_all_derivatives_to_reduce, (ba0_int_p*)0))
    {
	bad_reduce_product_by_regchain 
		(&R, &H, &Pbar, C, type_red, bad_all_derivatives_to_reduce);
	bad_invert_product_mod_regchain (&U, &G, &H, C, ddz);

	bap_mul_product_mpz (&Pbar, &U, &R);
	bap_mul_product_mpz (&Qbar, &Qbar, &G);
    }

    bap_gcd_product_mpz ((bap_product_mpz)0, &Pbar, &Qbar, &Pbar, &Qbar);
    ba0_pull_stack ();
    bap_expand_product_mpz (&NF->numer, &Pbar);
    bap_expand_product_mpz (&NF->denom, &Qbar);
    lc = bap_numeric_initial_polynom_mpz (&NF->denom);
    if (ba0_mpz_sgn (*lc) < 0)
    {	bap_neg_polynom_mpz (&NF->numer, &NF->numer);
	bap_neg_polynom_mpz (&NF->denom, &NF->denom);
    }
    ba0_restore (&M);
}

BAD_DLL void bad_normal_form_ratfrac_mod_regchain 
		(bap_ratfrac_mpz NF, bap_ratfrac_mpz A, 
		 bad_regchain C, bap_polynom_mpz* ddz)
{   struct bap_product_mpz Pbar, Qbar, H, R, U, G;
    enum bad_typeof_reduction type_red;
    struct ba0_mark M;
    ba0_mpz_t *lc;

    ba0_push_another_stack ();
    ba0_record (&M);

    type_red = bad_defines_a_differential_ideal_regchain (C) ? 
				bad_full_reduction : bad_algebraic_reduction;

    bap_init_product_mpz (&Pbar);
    bap_init_product_mpz (&Qbar);
    bap_init_product_mpz (&R);
    bap_init_product_mpz (&H);
    bap_init_product_mpz (&U);
    bap_init_product_mpz (&G);

    bad_invert_polynom_mod_regchain (&Pbar, &Qbar, &A->denom, C, ddz);
/*
 * Then: Pbar * A.denom = Qbar mod I
 */
    bap_mul_product_polynom_mpz (&Pbar, &Pbar, &A->numer, 1);
/*
 * Pbar := Pbar * A.numer, thus, Pbar * A.denom = Qbar * A.numer mod I
 *
 * Loop invariant: Pbar * A.denom = Qbar * A.numer mod I
 */
    while (bad_is_a_reducible_product_by_regchain 
      (&Pbar, C, type_red, bad_all_derivatives_to_reduce, (ba0_int_p*)0))
    {
	bad_reduce_product_by_regchain 
		(&R, &H, &Pbar, C, type_red, bad_all_derivatives_to_reduce);
	bad_invert_product_mod_regchain (&U, &G, &H, C, ddz);
/*
 * H * Pbar = R mod I
 * U * H    = G mod I
 *
 * Thus U * H * Pbar = G * Pbar = U * R
 */
	bap_mul_product_mpz (&Pbar, &U, &R);
	bap_mul_product_mpz (&Qbar, &Qbar, &G);
/*
 * Then Pbar := U * R
 *      Qbar := Qbar * G
 */
    }

    bap_gcd_product_mpz ((bap_product_mpz)0, &Pbar, &Qbar, &Pbar, &Qbar);
    ba0_pull_stack ();
    bap_expand_product_mpz (&NF->numer, &Pbar);
    bap_expand_product_mpz (&NF->denom, &Qbar);
    lc = bap_numeric_initial_polynom_mpz (&NF->denom);
    if (ba0_mpz_sgn (*lc) < 0)
    {	bap_neg_polynom_mpz (&NF->numer, &NF->numer);
	bap_neg_polynom_mpz (&NF->denom, &NF->denom);
    }
    ba0_restore (&M);
}

/*
 * Compute the normal form of A modulo the ideal defined by the
 * ideal tabC. The result is stored on the top of tabNF.
 */

BAD_DLL void bad_normal_form_ratfrac_mod_intersectof_regchain
        (bap_tableof_ratfrac_mpz tabNF, bap_ratfrac_mpz A,
         bad_intersectof_regchain tabC, bap_polynom_mpz* ddz)
{   ba0_int_p i;

    ba0_realloc2_table ((ba0_table)tabNF, tabNF->size + tabC->inter.size,
			(ba0_new_function*)&bap_new_ratfrac_mpz);
    for (i = 0; i < tabC->inter.size; i++)
    {	bad_normal_form_ratfrac_mod_regchain 
		(tabNF->tab [tabNF->size], A, tabC->inter.tab [i], ddz);
	tabNF->size += 1;
    }
}

/*
 * Compute the normal form of A modulo the ideal defined by the
 * regchain C on the top of tabC. In most of cases, the computation does
 * not raise any exception and the result is stored on the top of tabNF.
 * 
 * In general, the computation may raise exceptions (EXRNUL, EXRDDZ), 
 * leading to splitting cases. 
 *
 * If EXRDDZ is raised then the current chain is removed from tabC and
 * split in two chains which are stored on the top of tabC. Normal form
 * computations are restarted over each branch.
 *
 * If EXRNUL is raised then the current chain is removed from tabC and
 * stored on the top of tabNUL (if tabNUL is nonzero).
 *
 * The successfully computed normal forms are stored in tabNF in the
 * same order as tabC.
 */

BAD_DLL void bad_normal_form_handling_exceptions_ratfrac_mod_regchain 
	(bap_tableof_ratfrac_mpz tabNF, bad_intersectof_regchain tabC, 
	 bad_intersectof_regchain tabNUL, bap_ratfrac_mpz A)
{   struct bad_intersectof_regchain S;
    volatile bad_regchain C;	/* volatile needed here */
    struct bap_ratfrac_mpz NF;
    bap_polynom_mpz ddz;
    struct ba0_mark M;
    struct ba0_exception_code code;

    ba0_realloc2_table 
		((ba0_table)tabNF, tabNF->size + 1, 
		 (ba0_new_function*)&bap_new_ratfrac_mpz);
/*
 * The first case is handled separately to avoid copying regchains
 */
    BA0_PUSH_EXCEPTION (code);
    if (ba0_exception_is_set (code))
    {	C = tabC->inter.tab [tabC->inter.size-1];
	bad_normal_form_ratfrac_mod_regchain 
				(tabNF->tab [tabNF->size], A, C, &ddz);
	ba0_pull_exception (code);
	return;
    } else if (ba0_mesgerr == BAD_EXRNUL)
    {	if (tabNUL != (bad_intersectof_regchain)0)
	    bad_append_intersectof_regchain_regchain (tabNUL, C);
	tabC->inter.size -= 1;
	return;
    } else if (ba0_mesgerr != BAD_EXRDDZ)
	BA0_RAISE_EXCEPTION (ba0_mesgerr);

    ba0_push_another_stack ();
    ba0_record (&M);

    bad_init_intersectof_regchain (&S);
    C = tabC->inter.tab [tabC->inter.size-1];
    tabC->inter.size -= 1;
    bad_set_intersectof_regchain_regchain (&S, C);
    bap_init_ratfrac_mpz (&NF);

    bad_handle_splitting_exceptions_regchain 
					(&S, ddz, (bad_regchain)0, BAD_EXRDDZ);
    while (S.inter.size > 0)
    {	
	BA0_PUSH_EXCEPTION (code);
	if (ba0_exception_is_set (code))
	{   C = S.inter.tab [S.inter.size - 1];
	    bad_normal_form_ratfrac_mod_regchain (&NF, A, C, &ddz);
	    ba0_pull_exception (code);

	    ba0_pull_stack ();
	    bad_append_intersectof_regchain_regchain (tabC, C);
	    ba0_realloc2_table 
			((ba0_table)tabNF, tabNF->size + S.inter.size, 
			 (ba0_new_function*)&bap_new_ratfrac_mpz);
	    bap_set_ratfrac_mpz (tabNF->tab [tabNF->size++], &NF);
	    ba0_push_another_stack ();

	    S.inter.size -= 1;
	} else if (ba0_mesgerr == BAD_EXRNUL)
	{   if (tabNUL != (bad_intersectof_regchain)0)
	    {	ba0_pull_stack ();
		bad_append_intersectof_regchain_regchain (tabNUL, C);
		ba0_push_another_stack ();
	    }

	    S.inter.size -= 1;
	} else if (ba0_mesgerr == BAD_EXRDDZ)
	    bad_handle_splitting_exceptions_regchain
				(&S, ddz, (bad_regchain)0, BAD_EXRDDZ);
	else
	    BA0_RAISE_EXCEPTION (ba0_mesgerr);
    }

    ba0_pull_stack ();
    ba0_restore (&M);
}
