#ifndef BAD_SPLITTING_CONTROL_H
#define BAD_SPLITTING_CONTROL_H 1

#include "bad_common.h"
#include "bad_attchain.h"

BEGIN_C_DECLS

enum bad_typeof_dimension_lower_bound {
	bad_no_dimension_lower_bound,
	bad_algebraic_dimension_lower_bound,
	bad_ode_dimension_lower_bound,
	bad_pde_dimension_lower_bound };

/*
 * first_branch_only
 *        If true, stops at the first consistent regular chain (if any).
 *
 * dimlb
 * 	  Permits to cut branches which lead to regular chains
 * 	  of differential dimension strictly less than the
 * 	  number of input equations.
 *
 *	  This is correct in the algebraic case and, in the case
 *	  of a single input equation. In the other cases, it is a
 *	  conjecture (Ritt, Question for investigation number 10).
 *
 * one_eq = true
 * 	  Permits to apply the above optimization, in the case
 * 	  of a single equation.
 */

typedef struct bad_splitting_control {
   bool first_branch_only;
   enum bad_typeof_dimension_lower_bound dimlb;
   bool apply_dimlb_one_eq;
} *bad_splitting_control;

/*
 * first_branch_only = false
 * dimlb = bad_algebraic_dimension_lower_bound
 * apply_dimlb_one_eq = true
 */

extern BAD_DLL void bad_init_splitting_control (bad_splitting_control);
extern BAD_DLL bad_splitting_control bad_new_splitting_control (void);
extern BAD_DLL void bad_set_splitting_control 
			(bad_splitting_control, bad_splitting_control);

extern BAD_DLL void bad_set_first_branch_only_splitting_control
			(bad_splitting_control, bool);

extern BAD_DLL void bad_set_dimension_lower_bound_splitting_control
			(bad_splitting_control, 
			enum bad_typeof_dimension_lower_bound, bool);

extern BAD_DLL bool bad_apply_dimension_lower_bound_splitting_control
			(bad_splitting_control, bap_listof_polynom_mpz,
			 bad_attchain);
END_C_DECLS

#endif /* !BAD_SPLITTING_CONTROL_H */
