#include "bad.h"

/*
 * An example borrowed from 
 * Solving Ordinary Differential Equations II
 * Stiff and Differential-Algebraic Problems
 * chapter VII, page 454
 *
 * E. Hairer, G. Wanner.
 */

int main ()
{   struct bad_intersectof_regchain ideal;
    struct bap_tableof_polynom_mpz eqns, ineqns;
    struct bad_base_field K;
    struct bad_splitting_control C;
    bav_Iordering r;
    bool b;

    bad_restart (0, 0);
    ba0_sscanf2 ("ordering (derivations = [t], blocks = [[s, y, c, x, z]])",
		 "%ordering", &r);
    bav_R_push_ordering (r);
    bad_init_intersectof_regchain (&ideal);
    ba0_sscanf2 ("intersectof_regchain ([], \
	      	  [differential, primitive, autoreduced, normalized])", 
	     	  "%intersectof_regchain", &ideal);
    ba0_init_table ((ba0_table)&eqns);
    ba0_init_table ((ba0_table)&ineqns);
/*
 * The original input system is
 * x' = 0.7*y + sin (2.5 * z)
 * y' = 1.4*x + cos (2.5 * z)
 * x^2 + y^2 = 1
 *
 * Encoding: s = sin (2.5 * z), c = cos (2.5 * z)
 * Floating point coefficients are converted into rational.
 * Denominators are then cleared.
 */
    ba0_sscanf2 ("[10*x[t] - 7*y - 10*s, 10*y[t] - 14*x - 10*c, \
		    10*s[t] - 25*z[t]*c, 10*c[t] + 25*z[t]*s, \
		    x^2 + y^2 - 1, c^2 + s^2 - 1]",
		 "%t[%Az]", &eqns);
    bad_init_base_field (&K);
    bad_init_splitting_control (&C);
    bad_Rosenfeld_Groebner (&ideal, &eqns, &ineqns, &K, &C);
#define PRINT 1
#undef PRINT
#if defined (PRINT)
    ba0_printf ("result = %intersectof_regchain_equations\n", &ideal);
#endif
    b = ideal.inter.size == 1;
    bad_terminate (ba0_init_level);
    return (b ? 0 : 1);
}
