#ifndef BAI_ODEX_INTEGRATED_FUNCTION_H
#define BAI_ODEX_INTEGRATED_FUNCTION_H 1

/*
 * The signatures of the function which evaluate the righthand side
 * of the ODE system to be integrated by explicit ODE integrators and 
 * its jacobian matrix.
 *
 * Signatures are compatible with that defined in the Gnu Scientific Library.
 */

#include "bai_common.h"
#include "bai_odex_function.h"

BEGIN_C_DECLS

typedef enum bai_odex_exit_code bai_odex_integrated_function 
					(double, double*, double*, void*);

/* 
 * double t     the time 
 * double* x    the array which contains the current values of the vars
 * double* f    the array which receives f(x)
 * void* params if the function to be integrated is generated by
 * 	 	bai_odex_generate_C_code, then params is actually a 
 * 		pointer to a struct bai_parameters (see bai_odex_function.h).
 */

typedef enum bai_odex_exit_code bai_odex_jacobianof_integrated_function
		(double, double*, double*, double*, void*);

/* 
 * double t	the time
 * double* x	the array which contains the current values of the vars
 * double* J	the area which receives the Jacobian matrix df(i)/dx(j)
 * 		regarded as a row-ordered matrix. Same convention as in 
 * 		 the Gnu Scientific Library: first occur the elements of 
 * 		 the first row, then the elements of the second row and so on.
 * double* K	the array which receives the vector df(i)/dt
 * void* params	if the function to be integrated is generated by
 * 		bai_odex_generate_C_code, then params is actually a
 * 		pointer to a struct bai_parameters (see bai_odex_function.h).
 */

END_C_DECLS

#endif /* !BAI_ODEX_INTEGRATED_FUNCTION_H */
