// The C++ version of vdp_dop853.c

#include <blad.h>

#ifdef _MSC_VER
#    include <float.h>
#    define finite _finite
#endif

static enum bai_odex_exit_code van_der_pol 
			(double t, double* x, double* f, void* params)
{   double* p;
    bai_command_function** c;

    p = ((struct_bai_params*)params)->pars.values;
    c = ((struct_bai_params*)params)->cmds.fptrs;

    f[0] = x[1];
    if (! finite (f[0]))
        return bai_odex_non_finite;
    f[1] = -x[0]*x[0]*x[1]*p[0] - x[0] + x[1]*p[0];
    if (! finite (f[1]))
        return bai_odex_non_finite;
    return bai_odex_success;
}

/*
 * End of file vdp.c
 */

int main (int argc, char** argv)
{
    bai_odex_integrated_function* fcn;
    struct_bai_dop853_workspace w;
    struct_bai_dop853_initial_values iv;
    struct_bai_dop853_errtol tol;
    struct_bai_dop853_setting setting;
    struct_bai_dop853_dense_output dow;
    struct_bai_dop853_view view;
    struct_ba0_mark M;

    bool interpolate;
    double interpolate_value;
    double t, a, b;

    bai_odex_system S;
    struct_bai_params params;
    bav_Iordering r;

    bai_restart (0, 0);
    ba0_record (&M);

    S = bai_new_odex_system ();
    ba0_sscanf2 ("ordering (derivations = [t], blocks = [[y0, y1], epsilon])",
                 "%ordering", &r);
    bav_R_push_ordering (r);

    ba0_sscanf2 
	("odex (params = [epsilon], \
		eqns = [y0[t] = y1, y1[t] = epsilon*(1 - y0^2)*y1 - y0])", 
	 "%odex", S);
    ba0_garbage ("%odex", &M, &S);
/*
    bai_odex_generate_function_C_code (stdout, "van_der_pol", S);
*/
    fcn = &van_der_pol;

    bai_init_params (&params);
    bai_set_params_odex_system (&params, S);
    bai_set_params_parameter (&params, S->params.tab [0], 1.024);
/*
   Initial values
*/
    bai_dop853_init_initial_values (&iv);
    bai_dop853_set_initial_values_time (&iv, 0.0, 20.0);
    bai_dop853_set_initial_values_variable 
		(&iv, bav_order_zero_variable (S->lhs.tab [0]), S, -3.0);
    bai_dop853_set_initial_values_variable 
		(&iv, bav_order_zero_variable (S->lhs.tab [1]), S, 2.0);
/*
   Setting 
*/
    bai_dop853_init_setting (&setting);
    bai_dop853_set_default_setting (&setting);
/*
   Tolerance
*/
    bai_dop853_init_errtol (&tol);
    bai_dop853_set_default_errtol (&tol);
/*
   View
*/
    bai_dop853_init_view (&view);
    view.view_t = true;
    bai_dop853_set_view_variable 
			(&view, bav_order_zero_variable (S->lhs.tab [0]), S);
    bai_dop853_set_view_variable 
			(&view, bav_order_zero_variable (S->lhs.tab [1]), S);
/*
   Interpolation
*/
    bai_dop853_init_dense_output (&dow);
    bai_dop853_set_dense_output_variable 
			(&dow, bav_order_zero_variable (S->lhs.tab [0]), S);
    bai_dop853_set_dense_output_variable 
			(&dow, bav_order_zero_variable (S->lhs.tab [1]), S);
    interpolate = true;
    interpolate_value = 1e-1;
/*
   Workspace
*/
    bai_dop853_init_workspace (&w);
    bai_dop853_start_workspace (&w, fcn, &iv, &tol, &setting, &dow, &params);

    t = 1.936513e+01;
    a = bai_dop853_evaluate_variable 
			(bav_order_zero_variable (S->lhs.tab [0]), t, &w, S);
    if (fabs (a - 1.256257e+00) > 1e-3)
	return 1;

    t = 8.077999e+00;
    b = bai_dop853_evaluate_variable 
			(bav_order_zero_variable (S->lhs.tab [1]), t, &w, S);
    if (fabs (b - 9.572101e-02) / 1e-1 > 1e-3)
	return 1;

    t = 1.936513e+01;
    a = bai_dop853_evaluate_variable 
			(bav_order_zero_variable (S->lhs.tab [0]), t, &w, S);
    if (fabs (a - 1.256257e+00) > 1e-3)
        return 1;

    bai_terminate (ba0_init_level);

    return 0;
}
