#include "bap_polynom_COEFF.h"
#include "bap_add_polynom_COEFF.h"
#include "bap_mul_polynom_COEFF.h"
#include "bap_prem_polynom_COEFF.h"
#include "bap_Ducos_COEFF.h"

#define BAD_FLAG_COEFF
#if !defined (BAD_FLAG_mpfr) || defined (HAVE_MPFR_H)

/*
   This a C translation of the AXIOM functions of Lionel Ducos
   (file {\tt prs.spad}). Functions suffixed by~$2$ compute single values.
   Functions suffixed by~$3$ apply to vectors.
*/

/*
 * Function {\tt next_sousResultant2} of  Lionel Ducos.
 * The polynomial $P$ has positive degree in $v$ but $Q$ and $H$ do not
 * necessarily. The variable $v$ must be greater than or equal to all
 * the variables occuring in the polynomials.
 */

BAP_DLL void bap_nsr2_Ducos_polynom_COEFF 
	(bap_polynom_COEFF R, bap_polynom_COEFF P, bap_polynom_COEFF Q, 
	 bap_polynom_COEFF H, bap_polynom_COEFF s, bav_variable v)
{   struct bap_polynom_COEFF lcP, lcQ, lcH, PP, QQ, HH, A, 
						coeff, reductum, bunk;
    bav_Idegree degP, degQ, i;
    struct bav_term T;
    struct bav_rank rg;
    struct ba0_mark M;
/*
>   next_sousResultant2(P : polR, Q : polR, Z : polR, s : R) : polR ==
    Ici, Z est note H.
*/
    ba0_push_another_stack ();
    ba0_record (&M);

    rg.var = v;
    rg.deg = 1;
    bav_init_term (&T);
    bav_set_term_rank (&T, &rg);

/*
>      (lcP, c, se) := (LC(P), LC(Q), LC(Z))
>      (d, e) := (degree(P), degree(Q))
>      (P, Q, H) := (reductum(P), reductum(Q), - reductum(Z))
>      A : polR := coefficient(P, e) * H
*/
    bap_init_readonly_polynom_COEFF (&lcP);
    bap_init_readonly_polynom_COEFF (&lcQ);
    bap_init_readonly_polynom_COEFF (&lcH);

    bap_init_readonly_polynom_COEFF (&PP);
    bap_init_readonly_polynom_COEFF (&QQ);
    bap_init_polynom_COEFF (&HH);

    bap_init_polynom_COEFF (&A);
    bap_init_polynom_COEFF (&bunk);
    bap_init_readonly_polynom_COEFF (&coeff);
    bap_init_readonly_polynom_COEFF (&reductum);

    bap_initial_and_reductum_polynom_COEFF (&lcP, &PP, P);
    bap_initial_and_reductum2_polynom_COEFF (&lcQ, &QQ, Q, v);
    bap_initial_and_reductum2_polynom_COEFF (&lcH, &coeff, H, v);
    bap_neg_polynom_COEFF (&HH, &coeff);

    degP = bap_leading_degree_polynom_COEFF (P);
    degQ = bap_degree_polynom_COEFF (Q, v);

    bap_coeff2_polynom_COEFF (&coeff, &PP, v, degQ);
    bap_mul_polynom_COEFF (&A, &HH, &coeff);
/*
>      for i in e+1..d-1 repeat
>         H := if degree(H) = e-1 then
>                 X * reductum(H) - ((LC(H) * Q) exquo c)::polR
>              else
>                 X * H
>         -- H = s_e * X^i mod S_d-1
>         A := coefficient(P, i) * H + A
*/
/*
    rg.deg = 0;
    for (i = degQ + 1; i < degP; i++)
    {	if (bap_degree_polynom_COEFF (&HH, v) == degQ - 1)
	{
	    bap_initial_and_reductum2_polynom_COEFF (&coeff, &reductum, &HH, v);
	    bap_mul_polynom_COEFF (&bunk, &QQ, &coeff);
	    bap_exquo_polynom_COEFF (&bunk, &bunk, &lcQ);
	    rg.deg++;
	    bap_submulrk_polynom_COEFF (&HH, &reductum, &rg, &bunk);
	    rg.deg = 0;
	} else
	    rg.deg++;
	bap_coeff2_polynom_COEFF (&coeff, &PP, v, i);
	bap_mul_polynom_COEFF (&bunk, &HH, &coeff);
	if (rg.deg == 0)
	    bap_add_polynom_COEFF (&A, &bunk, &A);
	else
	    bap_addmulrk_polynom_COEFF (&A, &bunk, &rg, &A);
    }
    if (rg.deg > 0)
    {	struct bav_term T;
	bav_init_term (&T);
	bav_set_term_rank (&T, &rg);
	bap_mul_polynom_term_COEFF (&HH, &HH, &T);
    }
*/
    rg.deg = 1;
    for (i = degQ + 1; i < degP; i++)
    {   if (bap_degree_polynom_COEFF (&HH, v) == degQ - 1)
	{
            bap_initial_and_reductum2_polynom_COEFF (&coeff, &reductum, &HH, v);            bap_mul_polynom_COEFF (&bunk, &QQ, &coeff);
            bap_exquo_polynom_COEFF (&bunk, &bunk, &lcQ);
            bap_submulrk_polynom_COEFF (&HH, &reductum, &rg, &bunk);
	} else
	    bap_mul_polynom_term_COEFF (&HH, &HH, &T);
	bap_coeff2_polynom_COEFF (&coeff, &PP, v, i);
	bap_mul_polynom_COEFF (&bunk, &HH, &coeff);
	bap_add_polynom_COEFF (&A, &bunk, &A);
    }
/*
>      while degree(P) >= e repeat P := reductum(P)
*/
    while (bap_degree_polynom_COEFF (&PP, v) >= degQ)
	bap_initial_and_reductum2_polynom_COEFF 
				((bap_polynom_COEFF)0, &PP, &PP, v);
/*
>      A := A + se * P            --  A = s_e * reductum(P_0)       mod S_d-1
>      A := (A exquo lcP)::polR   --  A = s_e * reductum(S_d) / s_d mod S_d-1
*/
    bap_mul_polynom_COEFF (&bunk, &PP, &lcH);
    bap_add_polynom_COEFF (&A, &A, &bunk);
    bap_exquo_polynom_COEFF (&A, &A, &lcP);
/*
>      A := if degree(H) = e-1 then
>              c * (X * reductum(H) + A) - LC(H) * Q
>           else
>              c * (X * H + A)
>      A := (A exquo s)::polR                    -- A = +/- S_e-1
*/
    if (bap_degree_polynom_COEFF (&HH, v) == degQ - 1)
    {	bap_initial_and_reductum2_polynom_COEFF (&coeff, &HH, &HH, v);
	bap_mul_polynom_COEFF (&bunk, &QQ, &coeff);
	rg.deg = 1;
	bap_addmulrk_polynom_COEFF (&A, &HH, &rg, &A);
	bap_mul_polynom_COEFF (&A, &A, &lcQ);
	bap_sub_polynom_COEFF (&A, &A, &bunk);
    } else
    {	rg.deg = 1;
	bap_addmulrk_polynom_COEFF (&A, &HH, &rg, &A);
	bap_mul_polynom_COEFF (&A, &A, &lcQ);
    }
    bap_exquo_polynom_COEFF (&A, &A, s);
    ba0_pull_stack ();
/*  
>     return (if odd?(d-e) then A else - A)
*/
    if ((degP - degQ) % 2 == 1)
	bap_set_polynom_COEFF (R, &A);
    else
	bap_neg_polynom_COEFF (R, &A);
    ba0_restore (&M);
}

/* Function {\tt next_sousResultant3} of Lionel Ducos. */

BAP_DLL void bap_nsr3_Ducos_polynom_COEFF (bap_tableof_polynom_COEFF VR, 
		bap_tableof_polynom_COEFF VP, bap_tableof_polynom_COEFF VQ, 
		bap_polynom_COEFF s, bap_polynom_COEFF ss, bav_variable v)
{   struct bap_polynom_COEFF lcP, lcQ, bunk, coeff, coegg, r, rr, quot;
    struct ba0_table VPP;
    bap_polynom_COEFF P, Q;
    bav_Idegree e, delta;
    ba0_int_p i;
    struct bav_rank rg;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);

    rg.var = v;

    bap_init_polynom_COEFF (&bunk);
    bap_init_readonly_polynom_COEFF (&coeff);
    bap_init_readonly_polynom_COEFF (&coegg);

    ba0_init_table (&VPP);
    ba0_realloc2_table 
	((ba0_table)&VPP, VP->size, (ba0_new_function*)&bap_new_polynom_COEFF);
    VPP.size = VP->size;
/*
    next_sousResultant3(VP : Vector(polR), VQ : Vector(polR), s : R, ss : R) :
                                                      Vector(polR) ==
    -- P ~ S_d,  Q = S_d-1,  s = lc(S_d),  ss = lc(S_e)
       (P, Q) := (VP.1, VQ.1)
*/
    P = VP->tab [0];
    Q = VQ->tab [0];
/*
       (lcP, c) := (LC(P), LC(Q))
       ici c est nomme lcQ.
*/
    bap_init_readonly_polynom_COEFF (&lcP);
    bap_init_readonly_polynom_COEFF (&lcQ);
    bap_initial2_polynom_COEFF (&lcP, P, v);
    bap_initial2_polynom_COEFF (&lcQ, Q, v);
/*
       e : NNI := degree(Q)
       if one?(delta := degree(P) - e) then
*/
    e = bap_degree_polynom_COEFF (Q, v);
    delta = bap_degree_polynom_COEFF (P, v) - e;
    if (delta == 1)
    {	bap_coeff2_polynom_COEFF (&coeff, P, v, e);
	bap_coeff2_polynom_COEFF (&coegg, Q, v, e - 1);
	for (i = 0; i < VP->size; i++)
	{
/*
         VP := c * VP - coefficient(P, e) * VQ
*/
	    bap_mul_polynom_COEFF (&bunk, &coeff, VQ->tab [i]);
	    bap_mul_polynom_COEFF (VPP.tab [i], VP->tab [i], &lcQ);
	    bap_sub_polynom_COEFF (VPP.tab [i], VPP.tab [i], &bunk);
/*
         VP := VP exquo lcP
*/
	    bap_exquo_polynom_COEFF (VPP.tab [i], VPP.tab [i], &lcP);
/*
	         VP := c * (VP - X * VQ) + coefficient(Q, (e-1)::NNI) * VQ
*/
	    rg.deg = 1;
	    bap_submulrk_polynom_COEFF 
				(VPP.tab [i], VQ->tab [i], &rg, VPP.tab [i]);
	    bap_mul_polynom_COEFF (VPP.tab [i], VPP.tab [i], &lcQ);
	    bap_mul_polynom_COEFF (&bunk, VQ->tab [i], &coegg);
	    bap_sub_polynom_COEFF (VPP.tab [i], &bunk, VPP.tab [i]);
/*
         VP := VP exquo s
*/
	    bap_exquo_polynom_COEFF (VPP.tab [i], VPP.tab [i], s);
	} 
    } else
    {   
/*
       else                                    -- algorithm of Lickteig - Roy
         (r, rr) := (s * lcP, ss * c)
*/
	bap_init_polynom_COEFF (&r);
	bap_init_polynom_COEFF (&rr);
	bap_mul_polynom_COEFF (&r, s, &lcP);
	bap_mul_polynom_COEFF (&rr, ss, &lcQ);
/*
         divid := divide(rr * P, Q)
*/
	bap_mul_polynom_COEFF (&bunk, &rr, P);
	bap_init_polynom_COEFF (&quot);
	bap_rem_polynom_COEFF (&quot, VPP.tab [0], &bunk, Q, v);
/*
         VP.1 := (divid.remainder exquo r)::polR
*/
	bap_exquo_polynom_COEFF (VPP.tab [0], VPP.tab [0], &r);
/*
         for i in 2..#VP repeat
*/
	for (i = 1; i < VP->size; i++)
	{
/*
           VP.i := rr * VP.i - VQ.i * divid.quotient
*/
	    bap_mul_polynom_COEFF (VPP.tab [i], &rr, VP->tab [i]);
	    bap_mul_polynom_COEFF (&bunk, VQ->tab [i], &quot);
	    bap_sub_polynom_COEFF (VPP.tab [i], VPP.tab [i], &bunk);
/*
           VP.i := (VP.i exquo r)::polR
*/
	    bap_exquo_polynom_COEFF (VPP.tab [i], VPP.tab [i], &r);
	}
    }
    ba0_pull_stack ();
    for (i = 0; i < VR->size; i++)
	bap_set_polynom_COEFF (VR->tab [i], VPP.tab [i]);
    ba0_restore (&M);
}

/*
 * Function {\tt Lazard} of Lionel Ducos.
 *
 * R := x^n/y^(n-1)
 *
 * Alternating multiplications and divisions.
 */

BAP_DLL void bap_muldiv_Lazard_polynom_COEFF 
		(bap_polynom_COEFF R, bap_polynom_COEFF x, 
		 bap_polynom_COEFF y, bav_Idegree n)
{   struct bap_polynom_COEFF c;
    bav_Idegree a;
    struct ba0_mark M;
/*
>   Lazard(x : R, y : R, n : NNI) : R ==
>      zero?(n) => error("Lazard$PRS : n = 0")
>      one?(n) => x
*/
    if (n == 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    if (n == 1)
    {	if (R != x)
	    bap_set_polynom_COEFF (R, x);
    } else
    {
/*
>      a : NNI := 1
>      while n >= (b := 2*a) repeat a := b
*/
	ba0_push_another_stack ();
	ba0_record (&M);

	for (a = 1; n >= 2*a; a *= 2);
/*
   Comme n >= 2, on est sur que a > 1 et donc qu'on entre une fois
   dans la boucle. J'ai deplie la premiere iteration.
>      c : R := x
>      n := (n - a)::NNI
>      repeat                    --  c = x**i / y**(i-1),  i=n_0 quo a,  a=2**?
>         one?(a) => return c
>         a := a quo 2
>         c := ((c * c) exquo y)::R
>         if n >= a then ( c := ((c * x) exquo y)::R ; n := (n - a)::NNI )
*/
	n = n - a;
	a /= 2;
	bap_init_polynom_COEFF (&c);
	bap_mul_polynom_COEFF (&c, x, x);
	bap_exquo_polynom_COEFF (&c, &c, y);
	if (n >= a)
	{   bap_mul_polynom_COEFF (&c, &c, x);
	    bap_exquo_polynom_COEFF (&c, &c, y);
	    n = n - a;
	}
	while (a != 1)
	{   a /= 2;
	    bap_mul_polynom_COEFF (&c, &c, &c);
	    bap_exquo_polynom_COEFF (&c, &c, y);
	    if (n >= a)
	    {	bap_mul_polynom_COEFF (&c, &c, x);
		bap_exquo_polynom_COEFF (&c, &c, y);
		n = n - a;
	    }
	}
	ba0_pull_stack ();
	bap_set_polynom_COEFF (R, &c);
	ba0_restore (&M);
    } 
}

/*
 * Function {\tt Lazard2} of Lionel Ducos.
 * Sets $R$ to $\displaystyle{F\,\left(\frac x y\right)^{(n-1)}}$.
 */

BAP_DLL void bap_muldiv2_Lazard_polynom_COEFF 
	(bap_polynom_COEFF R, bap_polynom_COEFF F, bap_polynom_COEFF x, 
	 bap_polynom_COEFF y, bav_Idegree n)
{   struct bap_polynom_COEFF A;
    struct ba0_mark M;
/*
>   Lazard2(F : polR, x : R, y : R, n : NNI) : polR ==
>      zero?(n) => error("Lazard2$PRS : n = 0")
>      one?(n) => F
*/
    if (n == 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    if (n == 1)
    {	if (R != F)
	    bap_set_polynom_COEFF (R, F);
    } else
    {
/*
>      x := Lazard(x, y, (n-1)::NNI)
>      return ((x * F) exquo y)::polR
*/
	ba0_push_another_stack ();
	ba0_record (&M);
	bap_init_polynom_COEFF (&A);
	bap_muldiv_Lazard_polynom_COEFF (&A, x, y, n-1);
	bap_mul_polynom_COEFF (&A, &A, F);
	ba0_pull_stack ();
	bap_exquo_polynom_COEFF (R, &A, y);
	ba0_restore (&M);
    }
}

/*
 * VR := VP * (x/y)^(n-1)
 */

BAP_DLL void bap_muldiv3_Lazard_polynom_COEFF (bap_tableof_polynom_COEFF VR, 
		bap_tableof_polynom_COEFF VP, bap_polynom_COEFF x, 
		bap_polynom_COEFF y, bav_Idegree n)
{   struct bap_polynom_COEFF z, t;
    ba0_int_p i;
    struct ba0_mark M;
/*
    Lazard3(V : Vector(polR), x : R, y : R, n : NNI) : Vector(polR) ==
    -- computes x**(n-1) * V / y**(n-1)
       zero?(n) => error("Lazard2$prs : n = 0")
       one?(n) => V
       x := Lazard(x, y, (n-1)::NNI)
       return ((x * V) exquo y)
*/
    if (n == 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    if (n == 1)
    {	if (VR != VP)
	{   for (i = 0; i < VR->size; i++)
		bap_set_polynom_COEFF (VR->tab [i], VP->tab [i]);
	}
    } else
    {	ba0_push_another_stack ();
	ba0_record (&M);
	bap_init_polynom_COEFF (&z);
	bap_init_polynom_COEFF (&t);
	bap_muldiv_Lazard_polynom_COEFF (&z, x, y, n - 1);
	for (i = 0; i < VR->size; i++)
	{   bap_mul_polynom_COEFF (&t, &z, VP->tab [i]);
	    ba0_pull_stack ();
	    bap_exquo_polynom_COEFF (VR->tab [i], &t, y);
	    ba0_push_another_stack ();
	}
	ba0_pull_stack ();
	ba0_restore (&M);
    }
}


/*
 * Function algo_new of Lionel Ducos
 * Polynomials P and Q have nonzero degree in v
 * v is their leading variable
 */

static void algo_new_COEFF (bap_polynom_COEFF R, bap_polynom_COEFF P, 
			    bap_polynom_COEFF Q, bav_variable v)
{   struct bap_polynom_COEFF coeff, s, Z;
    bap_polynom_COEFF A, B;
    bav_Idegree delta;
    struct ba0_mark M;

    if (bap_leader_polynom_COEFF (P) != v || bap_leader_polynom_COEFF (Q) != v)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

/*
>   algo_new(P : polR, Q : polR) : R ==
>      delta : NNI := (degree(P) - degree(Q))::NNI
>      s : R := LC(Q)**delta
>      (P, Q) := (Q, pseudoRemainder(P, -Q))
*/
    ba0_push_another_stack ();
    ba0_record (&M);
/*
ba0_printf ("algo_new (%A_, %A_, %v)\n", P, Q, v);
*/
    bap_init_readonly_polynom_COEFF (&coeff);
    bap_init_polynom_COEFF (&s);
    bap_init_polynom_COEFF (&Z);

    bap_initial_polynom_COEFF (&coeff, Q);
    delta = bap_leading_degree_polynom_COEFF (P) - 
				bap_leading_degree_polynom_COEFF (Q);
    bap_pow_polynom_COEFF (&s, &coeff, delta);

    A = bap_new_polynom_COEFF ();
    B = bap_new_polynom_COEFF ();

    bap_set_polynom_COEFF (A, Q);

    {	bool b;

	ba0_pull_stack ();
	b = Q->readonly;
	Q->readonly = false;
	bap_neg_polynom_COEFF (Q, Q);
	ba0_push_another_stack ();
	bap_prem_polynom_COEFF (B, (bav_Idegree*)0, P, Q, v);
	ba0_pull_stack ();
	bap_neg_polynom_COEFF (Q, Q);
	Q->readonly = b;
	ba0_push_another_stack ();
    }
/*
>      repeat
>         -- P = S_c-1 (except the first turn : P ~ S_c-1),
>         -- Q = S_d-1,  s = lc(S_d)
>         zero?(Q) => return 0
>         delta := (degree(P) - degree(Q))::NNI
>         Z : polR := Lazard2(Q, LC(Q), s, delta)
>         -- Z = S_e ~ S_d-1
>         zero?(degree(Z)) => return LC(Z)
>         (P, Q) := (Q, next_sousResultant2(P, Q, Z, s))
>         s := LC(Z)
*/
    for (;;)
    {	
/*
ba0_printf ("%A_\n", B);
*/
	if (bap_is_zero_polynom_COEFF (B))
	{   ba0_pull_stack ();
	    bap_set_polynom_zero_COEFF (R);
	    goto fin;
	}
/*
   Attention a B qui peut ne pas contenir v
*/
	delta = bap_leading_degree_polynom_COEFF (A) - 
					bap_degree_polynom_COEFF (B, v);
	bap_initial2_polynom_COEFF (&coeff, B, v);
	bap_muldiv2_Lazard_polynom_COEFF (&Z, B, &coeff, &s, delta);
	if (! bap_depend_polynom_COEFF (&Z, v))
	{   ba0_pull_stack ();
	    bap_set_polynom_COEFF (R, &Z);
	    goto fin;
	}
	bap_nsr2_Ducos_polynom_COEFF (A, A, B, &Z, &s, v);
	BA0_SWAP (bap_polynom_COEFF, A, B);
	bap_lcoeff_polynom_COEFF (&s, &Z, v);
    }
fin:
/*
ba0_printf ("fin\n");
*/
    ba0_restore (&M);
}

/*
 * Function {\tt resultant} of Lionel Ducos.
 * Sets $R$ to the resultant of $P$ and $Q$ w.r.t.~$v$.
 */

BAP_DLL void bap_resultant2_Ducos_polynom_COEFF 
		(bap_polynom_COEFF R, bap_polynom_COEFF P, 
		 bap_polynom_COEFF Q, bav_variable v)
{   bav_Idegree degP, degQ;
    bool oppose, b = false;

    if (bap_is_zero_polynom_COEFF (P) || bap_is_zero_polynom_COEFF (Q))
	bap_set_polynom_zero_COEFF (R);
    else
    {   degP = bap_degree_polynom_COEFF (P, v);
	degQ = bap_degree_polynom_COEFF (Q, v);
	oppose = false;
	if (degP < degQ)
	{   BA0_SWAP (bap_polynom_COEFF, P, Q);
	    BA0_SWAP (bav_Idegree, degP, degQ);
	    if (degP % 2 == 1 && degQ % 2 == 1)
	    {	oppose = true;
		b = Q->readonly;
		Q->readonly = false;
		bap_neg_polynom_COEFF (Q, Q);
	    }
	}
	if (degQ == 0)
	    bap_pow_polynom_COEFF (R, Q, degP);
	algo_new_COEFF (R, P, Q, v);
	if (oppose)
	{   bap_neg_polynom_COEFF (Q, Q);
	    Q->readonly = b;
	}
    }
}

/*
 * Variant of {\tt lastSubResultantEuclidean} of Lionel Ducos.
 * The array $\mbox{\em VR}$ has size between $1$ and $3$.
 * In the case the size is $3$, sets $\mbox{\em VR}$ to the array
 * $(G,\,U,\,V)$ where $G$ is the last nonzero subresultant of $P$ and $Q$
 * and where $U$ and $V$ satisfy 
 * $$G = P\,U + Q\,V.$$
 * If the size is $2$, only computes $(G,\,U)$. 
 * If the size is $1$ only computes $(G)$.
 */

BAP_DLL void bap_lsr3_Ducos_polynom_COEFF 
		(bap_tableof_polynom_COEFF VR, bap_polynom_COEFF P, 
		 bap_polynom_COEFF Q, bav_variable v)
{   struct bap_tableof_polynom_COEFF VZ;
    bap_tableof_polynom_COEFF VP, VQ;
    struct bap_polynom_COEFF s, coeff;
    bav_Idegree degP, degQ, degZ;
    bool b, perm;
    ba0_int_p i;
    struct ba0_mark M;

/*
    lastSubResultantEuclidean(P : polR, Q : polR) :
                    Record(coef1 : polR, coef2 : polR, subResultant : polR) ==
       zero?(Q) or zero?(P) => construct(0::polR, 0::polR, 0::polR)
*/
    if (bap_is_zero_polynom_COEFF (P) || bap_is_zero_polynom_COEFF (Q))
    {	for (i = 0; i < VR->size; i++)
	    bap_set_polynom_zero_COEFF (VR->tab [i]);
	return;
    }
/*
       if degree(P) < degree(Q) then
          l := lastSubResultantEuclidean(Q, P)
          return construct(l.coef2, l.coef1, l.subResultant)
*/
    degP = bap_degree_polynom_COEFF (P, v);
    degQ = bap_degree_polynom_COEFF (Q, v);
    if (degP < degQ)
    {	BA0_SWAP (bap_polynom_COEFF, P, Q);
	BA0_SWAP (bav_Idegree, degP, degQ);
	perm = true;
    } else
	perm = false;
/*
       if zero?(degree(Q)) then
          degP : NNI := degree(P)
          zero?(degP) =>
              error("lastSubResultantEuclidean$PRS : bav_constant_symbol polynomials")
          s : R := LC(Q)**(degP-1)::NNI
          return construct(0::polR, s::polR, s * Q)
*/
    if (degQ == 0)
    {	if (degP == 0)
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
	if (VR->size == 1)
	    bap_pow_polynom_COEFF (VR->tab [0], Q, degP);
	else if (VR->size == 2)
	{   bap_pow_polynom_COEFF (VR->tab [0], Q, degP);
	    if (perm)
		bap_pow_polynom_COEFF (VR->tab [1], Q, degP - 1);
	    else 
		bap_set_polynom_zero_COEFF (VR->tab [1]);
	} else
	{   if (perm)
	    {   bap_pow_polynom_COEFF (VR->tab [1], Q, degP - 1);
		bap_mul_polynom_COEFF (VR->tab [0], VR->tab [1], Q);
		bap_set_polynom_zero_COEFF (VR->tab [2]);
	    } else
	    {	bap_pow_polynom_COEFF (VR->tab [2], Q, degP - 1);
		bap_mul_polynom_COEFF (VR->tab [0], VR->tab [2], Q);
		bap_set_polynom_zero_COEFF (VR->tab [1]);
	    }
	}
	return;
    }
/*
       s : R := LC(Q)**(degree(P) - degree(Q))::NNI
       VP : Vector(polR) := [Q, 0::polR, 1::polR]
       pdiv := pseudoDivide(P, -Q)
       VQ : Vector(polR) := [pdiv.remainder, pdiv.coef::polR, pdiv.quotient]
       VZ : Vector(polR) := copy(VP)
*/
    ba0_push_another_stack ();
    ba0_record (&M);

    bap_init_readonly_polynom_COEFF (&coeff);
    bap_init_polynom_COEFF (&s);
    VP = (bap_tableof_polynom_COEFF)ba0_new_table ();
    VQ = (bap_tableof_polynom_COEFF)ba0_new_table ();
    ba0_init_table ((ba0_table)&VZ);
    ba0_realloc2_table 
	((ba0_table)VP, VR->size, (ba0_new_function*)&bap_new_polynom_COEFF);
    ba0_realloc2_table 
	((ba0_table)VQ, VR->size, (ba0_new_function*)&bap_new_polynom_COEFF);
    ba0_realloc2_table 
	((ba0_table)&VZ, VR->size, (ba0_new_function*)&bap_new_polynom_COEFF);
    VP->size = VR->size;
    VQ->size = VR->size;
    VZ.size = VR->size;

    bap_initial2_polynom_COEFF (&coeff, Q, v);
    bap_pow_polynom_COEFF (&s, &coeff, degP - degQ);
    bap_set_polynom_COEFF (VP->tab [0], Q);
    bap_set_polynom_COEFF (VZ.tab [0], Q);
    if (VP->size == 2)
    {	if (perm)
	{   bap_set_polynom_one_COEFF (VP->tab [1]);
	    bap_set_polynom_one_COEFF (VZ.tab [1]);
	}
    } else if (VP->size == 3)
    {	if (perm)
	{   bap_set_polynom_one_COEFF (VP->tab [1]);
	    bap_set_polynom_one_COEFF (VZ.tab [1]);
	} else
	{   bap_set_polynom_one_COEFF (VP->tab [2]);
	    bap_set_polynom_one_COEFF (VZ.tab [2]);
	}
    }

    b = Q->readonly;
    Q->readonly = false;
    bap_neg_polynom_COEFF (Q, Q);
    if (VQ->size == 1)
	bap_pseudo_division_polynom_COEFF 
		((bap_polynom_COEFF)0, VQ->tab [0], (bav_Idegree*)0, P, Q, v);
    else if (VQ->size == 2)
    {	if (perm)
	{   bap_pseudo_division_polynom_COEFF 
		(VQ->tab [1], VQ->tab [0], (bav_Idegree*)0, P, Q, v);
	} else
	{   bav_Idegree e;

	    bap_pseudo_division_polynom_COEFF
			((bap_polynom_COEFF)0, VQ->tab [0], &e, P, Q, v);
	    bap_initial2_polynom_COEFF (&coeff, Q, v);
	    bap_pow_polynom_COEFF (VQ->tab [1], &coeff, e);
	}
    } else
    {	bav_Idegree e;

	if (perm)
	{   bap_pseudo_division_polynom_COEFF 
				(VQ->tab [1], VQ->tab [0], &e, P, Q, v);
	    bap_initial2_polynom_COEFF (&coeff, Q, v);
	    bap_pow_polynom_COEFF (VQ->tab [2], &coeff, e);
	} else
	{   bap_pseudo_division_polynom_COEFF 
				(VQ->tab [2], VQ->tab [0], &e, P, Q, v);
	    bap_initial2_polynom_COEFF (&coeff, Q, v);
	    bap_pow_polynom_COEFF (VQ->tab [1], &coeff, e);
	}
    }
    bap_neg_polynom_COEFF (Q, Q);
    Q->readonly = b;
/*
       repeat
          --  VZ.1 = S_d,  VP.1 = S_{c-1},  VQ.1 = S_{d-1},  s = lc(S_d)
          --  S_{c-1} = VP.2 P_0 + VP.3 Q_0
          --  S_{d-1} = VQ.2 P_0 + VQ.3 Q_0
          --  S_d     = VZ.2 P_0 + VZ.3 Q_0
          (Q, Z) := (VQ.1, VZ.1)
          zero?(Q) => return construct(VZ.2, VZ.3, VZ.1)
          VZ := Lazard3(VQ, LC(Q), s, (degree(Z) - degree(Q))::NNI)
          zero?(degree(Q)) => return construct(VZ.2, VZ.3, VZ.1)
          ss : R := LC(VZ.1)
          (VP, VQ) := (VQ, next_sousResultant3(VP, VQ, s, ss))
          s := ss
*/
    for (;;)
    {	if (bap_is_zero_polynom_COEFF (VQ->tab [0]))
	    break;
	bap_initial2_polynom_COEFF (&coeff, VQ->tab [0], v);
	degQ = bap_degree_polynom_COEFF (VQ->tab [0], v);
	degZ = bap_degree_polynom_COEFF (VZ.tab [0], v);
	bap_muldiv3_Lazard_polynom_COEFF (&VZ, VQ, &coeff, &s, degZ - degQ);
	if (degQ == 0)
	    break;
	bap_initial2_polynom_COEFF (&coeff, VZ.tab [0], v);
	bap_nsr3_Ducos_polynom_COEFF (VP, VP, VQ, &s, &coeff, v);
	BA0_SWAP (bap_tableof_polynom_COEFF, VP, VQ);
	bap_set_polynom_COEFF (&s, &coeff);
    }
    ba0_pull_stack ();
    for (i = 0; i < VR->size; i++)
	bap_set_polynom_COEFF (VR->tab [i], VZ.tab [i]);
    ba0_restore (&M);
    return;
}

#endif
#undef BAD_FLAG_COEFF
