#include "bap_common.h"
#include "bap_polynom_mpz.h"
#include "bap_gcd_polynom_mpz.h"
#include "bap_polynom_mpzm.h"
#include "bap_polynom_mpq.h"
#include "bap_polynom_mint_hp.h"
#include "bap_parse_polynom_mpz.h"
#include "bap_parse_polynom_mpzm.h"
#include "bap_parse_polynom_mpq.h"
#include "bap_parse_polynom_mint_hp.h"
#include "bap_product_mpz.h"
#include "bap_product_mpzm.h"
#include "bap_product_mpq.h"
#include "bap_product_mint_hp.h"
#include "bap_ratfrac_mpz.h"
#include "bap_global.h"

#if HAVE_MPFR_H
#include "bap_polynom_mpfr.h"
#include "bap_product_mpfr.h"
#include "bap_parse_polynom_mpfr.h"
#endif

BAP_DLL void bap_reset_all_settings (void)
{
    bav_reset_all_settings ();
}

BAP_DLL ba0_int_p bap_ceil_log2 (ba0_int_p s)
{   ba0_int_p i;

    for (i = 0; bap_2_pow [i] < s; i++);
    return i;
}

BAP_DLL void bap_restart (ba0_int_p time_limit, ba0_int_p memory_limit)
{   enum ba0_restart_level level = ba0_actual_restart_level;
    bool no_oom, no_oot;

    ba0_get_settings_no_oot (&no_oot);
    ba0_set_settings_no_oot (true);
    ba0_get_settings_no_oom (&no_oom);
    ba0_set_settings_no_oom (true);

    bav_restart (time_limit, memory_limit);

    switch (level)
    {	case ba0_init_level:	

            {   ba0_int_p i;
                bap_2_pow [0] = 1;
                for (i = 1; i < (ba0_int_p)BA0_NBBITS_INT_P; i++)
                    bap_2_pow [i] = 2 * bap_2_pow [i-1];
            }

	    ba0_define_format_with_sizelt ("Az", 
				sizeof (struct bap_polynom_mpz),
				&bap_scanf_polynom_mpz, 
				&bap_printf_polynom_mpz,
				&bap_garbage1_polynom_mpz, 
				&bap_garbage2_polynom_mpz,
				&bap_copy_polynom_mpz);
	
	    ba0_define_format_with_sizelt ("expanded_Az", 
				sizeof (struct bap_polynom_mpz),
				&bap_scanf_expanded_polynom_mpz, 
				&bap_printf_polynom_mpz,
				&bap_garbage1_polynom_mpz, 
				&bap_garbage2_polynom_mpz,
				&bap_copy_polynom_mpz);
	
	    ba0_define_format_with_sizelt ("careful_expanded_Az", 
				sizeof (struct bap_polynom_mpz),
				&bap_scanf_careful_expanded_polynom_mpz, 
				&bap_printf_polynom_mpz,
				&bap_garbage1_polynom_mpz, 
				&bap_garbage2_polynom_mpz,
				&bap_copy_polynom_mpz);
	
	    ba0_define_format_with_sizelt ("Qz", 
				sizeof (struct bap_ratfrac_mpz),
				&bap_scanf_ratfrac_mpz, 
				&bap_printf_ratfrac_mpz,
				&bap_garbage1_ratfrac_mpz, 
				&bap_garbage2_ratfrac_mpz,
				&bap_copy_ratfrac_mpz);
	
	    ba0_define_format_with_sizelt ("expanded_Qz", 
				sizeof (struct bap_ratfrac_mpz),
				&bap_scanf_expanded_ratfrac_mpz, 
				&bap_printf_ratfrac_mpz,
				&bap_garbage1_ratfrac_mpz, 
				&bap_garbage2_ratfrac_mpz,
				&bap_copy_ratfrac_mpz);
	
	    ba0_define_format_with_sizelt ("careful_expanded_Qz",
				sizeof (struct bap_ratfrac_mpz),
                                &bap_scanf_careful_expanded_ratfrac_mpz,
                                &bap_printf_ratfrac_mpz,
                                &bap_garbage1_ratfrac_mpz,
                                &bap_garbage2_ratfrac_mpz,
                                &bap_copy_ratfrac_mpz);

	    ba0_define_format_with_sizelt ("Azm", 
				sizeof (struct bap_polynom_mpzm),
				&bap_scanf_polynom_mpzm, 
				&bap_printf_polynom_mpzm,
				&bap_garbage1_polynom_mpzm, 
				&bap_garbage2_polynom_mpzm,
				&bap_copy_polynom_mpzm);
	
	    ba0_define_format_with_sizelt ("expanded_Azm", 
				sizeof (struct bap_polynom_mpz),
				&bap_scanf_expanded_polynom_mpzm, 
				&bap_printf_polynom_mpzm,
				&bap_garbage1_polynom_mpzm, 
				&bap_garbage2_polynom_mpzm,
				&bap_copy_polynom_mpzm);

	    ba0_define_format_with_sizelt ("careful_expanded_Azm", 
				sizeof (struct bap_polynom_mpzm),
				&bap_scanf_careful_expanded_polynom_mpzm, 
				&bap_printf_polynom_mpzm,
				&bap_garbage1_polynom_mpzm, 
				&bap_garbage2_polynom_mpzm,
				&bap_copy_polynom_mpzm);
	
	    ba0_define_format_with_sizelt ("Aq", 
				sizeof (struct bap_polynom_mpq),
				&bap_scanf_polynom_mpq, 
				&bap_printf_polynom_mpq,
				&bap_garbage1_polynom_mpq, 
				&bap_garbage2_polynom_mpq,
				&bap_copy_polynom_mpq);
	
	    ba0_define_format_with_sizelt ("expanded_Aq", 
				sizeof (struct bap_polynom_mpq),
				&bap_scanf_expanded_polynom_mpq, 
				&bap_printf_polynom_mpq,
				&bap_garbage1_polynom_mpq, 
				&bap_garbage2_polynom_mpq,
				&bap_copy_polynom_mpq);
	
	    ba0_define_format_with_sizelt ("careful_expanded_Aq", 
				sizeof (struct bap_polynom_mpq),
				&bap_scanf_careful_expanded_polynom_mpq, 
				&bap_printf_polynom_mpq,
				&bap_garbage1_polynom_mpq, 
				&bap_garbage2_polynom_mpq,
				&bap_copy_polynom_mpq);
	
	    ba0_define_format_with_sizelt ("Aim", 
				sizeof (struct bap_polynom_mint_hp),
				&bap_scanf_polynom_mint_hp, 
				&bap_printf_polynom_mint_hp,
				&bap_garbage1_polynom_mint_hp, 
				&bap_garbage2_polynom_mint_hp,
				&bap_copy_polynom_mint_hp);

	    ba0_define_format_with_sizelt ("expanded_Aim", 
				sizeof (struct bap_polynom_mint_hp),
				&bap_scanf_expanded_polynom_mint_hp, 
				&bap_printf_polynom_mint_hp,
				&bap_garbage1_polynom_mint_hp, 
				&bap_garbage2_polynom_mint_hp,
				&bap_copy_polynom_mint_hp);
	
	    ba0_define_format_with_sizelt ("careful_expanded_Aim", 
				sizeof (struct bap_polynom_mint_hp),
				&bap_scanf_careful_expanded_polynom_mint_hp, 
				&bap_printf_polynom_mint_hp,
				&bap_garbage1_polynom_mint_hp, 
				&bap_garbage2_polynom_mint_hp,
				&bap_copy_polynom_mint_hp);
	
	    ba0_define_format_with_sizelt ("Pz", 
				sizeof (struct bap_product_mpz),
				&bap_scanf_product_mpz, 
				&bap_printf_product_mpz,
				&bap_garbage1_product_mpz, 
				&bap_garbage2_product_mpz,
				&bap_copy_product_mpz);
	
	    ba0_define_format_with_sizelt ("Pzm", 
				sizeof (struct bap_product_mpzm),
				&bap_scanf_product_mpzm, 
				&bap_printf_product_mpzm,
				&bap_garbage1_product_mpzm, 
				&bap_garbage2_product_mpzm,
				&bap_copy_product_mpzm);
	
	    ba0_define_format_with_sizelt ("Pq", 
				sizeof (struct bap_product_mpq),
				&bap_scanf_product_mpq, 
				&bap_printf_product_mpq,
				&bap_garbage1_product_mpq, 
				&bap_garbage2_product_mpq,
				&bap_copy_product_mpq);
	
	    ba0_define_format_with_sizelt ("Pim", 
				sizeof (struct bap_product_mint_hp),
				&bap_scanf_product_mint_hp, 
				&bap_printf_product_mint_hp,
				&bap_garbage1_product_mint_hp, 
				&bap_garbage2_product_mint_hp,
				&bap_copy_product_mint_hp);
#if HAVE_MPFR_H
	    ba0_define_format_with_sizelt ("Afr",
				sizeof (struct bap_polynom_mpfr),
				&bap_scanf_polynom_mpfr,
				&bap_printf_polynom_mpfr,
				&bap_garbage1_polynom_mpfr,
				&bap_garbage2_polynom_mpfr,
				&bap_copy_polynom_mpfr);

	    ba0_define_format_with_sizelt ("expanded_Afr",
				sizeof (struct bap_polynom_mpfr),
				&bap_scanf_expanded_polynom_mpfr,
				&bap_printf_polynom_mpfr,
				&bap_garbage1_polynom_mpfr,
				&bap_garbage2_polynom_mpfr,
				&bap_copy_polynom_mpfr);

	    ba0_define_format_with_sizelt ("careful_expanded_Afr",
				sizeof (struct bap_polynom_mpfr),
				&bap_scanf_careful_expanded_polynom_mpfr,
				&bap_printf_polynom_mpfr,
				&bap_garbage1_polynom_mpfr,
				&bap_garbage2_polynom_mpfr,
				&bap_copy_polynom_mpfr);

	    ba0_define_format_with_sizelt ("Pfr",
				sizeof (struct bap_product_mpfr),
				&bap_scanf_product_mpfr,
				&bap_printf_product_mpfr,
				&bap_garbage1_product_mpfr,
				&bap_garbage2_product_mpfr,
				&bap_copy_product_mpfr);

	    ba0_define_format_with_sizelt ("gcd_data",
				sizeof (struct bap_gcd_data),
				0,
				&bap_printf_gcd_data,
				0, 0, 0);
#endif
	case ba0_reset_level:

	    ba0_push_exception_extra_stack 
		(&bap_clot_parameters_sp, (ba0_extra_restore_function*)0);
	    bap_reset_clot_parameters_stack ();
	    bap_push_clot_parameters_stack 
		(bap_zipping_termanager, bap_decreasing_monom_sequence);

	case ba0_done_level:
	    break;
    }
    ba0_set_settings_no_oom (no_oom);
    ba0_set_settings_no_oot (no_oot);
}

BAP_DLL void bap_terminate (enum ba0_restart_level level)
{
    bav_terminate (level);
}

