#include "bap_creator_COEFF.h"

#define BAD_FLAG_COEFF
#if !defined (BAD_FLAG_mpfr) || defined (HAVE_MPFR_H)

/****************************************************************************
 CREATORS OF MONOMIALS

 They permit us to create polynomials monomials by monomials.
 ****************************************************************************/

/*
 * Initializes the creator at the beginning of $A$.
 * If allocating some {\tt bap_mont_COEFF} in the clot of $A$ is necessary,
 * they will have size {\em nbmon_par_mont} monomials.
 * The parameter {\em type} indicates if $T$ is the exact total rank
 * of the resulting polynomial or is only a bound.
 */

BAP_DLL void bap_begin_creator_COEFF 
		       (bap_creator_COEFF crea, bap_polynom_COEFF A, 
			bav_term T, enum bap_typeof_total_rank type, 
			ba0_int_p nbmon_par_mont)
{
    if (A == (bap_polynom_COEFF)0 || A->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);
    bap_begin_creator_clot_COEFF (&crea->crea, A->clot, T, nbmon_par_mont);
    if (type == bap_exact_total_rank)
	bav_set_term (&A->rang_total, T);
    crea->polfi = A;
    crea->type = type;
}

/* Initializes the creator at the end of $A$. */

BAP_DLL void bap_append_creator_COEFF 
	(bap_creator_COEFF crea, bap_polynom_COEFF A, ba0_int_p nbmon_par_mont)
{
    if (A == (bap_polynom_COEFF)0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    if (A->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);
    bap_append_creator_clot_COEFF (&crea->crea, A->clot, nbmon_par_mont);
    crea->polfi = A;
    crea->type = bap_exact_total_rank;
}

/* Writes c*T on crea.  The coefficient c may be zero. */

BAP_DLL void bap_write_creator_COEFF 
			(bap_creator_COEFF crea, bav_term T, COEFF_t c)
{
    bap_write_creator_clot_COEFF (&crea->crea, T, c);
}

/* Writes - c*T on crea.  The coefficient c may be zero. */

BAP_DLL void bap_write_neg_creator_COEFF 
			(bap_creator_COEFF crea, bav_term T, COEFF_t c)
{
    bap_write_neg_creator_clot_COEFF (&crea->crea, T, c);
}

/*
 * Assume $A$ needs to be written on $\ast\mbox{\em crea}$ without
 * modifying the terms of $A$ (i.e. {\tt set}, {\tt mul_cste}, \dots). 
 * Returns true if this operation can be performed using the variants
 * of {\tt bap_write_all_creator_COEFF}. This function only checks the
 * structure of the polynomials. For some operations, some more operations
 * may be needed (checking the fact that the coefficient ring is domain
 * in the case of the multiplication by a constant).
 */

BAP_DLL bool bap_is_write_allable_creator_COEFF 
				(bap_creator_COEFF crea, bap_polynom_COEFF A)
{
    return crea->polfi->clot->ordering == A->clot->ordering &&
           crea->polfi->clot->sens == A->clot->sens &&
	   bap_are_disjoint_polynom_COEFF (crea->polfi, A) &&
           bap_equal_termanager (&crea->polfi->clot->tgest, &A->clot->tgest) &&
           A->acces == bap_sequential_monom_access &&
           bap_identity_selection (&A->select, A->clot->ordering);
}

/* Writes $A$ on $\ast\mbox{\em crea}$ efficiently. */

BAP_DLL void bap_write_all_creator_COEFF 
				(bap_creator_COEFF crea, bap_polynom_COEFF A)
{
    bap_write_all_creator_clot_COEFF 
			(&crea->crea, A->clot, A->seq.first, A->seq.after);
}

/* Write $- A$ on $\ast\mbox{\em crea}$ efficiently. */

BAP_DLL void bap_write_neg_all_creator_COEFF 
				(bap_creator_COEFF crea, bap_polynom_COEFF A)
{
    bap_write_neg_all_creator_clot_COEFF 
			(&crea->crea, A->clot, A->seq.first, A->seq.after);
}

/* Writes $c\,A$ on $\ast\mbox{\em crea}$ efficiently. */

BAP_DLL void bap_write_mul_all_creator_COEFF 
		(bap_creator_COEFF crea, bap_polynom_COEFF A, COEFF_t c)
{
    bap_write_mul_all_creator_clot_COEFF
			(&crea->crea, A->clot, c, A->seq.first, A->seq.after);
}

#if defined (BAD_FLAG_mpz)
/* Writes $A / c$  on $\ast\mbox{\em crea}$ efficiently. */

BAP_DLL void bap_write_exquo_all_creator_COEFF
		(bap_creator_COEFF crea, bap_polynom_COEFF A, COEFF_t c)
{
    bap_write_exquo_all_creator_clot_COEFF
                        (&crea->crea, A->clot, c, A->seq.first, A->seq.after);
}
#endif

/* Concludes the construction of the polynomial. */

BAP_DLL void bap_close_creator_COEFF (bap_creator_COEFF crea)
{   bap_polynom_COEFF A;

    bap_close_creator_clot_COEFF (&crea->crea);
    
    A = crea->polfi;

    A->acces = bap_sequential_monom_access;
    A->seq.first = 0;
    A->seq.after = A->clot->size;

    if (crea->type == bap_approx_total_rank)
	bap_set_total_rank_polynom_COEFF (A);

    bap_init_selection (&A->select, (bav_variable)-1, A->clot->ordering);
}

#endif
#undef BAD_FLAG_COEFF
