#include "bap_polynom_COEFF.h"
#include "bap_creator_COEFF.h"
#include "bap_itermon_COEFF.h"
#include "bap_itercoeff_COEFF.h"
#include "bap_add_polynom_COEFF.h"
#include "bap_mul_polynom_COEFF.h"
#include "bap_eval_polynom_COEFF.h"
#include "bap_geobucket_COEFF.h"
#include "bap__check_COEFF.h"

#define BAD_FLAG_COEFF
#if defined (BAD_FLAG_mpz) || defined (BAD_FLAG_mpzm) || defined (BAD_FLAG_mpq)

/*
 * Sets R to the polynomial obtained by evaluating A at v = c.
 */

BAP_DLL void bap_eval_polynom_numeric_COEFF 
	(bap_polynom_COEFF R, bap_polynom_COEFF A, bav_variable v, COEFF_t c)
{   struct bap_polynom_COEFF B, C, E;
    struct bap_itercoeff_COEFF  iter;
    struct bav_term T;
    bav_Idegree d;
    bav_Iordering r;
    struct ba0_mark M;

    bap__check_ordering_COEFF (A);
    if (R->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    r = bav_R_copy_ordering (bav_R_Iordering ());
    bav_R_push_ordering (r);
    bav_R_set_maximal_variable (v);

    ba0_push_another_stack ();
    ba0_record (&M);

    bav_init_term (&T);

    bap_init_polynom_COEFF (&B);
    bap_init_polynom_COEFF (&C);
    bap_init_polynom_COEFF (&E);
    bap_sort_polynom_COEFF (&B, A);

    if (! bap_is_numeric_polynom_COEFF (&B) && 
					bap_leader_polynom_COEFF (&B) == v)
    {	bap_begin_itercoeff_COEFF (&iter, &B, v);
	bap_coeff_itercoeff_COEFF (&C, &iter);
	bap_term_itercoeff_COEFF (&T, &iter);
	bap_set_polynom_COEFF (&E, &C);
/*
   Horner scheme
*/
	for (d = bav_degree_term (&T, v) - 1; d >= 0; d--)
	{   bap_mul_polynom_numeric_COEFF (&E, &E, c);
	    bav_set_term_variable (&T, v, d);
	    bap_seek_coeff_itercoeff_COEFF (&C, &iter, &T);
	    if (! bap_is_zero_polynom_COEFF (&C))
		bap_add_polynom_COEFF (&E, &E, &C);
	}
	bap_close_itercoeff_COEFF (&iter);
	ba0_pull_stack ();
	bap_set_polynom_COEFF (R, &E);
    } else
    {	ba0_pull_stack ();
	if (R != A)
	   bap_set_polynom_COEFF (R, A);
    }
    ba0_restore (&M);
    bav_R_pull_ordering ();
    bav_R_free_ordering (r);
    bap_physort_polynom_COEFF (R, bap_formal_transform);
}

BAP_DLL void bap_eval_polynom_polynom_COEFF 
	(bap_polynom_COEFF R, bap_polynom_COEFF A, 
	 bav_variable v, bap_polynom_COEFF val)
{   struct bap_polynom_COEFF G, B, coeff, value;
    struct bap_itercoeff_COEFF iter;
    struct bav_term T;
    bav_Idegree d;
    bav_Iordering r;
    struct ba0_mark M;

    bap__check_ordering_COEFF (A);
    if (R->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    r = bav_R_copy_ordering (bav_R_Iordering ());
    bav_R_push_ordering (r);
    bav_R_set_maximal_variable (v);

    ba0_push_another_stack ();
    ba0_record (&M);

    bav_init_term (&T);

    bap_init_readonly_polynom_COEFF (&B);
    bap_init_readonly_polynom_COEFF (&coeff);
    bap_init_readonly_polynom_COEFF (&value);
    bap_init_polynom_COEFF (&G);
    bap_sort_polynom_COEFF (&B, A);
    bap_sort_polynom_COEFF (&value, val);

    if (! bap_is_numeric_polynom_COEFF (&B) && 
					bap_leader_polynom_COEFF (&B) == v)
    {	bap_begin_itercoeff_COEFF (&iter, &B, v);
        bap_coeff_itercoeff_COEFF (&coeff, &iter);
        bap_term_itercoeff_COEFF (&T, &iter);
        bap_set_polynom_COEFF (&G, &coeff);
/*
   Horner scheme
*/
        for (d = bav_degree_term (&T, v) - 1; d >= 0; d--)
        {   bap_mul_polynom_COEFF (&G, &G, &value);
            bav_set_term_variable (&T, v, d);
            bap_seek_coeff_itercoeff_COEFF (&coeff, &iter, &T);
            if (! bap_is_zero_polynom_COEFF (&coeff))
                bap_add_polynom_COEFF (&G, &G, &coeff);
        }
	bap_close_itercoeff_COEFF (&iter);
	ba0_pull_stack ();
	bap_set_polynom_COEFF (R, &G);
    } else
    {	ba0_pull_stack ();
	if (R != A)
	   bap_set_polynom_COEFF (R, A);
    }
    ba0_restore (&M);
    bav_R_pull_ordering ();
    bav_R_free_ordering (r);
    bap_physort_polynom_COEFF (R, bap_definitive_transform);
}
#endif

#if defined (BAD_FLAG_mpz) || defined (BAD_FLAG_mpzm)
/*
 * Assigns A mod (x = alpha) to R
 */

BAP_DLL void bap_eval_polynom_ev_value_COEFF 
	(bap_polynom_COEFF R, bap_polynom_COEFF A, struct bav_ev_value val)
{   COEFF_t c;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);
    COEFF_init_set_si (c, val.value);
    ba0_pull_stack ();

    bap_eval_polynom_numeric_COEFF (R, A, val.var, c);
    ba0_restore (&M);
}

/*
 * Sets to point an evaluation point the variables of which being
 * that of A and the values of which being zero.
 * Variables appear by decreasing order.
 */

BAP_DLL void bap_set_ev_point_polynom_COEFF 
				(bav_ev_point point, bap_polynom_COEFF A)
{   ba0_int_p i;

    bav_realloc_ev_point (point, A->rang_total.size);
    for (i = 0; i < A->rang_total.size; i++)
    {	point->tab [i].var = A->rang_total.rg [i].var;
	point->tab [i].value = 0;
    }
    point->size = A->rang_total.size;
}

/*
 * Assigns A mod point to *value
 */
 
BAP_DLL void bap_eval_polynom_ev_point_COEFF 
		(COEFF_t* value, bap_polynom_COEFF A, bav_ev_point point)
{   struct bap_itermon_COEFF iter;
    COEFF_t v, p;
    struct bav_term T;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);

    bav_init_term (&T);
    COEFF_init (v);
    COEFF_init (p);

    bap_begin_itermon_COEFF (&iter, A);
    while (! bap_outof_itermon_COEFF (&iter))
    {	bap_term_itermon_COEFF (&T, &iter);
	bav_term_at_ev_point (p, &T, point);
	COEFF_mul (p, *bap_coeff_itermon_COEFF (&iter), p);
	COEFF_add (v, v, p);
	bap_next_itermon_COEFF (&iter);
    }
    ba0_pull_stack ();
    COEFF_set (*value, v);
    ba0_restore (&M);
}

/*
 * Assigns A mod point to R.
 * It is assumed that point gives values to all the variables of A
 * apart the leader.
 */

BAP_DLL void bap_evalcoeff_polynom_ev_point_COEFF 
		(bap_polynom_COEFF R, bap_polynom_COEFF A, bav_ev_point point)
{   struct bap_creator_COEFF crea;
    struct bap_itercoeff_COEFF iter;
    struct bap_polynom_COEFF C;
    bap_polynom_COEFF P;
    struct bav_term T;
    COEFF_t c;
    struct bav_rank rg;
    struct ba0_mark M;

    if (R->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    if (bap_is_numeric_polynom_COEFF (A))
    {	if (A != R)
	    bap_set_polynom_COEFF (R, A);
	return;
    }

    ba0_push_another_stack ();
    ba0_record (&M);

    bap_init_polynom_COEFF (&C);
    COEFF_init (c);

    rg = bap_rank_polynom_COEFF (A);
    bav_init_term (&T);
    bav_set_term_rank (&T, &rg);
    P = bap_new_polynom_COEFF ();
    bap_begin_creator_COEFF (&crea, P, &T, bap_approx_total_rank, rg.deg);
    bap_begin_itercoeff_COEFF (&iter, A, rg.var);
    while (! bap_outof_itercoeff_COEFF (&iter))
    {	bap_coeff_itercoeff_COEFF (&C, &iter);
	bap_term_itercoeff_COEFF (&T, &iter);
	bap_eval_polynom_ev_point_COEFF (&c, &C, point);
	if (COEFF_sgn (c) != 0)
	    bap_write_creator_COEFF (&crea, &T, c);
	bap_next_itercoeff_COEFF (&iter);
    }
    bap_close_creator_COEFF (&crea);
    ba0_pull_stack ();
    bap_set_polynom_COEFF (R, P);
    ba0_restore (&M);
}

#endif
#undef BAD_FLAG_COEFF
