#include "bap_add_polynom_COEFF.h"
#include "bap_mul_polynom_COEFF.h"
#include "bap_creator_COEFF.h"
#include "bap_itermon_COEFF.h"
#include "bap_geobucket_COEFF.h"
#include "bap_global.h"

#define BAD_FLAG_COEFF
#if !defined (BAD_FLAG_mpfr) || defined (HAVE_MPFR_H)

/*
 * A geobucket (JSC (1998) 25, 285-293, Thomas Yan) is an array of polynomials
 * representing the sum of these polynomials.
 * The goal consists in implementing lazy sums.
 * The polynomial at index~$i$ is a sum of at most $2^i$ monomials.
 * One manages to have $\mbox{\tt alloc} > \mbox{\tt size}$ except
 * for empty geobuckets.
 */

BAP_DLL void bap_init_geobucket_COEFF (bap_geobucket_COEFF geo)
{
    ba0_init_table ((ba0_table)geo);
}

BAP_DLL void bap_reset_geobucket_COEFF (bap_geobucket_COEFF geo)
{
    ba0_reset_table ((ba0_table)geo);
}

/*
 * geo = geo * A
 */

BAP_DLL void bap_mul_geobucket_COEFF 
				(bap_geobucket_COEFF geo, bap_polynom_COEFF A)
{   ba0_int_p i, j, n;

    if (bap_is_zero_polynom_COEFF (A))
	bap_reset_geobucket_COEFF (geo);
    else if (! bap_is_one_polynom_COEFF (A))
    {	
/*
 * First multiply
 */
	for (i = 0; i < geo->size; i++)
	    if (! bap_is_zero_polynom_COEFF (geo->tab [i]))
	    	bap_mul_polynom_COEFF (geo->tab [i], geo->tab [i], A);
/*
 * The numbers of monomials changed. Move entries.
 */
	i = geo->size - 1;
	while (i >= 0)
	{   
	    n = bap_nbmon_polynom_COEFF (geo->tab [i]);
	    if (n > 0)
	    {
/*
 * Move tab[i] to tab [j]. Find j.
 */
	        j = i;
	        while (n > bap_2_pow [j])
		    j += 1;

	        if (j >= geo->size)
	        {   ba0_realloc2_table ((ba0_table)geo, j + 2,
				(ba0_new_function*)&bap_new_polynom_COEFF);
		    geo->size = j+1;
	        }

	        if (j != i)
	        {   if (bap_is_zero_polynom_COEFF (geo->tab [j]))
		    {   BA0_SWAP (bap_polynom_COEFF, 
					geo->tab [i], geo->tab [j]);
		        i -= 1;
		    } else
		    {   bap_add_polynom_COEFF 
				(geo->tab [j], geo->tab [j], geo->tab [i]);
		        bap_set_polynom_zero_COEFF (geo->tab [i]);
/*
 * The sum may, also, have changed the number of monomials.
 * Restart from j.
 */
		        i = j;
		    }
		} else
		    i -= 1;
	    } else
		i -= 1;
	}
    }
#if defined (DEBUG)
    for (i = 0; i < geo->size; i++)
	if (nbmon_polynom_COEFF (geo->tab [i]) >= bap_2_pow [i])
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
#endif
}

/*
 * geo = geo * c. Assume a domain.
 */

BAP_DLL void bap_mul_geobucket_numeric_COEFF 
				(bap_geobucket_COEFF geo, COEFF_t c)
{   ba0_int_p i;

    if (COEFF_sgn (c) == 0)
	bap_reset_geobucket_COEFF (geo);
    else if (COEFF_cmp_ui (c, 1) != 0)
    {	for (i = 0; i < geo->size; i++)
	    if (! bap_is_zero_polynom_COEFF (geo->tab [i]))
		bap_mul_polynom_numeric_COEFF (geo->tab [i], geo->tab [i], c);
    }
}

/*
 * geo = geo + A
 */

BAP_DLL void bap_add_geobucket_COEFF 
				(bap_geobucket_COEFF geo, bap_polynom_COEFF A)
{   ba0_int_p i;

    i = bap_ceil_log2 (bap_nbmon_polynom_COEFF (A));
    if (i >= geo->size)
    {	ba0_realloc2_table ((ba0_table)geo, i + 2, 
				(ba0_new_function*)&bap_new_polynom_COEFF);
	bap_set_polynom_COEFF (geo->tab [i], A);
	geo->size = i + 1;
    } else
    {	bap_add_polynom_COEFF (geo->tab [i], geo->tab [i], A);
	while (bap_nbmon_polynom_COEFF (geo->tab [i]) > bap_2_pow [i])
	{   bap_add_polynom_COEFF 
				(geo->tab [i+1], geo->tab [i+1], geo->tab [i]);
	    bap_set_polynom_zero_COEFF (geo->tab [i]);
	    i += 1;
        }
/*
   i is the last index used
*/
	if (i >= geo->size)
	{   geo->size = i + 1;
	    ba0_realloc2_table ((ba0_table)geo, i + 2, 
				(ba0_new_function*)&bap_new_polynom_COEFF);
	}
    }
}

/*
 * geo = geo - A
 */

BAP_DLL void bap_sub_geobucket_COEFF 
				(bap_geobucket_COEFF geo, bap_polynom_COEFF A)
{   ba0_int_p i;

    i = bap_ceil_log2 (bap_nbmon_polynom_COEFF (A));
    if (i >= geo->size)
    {	ba0_realloc2_table ((ba0_table)geo, i + 2, 
				(ba0_new_function*)&bap_new_polynom_COEFF);
	bap_neg_polynom_COEFF (geo->tab [i], A);
	geo->size = i + 1;
    } else
    {	bap_sub_polynom_COEFF (geo->tab [i], geo->tab [i], A);
	while (bap_nbmon_polynom_COEFF (geo->tab [i]) > bap_2_pow [i])
	{   bap_add_polynom_COEFF 
				(geo->tab [i+1], geo->tab [i+1], geo->tab [i]);
	    bap_set_polynom_zero_COEFF (geo->tab [i]);
	    i += 1;
        }
/*
   i is the last used index
*/
	if (i >= geo->size)
	{   geo->size = i + 1;
	    ba0_realloc2_table ((ba0_table)geo, i + 2, 
				(ba0_new_function*)&bap_new_polynom_COEFF);
	}
    }
}

/*
 * A = geo
 */

BAP_DLL void bap_set_polynom_geobucket_COEFF 
				(bap_polynom_COEFF A, bap_geobucket_COEFF geo)
{   struct bap_creator_COEFF crea;
    bap_itermon_COEFF iter;
    struct bav_term *T, U;
    enum ba0_compare_code code;
    COEFF_t *lcm, *lc;
    ba0_int_p im, i, deb, fin;
    bool found;
    struct ba0_mark M;

    if (geo->size == 0)
    {	bap_set_polynom_zero_COEFF (A);
	return;
    }

    ba0_push_another_stack ();
    ba0_record (&M);

    iter = (bap_itermon_COEFF)ba0_alloc 
			(sizeof (struct bap_itermon_COEFF) * geo->size);
    T = (struct bav_term*)ba0_alloc (sizeof (struct bav_term) * geo->size);

    bav_init_term (&U);

    deb = 0; 
    while (deb < geo->size && bap_is_zero_polynom_COEFF (geo->tab [deb]))
	deb++;
    fin = geo->size - 1;
    while (fin >= deb && bap_is_zero_polynom_COEFF (geo->tab [fin]))
	fin--;

    for (i = deb; i <= fin; i++)
    {	bap_begin_itermon_COEFF (&iter [i], geo->tab [i]);
	bav_init_term (&T [i]);
	if (! bap_is_zero_polynom_COEFF (geo->tab [i]))
	    bap_term_itermon_COEFF (&T [i], &iter [i]);
	bav_lcm_term (&U, &U, &geo->tab [i]->rang_total);
    }

    ba0_pull_stack ();
    bap_begin_creator_COEFF (&crea, A, &U, bap_approx_total_rank, 
				bap_nbmon_polynom_COEFF (geo->tab [fin]));
    ba0_push_another_stack ();
/*
 * To avoid some pointless warnings (or use -Wno-uninitialized)
 */
    lcm = (COEFF_t*)0;
    im = 0;

    while (deb <= fin)
    {	found = false;
	while (! found && deb <= fin)
	{   im = deb;
	    if (bap_outof_itermon_COEFF (&iter [im]))
		BA0_RAISE_EXCEPTION (BA0_ERRALG);
	    lcm = bap_coeff_itermon_COEFF (&iter [im]);
	    found = true;
	    for (i = im + 1; found && i <= fin; i++)
	    {	if (bap_outof_itermon_COEFF (&iter [i]))
		    continue;
		code = bav_compare_term (&T [im], &T [i]);
		if (code == ba0_lt)
		{   im = i;
		    lcm = bap_coeff_itermon_COEFF (&iter [im]);
		} else if (code == ba0_eq)
		{   lc = bap_coeff_itermon_COEFF (&iter [i]);
		    COEFF_add (*lc, *lc, *lcm);
		    bap_next_itermon_COEFF (&iter [im]);
		    if (! bap_outof_itermon_COEFF (&iter [im]))
			bap_term_itermon_COEFF (&T [im], &iter [im]);
		    im = i;
		    lcm = lc;
		    if (COEFF_sgn (*lcm) == 0)
		    {	found = false;
			bap_next_itermon_COEFF (&iter [im]);
			if (! bap_outof_itermon_COEFF (&iter [im]))
			    bap_term_itermon_COEFF (&T [im], &iter [im]);
		    }
		    while (deb <= fin && bap_outof_itermon_COEFF (&iter [deb]))
			deb++;
		    while (deb <= fin && bap_outof_itermon_COEFF (&iter [fin]))
			fin--;
		}
	    }
	}
	if (found)
	{   ba0_pull_stack ();
	    bap_write_creator_COEFF (&crea, &T [im], *lcm);
	    ba0_push_another_stack ();
	    bap_next_itermon_COEFF (&iter [im]);
	    if (! bap_outof_itermon_COEFF (&iter [im]))
		bap_term_itermon_COEFF (&T [im], &iter [im]);
	    while (deb <= fin && bap_outof_itermon_COEFF (&iter [deb]))
		deb++;
	    while (deb <= fin && bap_outof_itermon_COEFF (&iter [fin]))
		fin--;
	}
    }

    ba0_pull_stack ();
    bap_close_creator_COEFF (&crea);
    ba0_restore (&M);
}

#endif
#undef BAD_FLAG_COEFF
