#include "bap_indexed_access.h"
#include "bap_iterator_index.h"

/*-
\paragraph{void bap_init_indexed_access (bap_indexed_access ind)}
Initializes $\ast\mbox{\em ind}$.
-*/

BAP_DLL void bap_init_indexed_access (bap_indexed_access ind)
{
    ind->alloc = 0;
    ind->size = 0;
    ba0_init_table ((ba0_table)&ind->tab_index);
}

/*-
\paragraph{void bap_realloc_indexed_access 
				(bap_indexed_access ind, ba0_int_p nbmon)}
Reallocs $\ast\mbox{\em ind}$ so that it can receive {\em nbmon} indices.
-*/

BAP_DLL void bap_realloc_indexed_access 
				(bap_indexed_access ind, ba0_int_p nbmon)
{   ba0_tableof_int_p tab;
    ba0_int_p n;

    n = nbmon - ind->tab_index.alloc;
    while (n > 0)
    {   ind->tab_index.size = ind->tab_index.alloc;
	ba0_realloc2_table ((ba0_table)&ind->tab_index, 
				ind->tab_index.size + 1, 
				(ba0_new_function*)&ba0_new_table);
	tab = ind->tab_index.tab [ind->tab_index.size++];
	ba0_t1_alloc (sizeof (ba0_int_p), n, (void**)&tab->tab, 
				(unsigned ba0_int_p*)&tab->alloc);
        ind->alloc += tab->alloc;
        n -= tab->alloc;
    }
}

/*-
\paragraph{void bap_reverse_indexed_access (bap_indexed_access ind)}
Reverts the array of indices of  $\ast\mbox{\em ind}$.
-*/

BAP_DLL void bap_reverse_indexed_access (bap_indexed_access ind)
{   struct bap_iterator_index deb, fin;

    bap_begin_iterator_index (&deb, ind);
    bap_end_iterator_index (&fin, ind);
    while (deb.num.global < fin.num.global)
    {	bap_swapindex_iterator_index (&deb, &fin);
	bap_next_iterator_index (&deb);
	bap_prev_iterator_index (&fin);
    }
}

/*
   ATTENTION : LES ACCES INDEXES SONT INLINES DANS LES ZONE_POLFI_COEFF
   CE QUI MODIFIE UN PEU l'ECRITURE DES CHOSES

   --> ba0_embedded
*/

static char _indexed_tab [] = "bap_indexed_access.tab_index";
static char _indexed_tab_i [] = "bap_indexed_access.tab_index [i]";
static char _indexed_tab_i_tab [] = "bap_indexed_access.tab_index [i].tab";

BAP_DLL ba0_int_p bap_garbage1_indexed_access 
					(void* AA, enum ba0_garbage_code code)
{   bap_indexed_access A = (bap_indexed_access)AA;
    ba0_int_p i, n = 0;

    if (code != ba0_embedded)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (A->tab_index.alloc > 0)
	n += ba0_new_gc_info 
		(A->tab_index.tab, sizeof(void*) * A->tab_index.alloc, 
		 _indexed_tab);

    for (i = 0; i < A->tab_index.alloc; i++)
    {	ba0_tableof_int_p tab = A->tab_index.tab [i];
	n += ba0_new_gc_info 
	    (tab, sizeof (struct ba0_tableof_int_p), _indexed_tab_i);
	n += ba0_new_gc_info 
	    (tab->tab, sizeof(ba0_int_p) * tab->alloc, _indexed_tab_i_tab);
    }

    return n;
}

BAP_DLL void* bap_garbage2_indexed_access (void* AA, enum ba0_garbage_code code)
{   bap_indexed_access A = (bap_indexed_access)AA;
    ba0_int_p i;

    if (code != ba0_embedded)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (A->tab_index.alloc > 0)
	A->tab_index.tab = (ba0_tableof_int_p*)ba0_new_addr_gc_info 
			(A->tab_index.tab, _indexed_tab);
    for (i = 0; i < A->tab_index.alloc; i++)
    {	ba0_tableof_int_p tab;
	A->tab_index.tab [i] = (ba0_tableof_int_p)ba0_new_addr_gc_info
		(A->tab_index.tab [i], _indexed_tab_i);
	tab = A->tab_index.tab [i];
	tab->tab = (ba0_int_p*)ba0_new_addr_gc_info
					(tab->tab, _indexed_tab_i_tab);
    }

    return (void*)0;
}
