#include "bap_polynom_COEFF.h"
#include "bap_add_polynom_COEFF.h"
#include "bap_mul_polynom_COEFF.h"
#include "bap_prem_polynom_COEFF.h"
#include "bap_itermon_COEFF.h"
#include "bap_creator_COEFF.h"
#include "bap_invert_COEFF.h"

#define BAD_FLAG_COEFF
#if !defined (BAD_FLAG_mpfr) || defined (HAVE_MPFR_H)

/*
	Polynomials with invertible numerical coefficients
*/

/* Makes $A$ numerically monic. Result in $R$. */

BAP_DLL void bap_numeric_initial_one_polynom_COEFF 
				(bap_polynom_COEFF R, bap_polynom_COEFF A)
{   COEFF_t a;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);
    COEFF_init (a);
    COEFF_invert (a, *bap_numeric_initial_polynom_COEFF (A));
    ba0_pull_stack ();
    bap_mul_polynom_numeric_COEFF (R, A, a);
    ba0_record (&M);
}

/*
 * Stores the quotient and the remainder of the Euclidean division
 * of $A$ by $B$ in $Q$ and $R$. The parameter $Q$ may be zero.
 * $A$ and $B$ are assumed to be univariate.
 */

BAP_DLL void bap_Euclidean_division_polynom_COEFF (bap_polynom_COEFF Q, 
	bap_polynom_COEFF R, bap_polynom_COEFF A, bap_polynom_COEFF B)
{   struct bap_creator_COEFF crea;
    struct bap_itermon_COEFF iterB, iterreste;
    struct bap_polynom_COEFF reste;
    bap_polynom_COEFF quotient = (bap_polynom_COEFF)0;
    COEFF_t b, q;
    struct bav_term TB, TQ, Treste;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);

    COEFF_init (b);
    bav_init_term (&TB);
/*
   Cas B = constante
*/
    bap_begin_itermon_COEFF (&iterB, B);
    bap_term_itermon_COEFF (&TB, &iterB);
    COEFF_invert (b, *bap_coeff_itermon_COEFF (&iterB));
    if (bav_is_one_term (&TB))
    {	ba0_pull_stack ();
	if (Q != (bap_polynom_COEFF)0)
	{   if (COEFF_is_one (b))
	    {	if (Q != A)
		    bap_set_polynom_COEFF (Q, A);
	    } else
		bap_mul_polynom_numeric_COEFF (Q, A, b);
	}
	bap_set_polynom_zero_COEFF (R);
	ba0_restore (&M);
	return;
    }
/*
   Cas ou deg A < deg B
*/
    if (! bap_is_zero_polynom_COEFF (A))
    {   bap_begin_itermon_COEFF (&iterreste, A);
    	bav_init_term (&Treste);
    	bap_term_itermon_COEFF (&Treste, &iterreste);
	bav_init_term (&TQ);
    }
    if (bap_is_zero_polynom_COEFF (A) || 
				! bav_is_factor_term (&Treste, &TB, &TQ))
    {	ba0_pull_stack ();
	if (Q != (bap_polynom_COEFF)0)
	    bap_set_polynom_zero_COEFF (Q);
	if (R != A)
	    bap_set_polynom_COEFF (R, A);
	ba0_restore (&M);
	return;
    }
/*
   Cas general : deg A >= deg B > 0
*/
    if (Q != (bap_polynom_COEFF)0)
    {	quotient = bap_new_polynom_COEFF ();
	bap_begin_creator_COEFF (&crea, quotient, &TQ, bap_exact_total_rank, 
				 bap_nbmon_polynom_COEFF (A));
    }
    COEFF_init (q);
    COEFF_mul (q, *bap_coeff_itermon_COEFF (&iterreste), b);
    bap_init_polynom_COEFF (&reste);
    bap_submulmon_polynom_COEFF (&reste, A, B, &TQ, q);
    if (Q != (bap_polynom_COEFF)0)
	bap_write_creator_COEFF (&crea, &TQ, q);

    for (;;)
    {	if (bap_is_zero_polynom_COEFF (&reste))
	    break;
	bap_begin_itermon_COEFF (&iterreste, &reste);
	bap_term_itermon_COEFF (&Treste, &iterreste);
	if (! bav_is_factor_term (&Treste, &TB, &TQ))
	    break;
	COEFF_mul (q, *bap_coeff_itermon_COEFF (&iterreste), b);
	bap_submulmon_polynom_COEFF (&reste, &reste, B, &TQ, q);
	if (Q != (bap_polynom_COEFF)0)
	    bap_write_creator_COEFF (&crea, &TQ, q);
    }
    if (Q != (bap_polynom_COEFF)0)
	bap_close_creator_COEFF (&crea);
    ba0_pull_stack ();
    if (Q != (bap_polynom_COEFF)0)
	bap_set_polynom_COEFF (Q, quotient);
    bap_set_polynom_COEFF (R, &reste);
    ba0_restore (&M);
}

/*
 * Sets $G$ to the monic gcd of $A$ and $B$.
 * The computation is performed by the basic Euclidean algorithm.
 * Polynomials are assumed to be univariate.
 */

BAP_DLL void bap_Euclid_polynom_COEFF 
		(bap_polynom_COEFF G, bap_polynom_COEFF A, bap_polynom_COEFF B)
{   bap_polynom_COEFF U3, V3, T3;
    struct ba0_mark M;

    if (bap_lt_rank_polynom_COEFF (A, B))
	BA0_SWAP (bap_polynom_COEFF, A, B);

    ba0_push_another_stack ();
    ba0_record (&M);

    U3 = bap_new_polynom_COEFF ();
    V3 = bap_new_polynom_COEFF ();
    T3 = bap_new_polynom_COEFF ();

    if (bap_is_zero_polynom_COEFF (B))
	bap_set_polynom_COEFF (U3, A);
    else
    {	bap_Euclidean_division_polynom_COEFF ((bap_polynom_COEFF)0, V3, A, B);
	bap_set_polynom_COEFF (U3, B);
	while (! bap_is_zero_polynom_COEFF (V3))
	{   bap_set_polynom_COEFF (T3, V3);
	    bap_Euclidean_division_polynom_COEFF 
					((bap_polynom_COEFF)0, V3, U3, V3);
	    BA0_SWAP (bap_polynom_COEFF, U3, T3);
	}
    }
    if (! bap_is_zero_polynom_COEFF (U3))
	bap_numeric_initial_one_polynom_COEFF (U3, U3);
    ba0_pull_stack ();
    bap_set_polynom_COEFF (G, U3);
    ba0_restore (&M);
}

/*
 * Stores in $U$, $V$ and $G$ three polynomials such that
 * $U\,A + V\,B = G = A \wedge B$. The gcd $G$ is monic.
 * The extended Euclidean algorithm is applied.
 * Polynomials are assumed to be univariate.
 */

BAP_DLL void bap_extended_Euclid_polynom_COEFF 
	(bap_polynom_COEFF U, bap_polynom_COEFF V, bap_polynom_COEFF G, 
	 bap_polynom_COEFF A, bap_polynom_COEFF B)
{   bap_polynom_COEFF U1, U3, V1, V3, T1, T3, Q;
    COEFF_t a;
    struct ba0_mark M;

    if (bap_lt_rank_polynom_COEFF (A, B))
    {	BA0_SWAP (bap_polynom_COEFF, A, B);
	BA0_SWAP (bap_polynom_COEFF, U, V);
    }

    ba0_push_another_stack ();
    ba0_record (&M);

    U1 = bap_new_polynom_one_COEFF ();
    U3 = bap_new_polynom_COEFF ();
    bap_set_polynom_COEFF (U3, A);
    V1 = bap_new_polynom_COEFF ();
    V3 = bap_new_polynom_COEFF ();
    bap_set_polynom_COEFF (V3, B);
    Q = bap_new_polynom_COEFF ();
    T1 = bap_new_polynom_COEFF ();
    T3 = bap_new_polynom_COEFF ();
    while (! bap_is_zero_polynom_COEFF (V3))
    {   bap_set_polynom_COEFF (T3, V3);
        bap_Euclidean_division_polynom_COEFF (Q, V3, U3, V3);
        BA0_SWAP (bap_polynom_COEFF, U3, T3);
        bap_set_polynom_COEFF (T1, V1);
        bap_mul_polynom_COEFF (Q, Q, V1);
        bap_sub_polynom_COEFF (V1, U1, Q);
        BA0_SWAP (bap_polynom_COEFF, U1, T1);
    }
    if (! bap_is_zero_polynom_COEFF (U3))
    {	COEFF_init (a);
	COEFF_invert (a, *bap_numeric_initial_polynom_COEFF (U3));
	bap_mul_polynom_numeric_COEFF (U3, U3, a);
	bap_mul_polynom_numeric_COEFF (U1, U1, a);
    }
    bap_mul_polynom_COEFF (T1, U1, A);
    bap_sub_polynom_COEFF (T1, U3, T1);
    ba0_pull_stack ();
    bap_set_polynom_COEFF (U, U1);
    bap_exquo_polynom_COEFF (V, T1, B);
    bap_set_polynom_COEFF (G, U3);
    ba0_restore (&M);
}

#endif
#undef BAD_FLAG_COEFF
