#include "bap_itercoeff_COEFF.h"
#include "bap_global.h"

#define BAD_FLAG_COEFF
#if !defined (BAD_FLAG_mpfr) || defined (HAVE_MPFR_H)

/****************************************************************************
 ITERATORS OF COEFFICIENTS
 ****************************************************************************/

/*
 * Let $A$ be a differential polynomial and
 * $X = \{ x_1 < \cdots < x_n \}$ the alphabet of the variables
 * it depends on. Let $1 \leq i \leq n$ be an index.
 * One wishes to look at the coefficients of $A$, viewed as a polynomial
 * over the alphabet
 * $\{ x_i, \ldots, x_n \}$ and with coefficients in the ring of the polynomials
 * over the alphabet
 * $\{ x_1, \ldots, x_{i-1} \}$.
 * The first coefficient of $A$ is the one which involves the first
 * monomial, in the sense defined above.
 * 
 * The field {\tt polfi} points to the polynomial to be read.
 * The field {\tt last_variable} gives the last variable (i.e.
 * $x_i$ over the example) of the alphabet of the terms.
 * The iterators {\tt debut} and {\tt fin} point to the first and
 * the last monomial which form the current coefficient.
 * The boolean {\tt outof} indicates if the iterator has exited from the
 * polynomial.
 * 
 */

#define itercoeff_last_number	bap_global.itercoeff.last_number
#define itercoeff_terme		bap_global.itercoeff.term

static enum bap_rank_code itercoeff_samerank_COEFF (bav_term R)
{   enum ba0_compare_code code;
/*
 * Readonly static data
 */
    static enum bap_rank_code tab [] = 
	{ bap_rank_too_low, bap_rank_ok, bap_rank_ok, bap_rank_too_high };

    code = 
	bav_compare_stripped_term (R, itercoeff_terme, itercoeff_last_number);
    return tab [code];
}

/*
 * Initializes the iterator on the first coefficient of $A$.
 * The variable $v$ provides the last variable of the alphabet of the terms.
 */

BAP_DLL void bap_begin_itercoeff_COEFF 
		(bap_itercoeff_COEFF I, bap_polynom_COEFF A, bav_variable v)
{   struct ba0_mark M;

    I->polfi = A;
    I->last_variable = v;

    if (bap_is_zero_polynom_COEFF (A))
    {	I->outof = true;
	return;
    }

    ba0_record (&M);
    itercoeff_terme = bav_new_term ();
    itercoeff_last_number = 
		v->number.tab [bav_R_Iordering ()];

    bap_begin_itermon_COEFF (&I->debut, A);
    bap_term_itermon_COEFF (itercoeff_terme, &I->debut);
    bap_begin_itermon_COEFF (&I->fin, A);
    bap_seeklast_itermon_COEFF (&I->fin, &itercoeff_samerank_COEFF);
    I->outof = false;

    ba0_restore (&M);
}

/* Returns true if the iterator has exited from the polynomial. */

BAP_DLL bool bap_outof_itercoeff_COEFF (bap_itercoeff_COEFF I)
{
    return I->outof;
}

/* One step. */

BAP_DLL void bap_next_itercoeff_COEFF (bap_itercoeff_COEFF I)
{   struct ba0_mark M;

    if (bap_outof_itercoeff_COEFF (I))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    ba0_record (&M);
    itercoeff_last_number = I->last_variable->number.tab [bav_R_Iordering ()];
    itercoeff_terme = bav_new_term ();

    bap_set_itermon_COEFF (&I->debut, &I->fin);
    bap_next_itermon_COEFF (&I->debut);
    if (! bap_outof_itermon_COEFF (&I->debut))
    {	bap_term_itermon_COEFF (itercoeff_terme, &I->debut);
	bap_seeklast_itermon_COEFF (&I->fin, &itercoeff_samerank_COEFF);
    } else
	I->outof = true;

    ba0_restore (&M);
}

/* Sets the current term to $T$. */

BAP_DLL void bap_term_itercoeff_COEFF (bav_term T, bap_itercoeff_COEFF I)
{
    if (bap_outof_itercoeff_COEFF (I))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    bap_term_itermon_COEFF (T, &I->debut);
    bav_strip_term (T, T, I->last_variable->number.tab [bav_R_Iordering ()]);
}

/* Sets the current coefficient to $A$ as a readonly polynomial. */

BAP_DLL void bap_coeff_itercoeff_COEFF 
				(bap_polynom_COEFF A, bap_itercoeff_COEFF I)
{   bav_term T;
    bav_Inumber n;
    ba0_int_p j;

    if (bap_outof_itercoeff_COEFF (I))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
/*
   Calcul des numeros de monomes
*/
    if (I->polfi->acces == bap_sequential_monom_access)
    {	A->clot = I->polfi->clot;
	A->acces = bap_sequential_monom_access;
	A->seq.first = bap_number_itermon_clot_COEFF (&I->debut.iter);
	A->seq.after = bap_number_itermon_clot_COEFF (&I->fin.iter) + 1;
    } else
    {	struct bap_creator_index crea;
	struct bap_iterator_index iter;
	ba0_int_p i, nbmon;

	A->clot = I->polfi->clot;
	A->acces = bap_indexed_monom_access;

	nbmon = I->fin.iter_ix.num.global - I->debut.iter_ix.num.global + 1;
	bap_realloc_indexed_access (&A->ind, nbmon);
	bap_begin_creator_index (&crea, &A->ind);

	bap_set_iterator_index (&iter, &I->debut.iter_ix);
	for (i = 0; i < nbmon; i++)
	    bap_write_creator_index (&crea, bap_read_iterator_index (&iter));
	bap_close_creator_index (&crea);
    }
    bap_set_selection (&A->select, &I->polfi->select);
/*
   Calcul de la bav_variable maximale
*/
    T = &I->polfi->rang_total;
    n = I->last_variable->number.tab [bav_R_Iordering ()];
    j = 0;
    while (j < T->size && bav_R_variable_number (T->rg [j].var) >= n)
	j++;
    bap_change_variable_selection 
	(&A->select, j == T->size ? BAV_NOT_A_VARIABLE : T->rg [j].var);
/*
   Calcul du struct bav_rank total
*/
    bap_set_total_rank_polynom_COEFF (A);
/*
   Calcul du booleen coeff
*/
    A->readonly = true;
}

/*
 * Sets the coefficient of $T$ to $A$ as a readonly polynomial.
 * If $T$ involves variables strictly less than the last variable
 * of the alphabet of the terms, they are ignored.
 */

BAP_DLL void bap_seek_coeff_itercoeff_COEFF 
		(bap_polynom_COEFF A, bap_itercoeff_COEFF I, bav_term T)
{   struct bav_term U;
    enum ba0_compare_code code;
    struct ba0_mark M;

    if (bap_outof_itercoeff_COEFF (I))
        BA0_RAISE_EXCEPTION (BA0_ERRALG);
    itercoeff_terme = T;
    itercoeff_last_number = I->last_variable->number.tab [bav_R_Iordering ()];

    bap_seekfirst_itermon_COEFF (&I->debut, &itercoeff_samerank_COEFF);

    ba0_record (&M);
    bav_init_term (&U);
    bap_term_itermon_COEFF (&U, &I->debut);
    code = bav_compare_stripped_term (T, &U, itercoeff_last_number);
    ba0_restore (&M);

    if (code != ba0_eq)
	bap_set_polynom_zero_COEFF (A);
    else
    {	bap_seeklast_itermon_COEFF (&I->fin, &itercoeff_samerank_COEFF);
	bap_coeff_itercoeff_COEFF (A, I);
    }
}

BAP_DLL void bap_close_itercoeff_COEFF (bap_itercoeff_COEFF iter)
{
    iter = (bap_itercoeff_COEFF)0;
}

#endif
#undef BAD_FLAG_COEFF
