#include "bap_itermon_COEFF.h"

#define BAD_FLAG_COEFF
#if !defined (BAD_FLAG_mpfr) || defined (HAVE_MPFR_H)

/****************************************************************************
 ITERATORS OF MONOMIALS
 ****************************************************************************/

/*
 * The field {\tt polfi} is the polynomial to be read.
 * The first monomial is the monomial of least number in the
 * case of a sequential access and the one of index zero
 * in the case of an indexed access.
 * The field {\tt itx} makes sense in the case of indexed access only.
 * It then gives the current index in the array {\tt polfi->ind.tab_index}.
 */

/* Initializes the iterator on the first monomial of $A$. */

BAP_DLL void bap_begin_itermon_COEFF (bap_itermon_COEFF I, bap_polynom_COEFF A)
{   
    I->polfi = A;
    bap_begin_itermon_clot_COEFF (&I->iter, I->polfi->clot);
    if (! bap_is_zero_polynom_COEFF (A))
    {	if (I->polfi->acces == bap_sequential_monom_access)
    	    bap_goto_itermon_clot_COEFF (&I->iter, I->polfi->seq.first);
	else
	{   bap_begin_iterator_index (&I->iter_ix, &I->polfi->ind);
	    bap_goto_itermon_clot_COEFF 
			(&I->iter, bap_index_iterator_index (&I->iter_ix));
        }
    }
}

/* Initializes the iterator on the last monomial of $A$.  */

BAP_DLL void bap_end_itermon_COEFF (bap_itermon_COEFF I, bap_polynom_COEFF A)
{   
    I->polfi = A;
    bap_begin_itermon_clot_COEFF (&I->iter, I->polfi->clot);
    if (I->polfi->acces == bap_sequential_monom_access)
	bap_goto_itermon_clot_COEFF (&I->iter, I->polfi->seq.after - 1);
    else
    {	bap_end_iterator_index (&I->iter_ix, &I->polfi->ind);
	if (! bap_is_zero_polynom_COEFF (A))
	    bap_goto_itermon_clot_COEFF 
			(&I->iter, bap_index_iterator_index (&I->iter_ix));
    }
}

/* Sets $\ast I$ to $\ast J$.  */

BAP_DLL void bap_set_itermon_COEFF (bap_itermon_COEFF I, bap_itermon_COEFF J)
{
    *I = *J;
}

/* Returns true if the iterator has exited from the polynomial.  */

BAP_DLL bool bap_outof_itermon_COEFF (bap_itermon_COEFF I)
{   
    if (I->polfi->acces == bap_sequential_monom_access)
    {	ba0_int_p n = bap_number_itermon_clot_COEFF (&I->iter);
	return n < I->polfi->seq.first || n >= I->polfi->seq.after;
    } else
	return bap_outof_iterator_index (&I->iter_ix);
}

/* One step.  */

BAP_DLL void bap_next_itermon_COEFF (bap_itermon_COEFF I)
{   if (I->polfi->acces == bap_sequential_monom_access)
    	bap_next_itermon_clot_COEFF (&I->iter);
    else
    {	bap_next_iterator_index (&I->iter_ix);
	if (! bap_outof_iterator_index (&I->iter_ix))
	    bap_goto_itermon_clot_COEFF 
			(&I->iter, bap_index_iterator_index (&I->iter_ix));
    }
}

/* One step backwards.  */

BAP_DLL void bap_prev_itermon_COEFF (bap_itermon_COEFF I)
{   if (I->polfi->acces == bap_sequential_monom_access)
    	bap_prev_itermon_clot_COEFF (&I->iter);
    else
    {	bap_prev_iterator_index (&I->iter_ix);
	if (! bap_outof_iterator_index (&I->iter_ix))
	    bap_goto_itermon_clot_COEFF 
			(&I->iter, bap_index_iterator_index (&I->iter_ix));
    }
}

/* Moves the iterator to the monomial with number $n$.  */

BAP_DLL void bap_goto_itermon_COEFF (bap_itermon_COEFF I, ba0_int_p n)
{
    if (n < 0 || n >= bap_nbmon_polynom_COEFF (I->polfi))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (I->polfi->acces == bap_sequential_monom_access)
	bap_goto_itermon_clot_COEFF (&I->iter, I->polfi->seq.first + n);
    else
    {	bap_goto_iterator_index (&I->iter_ix, n);
	bap_goto_itermon_clot_COEFF 
			(&I->iter, bap_index_iterator_index (&I->iter_ix));
    }
}

/* Returns the address of the current coefficient.  */

BAP_DLL COEFF_t* bap_coeff_itermon_COEFF (bap_itermon_COEFF I)
{
    return bap_coeff_itermon_clot_COEFF (&I->iter);
}

/* Sets the current term to~$T$.  */

BAP_DLL void bap_term_itermon_COEFF (bav_term T, bap_itermon_COEFF I)
{
    if (bap_outof_itermon_COEFF (I))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    bap_term_itermon_clot_COEFF (T, &I->iter);
    bap_term_selection (T, I->polfi->clot->ordering, &I->polfi->select);
}

/*
 * Sets $A$ to the sum of all the monomials following the current monomial
 * (this one included). The result is readonly. The case I->polfi == A is
 * allowed.
 */

BAP_DLL void bap_reductum_itermon_COEFF 
				(bap_itermon_COEFF I, bap_polynom_COEFF A)
{
    if (bap_outof_itermon_COEFF (I))
	bap_set_polynom_zero_COEFF (A);
    else if (A != I->polfi)
    {   A->clot = I->polfi->clot;
        if (I->polfi->acces == bap_sequential_monom_access)
        {   A->acces = bap_sequential_monom_access;
    	    A->seq.first = bap_number_itermon_clot_COEFF (&I->iter);
    	    A->seq.after = I->polfi->seq.after;
        } else
        {   struct bap_creator_index crea;
	    struct bap_iterator_index iter;
	    ba0_int_p i, nbmon;

	    A->acces = bap_indexed_monom_access;
	    bap_set_iterator_index (&iter, &I->iter_ix);
	    nbmon = iter.ind->size - iter.num.global;
	    bap_realloc_indexed_access (&A->ind, nbmon);
	    bap_begin_creator_index (&crea, &A->ind);
	    for (i = 0; i < nbmon; i++)
		bap_write_creator_index 
				(&crea, bap_read_iterator_index (&iter));
	    bap_close_creator_index (&crea);
        }
	bap_set_selection (&A->select, &I->polfi->select);
	bap_set_total_rank_polynom_COEFF (A);
	A->readonly = true;
    } else
/*
   F. Boulier 2008. Handling separately the case A == I->polfi
*/
    {   
        if (I->polfi->acces == bap_sequential_monom_access)
        {   
    	    A->seq.first = bap_number_itermon_clot_COEFF (&I->iter);
    	    A->seq.after = I->polfi->seq.after;
        } else
        {   struct bap_creator_index crea;
	    struct bap_iterator_index iter;
	    ba0_int_p i, nbmon;

	    bap_set_iterator_index (&iter, &I->iter_ix);
	    nbmon = iter.ind->size - iter.num.global;
	    bap_begin_creator_index (&crea, &A->ind);
	    for (i = 0; i < nbmon; i++)
		bap_write_creator_index 
				(&crea, bap_read_iterator_index (&iter));
	    bap_close_creator_index (&crea);
        }
	bap_set_total_rank_polynom_COEFF (A);
	A->readonly = true;
    }
}

static void seekfirst_sequentiel_itermon_polfi_COEFF 
		(bap_itermon_COEFF I, bap_rank_code_function* f)
{   ba0_int_p left, right, middle;
    enum bap_rank_code code;
    enum bap_typeof_monom_sequence sens;
    struct bav_term T;

    bav_init_term (&T);
    sens = I->polfi->clot->sens;
    bap_term_itermon_COEFF (&T, I);
    code = (*f) (&T);
    if ((code == bap_rank_too_high && sens == bap_decreasing_monom_sequence)
		|| (code == bap_rank_too_low && 
					sens == bap_increasing_monom_sequence))
    {
/*
   look on the right
*/
	left = bap_number_itermon_clot_COEFF (&I->iter);
	right = I->polfi->seq.after - 1;
    } else
    {
/*
   look on the left
*/
	left = I->polfi->seq.first;
	right = bap_number_itermon_clot_COEFF (&I->iter);
    } 
    while (right - left > 1)
    {	middle = (right + left) / 2;
	bap_goto_itermon_clot_COEFF (&I->iter, middle);
	bap_term_itermon_COEFF (&T, I);
	code = (*f) (&T);
	if ((code == bap_rank_too_high && 
		sens == bap_decreasing_monom_sequence) ||
		(code == bap_rank_too_low && sens == 
					bap_increasing_monom_sequence))
	    left = middle;
	else 
	    right = middle;
    }
/*
  When the two bounds are different, one considers the left one.
  If the rank is correct, one goes on it, else one goes to the right one.
*/
    if (left != right)
    {	if (left != bap_number_itermon_clot_COEFF (&I->iter))
	    bap_prev_itermon_clot_COEFF (&I->iter);
	bap_term_itermon_COEFF (&T, I);
	code = (*f) (&T);
	if (code != bap_rank_ok)
	    bap_next_itermon_clot_COEFF (&I->iter);
    } 
}

static void seekfirst_indexe_itermon_polfi_COEFF 
		(bap_itermon_COEFF I, bap_rank_code_function* f)
{   ba0_int_p left, right, middle;
    enum bap_rank_code code;
    enum bap_typeof_monom_sequence sens;
    struct bav_term T;

    bav_init_term (&T);
    sens = I->polfi->clot->sens;
    bap_term_itermon_COEFF (&T, I);
    code = (enum bap_rank_code)(*f) (&T);
    if ((code == bap_rank_too_high && sens == bap_decreasing_monom_sequence)
		|| (code == bap_rank_too_low && 
					sens == bap_increasing_monom_sequence))
    {
/*
   look on the right
*/
	left = I->iter_ix.num.global;
	right = I->polfi->ind.size - 1;
    } else
    {
/*
   look on the left
*/
	left = 0;
	right = I->iter_ix.num.global;
    } 
    while (right - left > 1)
    {	middle = (right + left) / 2;
	bap_goto_iterator_index (&I->iter_ix, middle);
	bap_goto_itermon_clot_COEFF 
			(&I->iter, bap_index_iterator_index (&I->iter_ix));
	bap_term_itermon_COEFF (&T, I);
	code = (*f) (&T);
	if ((code == bap_rank_too_high && 
		sens == bap_decreasing_monom_sequence) ||
			(code == bap_rank_too_low && 
				sens == bap_increasing_monom_sequence))
	    left = middle;
	else 
	    right = middle;
    }
/*
  When the two bounds are different, one considers the left one.
  If the rank is correct, one goes on it, else one goes to the right one.
*/
    if (left != right)
    {	if (left != I->iter_ix.num.global)
	{   bap_prev_iterator_index (&I->iter_ix);
	    bap_goto_itermon_clot_COEFF 
			(&I->iter, bap_index_iterator_index (&I->iter_ix));
	}
	bap_term_itermon_COEFF (&T, I);
	code = (*f) (&T);
	if (code != bap_rank_ok)
	{   bap_next_iterator_index (&I->iter_ix);
	    bap_goto_itermon_clot_COEFF 
			(&I->iter, bap_index_iterator_index (&I->iter_ix));
	}
    }
}
 
/*
 * Moves the iterator to the first monomial $M$ such that $f(M)$
 * returns {\tt bap_rank_ok}. Dichotomic search.
*/

BAP_DLL void bap_seekfirst_itermon_COEFF 
		(bap_itermon_COEFF I, bap_rank_code_function* f)
{   struct ba0_mark M;

    if (bap_is_zero_polynom_COEFF (I->polfi))
        BA0_RAISE_EXCEPTION (BAP_ERRNUL);
    if (I->polfi->clot->sens != bap_decreasing_monom_sequence && 
			I->polfi->clot->sens != bap_increasing_monom_sequence)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    ba0_record (&M);
    if (bap_outof_itermon_COEFF (I))
	bap_begin_itermon_COEFF (I, I->polfi);
    if (I->polfi->acces == bap_sequential_monom_access)
	seekfirst_sequentiel_itermon_polfi_COEFF (I, f);
    else
	seekfirst_indexe_itermon_polfi_COEFF (I, f);
    ba0_restore (&M);
}

static void seeklast_sequentiel_itermon_polfi_COEFF 
		(bap_itermon_COEFF I, bap_rank_code_function* f)
{   ba0_int_p left, right, middle;
    enum bap_rank_code code;
    enum bap_typeof_monom_sequence sens;
    struct bav_term T;

    bav_init_term (&T);
    sens = I->polfi->clot->sens;
    bap_term_itermon_COEFF (&T, I);
    code = (*f) (&T);
    if ((code == bap_rank_too_low && sens == bap_decreasing_monom_sequence) 
		|| (code == bap_rank_too_high && sens == 
					bap_increasing_monom_sequence))
    {
/*
   look on the left
*/
	left = I->polfi->seq.first;
        right = bap_number_itermon_clot_COEFF (&I->iter);
    } else
    {
/*
   look on the right
*/
	left = bap_number_itermon_clot_COEFF (&I->iter);
	right = I->polfi->seq.after - 1;
    }
    while (right - left > 1)
    {	middle = (right + left) / 2;
	bap_goto_itermon_clot_COEFF (&I->iter, middle);
	bap_term_itermon_COEFF (&T, I);
	code = (*f) (&T);
	if ((code == bap_rank_too_low && sens == bap_decreasing_monom_sequence)
			|| (code == bap_rank_too_high && 
				sens == bap_increasing_monom_sequence))
	    right = middle;
	else 
	    left = middle;
    }
/*
   If the two bounds are different, one considers the right one.
   If the rank is correct then one goes on it else one goes to the left one.
*/
    if (left != right)
    {	if (right != bap_number_itermon_clot_COEFF (&I->iter))
	    bap_next_itermon_clot_COEFF (&I->iter);
	bap_term_itermon_COEFF (&T, I);
	code = (*f) (&T);
	if (code != bap_rank_ok)
	    bap_prev_itermon_clot_COEFF (&I->iter);
    } 
}

static void seeklast_indexe_itermon_polfi_COEFF 
		(bap_itermon_COEFF I, bap_rank_code_function* f)
{   ba0_int_p left, right, middle;
    enum bap_rank_code code;
    enum bap_typeof_monom_sequence sens;
    struct bav_term T;

    bav_init_term (&T);
    sens = I->polfi->clot->sens;
    bap_term_itermon_COEFF (&T, I);
    code = (*f) (&T);
    if ((code == bap_rank_too_low && sens == bap_decreasing_monom_sequence)
		|| (code == bap_rank_too_high && sens == 
					bap_increasing_monom_sequence))
    {
/*
   look on the left
*/
	left = 0;
        right = I->iter_ix.num.global;
    } else
    {
/*
   look on the right
*/
	left = I->iter_ix.num.global;
	right = I->polfi->ind.size - 1;
    }
    while (right - left > 1)
    {	middle = (right + left) / 2;
	bap_goto_iterator_index (&I->iter_ix, middle);
	bap_goto_itermon_clot_COEFF 
			(&I->iter, bap_index_iterator_index (&I->iter_ix));
	bap_term_itermon_COEFF (&T, I);
	code = (*f) (&T);
	if ((code == bap_rank_too_low && sens == bap_decreasing_monom_sequence)
		|| (code == bap_rank_too_high && 
			sens == bap_increasing_monom_sequence))
	    right = middle;
	else 
	    left = middle;
    }
/*
   If the two bounds are different, one considers the right one.
   If the rank is correct then one goes on it else one goes to the left one.
*/
    if (left != right)
    {	if (right != I->iter_ix.num.global)
	{   bap_next_iterator_index (&I->iter_ix);
	    bap_goto_itermon_clot_COEFF 
			(&I->iter, bap_index_iterator_index (&I->iter_ix));
	}
	bap_term_itermon_COEFF (&T, I);
	code = (*f) (&T);
	if (code != bap_rank_ok)
	{   bap_prev_iterator_index (&I->iter_ix);
	    bap_goto_itermon_clot_COEFF 
			(&I->iter, bap_index_iterator_index (&I->iter_ix));
	}
    }
}

/*
 * Moves the iterator to the last monomial $M$ such that $f(M)$
 * returns {\tt bap_rank_ok}. Dichotomic search.
*/

BAP_DLL void bap_seeklast_itermon_COEFF 
		(bap_itermon_COEFF I, bap_rank_code_function* f)
{   struct ba0_mark M;

    if (bap_is_zero_polynom_COEFF (I->polfi))
        BA0_RAISE_EXCEPTION (BAP_ERRNUL);
    if (I->polfi->clot->sens != bap_decreasing_monom_sequence && 
			I->polfi->clot->sens != bap_increasing_monom_sequence)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    ba0_record (&M);
    if (bap_outof_itermon_COEFF (I))
	bap_begin_itermon_COEFF (I, I->polfi);
    if (I->polfi->acces == bap_sequential_monom_access)
	seeklast_sequentiel_itermon_polfi_COEFF (I, f);
    else
	seeklast_indexe_itermon_polfi_COEFF (I, f);
    ba0_restore (&M);
}

BAP_DLL void bap_close_itermon_COEFF (bap_itermon_COEFF iter)
{
    iter = (bap_itermon_COEFF)0;
}

#endif
#undef BAD_FLAG_COEFF
