#include "bap_polynom_COEFF.h"
#include "bap_creator_COEFF.h"
#include "bap_itermon_COEFF.h"
#include "bap_itercoeff_COEFF.h"
#include "bap_add_polynom_COEFF.h"
#include "bap_mul_polynom_COEFF.h"
#include "bap_prem_polynom_COEFF.h"
#include "bap__check_COEFF.h"
#include "bap_geobucket_COEFF.h"
#include "bap_product_COEFF.h"

#define BAD_FLAG_COEFF
#if !defined (BAD_FLAG_mpfr) || defined (HAVE_MPFR_H)

/*
 * Returns true if c divides A.
 * If so and Q is nonzsero, then Q = A/c.
 */

#if ! defined (BAD_FLAG_mpz)

/* The mpz_t case is handled in a separate file */

BAP_DLL bool bap_is_numeric_factor_polynom_COEFF 
		(bap_polynom_COEFF A, COEFF_t c, bap_polynom_COEFF Q)
{   COEFF_t cbar;
    struct ba0_mark M;

    if (Q != BAP_NOT_A_POLYNOM_COEFF && Q->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    if (Q == BAP_NOT_A_POLYNOM_COEFF)
	return true;
    ba0_push_another_stack ();
    ba0_record (&M);
    COEFF_init (cbar);
    COEFF_invert (cbar, c);
    ba0_pull_stack ();
    bap_mul_polynom_numeric_COEFF (Q, A, cbar);
    ba0_restore (&M);
    return true;
}

#endif

/*
 * Returns true if v divides A. If so and e is nonzero then *e
 * is the greatest degree d such that v^d divides A.
 */

BAP_DLL bool bap_is_variable_factor_polynom_COEFF 
			(bap_polynom_COEFF A, bav_variable v, bav_Idegree* e)
{   struct bap_itermon_COEFF iter;
    struct bav_term T;
    bool found;
    ba0_int_p i;
    bav_Idegree d;
    struct ba0_mark M;

    if (bap_is_numeric_polynom_COEFF (A))
    {	if (e != (bav_Idegree*)0)
	    *e = 0;
	return false;
    }
    d = BAV_MAX_IDEGRE;
    found = true;
    ba0_record (&M);
    bap_begin_itermon_COEFF (&iter, A);
    bav_init_term (&T);
    while (found && ! bap_outof_itermon_COEFF (&iter))
    {	found = false;
	bap_term_itermon_COEFF (&T, &iter);
	for (i = 0; ! found && i < T.size; i++)
	{   if (T.rg [i].var == v)
	    {	d = BA0_MIN (d, T.rg [i].deg);
		found = true;
	    }
	}
	bap_next_itermon_COEFF (&iter);
    }
    if (e != (bav_Idegree*)0)
        *e = found ? d : 0;
    ba0_restore (&M);
    return found;
}

/* 
 * Elementary version of bap_is_factor_polynom_COEFF
 */

static bool bap_is_factor_elem_polynom_COEFF 
		(bap_polynom_COEFF A, bap_polynom_COEFF B, bap_polynom_COEFF Q)
{   struct bap_itermon_COEFF iterR, iterB;
    struct bap_creator_COEFF creaQ;
    bap_polynom_COEFF P = (bap_polynom_COEFF)0, R;
    struct bav_term TQ, TR, TB, TQmax;
    COEFF_t q, r, *cR, *cB;
    bool divisible, first = true;
    struct ba0_mark  M;

    ba0_push_another_stack ();
    ba0_record (&M);

    bav_init_term (&TQmax);
    bav_init_term (&TQ);
/*
 * False if the coefficient ring is not a domain.
 */
    if (! bav_is_factor_term (&A->rang_total, &B->rang_total, &TQmax))
    {   divisible = false;
	goto fin;
    }

    if (Q != BAP_NOT_A_POLYNOM_COEFF)
    {   P = bap_new_polynom_COEFF ();
	bap_begin_creator_COEFF (&creaQ, P, &TQmax, bap_exact_total_rank, 
						bap_nbmon_polynom_COEFF (A));
    }

    divisible = true;

    R = bap_new_polynom_COEFF ();
    bav_init_term (&TR);
    bav_init_term (&TB);
    COEFF_init (q);
    COEFF_init (r);
    bap_begin_itermon_COEFF (&iterB, B);
    cB = bap_coeff_itermon_COEFF (&iterB);
    bap_term_itermon_COEFF (&TB, &iterB);
    while (divisible && (first || ! bap_is_zero_polynom_COEFF (R)))
    {	bap_begin_itermon_COEFF (&iterR, first ? A : R);
	bap_term_itermon_COEFF (&TR, &iterR);
	if (! bav_is_factor_term (&TR, &TB, &TQ))
	{   divisible = false;
	    break;
	}
	cR = bap_coeff_itermon_COEFF (&iterR);
#if defined (BAD_FLAG_mint_hp)
	COEFF_div (q, *cR, *cB);
	r = 0; /* pour eviter un warning */
#elif defined (BAD_FLAG_mpzm) || defined (BAD_FLAG_mpq)
	COEFF_div (q, *cR, *cB);
#elif defined (BAD_FLAG_mpz)
        mpz_tdiv_qr (q, r, *cR, *cB);
        if (mpz_sgn (r) != 0)
	{   divisible = false;
	    break;
	}
#endif
	bap_submulmon_polynom_COEFF (R, first ? A : R, B, &TQ, q);
/*
 * False if the numerical coefficient ring is not a domain
 */
	if (! bav_is_factor_term (&TQmax, &TQ, (bav_term)0))
	{   divisible = false;
	    break;
	}
	if (Q != BAP_NOT_A_POLYNOM_COEFF)
	    bap_write_creator_COEFF (&creaQ, &TQ, q);
	first = false;
    }
fin:
    if (divisible && Q != BAP_NOT_A_POLYNOM_COEFF)
    {	bap_close_creator_COEFF (&creaQ);
	ba0_pull_stack ();
	bap_set_polynom_COEFF (Q, P);
    } else
	ba0_pull_stack ();
    ba0_restore (&M);
    return divisible;
}

/*
 * Returns true if B divides A. If so and Q is nonzero then Q = A/B.
 * Does not work for coefficient rings which are not domains.
 */

BAP_DLL bool bap_is_factor_polynom_COEFF 
		(bap_polynom_COEFF A, bap_polynom_COEFF B, bap_polynom_COEFF Q)
{   struct bap_itercoeff_COEFF iter;
    struct bap_itermon_COEFF itermon;
    struct bap_creator_COEFF crea;
    struct bap_polynom_COEFF q, c;
    bap_polynom_COEFF AA = (bap_polynom_COEFF)0, P = (bap_polynom_COEFF)0;
    struct bav_term T, U;
    bav_variable v, w;
    bool divisible;
    bav_Iordering r;
    ba0_int_p i;
    struct ba0_mark M;

    if (bap_is_zero_polynom_COEFF (B))
	BA0_RAISE_EXCEPTION (BAP_ERRNUL);
    bap__check_compatible_COEFF (A, B);
    if (Q && Q->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);
/*
 * Get rid of the case A == 0
 */
    if (bap_is_zero_polynom_COEFF (A))
    {	if (Q != BAP_NOT_A_POLYNOM_COEFF)
	    bap_set_polynom_zero_COEFF (Q);
	return true;
    }

    ba0_push_another_stack ();
    ba0_record (&M);
    bav_init_term (&T);
/*
 * Get rid of the case B is a constant
 */
    if (bap_is_numeric_polynom_COEFF (B))
    {	COEFF_t c;
	COEFF_init_set (c, *bap_numeric_initial_polynom_COEFF (B));
	ba0_pull_stack ();
	divisible = bap_is_numeric_factor_polynom_COEFF (A, c, Q);
	ba0_restore (&M);
	return divisible;
    }
/*
 * A quick test, which is false if the coefficient ring is not a domain
 */
    if (! bav_is_factor_term (&A->rang_total, &B->rang_total, &T))
    {	ba0_pull_stack ();
	ba0_restore (&M);
	return false;
    }
/*
 * Does A depend on variables which do not occur in B ?
 */
    bav_R_mark_variables (false);
    bap_mark_indets_polynom_COEFF (B);
    i = A->rang_total.size - 1;
    while (i >= 0 && A->rang_total.rg [i].var->hack)
	i--;
/*
 * If A does not, apply the elementary algorithm.
 */
    if (i < 0)
    {	ba0_pull_stack ();
	divisible = bap_is_factor_elem_polynom_COEFF (A, B, Q);
	ba0_restore (&M);
	return divisible;
    }
/*
 * If it does, one splits the set of variables of A into { L, N } where
 * L are the variables which do not occur in B, and N are the other ones.
 *
 * One views A as a linear combination c_i * t_i where the t_i are terms
 * over L. 
 */
    v = A->rang_total.rg [i].var;
    r = bav_R_copy_ordering (bav_R_Iordering ());
    bav_R_push_ordering (r);
    while (i >= 0)
    {	w = A->rang_total.rg [i].var;
	if (w->hack == false)
	    bav_R_set_maximal_variable (w);
	i--;
    }

    if (A->acces == bap_sequential_monom_access)
    {	AA = A;
	bap_physort_polynom_COEFF (AA, bap_invertible_transform);
    } else
    {	AA = bap_new_readonly_polynom_COEFF ();
	bap_sort_polynom_COEFF (AA, A);
    }
    bap_begin_itercoeff_COEFF (&iter, AA, v);

    if (Q != BAP_NOT_A_POLYNOM_COEFF)
    {	bav_init_term (&U);
	bap_init_polynom_COEFF (&q);
/*
 * T was initialized above (wrong if the coefficient ring is not a domain)
 */
	bav_sort_term (&T);
	bap_push_clot_parameters_stack 
	    (bap_current_typeof_termanager (), bap_random_monom_sequence);
	P = bap_new_polynom_COEFF ();
	bap_begin_creator_COEFF 
	    (&crea, P, &T, bap_exact_total_rank, bap_nbmon_polynom_COEFF (A));
	bap_pull_clot_parameters_stack ();
    }
    bap_init_readonly_polynom_COEFF (&c);
    divisible = true;
    while (divisible && ! bap_outof_itercoeff_COEFF (&iter))
    {	bap_coeff_itercoeff_COEFF (&c, &iter);
	if (Q != BAP_NOT_A_POLYNOM_COEFF)
	{   divisible = bap_is_factor_elem_polynom_COEFF (&c, B, &q);
	    if (divisible)
	    {	bap_term_itercoeff_COEFF (&T, &iter);
	    	bap_begin_itermon_COEFF (&itermon, &q);
	    	while (! bap_outof_itermon_COEFF (&itermon))
	        {   bap_term_itermon_COEFF (&U, &itermon);
		    bav_mul_term (&U, &U, &T);
		    bap_write_creator_COEFF 
			(&crea, &U, *bap_coeff_itermon_COEFF (&itermon));
		    bap_next_itermon_COEFF (&itermon);
		}
	    }
	} else
	    divisible = bap_is_factor_elem_polynom_COEFF 
					(&c, B, BAP_NOT_A_POLYNOM_COEFF);
	if (divisible)
	    bap_next_itercoeff_COEFF (&iter);
    }

    i = bap_nbmon_polynom_COEFF (A);

    if (divisible && Q != BAP_NOT_A_POLYNOM_COEFF)
    {	bap_close_creator_COEFF (&crea);
	bav_R_pull_ordering ();

	bap_physort_polynom_COEFF (P, bap_definitive_transform);

	ba0_pull_stack ();
	bap_set_polynom_COEFF (Q, P);
	if (A->acces == bap_sequential_monom_access && 
				bap_are_disjoint_polynom_COEFF (A, Q))
	    bap_physort_polynom_COEFF (A, bap_inverse_transform);
    } else
    {   bav_R_pull_ordering ();
	ba0_pull_stack ();
	if (A->acces == bap_sequential_monom_access)
	    bap_physort_polynom_COEFF (A, bap_inverse_transform);
    }

    bav_R_free_ordering (r);
    ba0_restore (&M);
    return divisible;
}
 
/*
 * Returns true if A divides each element of T
 */

BAP_DLL bool bap_is_factor_tableof_polynom_COEFF
		(bap_polynom_COEFF A, bap_tableof_polynom_COEFF T)
{   ba0_int_p i;
    bool b;

    b = true;
    for (i = 0; i < T->size && b; i++)
	b = bap_is_factor_polynom_COEFF 
				(T->tab [i], A, BAP_NOT_A_POLYNOM_COEFF);
    return b;
}

/*
 * R = A/T
 */

BAP_DLL void bap_exquo_polynom_term_COEFF 
		(bap_polynom_COEFF R, bap_polynom_COEFF A, bav_term T)
{   struct bap_itermon_COEFF iter;
    struct bap_creator_COEFF crea;
    bap_polynom_COEFF P;
    struct bav_term U;
    struct ba0_mark M;

    bap__check_ordering_COEFF (A);
    if (R->readonly)
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    if (bav_is_one_term (T))
    {	if (R != A)
	    bap_set_polynom_COEFF (R, A);
	return;
    }

    ba0_push_another_stack ();
    ba0_record (&M);
    bav_init_term (&U);
    bav_set_term (&U, &A->rang_total);
    bav_exquo_term (&U, &U, T);

    bap_begin_itermon_COEFF (&iter, A);
    P = bap_new_polynom_COEFF ();
    bap_begin_creator_COEFF 
	(&crea, P, &U, bap_exact_total_rank, bap_nbmon_polynom_COEFF (A));
    while (! bap_outof_itermon_COEFF (&iter))
    {	bap_term_itermon_COEFF (&U, &iter);
	bav_exquo_term (&U, &U, T);
	bap_write_creator_COEFF 
			(&crea, &U, *bap_coeff_itermon_COEFF (&iter));
	bap_next_itermon_COEFF (&iter);
    }
    bap_close_creator_COEFF (&crea);
    ba0_pull_stack ();
    bap_set_polynom_COEFF (R, P);
    ba0_restore (&M);
}

/*
 * R = A/B
 */

BAP_DLL void bap_exquo_polynom_COEFF 
	(bap_polynom_COEFF R, bap_polynom_COEFF A, bap_polynom_COEFF B)
{   
    if (! bap_is_factor_polynom_COEFF (A, B, R))
	BA0_RAISE_EXCEPTION (BAV_EXEXQO);
}

/*
 * R = A/P
 */

BAP_DLL void bap_exquo_polynom_product_COEFF
	(bap_polynom_COEFF R, bap_polynom_COEFF A, struct bap_product_COEFF* P)
{   struct bap_polynom_COEFF B;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);
    bap_init_polynom_COEFF (&B);
    bap_expand_product_COEFF (&B, P);
    ba0_pull_stack ();
    bap_exquo_polynom_COEFF (R, A, &B);
    ba0_restore (&M);
}

/**************************************************************************
 PSEUDO DIVISION
 **************************************************************************/

/*
 * This function stores in R the sum of the C [i] * v^i
 */

static void pseudo_division_rebuild_COEFF 
	(bap_polynom_COEFF R, bap_tableof_polynom_COEFF C, bav_variable v)
{   struct bap_creator_COEFF crea;
    struct bap_itermon_COEFF iter;
    COEFF_t *lc;
    struct bav_term T;
    struct bav_rank rg;
    bav_Idegree d, j;
    ba0_int_p nbmon;
    struct ba0_mark M;
/*
 * One computes the number of monomials and the total rank of the result
 */
    ba0_push_another_stack ();
    ba0_record (&M);

    nbmon = 0;
    bav_init_term (&T);
    d = C->size - 1;
    while (d >= 0 && bap_is_zero_polynom_COEFF (C->tab [d]))
	d--;
    for (j = 0; j <= d; j++)
    {	nbmon += bap_nbmon_polynom_COEFF (C->tab [j]);
	bav_lcm_term (&T, &T, &C->tab [j]->rang_total);
    }
    rg.var = v;
    if (d > 0)
    {	rg.deg = d;
	bav_mul_term_rank (&T, &T, &rg);
    }

    ba0_pull_stack ();
    bap_begin_creator_COEFF (&crea, R, &T, bap_exact_total_rank, nbmon);

    for (j = d; j >= 0; j--)
    {	rg.deg = j;
	if (! bap_is_zero_polynom_COEFF (C->tab [j]))
	{   bap_begin_itermon_COEFF (&iter, C->tab [j]);
	    while (! bap_outof_itermon_COEFF (&iter))
	    {	bap_term_itermon_COEFF (&T, &iter);
		lc = bap_coeff_itermon_COEFF (&iter);
		if (j > 0)
		    bav_mul_term_rank (&T, &T, &rg);
	  	bap_write_creator_COEFF (&crea, &T, *lc);
		bap_next_itermon_COEFF (&iter);
	    }
	}
    }
    bap_close_creator_COEFF (&crea);
    ba0_restore (&M);
}

/*
 * Performs the pseudo-division of A by B.
 * One assumes that, if deg (A, ld B) > 0 then ld B = ld A.
 * The pseudo-quotient and pseudo-remainder are stored in Q and R.
 * Parameters Q and R may be zero.
 * The table ibp contains the powers of the initial of B
 * x is a variable greater than or equal to the leader of B
 */

static void bap_pseudo_division_elem_polynom_COEFF 
		(bap_polynom_COEFF Q, bap_polynom_COEFF R, 
		 bap_polynom_COEFF A, bap_polynom_COEFF B, 
		 bav_variable x, bap_tableof_polynom_COEFF ibp)
{   struct bap_itercoeff_COEFF iterA, iterB;
    struct bap_itermon_COEFF itermon;
    struct bap_creator_COEFF crea;
    struct bap_tableof_polynom_COEFF C;
    struct bap_polynom_COEFF coeff_A, coeff_B, tmp1, tmp2;
    bap_polynom_COEFF P = (bap_polynom_COEFF)0;
    struct bav_term T;
    bav_Idegree degA, degB, j, k, t;
    struct ba0_mark M;

    degB = bap_degree_polynom_COEFF (B, x);
    degA = bap_degree_polynom_COEFF (A, x);

    if (degA < degB)
    {	if (R != BAP_NOT_A_POLYNOM_COEFF && R != A)
	    bap_set_polynom_COEFF (R, A);
	if (Q != BAP_NOT_A_POLYNOM_COEFF)
	    bap_set_polynom_zero_COEFF (Q);
	return;
    }
/*
 * degree (A, ld B) >= degree (B, ld B)
 */
    ba0_push_another_stack ();
    ba0_record (&M);

    bav_init_term (&T);

    bap_begin_itercoeff_COEFF (&iterB, B, x);
    bap_begin_itercoeff_COEFF (&iterA, A, x);

    if (Q != BAP_NOT_A_POLYNOM_COEFF)
    {	ba0_int_p nbmon, nbmonA, nbmonB;
	struct bav_term U;

	if (degB > 0)
	{   bav_shift_term (&T, &B->rang_total);
	    bav_pow_term (&T, &T, degA - degB);
	} else
	    bav_pow_term (&T, &B->rang_total, degA);

	if (degA > 0)
	{   bav_init_term (&U);
	    bav_shift_term (&U, &A->rang_total);
	    bav_mul_term (&T, &T, &U);
	} else
	    bav_mul_term (&T, &T, &A->rang_total);

	bav_mul_term_variable (&T, &T, x, degA - degB);

	nbmonA = bap_nbmon_polynom_COEFF (A) / BA0_MAX (1, degA);
	nbmonB = bap_nbmon_polynom_COEFF (B) / BA0_MAX (1, degB);
	nbmon = BA0_MAX (nbmonA, nbmonB);
	nbmon = BA0_MAX (nbmon, 1);
	nbmon *= degA - degB + 1;

	P = bap_new_polynom_COEFF ();
	bap_begin_creator_COEFF (&crea, P, &T, bap_approx_total_rank, nbmon);
    }
/*
 * Denote B [k] the kth coefficient of B.
 *
 * R0.
 *
 * Assign the coefficients of A to C [degA] .. C [0].
 * Apply Knuth remark: replacing C [t] by
 *    B [1]^(degA - degB - t) C [t] for t = 0 .. degA - degB - 1.
 *
 * To avoid side effects, if C [t] is zero, take the coefficient of x^t in A.
 */
    bap_init_readonly_polynom_COEFF (&coeff_A);
    bap_init_readonly_polynom_COEFF (&coeff_B);

    ba0_init_table ((ba0_table)&C);
    ba0_realloc_table ((ba0_table)&C, degA + 1);
    C.size = degA + 1; 		/* la taille de C sera reduite a la fin */
    for (t = 0; t < degA - degB; t++)
    {	C.tab [t] = bap_new_polynom_COEFF ();
	bav_set_term_variable (&T, x, t);
	bap_seek_coeff_itercoeff_COEFF (&coeff_A, &iterA, &T);
	bap_mul_polynom_COEFF (C.tab [t], ibp->tab [degA - degB - t], &coeff_A);
    }
    for (t = degA - degB; t <= degA; t++)
	C.tab [t] = BAP_NOT_A_POLYNOM_COEFF;
/*
 * R1.
 *
 * Do step R2 for k = degB - degA .. 0.
 * The algorithm terminates with R = C [degB - 1] .. C [0].
 */
    bap_init_polynom_COEFF (&tmp1);
    bap_init_polynom_COEFF (&tmp2);
    for (k = degA - degB; k >= 0; k--)
    {
/*
 * R2.
 *
 * Set Q [k] = C [degB + k] * ibp [k]
 * Here, Q [k] * x^k is stored in the creator.
 */
	if (Q != BAP_NOT_A_POLYNOM_COEFF)
	{   if (C.tab [degB + k] == BAP_NOT_A_POLYNOM_COEFF)
	    {	bav_set_term_variable (&T, x, degB + k);
		bap_seek_coeff_itercoeff_COEFF (&coeff_A, &iterA, &T);
		bap_mul_polynom_COEFF (&tmp1, ibp->tab [k], &coeff_A);
	    } else
		bap_mul_polynom_COEFF (&tmp1, ibp->tab [k], C.tab [degB + k]);
	    bap_begin_itermon_COEFF (&itermon, &tmp1);
	    while (! bap_outof_itermon_COEFF (&itermon))
	    {	COEFF_t *lc;

		lc = bap_coeff_itermon_COEFF (&itermon);
		bap_term_itermon_COEFF (&T, &itermon);
		bav_mul_term_variable (&T, &T, x, k);
		bap_write_creator_COEFF (&crea, &T, *lc);
		bap_next_itermon_COEFF (&itermon);
	    }
	}
/*
 * Set C [j] = B [1] C [j] - C [degB + k] B [j - k] for j = degB + k - 1 .. 0.
 * When j < k this means that C [j] = B [1] C [j] since we treat B [-1], B [-2]
 * as zero. These multiplications could have been avoided if we had started
 * the algorithm by replacing C [t] by B [1]^(degA - degB - t) C [t] for
 * t = 0 .. degA - degB - 1.
 */
	for (j = degB + k - 1; j >= k; j--)
	{   if (C.tab [j] == BAP_NOT_A_POLYNOM_COEFF)
	    {	bav_set_term_variable (&T, x, j);
		bap_seek_coeff_itercoeff_COEFF (&coeff_A, &iterA, &T);
		bap_mul_polynom_COEFF (&tmp1, ibp->tab [1], &coeff_A);
		C.tab [j] = bap_new_polynom_COEFF ();
	    } else
		bap_mul_polynom_COEFF (&tmp1, ibp->tab [1], C.tab [j]);
	    bav_set_term_variable (&T, x, j - k);
	    bap_seek_coeff_itercoeff_COEFF (&coeff_B, &iterB, &T);
	    if (C.tab [degB + k] == BAP_NOT_A_POLYNOM_COEFF)
	    {	bav_set_term_variable (&T, x, degB + k);
		bap_seek_coeff_itercoeff_COEFF (&coeff_A, &iterA, &T);
		bap_mul_polynom_COEFF (&tmp2, &coeff_A, &coeff_B);
	    } else
		bap_mul_polynom_COEFF (&tmp2, C.tab [degB + k], &coeff_B);
	    bap_sub_polynom_COEFF (C.tab [j], &tmp1, &tmp2);
	}
    }
/*
 * The algorithm terminates with R = C [degB - 1] .. C [0].
 *
 * Creation of the remainder
 * One first recovers the intermediate data from the main stack
 */
    if (Q != BAP_NOT_A_POLYNOM_COEFF)
	bap_close_creator_COEFF (&crea);

    ba0_pull_stack ();

    if (R != BAP_NOT_A_POLYNOM_COEFF)
    {	C.size = degB;
	pseudo_division_rebuild_COEFF (R, &C, x);
    }
/*
 * The quotient, now
 */
    if (Q != BAP_NOT_A_POLYNOM_COEFF)
	bap_set_polynom_COEFF (Q, P);

    ba0_restore (&M);
}

/*
 * Pseudo-division of A by B, both polynomials being viewed as
 * polynomials in the leader of B. The pseudo-quotient and the 
 * pseudo-remainder are stored in Q and R. The integer *e
 * receives the power to which the initial of B was raised to perform
 * the operation. Parameters Q, R and e may be zero.
 */

BAP_DLL void bap_pseudo_division_polynom_COEFF 
		(bap_polynom_COEFF Q, bap_polynom_COEFF R,
                 bav_Idegree* e, bap_polynom_COEFF A, bap_polynom_COEFF B,
		 bav_variable x)
{   struct bap_itercoeff_COEFF iter;
    struct bap_creator_COEFF crea_quotient, crea_reste;
    struct bap_itermon_COEFF itermon;
    struct bap_tableof_polynom_COEFF ibp;
    struct bap_polynom_COEFF lambda;
    bap_polynom_COEFF AA, quotient, reste;
    bap_polynom_COEFF Pquotient = (bap_polynom_COEFF)0, 
		      Preste = (bap_polynom_COEFF)0;
    COEFF_t *lc;
    struct bav_term T, U;
    bav_variable v, w;
    bav_Iordering r;
    bav_Idegree degA, degB, deg_lambda, k;
    ba0_int_p nbmonA, nbmonB, nbmon;
    ba0_int_p i;
    struct ba0_mark M;

    bap__check_ordering_COEFF (B);
    if (bap_is_zero_polynom_COEFF (B))
	BA0_RAISE_EXCEPTION (BAP_ERRNUL);
    if (x == BAV_NOT_A_VARIABLE && bap_is_numeric_polynom_COEFF (B))
        BA0_RAISE_EXCEPTION (BAP_ERRCST);
    if ((Q != BAP_NOT_A_POLYNOM_COEFF && Q->readonly) || 
				(R != BAP_NOT_A_POLYNOM_COEFF && R->readonly))
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);

    if (x == BAV_NOT_A_VARIABLE)
	x = bap_leader_polynom_COEFF (B);

    degB = bap_degree_polynom_COEFF (B, x);
    degA = bap_degree_polynom_COEFF (A, x);

    if (degA < degB)
    {   if (R != BAP_NOT_A_POLYNOM_COEFF && R != A)
            bap_set_polynom_COEFF (R, A);
        if (Q != BAP_NOT_A_POLYNOM_COEFF)
            bap_set_polynom_zero_COEFF (Q);
	if (e != (bav_Idegree*)0)
	    *e = 0;
        return;
    } 
/*
 * degree (A, ld B) >= degree (B, ld B)
 *
 * Building of the table of the powers of B
 */
    ba0_push_another_stack ();
    ba0_record (&M);

    ba0_init_table ((ba0_table)&ibp);
    ba0_realloc2_table ((ba0_table)&ibp, degA - degB + 2, 
				(ba0_new_function*)&bap_new_polynom_COEFF);
    ibp.size = degA - degB + 2;

    bap_set_polynom_one_COEFF (ibp.tab [0]);
 
    bap_initial2_polynom_COEFF (ibp.tab [1], B, x);

    for (k = 2; k < ibp.size; k++)
	bap_mul_polynom_COEFF (ibp.tab [k], ibp.tab [k-1], ibp.tab [1]);
/*
 * Does A involve variables which do not occur in B ?
 */
    bav_R_mark_variables (false);
    bap_mark_indets_polynom_COEFF (B);
    i = A->rang_total.size - 1;
    while (i >= 0 && A->rang_total.rg [i].var->hack)
	i--;
/*
 * If not, apply the basic method
 */
    if (i < 0)
    {	ba0_pull_stack ();
	bap_pseudo_division_elem_polynom_COEFF (Q, R, A, B, x, &ibp);
	if (e != (bav_Idegree*)0)
	    *e = degA - degB + 1;
	ba0_restore (&M);
	return;
    }
/*
 * If yes, view A as the sum of the lambda_i * T_i  where the T_i are
 * terms over the alphabet of the non shared variables.
 *
 * Apply the elementary method over the lambda_i and adjust.
 * The quotient and the remainder are created using creators.
 * Non shared variables are greater than the other ones.
 */
    v = A->rang_total.rg [i].var;
    r = bav_R_copy_ordering (bav_R_Iordering ());
    bav_R_push_ordering (r);
    while (i >= 0)
    {	w = A->rang_total.rg [i].var;
	if (w->hack == false)
	    bav_R_set_maximal_variable (w);
	i--;
    }

    if (A->acces == bap_sequential_monom_access)
    {	AA = A;
	bap_physort_polynom_COEFF (AA, bap_invertible_transform);
    } else
    {	AA = bap_new_readonly_polynom_COEFF ();
	bap_sort_polynom_COEFF (AA, A);
    }
    bap_begin_itercoeff_COEFF (&iter, AA, v);

    bav_init_term (&T);
    bav_init_term (&U);
    if (degB > 0)
    {	bav_shift_term (&U, &B->rang_total);
    	bav_pow_term (&U, &U, degA - degB);
    } else
	bav_pow_term (&U, &B->rang_total, degA);
    bav_mul_term (&T, &U, &AA->rang_total);

    nbmonA = bap_nbmon_polynom_COEFF (A) / BA0_MAX (1, degA);
    nbmonB = bap_nbmon_polynom_COEFF (B) / BA0_MAX (1, degB);
    nbmon = BA0_MAX (nbmonA, nbmonB);
    nbmon = BA0_MAX (1, nbmon);
    if (Q != BAP_NOT_A_POLYNOM_COEFF)
    {	bap_push_clot_parameters_stack 
	    (bap_current_typeof_termanager (), bap_random_monom_sequence);
	Pquotient = bap_new_polynom_COEFF ();
	bap_begin_creator_COEFF 
		(&crea_quotient, Pquotient, &T, bap_approx_total_rank, 
		 nbmon * (degA - degB + 1));
	bap_pull_clot_parameters_stack ();
    }

    if (R != BAP_NOT_A_POLYNOM_COEFF)
    {	if (degB > 0)
	{   bav_shift_term (&U, &B->rang_total);
	    bav_mul_term (&T, &T, &U);
	} else
	    bav_mul_term (&T, &T, &B->rang_total);
	bap_push_clot_parameters_stack 
	    (bap_current_typeof_termanager (), bap_random_monom_sequence);
	Preste = bap_new_polynom_COEFF ();
	bap_begin_creator_COEFF 
			(&crea_reste, Preste, &T, bap_approx_total_rank, 
			 nbmon * (degB + 1));
	bap_pull_clot_parameters_stack ();
    }

    bap_init_readonly_polynom_COEFF (&lambda);
    quotient = Q == BAP_NOT_A_POLYNOM_COEFF ? Q : bap_new_polynom_COEFF ();
    reste = R == BAP_NOT_A_POLYNOM_COEFF ? R : bap_new_polynom_COEFF ();

    while (! bap_outof_itercoeff_COEFF (&iter))
    {	bap_coeff_itercoeff_COEFF (&lambda, &iter);
	deg_lambda = bap_degree_polynom_COEFF (&lambda, x);
	bap_term_itercoeff_COEFF (&T, &iter);
	if (deg_lambda >= degB)
	    bap_pseudo_division_elem_polynom_COEFF 
					(quotient, reste, &lambda, B, x, &ibp);
	if (Q != BAP_NOT_A_POLYNOM_COEFF && deg_lambda >= degB)
	{   bap_mul_polynom_COEFF 
		(quotient, quotient, ibp.tab [degA - deg_lambda]);
	    bap_begin_itermon_COEFF (&itermon, quotient);
	    while (! bap_outof_itermon_COEFF (&itermon))
	    {	lc = bap_coeff_itermon_COEFF (&itermon);
		bap_term_itermon_COEFF (&U, &itermon);
		bav_mul_term (&U, &U, &T);
		bap_write_creator_COEFF (&crea_quotient, &U, *lc);
		bap_next_itermon_COEFF (&itermon);
	    }
	}
	if (R != BAP_NOT_A_POLYNOM_COEFF)
	{   if (deg_lambda >= degB)
	    {	bap_mul_polynom_COEFF 
			(reste, reste, ibp.tab [degA - deg_lambda]);
	    } else
		bap_mul_polynom_COEFF 
			(reste, &lambda, ibp.tab [degA - degB + 1]);
	    bap_begin_itermon_COEFF (&itermon, reste);
	    while (! bap_outof_itermon_COEFF (&itermon))
	    {	lc = bap_coeff_itermon_COEFF (&itermon);
		bap_term_itermon_COEFF (&U, &itermon);
		bav_mul_term (&U, &U, &T);
		bap_write_creator_COEFF (&crea_reste, &U, *lc);
		bap_next_itermon_COEFF (&itermon);
	    }
	}
	bap_next_itercoeff_COEFF (&iter);
    }

    if (Q != BAP_NOT_A_POLYNOM_COEFF)
    {	bap_close_creator_COEFF (&crea_quotient);
	bav_R_pull_ordering ();
	bap_physort_polynom_COEFF (Pquotient, bap_definitive_transform);
	bav_R_push_ordering (r);
    }
    if (R != BAP_NOT_A_POLYNOM_COEFF)
    {	bap_close_creator_COEFF (&crea_reste);
	bav_R_pull_ordering ();
	bap_physort_polynom_COEFF (Preste, bap_definitive_transform);
	bav_R_push_ordering (r);
    }

    if (A->acces == bap_sequential_monom_access)
    {	bav_R_pull_ordering ();
	bap_physort_polynom_COEFF (A, bap_inverse_transform);
	bav_R_push_ordering (r);
    }

    bav_R_pull_ordering ();
    bav_R_free_ordering (r);

    ba0_pull_stack ();
    if (Q != BAP_NOT_A_POLYNOM_COEFF)
	bap_set_polynom_COEFF (Q, Pquotient);
    if (R != BAP_NOT_A_POLYNOM_COEFF)
	bap_set_polynom_COEFF (R, Preste);
    if (e != (bav_Idegree*)0)
	*e = degA - degB + 1;

    ba0_restore (&M);
}

/* Variant of the above function.  */

BAP_DLL void bap_prem_polynom_COEFF 
		(bap_polynom_COEFF R, bav_Idegree* e, 
		 bap_polynom_COEFF A, bap_polynom_COEFF B, bav_variable x)
{
    bap_pseudo_division_polynom_COEFF (BAP_NOT_A_POLYNOM_COEFF, R, e, A, B, x);
}

/* Variant of the above function.  */

BAP_DLL void bap_pquo_polynom_COEFF 
		(bap_polynom_COEFF Q, bav_Idegree* e, 
		 bap_polynom_COEFF A, bap_polynom_COEFF B, bav_variable x)
{
    bap_pseudo_division_polynom_COEFF (Q, BAP_NOT_A_POLYNOM_COEFF, e, A, B, x);
}

/*
 * Division of A by B viewed as polynomials in v.
 * In this function, it is assumed that the initial of B divides 
 * exactly the initial of the current remainder while the degree of this
 * remainder (w.r.t. v) is greater than or equal to that of B.
 */

BAP_DLL void bap_rem_polynom_COEFF (bap_polynom_COEFF Q, bap_polynom_COEFF R, 
		bap_polynom_COEFF A, bap_polynom_COEFF B, bav_variable v)
{   struct bap_polynom_COEFF reste, lcR, redR, lcB, redB, C;
    struct bap_geobucket_COEFF quotient;
    struct bav_term T;
    bav_Idegree ddeg;
    bool first;
    struct bav_rank rg;
    struct ba0_mark M;

    if (bap_is_zero_polynom_COEFF (B))
	BA0_RAISE_EXCEPTION (BAP_ERRCST);
    if (v == BAV_NOT_A_VARIABLE && bap_is_numeric_polynom_COEFF (B))
        BA0_RAISE_EXCEPTION (BAP_ERRCST);
    if ((Q != BAP_NOT_A_POLYNOM_COEFF && Q->readonly) || 
				(R != BAP_NOT_A_POLYNOM_COEFF && R->readonly))
	BA0_RAISE_EXCEPTION (BAP_ERRCOE);
    if (v == BAV_NOT_A_VARIABLE)
	v = bap_leader_polynom_COEFF (B);
    ddeg = bap_degree_polynom_COEFF (A, v) - bap_degree_polynom_COEFF (B, v);
    if (ddeg < 0)
    {	bap_set_polynom_zero_COEFF (Q);
	if (R != A)
	    bap_set_polynom_COEFF (R, A);
	return;
    }

    ba0_push_another_stack ();
    ba0_record (&M);

    bap_init_readonly_polynom_COEFF (&lcB);
    bap_init_readonly_polynom_COEFF (&redB);
    bap_initial_and_reductum2_polynom_COEFF (&lcB, &redB, B, v);

    bap_init_geobucket_COEFF (&quotient);
    bap_init_polynom_COEFF (&reste);
    bap_init_readonly_polynom_COEFF (&lcR);
    bap_init_readonly_polynom_COEFF (&redR);
    bap_init_polynom_COEFF (&C);
    bav_init_term (&T);
    rg.var = v;

    first = true;
    while (ddeg >= 0)
    {	bap_initial_and_reductum2_polynom_COEFF 
					(&lcR, &redR, first ? A : &reste, v);
	bap_exquo_polynom_COEFF (&C, &lcR, &lcB);
    	rg.deg = ddeg;
	bav_set_term_rank (&T, &rg);
	bap_mul_polynom_term_COEFF (&C, &C, &T);
	bap_add_geobucket_COEFF (&quotient, &C);
	bap_mul_polynom_COEFF (&C, &C, &redB);
	bap_sub_polynom_COEFF (&reste, &redR, &C);
	ddeg = bap_degree_polynom_COEFF (&reste, v) - 
					bap_degree_polynom_COEFF (B, v);
	first = false;
    }
    ba0_pull_stack ();
    bap_set_polynom_COEFF (R, &reste);
    bap_set_polynom_geobucket_COEFF (Q, &quotient);
    ba0_restore (&M);
}

#endif
#undef BAD_FLAG_COEFF
