#include "bap_polyspec_mpz.h"
#include "bap_gcd_polynom_mpz.h"
#include "bap_add_polynom_mpz.h"
#include "bap_mul_polynom_mpz.h"
#include "bap_polynom_mpq.h"
#include "bap_mul_polynom_mpq.h"
#include "bap_polyspec_mpq.h"
#include "bap_diff_polynom_mpz.h"
#include "bap_parse_polynom_mpz.h"
#include "bap_parse_polynom_mpq.h"
#include "bap_ratfrac_mpz.h"

/*
 * Initializes R to zero.
 */

BAP_DLL void bap_init_ratfrac_mpz (bap_ratfrac_mpz R)
{
    bap_init_polynom_mpz (&R->numer);
    bap_init_polynom_one_mpz (&R->denom);
}

BAP_DLL void bap_init_readonly_ratfrac_mpz (bap_ratfrac_mpz R)
{
    bap_init_readonly_polynom_mpz (&R->numer);
    bap_init_polynom_one_mpz (&R->denom);
}

/* 
 * Allocates a new fraction and initializes it to zero.
 */

BAP_DLL bap_ratfrac_mpz bap_new_ratfrac_mpz (void)
{   bap_ratfrac_mpz A;

    A = (bap_ratfrac_mpz)ba0_alloc (sizeof (struct bap_ratfrac_mpz));
    bap_init_ratfrac_mpz (A);
    return A;
}

/* 
 * Sets A to B.
 */

BAP_DLL void bap_set_ratfrac_mpz (bap_ratfrac_mpz A, bap_ratfrac_mpz B)
{
    if (A != B)
    {	bap_set_polynom_mpz (&A->numer, &B->numer);
    	bap_set_polynom_mpz (&A->denom, &B->denom);
    }
}

/* 
 * Sets A to zero.
 */

BAP_DLL void bap_set_ratfrac_zero_mpz (bap_ratfrac_mpz A)
{
    bap_set_polynom_zero_mpz (&A->numer);
    bap_set_polynom_one_mpz (&A->denom);
}

/* 
 * Sets A to one.
 */

BAP_DLL void bap_set_ratfrac_one_mpz (bap_ratfrac_mpz A)
{
    bap_set_polynom_one_mpz (&A->numer);
    bap_set_polynom_one_mpz (&A->denom);
}

/*
 * Sets R to A/1.
 */

BAP_DLL void bap_set_ratfrac_polynom_mpz (bap_ratfrac_mpz R, bap_polynom_mpz A)
{
    bap_set_polynom_mpz (&R->numer, A);
    bap_set_polynom_one_mpz (&R->denom);
}

/*
 * Sets R to P / Q.
 */

BAP_DLL void bap_set_ratfrac_fraction_mpz
		(bap_ratfrac_mpz R, bap_polynom_mpz P, bap_polynom_mpz Q)
{
    if (bap_is_zero_polynom_mpz (P))
	bap_set_ratfrac_zero_mpz (R);
    else if (bap_is_zero_polynom_mpz (Q))
	BA0_RAISE_EXCEPTION (BA0_ERRIVZ);
    else
    {	bap_set_polynom_mpz (&R->numer, P);
	bap_set_polynom_mpz (&R->denom, Q);
    }
}

BAP_DLL void* bap_scanf_ratfrac_eqn_mpz (void* R)
{   bap_ratfrac_mpz A = (bap_ratfrac_mpz)R;
    struct bap_ratfrac_mpz B, C;
    struct ba0_mark M;

    if (A == (bap_ratfrac_mpz)0)
        A = bap_new_ratfrac_mpz ();

    ba0_push_another_stack ();
    ba0_record (&M);

    bap_init_ratfrac_mpz (&B);
    bap_scanf_ratfrac_mpz (&B);

    if (! ba0_sign_token_analex ("="))
    {	ba0_pull_stack ();
	bap_set_ratfrac_mpz (A, &B);
    } else
    {	bap_init_ratfrac_mpz (&C);
	ba0_get_token_analex ();
	bap_scanf_ratfrac_mpz (&C);
	ba0_pull_stack ();
	bap_sub_ratfrac_mpz (A, &B, &C);
    }

    ba0_restore (&M);
    return A;
}

static void* bap_scanf2_ratfrac_mpz (void*);

BAP_DLL void* bap_scanf_ratfrac_mpz (void* R)
{   bap_ratfrac_mpz A;

    A = (bap_ratfrac_mpz)bap_scanf2_ratfrac_mpz (R);
    bap_reduce_ratfrac_mpz (A, A);
    return A;
}

static void* bap_scanf2_ratfrac_mpz (void* R)
{   bap_ratfrac_mpz A = (bap_ratfrac_mpz)R;
    struct bap_ratfrac_mpz S, P;
    struct ba0_mark M;

    if (A == (bap_ratfrac_mpz)0)
	A = bap_new_ratfrac_mpz ();

    ba0_push_another_stack ();
    ba0_record (&M);
    bap_init_ratfrac_mpz (&S);
    bap_init_ratfrac_mpz (&P);

    if (! ba0_sign_token_analex ("-"))
    {   bap_scanf_product_ratfrac_mpz (&S);
        ba0_get_token_analex ();
    }
    while (ba0_sign_token_analex ("+") || ba0_sign_token_analex ("-"))
    {   if (ba0_sign_token_analex ("+"))
        {   ba0_get_token_analex ();
            bap_scanf_product_ratfrac_mpz (&P);
            ba0_get_token_analex ();
	    bap_add_ratfrac_mpz (&S, &S, &P);
        } else
        {   ba0_get_token_analex ();
            bap_scanf_product_ratfrac_mpz (&P);
            ba0_get_token_analex ();
	    bap_sub_ratfrac_mpz (&S, &S, &P);
        }
    }
    ba0_unget_token_analex ();
    ba0_pull_stack ();
    bap_set_ratfrac_mpz (A, &S);
    ba0_restore (&M);
    return A;
}

static ba0_scanf_function bap_scanf_power_ratfrac_mpz;

BAP_DLL void* bap_scanf_product_ratfrac_mpz (void* AA)
{   bap_ratfrac_mpz A;
    struct bap_ratfrac_mpz B;
    struct ba0_mark M;

    if (AA == (void*)0)
        A = bap_new_ratfrac_mpz ();
    else
        A = (bap_ratfrac_mpz)AA;

    ba0_push_another_stack ();
    ba0_record (&M);
    bap_init_ratfrac_mpz (&B);
    ba0_pull_stack ();

    bap_scanf_power_ratfrac_mpz (A);

    ba0_get_token_analex ();
    while (ba0_sign_token_analex ("*") || ba0_sign_token_analex ("/"))
    {	if (ba0_sign_token_analex ("*"))
	{   ba0_get_token_analex ();
	    ba0_push_another_stack ();
	    bap_scanf_power_ratfrac_mpz (&B);
	    ba0_pull_stack ();
	    bap_mul_ratfrac_mpz (A, A, &B);
	} else
	{   ba0_get_token_analex ();
	    ba0_push_another_stack ();
	    bap_scanf_power_ratfrac_mpz (&B);
	    ba0_pull_stack ();
	    bap_div_ratfrac_mpz (A, A, &B);
	}
	ba0_get_token_analex ();
    }

    ba0_unget_token_analex ();

    ba0_restore (&M);
    return A;
}

static ba0_scanf_function bap_scanf_atomic_ratfrac_mpz;

static void* bap_scanf_power_ratfrac_mpz (void* AA)
{   bap_ratfrac_mpz A;
    bav_Idegree d;

    if (AA == (void*)0)
        A = bap_new_ratfrac_mpz ();
    else
	A = (bap_ratfrac_mpz)AA;

    bap_scanf_atomic_ratfrac_mpz (A);
    ba0_get_token_analex ();
    if (ba0_sign_token_analex ("^"))
    {   ba0_get_token_analex ();
        if (ba0_type_token_analex () != ba0_integer_token)
            BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);
	d = (bav_Idegree)atoi (ba0_value_token_analex ());
	bap_pow_ratfrac_mpz (A, A, d);
    } else
	ba0_unget_token_analex ();

    return A;
}

static void* bap_scanf_atomic_ratfrac_mpz (void* AA)
{   bap_ratfrac_mpz A;
    struct bap_polynom_mpz P;
    bav_variable v;
    ba0_mpz_t c;
    struct ba0_mark M;

    if (AA == (void*)0)
	A = bap_new_ratfrac_mpz ();
    else
	A = (bap_ratfrac_mpz)AA;

    if (ba0_type_token_analex () == ba0_integer_token)
    {   struct bav_rank rg;

	ba0_push_another_stack ();
	ba0_record (&M);
	ba0_mpz_init (c);
	ba0_scanf ("%z", c);
	bap_init_polynom_mpz (&P);
	rg = bav_constant_rank ();
	bap_set_polynom_crk_mpz (&P, c, &rg);
	ba0_pull_stack ();
	bap_set_ratfrac_polynom_mpz (A, &P);
	ba0_restore (&M);
    } else if (ba0_type_token_analex () == ba0_string_token)
    {	ba0_push_another_stack ();
	ba0_record (&M);
	bap_init_polynom_mpz (&P);
	ba0_scanf ("%v", &v);
	bap_set_polynom_variable_mpz (&P, v, 1);
	ba0_pull_stack ();
	bap_set_ratfrac_polynom_mpz (A, &P);
	ba0_restore (&M);
    } else if (ba0_sign_token_analex ("("))
    {	ba0_get_token_analex ();
	bap_scanf2_ratfrac_mpz (A);
	ba0_get_token_analex ();
	if (! ba0_sign_token_analex (")"))
	    BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);
    } else
	BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);
    return A;
}

/*
 * Should be able to parse an "expanded" rational fractional such
 * as the ones produced by the MAPLE command:
 *
 * lprint (expand (numer (ratfrac)) / expand (denom (ratfrac)))
 *
 * The format of such expressions is not that clear.
 * There may be a leading numerical factor.
 * There may also be rational numbers within the numerator and the denominator
 */

BAP_DLL void* bap_scanf_expanded_ratfrac_mpz (void* AA)
{   bap_ratfrac_mpz A;
    struct bap_polynom_mpq Anum, Aden;
    struct ba0_mark M;
    ba0_mpq_t q;
    ba0_mpz_t qnum, qden;
    ba0_mpz_t den_numer, den_denom;
    bool minus, coeffnum;
    ba0_int_p n;

    if (AA == (void*)0)
	A = bap_new_ratfrac_mpz ();
    else
	A = (bap_ratfrac_mpz)AA;

    ba0_push_another_stack ();
    ba0_record (&M);
/*
 * Read a possible starting minus sign.
 * This minus sign will only matter if there is a leading numerical factor.
 * It will then apply to this factor.
 */

    if (ba0_sign_token_analex ("-"))
    {	ba0_get_token_analex ();
	minus = true;
	n = 1;
    } else
    {	minus = false;
	n = 0;
    }
/*
 * Read a possible leading numerical factor.
 * We are not yet sure that the numerical factor is a global factor.
 *
 * Observe that we cannot read directly a rational number for
 * a rational fraction of the form 1/u would raise an exception.
 */
    ba0_mpz_init_set_ui (qnum, 1);
    ba0_mpz_init_set_ui (qden, 1);
    if (ba0_type_token_analex () == ba0_integer_token)
    {	coeffnum = true;
	ba0_scanf_mpz (qnum);
	ba0_get_token_analex ();
	n += 1;
	if (ba0_sign_token_analex ("/"))
	{   ba0_get_token_analex ();
	    n += 1;
	    if (ba0_type_token_analex () == ba0_integer_token)
	    {	ba0_scanf_mpz (qden);
		ba0_get_token_analex ();
		n += 1;
	    } else
	    {   ba0_unget_token_analex ();
		n -= 1;
	    }
	}
    } else
	coeffnum = false;
    ba0_mpq_init (q);
    mpq_set_num (q, qnum);
    mpq_set_den (q, qden);
    mpq_canonicalize (q);
    if (minus)
	ba0_mpq_neg (q, q);
/*
 * If there is a leading numerical factor then, it is in q.
 * The minus sign is taken into account.
 * 
 * If there is no leading numerical factor then we will have to
 * undo all ba0_get_token before reding the numerator. In this case,
 * the number of ba0_unget_token to perform is stored in n.
 */
    if (!coeffnum)
    {	if (ba0_sign_token_analex ("*") || ba0_sign_token_analex ("/") ||
		ba0_sign_token_analex ("+") || ba0_sign_token_analex ("-"))
	    BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);
    }
/*
 * Let us read the numerator and denominator of the rational fraction
 */
    bap_init_polynom_mpq (&Anum);
    bap_init_polynom_mpq (&Aden);
    bap_set_polynom_one_mpq (&Anum);
    bap_set_polynom_one_mpq (&Aden);

    if ((coeffnum && ba0_sign_token_analex ("*")) || !coeffnum)
    {	if (ba0_sign_token_analex ("*"))
	{   ba0_get_token_analex ();
            n += 1;
	}
	if (! ba0_sign_token_analex ("("))
	{
/*
 * The leading rational is actually not a leading numerical factor.
 * Undo all ba0_get_token_analex and reset q to 1/1.
 */
	    while (n > 0)
	    {	ba0_unget_token_analex ();
		n -= 1;
	    }
	    ba0_mpq_set_ui (q, 1);
	}
/*
 * Read numerator and denominator
 */
	bap_scanf_expanded_polynom_mpq (&Anum);
	ba0_get_token_analex ();
	if (ba0_sign_token_analex ("/"))
	{   ba0_get_token_analex ();
	    bap_scanf_expanded_polynom_mpq (&Aden);
	} else
	    ba0_unget_token_analex ();
    } else if (ba0_sign_token_analex ("+") || ba0_sign_token_analex ("-"))
    {
/*
 * numerical factor +/- ...
 */
	while (n > 0)
	{   ba0_unget_token_analex ();
	    n -= 1;
        }
        ba0_mpq_set_ui (q, 1);
/*
 * Read numerator 
 */
        bap_scanf_expanded_polynom_mpq (&Anum);
    } else if (ba0_sign_token_analex ("/"))
    {   ba0_get_token_analex ();
	bap_scanf_expanded_polynom_mpq (&Aden);
    } else
	ba0_unget_token_analex ();
/*
 * Incorporate the numerical factor q in Anum.
 */
    bap_mul_polynom_numeric_mpq (&Anum, &Anum, q);
/*
 * Let us compute the result
 */
    ba0_pull_stack ();

    ba0_mpz_init (den_numer);
    ba0_mpz_init (den_denom);
    bap_numer_polynom_mpq (&A->numer, den_numer, &Anum);
    bap_numer_polynom_mpq (&A->denom, den_denom, &Aden);
    bap_mul_polynom_numeric_mpz (&A->numer, &A->numer, den_denom);
    bap_mul_polynom_numeric_mpz (&A->denom, &A->denom, den_numer);
    if (ba0_mpz_sgn (*bap_numeric_initial_polynom_mpz (&A->denom)) < 0)
    {	bap_neg_polynom_mpz (&A->numer, &A->numer);
	bap_neg_polynom_mpz (&A->denom, &A->denom);
    }
    ba0_restore (&M);
    return A;
}

BAP_DLL void* bap_scanf_careful_expanded_ratfrac_mpz (void* AA)
{   bap_ratfrac_mpz A;
/*
 * Fix me.
 * not a call to scanf_expanded_ratfrac which misses some cases.
 * The call to bap_scanf2_ratfrac_mpz catches u[x,y] - u[y,x] = 0
 */
    A = bap_scanf2_ratfrac_mpz (AA);
    if (! bav_depends_on_zero_derivatives_of_parameter_term
				(&A->numer.rang_total, &bav_parameters) &&
		! bav_depends_on_zero_derivatives_of_parameter_term
				(&A->denom.rang_total, &bav_parameters))
	return A;

    if (bav_depends_on_zero_derivatives_of_parameter_term 
				(&A->denom.rang_total, &bav_parameters))
    {	bap_simplify_zero_derivatives_of_parameter_polynom_mpz
				(&A->denom, &A->denom, &bav_parameters);
	if (bap_is_zero_polynom_mpz (&A->denom))
	    BA0_RAISE_EXCEPTION (BA0_ERRIVZ);
/*
 * In the cae of an exception, A gets corrupted.
 */
    } 
    if (bav_depends_on_zero_derivatives_of_parameter_term 
				(&A->numer.rang_total, &bav_parameters))
	bap_simplify_zero_derivatives_of_parameter_polynom_mpz 
				(&A->numer, &A->numer, &bav_parameters);
    bap_reduce_ratfrac_mpz (A, A);
    return A;
}

BAP_DLL void bap_printf_ratfrac_mpz (void* AA)
{   bap_ratfrac_mpz A = (bap_ratfrac_mpz)AA;

    if (bap_is_zero_ratfrac_mpz (A))
	ba0_put_char ('0');
    else if (bap_is_one_polynom_mpz (&A->denom))
	bap_printf_polynom_mpz (&A->numer);
    else
    {   ba0_put_char ('(');
	bap_printf_polynom_mpz (&A->numer);
	ba0_put_string (")/(");
	bap_printf_polynom_mpz (&A->denom);
	ba0_put_char (')');
    }
}

/*
 * Readonly static data
 */

static char _struct_ratfrac [] = "struct bap_ratfrac_mpz";

BAP_DLL ba0_int_p bap_garbage1_ratfrac_mpz 
					(void* AA, enum ba0_garbage_code code)
{   bap_ratfrac_mpz A = (bap_ratfrac_mpz)AA;
    ba0_int_p n = 0;

    if (code == ba0_isolated)
	n += ba0_new_gc_info 
		(A, sizeof(struct bap_ratfrac_mpz), _struct_ratfrac);

    n += bap_garbage1_polynom_mpz (&A->numer, ba0_embedded);
    n += bap_garbage1_polynom_mpz (&A->denom, ba0_embedded);

    return n;
}

BAP_DLL void* bap_garbage2_ratfrac_mpz (void* AA, enum ba0_garbage_code code)
{   bap_ratfrac_mpz A;

    if (code == ba0_isolated)
	A = ba0_new_addr_gc_info (AA, _struct_ratfrac);
    else
	A = (bap_ratfrac_mpz)AA;

    bap_garbage2_polynom_mpz (&A->numer, ba0_embedded);
    bap_garbage2_polynom_mpz (&A->denom, ba0_embedded);
    return A;
}

BAP_DLL void* bap_copy_ratfrac_mpz (void* AA)
{   bap_ratfrac_mpz B;

    B = bap_new_ratfrac_mpz ();
    bap_set_ratfrac_mpz (B, (bap_ratfrac_mpz)AA);
    return B;
}

BAP_DLL void bap_sort_ratfrac_mpz (bap_ratfrac_mpz R, bap_ratfrac_mpz A)
{
    bap_sort_polynom_mpz (&R->numer, &A->numer);
    bap_sort_polynom_mpz (&R->denom, &A->denom);
}

BAP_DLL void bap_physort_ratfrac_mpz 
			(bap_ratfrac_mpz R, enum bap_typeof_transform type)
{
    bap_physort_polynom_mpz (&R->numer, type);
    bap_physort_polynom_mpz (&R->denom, type);
}

/*
 * Returns true if R is zero.
 */

BAP_DLL bool bap_is_zero_ratfrac_mpz (bap_ratfrac_mpz R)
{
    return bap_is_zero_polynom_mpz (&R->numer);
}

/* 
 * Returns true if R is one.
 */

BAP_DLL bool bap_is_one_ratfrac_mpz (bap_ratfrac_mpz R)
{
    return bap_equal_polynom_mpz (&R->numer, &R->denom);
}

BAP_DLL bool bap_is_numeric_ratfrac_mpz (bap_ratfrac_mpz R)
{
    return bap_is_numeric_polynom_mpz (&R->numer) &&
		bap_is_numeric_polynom_mpz (&R->denom);
}

BAP_DLL bool bap_equal_ratfrac_mpz (bap_ratfrac_mpz A, bap_ratfrac_mpz B)
{
    return bap_equal_polynom_mpz (&A->numer, &B->numer) &&
	   bap_equal_polynom_mpz (&A->denom, &B->denom);
}

BAP_DLL bool bap_depend_ratfrac_mpz (bap_ratfrac_mpz A, bav_variable v)
{
    return bap_depend_polynom_mpz (&A->numer, v) ||
		bap_depend_polynom_mpz (&A->denom, v);
}

BAP_DLL bav_variable bap_leader_ratfrac_mpz (bap_ratfrac_mpz R)
{   bav_variable w = BAV_NOT_A_VARIABLE;

    if (bap_is_numeric_polynom_mpz (&R->numer))
    {	if (bap_is_numeric_polynom_mpz (&R->denom))
	    BA0_RAISE_EXCEPTION (BAP_ERRCST);
	else
	    w = bap_leader_polynom_mpz (&R->denom);
    } else if (bap_is_numeric_polynom_mpz (&R->denom))
	w = bap_leader_polynom_mpz (&R->numer);
    else
    {	bav_variable u, v;
	u = bap_leader_polynom_mpz (&R->numer);
	v = bap_leader_polynom_mpz (&R->denom);
	w = bav_R_variable_number (u) > bav_R_variable_number (v) ? u : v;
    }
    return w;
}

/* 
 * Sets A to the numerator of B.
 */

BAP_DLL void bap_numer_ratfrac_mpz (bap_polynom_mpz A, bap_ratfrac_mpz R)
{
    bap_set_polynom_mpz (A, &R->numer);
}

/* 
 * Sets A to the denominator of B.
 */

BAP_DLL void bap_denom_ratfrac_mpz (bap_polynom_mpz A, bap_ratfrac_mpz R)
{
    bap_set_polynom_mpz (A, &R->denom);
}

BAP_DLL void bap_reduce_numeric_ratfrac_mpz 
				(bap_ratfrac_mpz R, bap_ratfrac_mpz A)
{   ba0_mpz_t contnum, contden, gcd;
    struct ba0_mark M;

    if (bap_is_zero_ratfrac_mpz (A))
	bap_set_ratfrac_zero_mpz (R);
    else
    {	ba0_push_another_stack ();
	ba0_record (&M);
	ba0_mpz_init (contnum);
	ba0_mpz_init (contden);
	ba0_mpz_init (gcd);
	bap_numeric_content_polynom_mpz (contnum, &A->numer);
	bap_numeric_content_polynom_mpz (contden, &A->denom);
	mpz_gcd (gcd, contnum, contden);
	if (ba0_mpz_sgn (*bap_numeric_initial_polynom_mpz (&A->denom)) < 0)
	    mpz_neg (gcd, gcd);
	ba0_pull_stack ();
	bap_exquo_polynom_numeric_mpz (&R->numer, &A->numer, gcd);
	bap_exquo_polynom_numeric_mpz (&R->denom, &A->denom, gcd);
	ba0_restore (&M);
    }
}

/* 
 * Sets R to the reduced fraction A.
 * Factors the gcd of the numerator and the denominator of A out.
 * The numerical leading coefficient of the denominator of R is positive.
 */

BAP_DLL void bap_reduce_ratfrac_mpz (bap_ratfrac_mpz R, bap_ratfrac_mpz A)
{
    if (bap_is_zero_polynom_mpz (&A->numer))
	bap_set_ratfrac_zero_mpz (R);
    else if (bap_is_numeric_polynom_mpz (&A->denom))
	bap_reduce_numeric_ratfrac_mpz (R, A);
    else
    {	bap_gcd_polynom_mpz 
	    (BAP_NOT_A_POLYNOM_mpz, &R->numer, &R->denom, &A->numer, &A->denom);
	if (ba0_mpz_sgn (*bap_numeric_initial_polynom_mpz (&R->denom)) < 0)
	{   bap_neg_polynom_mpz (&R->numer, &R->numer);
	    bap_neg_polynom_mpz (&R->denom, &R->denom);
	}
    }
}

BAP_DLL void bap_neg_ratfrac_mpz (bap_ratfrac_mpz R, bap_ratfrac_mpz A)
{
    bap_neg_polynom_mpz (&R->numer, &A->numer);
    if (R != A)
	bap_set_polynom_mpz (&R->denom, &A->denom);
}

BAP_DLL void bap_add_ratfrac_mpz 
		(bap_ratfrac_mpz R, bap_ratfrac_mpz A, bap_ratfrac_mpz B)
{   struct bap_polynom_mpz Abar, Bbar;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);

    bap_init_polynom_mpz (&Abar);
    bap_init_polynom_mpz (&Bbar);
    bap_mul_polynom_mpz (&Bbar, &A->numer, &B->denom);
    bap_mul_polynom_mpz (&Abar, &B->numer, &A->denom);

    ba0_pull_stack ();

    bap_add_polynom_mpz (&R->numer, &Abar, &Bbar);
    if (bap_is_zero_polynom_mpz (&R->numer))
	bap_set_ratfrac_zero_mpz (R);
    else
	bap_mul_polynom_mpz (&R->denom, &A->denom, &B->denom);

    ba0_restore (&M);
}

BAP_DLL void bap_sub_ratfrac_mpz 
		(bap_ratfrac_mpz R, bap_ratfrac_mpz A, bap_ratfrac_mpz B)
{   struct bap_polynom_mpz Abar, Bbar;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);

    bap_init_polynom_mpz (&Abar);
    bap_init_polynom_mpz (&Bbar);
    bap_mul_polynom_mpz (&Bbar, &A->numer, &B->denom);
    bap_mul_polynom_mpz (&Abar, &B->numer, &A->denom);

    ba0_pull_stack ();

    bap_sub_polynom_mpz (&R->numer, &Bbar, &Abar);
    if (bap_is_zero_polynom_mpz (&R->numer))
	bap_set_ratfrac_zero_mpz (R);
    else
	bap_mul_polynom_mpz (&R->denom, &A->denom, &B->denom);

    ba0_restore (&M);
}

BAP_DLL void bap_mul_ratfrac_numeric_mpz 
			(bap_ratfrac_mpz R, bap_ratfrac_mpz A, ba0_mpz_t c)
{
    if (bap_is_zero_ratfrac_mpz (A) || ba0_mpz_sgn (c) == 0)
	bap_set_ratfrac_zero_mpz (R);
    else 
    {	if (ba0_mpz_cmp_ui (c, 1) != 0)
	    bap_mul_polynom_numeric_mpz (&R->numer, &A->numer, c);
	if (R != A)
	    bap_set_polynom_mpz (&R->denom, &A->denom);
    }
}

BAP_DLL void bap_mul_ratfrac_mpz 
		(bap_ratfrac_mpz R, bap_ratfrac_mpz A, bap_ratfrac_mpz B)
{
    bap_mul_polynom_mpz (&R->numer, &A->numer, &B->numer);
    if (bap_is_zero_polynom_mpz (&R->numer))
	bap_set_ratfrac_zero_mpz (R);
    else
	bap_mul_polynom_mpz (&R->denom, &A->denom, &B->denom);
}

BAP_DLL void bap_invert_ratfrac_mpz (bap_ratfrac_mpz R, bap_ratfrac_mpz A)
{
    if (bap_is_zero_ratfrac_mpz (A))
        BA0_RAISE_EXCEPTION (BA0_ERRIVZ);
    if (R == A)
    {	BA0_SWAP (struct bap_polynom_mpz, R->numer, R->denom);
    } else
    {	bap_set_polynom_mpz (&R->numer, &A->denom);
	bap_set_polynom_mpz (&R->denom, &A->numer);
    }
}

BAP_DLL void bap_div_ratfrac_mpz
		(bap_ratfrac_mpz R, bap_ratfrac_mpz A, bap_ratfrac_mpz B)
{   struct bap_ratfrac_mpz C;

    if (bap_is_zero_ratfrac_mpz (B))
	BA0_RAISE_EXCEPTION (BA0_ERRIVZ);

    C.numer = B->denom;
    C.denom = B->numer;

    bap_mul_ratfrac_mpz (R, A, &C);
}

BAP_DLL void bap_pow_ratfrac_mpz 
			(bap_ratfrac_mpz R, bap_ratfrac_mpz A, bav_Idegree d)
{
    if (d == 0)
	bap_set_ratfrac_one_mpz (R);
    else if (d == 1)
    {	if (R != A)
	    bap_set_ratfrac_mpz (R, A);
    } else
    {	bap_pow_polynom_mpz (&R->numer, &A->numer, d);
	bap_pow_polynom_mpz (&R->denom, &A->denom, d);
    }
}

BAP_DLL void bap_separant_ratfrac_mpz (bap_ratfrac_mpz R, bap_ratfrac_mpz A)
{
    struct bap_polynom_mpz sep_num_A;
    struct bap_polynom_mpz sep_den_A;
    bav_variable u;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);

    u = bap_leader_ratfrac_mpz (A);

    if (bap_is_numeric_polynom_mpz (&A->numer) ||
				bap_leader_polynom_mpz (&A->numer) != u)
    {   bap_init_polynom_mpz (&sep_den_A);
	bap_separant_polynom_mpz (&sep_den_A, &A->denom);
	bap_mul_polynom_mpz (&sep_den_A, &sep_den_A, &A->numer);
	ba0_pull_stack ();
	bap_neg_polynom_mpz (&R->numer, &sep_den_A);
	bap_pow_polynom_mpz (&R->denom, &A->denom, 2);
    } else if (bap_is_numeric_polynom_mpz (&A->denom) ||
				bap_leader_polynom_mpz (&A->denom) != u)
    {	ba0_pull_stack ();
	bap_separant_polynom_mpz (&R->numer, &A->numer);
	if (A != R)
	    bap_set_polynom_mpz (&R->denom, &A->denom);
    } else
    {	bap_init_polynom_mpz (&sep_num_A);
	bap_init_polynom_mpz (&sep_den_A);
	bap_separant_polynom_mpz (&sep_num_A, &A->numer);
	bap_separant_polynom_mpz (&sep_den_A, &A->denom);
	bap_mul_polynom_mpz (&sep_num_A, &sep_num_A, &A->denom);
	bap_mul_polynom_mpz (&sep_den_A, &sep_den_A, &A->numer);

	ba0_pull_stack ();

	bap_sub_polynom_mpz (&R->numer, &sep_num_A, &sep_den_A);
	bap_pow_polynom_mpz (&R->denom, &A->denom, 2);
    }
    ba0_restore (&M);
    bap_reduce_ratfrac_mpz (R, R);
}

BAP_DLL void bap_separant2_ratfrac_mpz 
			(bap_ratfrac_mpz R, bap_ratfrac_mpz A, bav_variable v)
{   struct bap_ratfrac_mpz B;
    bav_Iordering r;
    struct ba0_mark M;

    if (! bap_depend_ratfrac_mpz (A, v))
	bap_set_ratfrac_zero_mpz (R);
    else if (bap_leader_ratfrac_mpz (A) == v)
	bap_separant_ratfrac_mpz (R, A);
    else
    {	
	ba0_push_another_stack ();
	ba0_record (&M);
	r = bav_R_copy_ordering (bav_R_Iordering ());
	bav_R_push_ordering (r);
	bav_R_set_maximal_variable (v);
	bap_init_readonly_ratfrac_mpz (&B);
	bap_sort_ratfrac_mpz (&B, A);
	ba0_pull_stack ();
	bap_separant_ratfrac_mpz (R, &B);
	bav_R_pull_ordering ();
	bap_physort_ratfrac_mpz (R, bap_definitive_transform);
	bav_R_free_ordering (r);
	ba0_restore (&M);
    }
}

BAP_DLL void bap_diff_ratfrac_mpz (bap_ratfrac_mpz R, bap_ratfrac_mpz A, 
				bav_symbol s, bav_tableof_variable nulles)
{
    struct bap_polynom_mpz B;
    struct bap_polynom_mpz C;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);

    bap_init_polynom_mpz (&B);
    bap_init_polynom_mpz (&C);

    bap_diff_polynom_mpz (&B, &A->numer, s, nulles);
    bap_diff_polynom_mpz (&C, &A->denom, s, nulles);

    bap_mul_polynom_mpz (&B, &B, &A->denom);
    bap_mul_polynom_mpz (&C, &C, &A->numer);

    ba0_pull_stack ();

    bap_sub_polynom_mpz (&R->numer, &B, &C);
    if (bap_is_zero_polynom_mpz (&R->numer))
	bap_set_ratfrac_zero_mpz (R);
    else
	bap_pow_polynom_mpz (&R->denom, &A->denom, 2);

    ba0_restore (&M);
}

