#include "bav_common.h"
#include "bav_block.h"
#include "bav_symbol.h"
#include "bav_parameter.h"
#include "bav_variable.h"
#include "bav_ordering.h"
#include "bav_differential_ring.h"
#include "bav_ev_point.h"
#include "bav_rank.h"
#include "bav_term.h"
#include "bav_hterm.h"
#include "bav_term_ordering.h"
#include "bav_global.h"

BAV_DLL void bav_reset_all_settings (void)
{
    bav_set_settings_common (0);
    bav_set_settings_symbol (0, 0);
    bav_set_settings_variable (0, 0, 0);
    bav_set_settings_rank (0);
    bav_set_settings_ordering (0);

    ba0_reset_all_settings ();
}

/*
 * The default function called when a ba0_indexed is not recognized.
 */

BAV_DLL void bav_unknown_default (ba0_indexed s)
{   char* t;

    t = ba0_indexed_to_string (s);
    strncpy (bav_unknown, t, BA0_BUFSIZE-1);
}

BAV_DLL void bav_set_settings_common (ba0_indexed_function* u)
{
    bav_custom_unknown = u ? u : &bav_unknown_default;
}

BAV_DLL void bav_get_settings_common (ba0_indexed_function** u)
{
    if (u) *u = bav_custom_unknown;
}

BAV_DLL void bav_restart (ba0_int_p time_limit, ba0_int_p memory_limit)
{   enum ba0_restart_level level = ba0_actual_restart_level;
    bool no_oom, no_oot;

    ba0_get_settings_no_oot (&no_oot);
    ba0_set_settings_no_oot (true);
    ba0_get_settings_no_oom (&no_oom);
    ba0_set_settings_no_oom (true);

    ba0_restart (time_limit, memory_limit);

    switch (level)
    {
	case ba0_init_level:
	    ba0_init_stack (&bav_term_stack);
	    bav_term_stack.ident = "bav_term_stack";
	    bav_init_htableof_term (&bav_hterm);

	    bav_init_parameters (&bav_parameters);

	    ba0_define_format ("ordering", 
				&bav_scanf_ordering,
				&bav_printf_ordering,
				(ba0_garbage1_function*)0,
				(ba0_garbage2_function*)0,
				(ba0_copy_function*)0);

	    ba0_define_format ("b", &bav_scanf_block, 
			&bav_printf_block,
			(ba0_garbage1_function*)-1, 
			(ba0_garbage2_function*)-1,
			(ba0_copy_function*)-1);

	    ba0_define_format ("y", &bav_scanf_symbol, 
			&bav_printf_symbol,
			(ba0_garbage1_function*)0,
			(ba0_garbage2_function*)0,
			(ba0_copy_function*)0);

	    ba0_define_format ("v", &bav_scanf_variable, 
			&bav_printf_variable,
			(ba0_garbage1_function*)0,
			(ba0_garbage2_function*)0,
			(ba0_copy_function*)0);

	    ba0_define_format ("ev_point", 
			&bav_scanf_ev_point, &bav_printf_ev_point,
			&bav_garbage1_ev_point, &bav_garbage2_ev_point,
			&bav_copy_ev_point);

	    ba0_define_format ("rank",
			&bav_scanf_rank, &bav_printf_rank,
			(ba0_garbage1_function*)0,
                        (ba0_garbage2_function*)0,
                        (ba0_copy_function*)0);

	    ba0_define_format_with_sizelt 
			("term", sizeof (struct bav_term),
			&bav_scanf_term, 
			&bav_printf_term,
			&bav_garbage1_term, 
			&bav_garbage2_term,
			&bav_copy_term);

	    ba0_define_format_with_sizelt
			("param", sizeof (struct bav_parameter),
			&bav_scanf_parameter,
			&bav_printf_parameter,
			(ba0_garbage1_function*)0,
			(ba0_garbage2_function*)0,
			(ba0_copy_function*)0);

	case ba0_reset_level:
	    bav_reset_htableof_term (&bav_hterm);
	    bav_reset_parameters (&bav_parameters);
	    ba0_push_exception_extra_stack 
		(&bav_R.ord_stack.size, (ba0_extra_restore_function*)0);
	    ba0_push_exception_extra_stack
		(&bav_R.ords.size, &bav_R_restore_ords_size);
	    bav_R_init ();
	    bav_set_term_ordering ("lex");
	case ba0_done_level:
	    break;
    }
    ba0_set_settings_no_oom (no_oom);
    ba0_set_settings_no_oot (no_oot);
}

BAV_DLL void bav_terminate (enum ba0_restart_level level)
{
    if (level <= ba0_reset_level)
	ba0_clear_cells_stack (&bav_term_stack);

    if (level == ba0_init_level)
    {	ba0_clear_stack (&bav_term_stack);
	bav_clear_htableof_term (&bav_hterm);
    }

    ba0_terminate (level);
}

