#ifndef BAV_COMMON_H
#define BAV_COMMON_H 1

#include <ba0.h>

/*
 * The _MSC_VER flag is set if the code is compiled under WINDOWS
 * by using Microsoft Visual C (through Visual Studio 2008).
 *
 * In that case, some specific annotations must be added for DLL exports
 * Beware to the fact that this header file is going to be used either
 * for/while building BLAD or for using BLAD from an outer software.
 *
 * In the first case, functions are exported.
 * In the second one, they are imported.
 *
 * The flag BAV_BLAD_BUILDING must thus be set in the Makefile and passed
 * to the C preprocessor at BAV building time. Do not set it when using BAV.
 *
 * When compiling static libraries under Windows, set BA0_STATIC.
 */

#if defined (_MSC_VER) && ! defined (BA0_STATIC)
#    if defined (BAV_BLAD_BUILDING)
#       define BAV_DLL  __declspec(dllexport)
#    else
#       define BAV_DLL  __declspec(dllimport)
#    endif
#else
#    define BAV_DLL
#endif

#include "bav_mesgerr.h"

BEGIN_C_DECLS

typedef ba0_int_p bav_Idegree;
#define BAV_MAX_IDEGRE BA0_MAX_INT_P

typedef ba0_int_p bav_Iordering;
typedef ba0_int_p bav_Inumber;
typedef ba0_int_p bav_Iorder;

#define bav_tableof_Inumber     ba0_tableof_int_p
#define bav_tableof_Iorder      ba0_tableof_int_p
#define bav_tableof_Iordering   ba0_tableof_int_p
#define bav_tableof_Idegree	ba0_tableof_int_p

#ifdef __cplusplus
#define struct_bav_tableof_Inumber     struct_ba0_tableof_int_p
#define struct_bav_tableof_Iorder      struct_ba0_tableof_int_p
#define struct_bav_tableof_Iordering   struct_ba0_tableof_int_p
#define struct_bav_tableof_Idegree     struct_ba0_tableof_int_p
#endif

/*
 * ba0_indexed not recognized by parsers.
 * The function pointer, a buffer and the default value for the pointer
 */

#define bav_custom_unknown	bav_initialized_global.common.unknown
#define bav_unknown		bav_global.common.unknown
extern BAV_DLL ba0_indexed_function bav_unknown_default;

extern BAV_DLL void bav_set_settings_common (ba0_indexed_function*);
extern BAV_DLL void bav_get_settings_common (ba0_indexed_function**);

extern BAV_DLL void bav_reset_all_settings (void);
extern BAV_DLL void bav_restart (ba0_int_p, ba0_int_p);
extern BAV_DLL void bav_terminate (enum ba0_restart_level);

END_C_DECLS

#endif /* !BAV_COMMON_H */
