#ifndef BAV_GLOBAL_H
#define BAV_GLOBAL_H 1

#include "bav_common.h"
#include "bav_differential_ring.h"
#include "bav_parameter.h"
#include "bav_term.h"
#include "bav_hterm.h"

BEGIN_C_DECLS

struct bav_global {
    struct {
/*
 * Buffer for the default function called when an unknown 
 * variable/symbol/parameter is parsed. Receives the faulty string.
 */
	char  unknown [BA0_BUFSIZE];
    } common;
    struct {
/*
 * The main data structure (symbols, variables, rankings)
 */
	struct bav_differential_ring R;
    } differential_ring;
    struct {
/*
 * When set, its value is "diff" or "Diff" to display derivatives
 * following the MAPLE notation.
 *
 * Local to bav_variable.
 */
	char* diff_string;
    } variable;
    struct {
/*
 * Table of (variable, dependencies).
 * Given a variable v, provides the independent vars. v depends on.
 * Only used for displaying derivatives.
 *
 * Local to bav_parameter, bav_variable, bav_common.
 */
	struct bav_tableof_parameter table;
    } parameter;
    struct {
/*
 * Function pointers for comparing terms.
 * Used by bav_compare_term and bav_compare_stripped_term.
 *
 * Local to bav_term_ordering.
 */
	enum ba0_compare_code (*compare) (bav_term, bav_term);
	enum ba0_compare_code (*compare_stripped) 
					(bav_term, bav_term, bav_Inumber);
    } term_ordering;
    struct {
/*
 * stack  = the stack where htable is located
 * htable = the htable of terms.
 *
 * Both are local to bav_common and bav_hterm.
 */
	struct ba0_stack stack;
	struct bav_htableof_term htable;
    } hterm;
};

struct bav_initialized_global {
    struct {
/*
 * Function called when an unknown symbol/variable/parameter is parsed
 */
	ba0_indexed_function* unknown;
    } common;
    struct {
/*
 * Functions pointers for customizing symbol parsing and printing
 */
	ba0_scanf_function* scanf;
	ba0_printf_function* printf;
    } symbol;
    struct {
/*
 * Functions pointers for customizing variable parsing and printing
 */
	ba0_scanf_function* scanf;
	ba0_printf_function* printf;
/*
 * The prefix of temporary variables
 */
	char* temp_string;
    } variable;
    struct {
/*
 * Function pointer for customizing the way ranks are printed.
 */
	ba0_printf_function* printf;
    } rank;
    struct {
/*
 * The string for displaying orderings
 */
	char* ordering_string;
    } ordering;
};

extern BAV_DLL struct bav_global bav_global;
extern BAV_DLL struct bav_initialized_global bav_initialized_global;

END_C_DECLS

#endif /* !BAV_GLOBAL_H */
