#include "bav_hterm.h"
#include "bav_global.h"

/* 
 * Prime numbers. Possible sizes for bav_hterm
 * Readonly static data.
 */

static ba0_int_p size [] = {1021, 2039, 4093, 8191, 16381, 32749, 65521,
	131071, 262139, 524287, 1048573, 2097143, 4194301, 8388593,
	16777213, 33554393, 67108859, 134217689, 268435399, 536870909,
	1073741789, 2147483647 };

BAV_DLL void bav_init_htableof_term (bav_htableof_term ht)
{
    ht->G.alloc = 0;
    ht->G.size = 0;
    ht->G.index_in_H = (ba0_int_p*)0;
    ht->G.hack = (ba0_int_p*)0;
    ht->H.alloc = 0;
    ht->H.size = 0;
    ht->H.code = (ba0_int_p*)0;
    ht->H.term = (bav_term*)0;
}

BAV_DLL void bav_clear_htableof_term (bav_htableof_term ht)
{
    if (bav_hterm.G.index_in_H)
	ba0_free (bav_hterm.G.index_in_H);
    ba0_malloc_counter -= sizeof(ba0_int_p)*bav_hterm.G.alloc;
    if (bav_hterm.G.hack)
	ba0_free (bav_hterm.G.hack);
    ba0_malloc_counter -= sizeof(ba0_int_p)*bav_hterm.G.alloc;
    if (bav_hterm.H.term)
	ba0_free (bav_hterm.H.term);
    ba0_malloc_counter -= sizeof(bav_term)*bav_hterm.H.alloc;
    if (bav_hterm.H.code)
	ba0_free (bav_hterm.H.code);
    ba0_malloc_counter -= sizeof(ba0_int_p)*bav_hterm.H.alloc;
    bav_init_htableof_term (ht);
}

BAV_DLL void bav_reset_htableof_term (bav_htableof_term ht)
{
    ht->G.size = 0;
    memset (ht->H.term, 0, ht->H.alloc * sizeof(bav_term));
}

BAV_DLL ba0_int_p bav_nbelem_htableof_term (bav_htableof_term ht)
{
    return ht->G.size;
}

static bool bav_is_free (bav_htableof_term ht, ba0_int_p i)
{
    return ht->H.term [i] == 0;
}

static ba0_int_p primary_key (bav_htableof_term ht, bav_term T)
{   unsigned ba0_int_p t;
    ba0_int_p i, j;

    t = 10737417 % ht->H.size;
    for (i = 0; i < T->size; i++)
    {	for (j = 0; j < T->rg [i].deg; j++)
	{   t *= (T->rg [i].var->index + 3);
	    t %= ht->H.size;
	}
    }

    return t;
}

static ba0_int_p secondary_key (bav_htableof_term ht, bav_term T)
{   unsigned ba0_int_p t;
    ba0_int_p i, j;

    t = 71473701 % ht->H.size;
    for (i = 0; i < T->size; i++)
    {   for (j = 0; j < T->rg [i].deg; j++)
        {   t *= (T->rg [i].var->index + 5);
            t %= ht->H.size;
        }
    }
    if (t == 0) t = 1; /* to avoid infinite loops */
    return t;
}
    
static bav_hash_code bav_hash_term2 
		(bav_htableof_term ht, bav_term T, bool duplicate_term)
{   ba0_int_p c1, c2, i;

    if (3 * bav_nbelem_htableof_term (ht) >= ht->H.size)
	bav_resize_htableof_term (ht, ht->H.size + 1);

    c1 = primary_key (ht, T);
    c2 = secondary_key (ht, T);

    i = c1; 
    while (! bav_is_free (ht, i) && ! bav_equal_term (ht->H.term [i], T))
	i = (i + c2) % ht->H.size;

    if (! bav_is_free (ht, i))
	return ht->H.code [i];

    if (duplicate_term)
    {	ba0_push_stack (&bav_term_stack);
	ht->H.term [i] = (bav_term)bav_copy_term (T);
	ba0_pull_stack ();
    } else
	ht->H.term [i] = T;

    ht->H.code [i] = ht->G.size;
    ht->G.index_in_H [ht->G.size++] = i;

    return ht->H.code [i];
}

BAV_DLL void bav_resize_htableof_term 
				(bav_htableof_term ht, ba0_int_p lower_bound)
{
    struct bav_htableof_term new_ht;
    ba0_int_p i;

    bav_init_htableof_term (&new_ht);

    i = 0; 
    while (size [i] < lower_bound)
	i++;

    if (ht->H.size == 0)
    {   new_ht.G.alloc = size [i] / 2;
	new_ht.G.size = 0;
	new_ht.G.index_in_H = 
		(ba0_int_p*)ba0_malloc 
		   (sizeof (ba0_int_p) * new_ht.G.alloc);
	new_ht.G.hack = 
		(ba0_int_p*)ba0_malloc (sizeof (ba0_int_p) * new_ht.G.alloc);
    } else
    {	new_ht.G.alloc = ht->H.size;
	new_ht.G.size = 0;
	new_ht.G.index_in_H = ht->H.code;
	new_ht.G.hack = (ba0_int_p*)ht->H.term;
    }
    new_ht.H.alloc = size [i];
    new_ht.H.size = size [i];
    new_ht.H.code = ba0_malloc (sizeof (bav_hash_code) * size [i]);
    new_ht.H.term = ba0_malloc (sizeof (bav_term) * size [i]);

    memset (new_ht.H.term, 0, sizeof (bav_term) * size [i]);

    for (i = 0; i < ht->G.size; i++)
	bav_hash_term2 (&new_ht, ht->H.term [ht->G.index_in_H [i]], false);

    if (ht->H.size != 0)
	memcpy (new_ht.G.hack, ht->G.hack, ht->G.size * sizeof(ba0_int_p));

    if (ht->G.index_in_H)
	ba0_free (ht->G.index_in_H);
    ba0_malloc_counter -= sizeof(ba0_int_p)*ht->G.alloc;
    if (ht->G.hack)
	ba0_free (ht->G.hack);
    ba0_malloc_counter -= sizeof(ba0_int_p)*ht->G.alloc;

    *ht = new_ht;
}

BAV_DLL bav_hash_code bav_hash_term (bav_term T)
{
    return bav_hash_term2 (&bav_hterm, T, true);
}

BAV_DLL void bav_unhash_term (bav_term T, bav_hash_code code)
{
    if (code < 0 || code >= bav_hterm.G.size)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    if (bav_is_free (&bav_hterm, bav_hterm.G.index_in_H [code]))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    bav_set_term (T, bav_hterm.H.term [bav_hterm.G.index_in_H [code]]);
}


