#include "bav_differential_ring.h"
#include "bav_rank.h"
#include "bav_global.h"

#define bav_custom_printf_rank  bav_initialized_global.rank.printf

BAV_DLL void bav_set_settings_rank (ba0_printf_function* p)
{
    bav_custom_printf_rank = p ? p : &bav_printf_default_rank;
}

BAV_DLL void bav_get_settings_rank (ba0_printf_function** p)
{
    if (p) *p = bav_custom_printf_rank;
}

BAV_DLL bool bav_is_zero_rank (bav_rank rg)
{
    return rg->deg == -1;
}

BAV_DLL void bav_init_rank (bav_rank rg)
{
    rg = (bav_rank)0;	/* to avoid an annoying warning */
}

BAV_DLL bav_rank bav_new_rank (void)
{   bav_rank rg;

    rg = (bav_rank)ba0_alloc (sizeof (struct bav_rank));
    bav_init_rank (rg);
    return rg;
}
    
BAV_DLL bool bav_is_constant_rank (bav_rank rg)
{
    return rg->deg == 0;
}

BAV_DLL bool bav_equal_rank (bav_rank rg, bav_rank sg)
{
    return rg->var == sg->var && rg->deg == sg->deg;
}

BAV_DLL bool bav_lt_rank (bav_rank rg, bav_rank sg)
{   bav_Inumber nr, ns;

    if (bav_is_zero_rank (rg))
	return ! bav_is_zero_rank (sg);
    if (bav_is_constant_rank (rg))
	return ! bav_is_zero_rank (sg) && ! bav_is_constant_rank (sg);
    if (bav_is_zero_rank (sg) || bav_is_constant_rank (sg))
	return false;
    nr = bav_R_variable_number (rg->var);
    ns = bav_R_variable_number (sg->var);
    if (nr < ns)
	return true;
    else if (nr == ns)
	return rg->deg < sg->deg;
    else
	return false;
}

BAV_DLL bool bav_gt_rank (bav_rank rg, bav_rank sg)
{   bav_Inumber nr, ns;

    if (bav_is_zero_rank (sg))
	return ! bav_is_zero_rank (rg);
    if (bav_is_constant_rank (sg))
	return ! bav_is_zero_rank (rg) && ! bav_is_constant_rank (rg);
    if (bav_is_zero_rank (rg) || bav_is_constant_rank (rg))
	return false;
    ns = bav_R_variable_number (sg->var);
    nr = bav_R_variable_number (rg->var);
    if (ns < nr)
	return true;
    else if (ns == nr)
	return sg->deg < rg->deg;
    else
	return false;
}

/*
 * Readonly static data
 */

BAV_DLL struct bav_rank bav_zero_rank (void)
{   static struct bav_rank rg = { BAV_NOT_A_VARIABLE, (bav_Idegree)-1 };
    return rg;
}

/*
 * Readonly static data
 */

BAV_DLL struct bav_rank bav_constant_rank (void)
{   static struct bav_rank rg = { BAV_NOT_A_VARIABLE, (bav_Idegree)0 };
    return rg;
}

BAV_DLL void* bav_scanf_rank (void* z)
{   bav_rank rg;

    if (z == (void*)0)
	rg = bav_new_rank ();
    else
	rg = (bav_rank)z;

    if (ba0_type_token_analex () == ba0_integer_token)
    {	if (strcmp (ba0_value_token_analex (), "0") == 0)
	    *rg = bav_zero_rank ();
	else 
	    *rg = bav_constant_rank ();
    } else
    {	ba0_scanf ("%v", &rg->var);
	ba0_get_token_analex ();
	if (ba0_sign_token_analex ("^"))
	{   ba0_get_token_analex ();
	    if (ba0_type_token_analex () != ba0_integer_token)
		BA0_RAISE_PARSER_EXCEPTION (BAV_ERRTER);
	    rg->deg = (bav_Idegree) atoi (ba0_value_token_analex ());
	    if (rg->deg <= 0)
		BA0_RAISE_PARSER_EXCEPTION (BAV_ERRTER);
	} else
	{   ba0_unget_token_analex ();
	    rg->deg = 1;
	}
    }
    return rg;
}

BAV_DLL void bav_printf_default_rank (void* z)
{   bav_rank rg = (bav_rank)z;

    if (bav_is_zero_rank (rg))
	ba0_printf ("0");
    else if (bav_is_constant_rank (rg))
	ba0_printf ("1");
    else if (rg->deg == 1)
	ba0_printf ("%v", rg->var);
    else
	ba0_printf ("%v^%d", rg->var, rg->deg);
}

BAV_DLL void bav_printf_rank (void* z)
{   
    if ((bav_custom_printf_rank) != (ba0_printf_function*)0)
	(*bav_custom_printf_rank) (z);
    else
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
}

BAV_DLL int bav_inv_comp_rank (const void* x, const void* y)
{   bav_rank r = (bav_rank)x;
    bav_rank s = (bav_rank)y;
    bav_Inumber n, m;

    if (r->var == s->var && r->deg == s->deg)
	return 0;

    n = bav_R_variable_number (r->var);
    m = bav_R_variable_number (s->var);

    return (n < m || (n == m && r->deg < s->deg)) ? 1 : -1;
}
