#include "bav_differential_ring.h"
#include "bav_term.h"
#include "bav_term_ordering.h"
#include "bav_global.h"
#include "bav_parameter.h"

BAV_DLL void bav_realloc_term (bav_term T, ba0_int_p n)
{   struct bav_rank* tab;

    if (T->alloc < n)
    {	tab = (struct bav_rank*)ba0_alloc (sizeof (struct bav_rank) * n);
	memcpy (tab, T->rg, T->size * sizeof (struct bav_rank));
	T->rg = tab;
	T->alloc = n;
    }
}

BAV_DLL void bav_init_term (bav_term T)
{   
    T->alloc = 0;
    T->size = 0;
    T->rg = (struct bav_rank*)0;
}

BAV_DLL bav_term bav_new_term (void)
{   bav_term T = (bav_term)ba0_alloc (sizeof (struct bav_term));
    bav_init_term (T);
    return T;
}

BAV_DLL void bav_set_term_one (bav_term T)
{
    T->size = 0;
}

BAV_DLL void bav_set_term_rank (bav_term T, bav_rank rg)
{
    if (bav_is_zero_rank (rg))
	BA0_RAISE_EXCEPTION (BAV_ERRRGZ);
    if (bav_is_constant_rank (rg))
	bav_set_term_one (T);
    else
    {	bav_realloc_term (T, 1);
	T->rg [0] = *rg;
	T->size = 1;
    }
}

BAV_DLL void bav_set_term_variable (bav_term T, bav_variable v, bav_Idegree d)
{
    if (d == 0)
	bav_set_term_one (T);
    else
    {	bav_realloc_term (T, 1);
	T->rg [0].var = v;
	T->rg [0].deg = d;
	T->size = 1;
    }
}

BAV_DLL void bav_set_term (bav_term T, bav_term U)
{
    if (T != U)
    {	bav_realloc_term (T, U->size);
	memcpy (T->rg, U->rg, U->size * sizeof (struct bav_rank));
	T->size = U->size;
    }
}

BAV_DLL void bav_shift_term (bav_term T, bav_term U)
{
    if (bav_is_one_term (U))
	BA0_RAISE_EXCEPTION (BAV_ERRTEU);
    bav_realloc_term (T, U->size - 1);
    memmove (T->rg, U->rg + 1, (U->size - 1) * sizeof (struct bav_rank));
    T->size = U->size - 1;
}

BAV_DLL void bav_strip_term (bav_term T, bav_term U, bav_Inumber n)
{   ba0_int_p i;

    i = 0;
    while (i < U->size && bav_R_variable_number (U->rg [i].var) >= n)
	i++;
    if (T == U)
	U->size = i;
    else
    {	bav_realloc_term (T, i);
	T->size = i;
	memcpy (T->rg, U->rg, T->size * sizeof (struct bav_rank));
    }
}

BAV_DLL bool bav_is_one_term (bav_term T)
{
    return T->size == 0;
}

BAV_DLL bav_variable bav_leader_term (bav_term T)
{
    if (bav_is_one_term (T))
	BA0_RAISE_EXCEPTION (BAV_ERRTEU);
    return T->rg [0].var;
}

BAV_DLL bav_Idegree bav_leading_degree_term (bav_term T)
{
    if (bav_is_one_term (T))
        BA0_RAISE_EXCEPTION (BAV_ERRTEU);
    return T->rg [0].deg;
}

BAV_DLL bav_Idegree bav_total_degree_term (bav_term T)
{   ba0_int_p i;
    bav_Idegree deg;

    for (i = 0, deg = 0; i < T->size; i++)
	deg += T->rg [i].deg;
    return deg;
}

BAV_DLL bav_Idegree bav_maximal_degree_term (bav_term T)
{   ba0_int_p i;
    bav_Idegree deg;

    for (i = 0, deg = 0; i < T->size; i++)
	if (deg < T->rg [i].deg)
	    deg = T->rg [i].deg;
    return deg;
}

BAV_DLL bav_Idegree bav_degree_term (bav_term T, bav_variable v)
{   ba0_int_p i;

    i = 0;
    while (i < T->size && T->rg [i].var != v)
	i++;
    return i < T->size ? T->rg [i].deg : 0;
}

BAV_DLL bav_Iorder bav_total_order_term (bav_term T)
{   ba0_int_p i;
    bav_Iorder ord, ordmax;

    ordmax = 0;
    for (i = 0; i < T->size; i++)
    {	if (T->rg [i].var->root->type != bav_independent_symbol)
	{   ord = bav_total_order_variable (T->rg [i].var);
	    if (ord > ordmax)
		ordmax = ord;
	}
    }
    return ordmax;
}

BAV_DLL struct bav_rank bav_rank_term (bav_term T)
{
    if (bav_is_one_term (T))
        return bav_constant_rank ();
    else
	return T->rg [0];
}

BAV_DLL bool bav_equal_term (bav_term T, bav_term U)
{
    if (T->size != U->size)
	return false;
    if (T == U)
	return true;
    return memcmp (T->rg, U->rg, T->size * sizeof (struct bav_rank)) == 0;
}

BAV_DLL bool bav_disjoint_term  (bav_term T, bav_term U)
{   ba0_int_p i, j;

    for (i = 0; i < T->size; i++)
	for (j = 0; j < U->size; j++)
	    if (T->rg [i].var == U->rg [j].var)
		return false;
    return true;
}

BAV_DLL void bav_sort_term (bav_term T)
{
    qsort (T->rg, T->size, sizeof (struct bav_rank), &bav_inv_comp_rank);
}

BAV_DLL void bav_lcm_term (bav_term R, bav_term T, bav_term U)
{   struct bav_term S;
    ba0_int_p t, u, s;
    bav_Inumber num_t, num_u;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);
    bav_init_term (&S);
    bav_realloc_term (&S, T->size + U->size);

    t = u = s = 0;
    while (t < T->size && u < U->size)
    {	if (T->rg [t].var == U->rg [u].var)
	{   S.rg [s].var = T->rg [t].var;
	    S.rg [s].deg = BA0_MAX (T->rg [t].deg, U->rg [u].deg);
	    t++; u++; s++;
	} else
	{   num_t = bav_R_variable_number (T->rg [t].var);
	    num_u = bav_R_variable_number (U->rg [u].var);
	    S.rg [s++] = num_t > num_u ? T->rg [t++] : U->rg [u++];
	}
    }
    while (t < T->size)
	S.rg [s++] = T->rg [t++];
    while (u < U->size)
	S.rg [s++] = U->rg [u++];
    S.size = s;
    ba0_pull_stack ();
    bav_set_term (R, &S);
    ba0_restore (&M);
}

BAV_DLL void bav_gcd_term (bav_term R, bav_term T, bav_term U)
{   struct bav_term S;
    ba0_int_p t, u, s;
    bav_Inumber num_t, num_u;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);
    bav_init_term (&S);
    bav_realloc_term (&S, BA0_MAX (T->size, U->size));

    t = u = s = 0;
    while (t < T->size && u < U->size)
    {	if (T->rg [t].var == U->rg [u].var)
	{   S.rg [s].var = T->rg [t].var;
	    S.rg [s].deg = BA0_MIN (T->rg [t].deg, U->rg [u].deg);
	    t++; u++; s++;
	} else
	{   num_t = bav_R_variable_number (T->rg [t].var);
	    num_u = bav_R_variable_number (U->rg [u].var);
	    if (num_t > num_u)
		t++;
	    else
		u++;
	}
    }
    S.size = s;
    ba0_pull_stack ();
    bav_set_term (R, &S);
    ba0_restore (&M);
}

BAV_DLL void bav_mul_term (bav_term R, bav_term T, bav_term U)
{   struct bav_term S;
    ba0_int_p t, u, s;
    bav_Inumber num_t, num_u;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);
    bav_init_term (&S);
    bav_realloc_term (&S, T->size + U->size);

    t = u = s = 0;
    while (t < T->size && u < U->size)
    {	if (T->rg [t].var == U->rg [u].var)
	{   S.rg [s].var = T->rg [t].var;
	    S.rg [s].deg = T->rg [t].deg + U->rg [u].deg;
	    t++; u++; s++;
	} else
	{   num_t = bav_R_variable_number (T->rg [t].var);
	    num_u = bav_R_variable_number (U->rg [u].var);
	    S.rg [s++] = num_t > num_u ? T->rg [t++] : U->rg [u++];
	}
    }
    while (t < T->size)
	S.rg [s++] = T->rg [t++];
    while (u < U->size)
	S.rg [s++] = U->rg [u++];
    S.size = s;
    ba0_pull_stack ();
    bav_set_term (R, &S);
    ba0_restore (&M);
}

BAV_DLL void bav_mul_term_rank (bav_term R, bav_term T, bav_rank rg)
{   struct bav_term S;
    ba0_int_p t, s;
    bav_Inumber num;
    struct ba0_mark M;

    if (bav_is_zero_rank (rg))
	BA0_RAISE_EXCEPTION (BAV_ERRRGZ);
    if (bav_is_constant_rank (rg))
	bav_set_term (R, T);
    else
    {   ba0_push_another_stack ();
	ba0_record (&M);
	bav_init_term (&S);
	bav_realloc_term (&S, T->size + 1);
	num = bav_R_variable_number (rg->var);
	t = s = 0; 
	while (t < T->size && bav_R_variable_number (T->rg [t].var) > num)
	    S.rg [s++] = T->rg [t++];
	if (t < T->size && T->rg [t].var == rg->var)
	{   S.rg [s].var = rg->var;
	    S.rg [s].deg = T->rg [t].deg + rg->deg;
	    t++;
	} else
	    S.rg [s] = *rg;
	s++;
	while (t < T->size)
	    S.rg [s++] = T->rg [t++];
	S.size = s;
	ba0_pull_stack ();
	bav_set_term (R, &S);
	ba0_restore (&M);
    }
}

BAV_DLL void bav_mul_term_variable 
		(bav_term T, bav_term U, bav_variable v, bav_Idegree d)
{   struct bav_rank rg;

    rg.var = v;
    rg.deg = d;
    bav_mul_term_rank (T, U, &rg);
}

BAV_DLL void bav_pow_term (bav_term T, bav_term U, bav_Idegree d)
{   ba0_int_p i;

    if (d == 0)
	bav_set_term_one (T);
    else
    {	bav_realloc_term (T, U->size);
	for (i = 0; i < U->size; i++)
	{   T->rg [i].var = U->rg [i].var;
	    T->rg [i].deg = U->rg [i].deg * d;
	}
	T->size = U->size;
    }
}

BAV_DLL bool bav_is_factor_term (bav_term T, bav_term U, bav_term R)
{   struct bav_term S;
    ba0_int_p  t, u, s;
    bool divisible;
    struct ba0_mark M;

    if (R != (bav_term)0)
    {	ba0_push_another_stack ();
	ba0_record (&M);
	bav_init_term (&S);
	bav_realloc_term (&S, T->size);
    }
    divisible = true;
    t = u = s = 0;
    while (divisible && u < U->size)
    {	if (t == T->size)
	    divisible = false;
	else if (T->rg [t].var == U->rg [u].var)
	{   if (T->rg [t].deg < U->rg [u].deg)
		divisible = false;
	    else if (T->rg [t].deg > U->rg [u].deg && R != (bav_term)0)
	    {   S.rg [s].var = T->rg [t].var;
		S.rg [s].deg = T->rg [t].deg - U->rg [u].deg;
		s++;
	    }
	    t++; u++;
	} else if (bav_R_variable_number (T->rg [t].var) <
			bav_R_variable_number (U->rg [u].var))
	    divisible = false;
	else if (R != (bav_term)0)
	    S.rg [s++] = T->rg [t++];
	else
	    t++;
    }
    if (R != (bav_term)0)
    {	if (divisible)
	{   while (t < T->size)
		S.rg [s++] = T->rg [t++];
	    S.size = s;
	    ba0_pull_stack ();
	    bav_set_term (R, &S);
	    ba0_restore (&M);
	} else
	{   ba0_pull_stack ();
	    ba0_restore (&M);
	}
    }
    return divisible;
}

BAV_DLL void bav_exquo_term (bav_term R, bav_term T, bav_term U)
{
    if (! bav_is_factor_term (T, U, R))
	BA0_RAISE_EXCEPTION (BAV_EXEXQO);
}

BAV_DLL void bav_exquo_term_variable 
		(bav_term R, bav_term T, bav_variable x, bav_Idegree d)
{   ba0_int_p i;

    if (d == 0)
    {	if (R != T)
	    bav_set_term (R, T);
	return;
    }

    i = 0;
    while (i < T->size && T->rg [i].var != x)
	i++;
    if (i == T->size || T->rg [i].deg < d)
	BA0_RAISE_EXCEPTION (BAV_EXEXQO);
    if (R == T)
    {	if (R->rg [i].deg > d)
	    R->rg [i].deg -= d;
	else
	{   memmove (R->rg + i, T->rg + i + 1, 
				(T->size - i - 1) * sizeof (struct bav_rank));
	    R->size--;
	}
    } else
    {	bav_set_term_one (R);
	bav_realloc_term (R, T->rg [i].deg > d ? T->size : T->size - 1);
	memcpy (R->rg, T->rg, i * sizeof (struct bav_rank));
	if (T->rg [i].deg > d)
	{   R->rg [i].var = T->rg [i].var;
	    R->rg [i].deg = T->rg [i].deg - d;
	    memcpy (R->rg + i + 1, T->rg + i + 1, 
				(T->size - i - 1) * sizeof (struct bav_rank));
	    R->size = T->size;
	} else
	{   memcpy (R->rg + i, T->rg + i + 1, 
				(T->size - i - 1) * sizeof (struct bav_rank));
	    R->size = T->size - 1;
	}
    }
}

BAV_DLL void bav_diff_term 
	(bav_term R, bav_term T, bav_symbol x, bav_tableof_variable nulles)
{   bav_variable v;
    enum bav_typeof_symbol type;
    ba0_int_p t;

    if (T->size == 0)
	bav_set_term_one (R);
    else if (T->size == 1 && T->rg [0].deg == 1)
    {	type = T->rg [0].var->root->type;
	if (type == bav_dependent_symbol || type == bav_operator_symbol)
	{   v = bav_diff_variable (T->rg [0].var, x);
	    if (bav_is_derivative2 (v, nulles))
		bav_set_term_one (R);
	    else
		bav_set_term_variable (R, v, 1);
	} else
	    bav_set_term_one (R);
    } else
    {	for (t = 0; t < T->size; t++)
	{   v = T->rg [t].var;
	    type = T->rg [t].var->root->type;
	    if (type == bav_dependent_symbol || type == bav_operator_symbol)
		bav_diff_variable (T->rg [t].var, x);
	}
	bav_R_mark_variables (0);
	for (t = 0; t < T->size; t++)
	    T->rg [t].var->hack = T->rg [t].deg;
	for (t = 0; t < T->size; t++)
	{   v = T->rg [t].var;
	    type = T->rg [t].var->root->type;
	    if (type == bav_dependent_symbol || type == bav_operator_symbol)
	    {	v = bav_diff_variable (v, x);
		if (! bav_is_derivative2 (v, nulles))
		    v->hack += 1;
	    }
	}
	bav_marked_ranks_term (R);
	bav_sort_term (R);
    }
}

BAV_DLL void bav_marked_ranks_term (bav_term T)
{   ba0_int_p i, n;
    bav_variable v;

    for (i = 0, n = 0; i < bav_R.vars.size; i++)
    {	v = bav_R.vars.tab [i];
	if (v->hack != 0)
	    n++;
    }
    bav_realloc_term (T, n);
    for (i = 0, n = 0; i < bav_R.vars.size; i++)
    {	v = bav_R.vars.tab [i];
	if (v->hack != 0)
	{   T->rg [n].var = v;
	    T->rg [n].deg = v->hack;
	    n++;
	}
    }
    T->size = n;
}

/*
 * Readonly static data
 */

static char _struct_term [] = "struct bav_term";
static char _struct_term_rg [] = "bav_term->rg";

BAV_DLL ba0_int_p bav_garbage1_term (void* U, enum ba0_garbage_code code)
{   bav_term T = (bav_term)U;
    ba0_int_p n = 0;

    if (code == ba0_isolated)
	n += ba0_new_gc_info (T, sizeof (struct bav_term), _struct_term);
    n += ba0_new_gc_info 
		(T->rg, sizeof (struct bav_rank) * T->alloc, _struct_term_rg);
    return n;
}

BAV_DLL void* bav_garbage2_term (void* U, enum ba0_garbage_code code)
{   bav_term T;

    if (code == ba0_isolated)
	T = (bav_term)ba0_new_addr_gc_info (U, _struct_term);
    else
	T = (bav_term)U;

    T->rg = (struct bav_rank*)ba0_new_addr_gc_info (T->rg, _struct_term_rg);
    return T;
}

BAV_DLL void* bav_copy_term (void* U)
{   bav_term T = (bav_term)U;
    bav_term V;

    V = bav_new_term ();
    bav_set_term (V, T);
    return V;
}

BAV_DLL void* bav_scanf_term (void* U)
{   bav_term R = (bav_term)U;
    struct bav_term T;
    struct bav_rank  rg;
    struct ba0_mark  M;

    ba0_push_another_stack ();
    ba0_record (&M);
    bav_init_term (&T);
    bav_realloc_term (&T, 20); /* optimisation heuristique */

    if (ba0_type_token_analex () == ba0_integer_token)
    {   if (strcmp (ba0_value_token_analex (), "1") != 0)
            BA0_RAISE_PARSER_EXCEPTION (BAV_ERRTER);
    } else
    {   for (;;)
        {   bav_scanf_variable (&rg.var);
            ba0_get_token_analex ();
            if (ba0_sign_token_analex ("^"))
            {   ba0_get_token_analex ();
                if (ba0_type_token_analex () != ba0_integer_token)
                    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRTER);
                rg.deg = (bav_Idegree) atoi (ba0_value_token_analex ());
		if (rg.deg <= 0)
		    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRTER);
            } else
            {   ba0_unget_token_analex ();
                rg.deg = 1;
            }
            bav_mul_term_rank (&T, &T, &rg);
            ba0_get_token_analex ();
            if (! ba0_sign_token_analex ("*"))
                break;
            ba0_get_token_analex ();
        }
        ba0_unget_token_analex ();
    }
    ba0_pull_stack ();
    if (R == (bav_term)0)
	R = bav_new_term ();
    bav_set_term (R, &T);
    ba0_restore (&M);
    return R;
}

BAV_DLL void bav_printf_term (void* U)
{   bav_term T = (bav_term)U;
    bool first = true;
    ba0_int_p  i;

    if (T->size == 0)
        ba0_put_char ('1');

    for (i = 0; i < T->size; i++)
    {   if (! first && ! ba0_LaTeX_flag) 
	    ba0_put_char ('*'); 
	else 
	    first = false;
	ba0_printf ("%rank", &T->rg [i]);
    }
}

/*
static ba0_int_p pow_int_p (ba0_int_p a, bav_Idegree d)
{   ba0_int_p e;
    bav_Idegree p;

    e = 1;
    if (d == 0)
        return e;
    for (p = d; p != 1; p /= 2)
    {   if (p % 2 == 1)
            e *= a;
	a *= a;
    }
    e *= a;
    return e;
}
*/

/*
 * Assigns T mod ev_point to value
 */

BAV_DLL void bav_term_at_ev_point 
			(ba0_mpz_t value, bav_term T, bav_ev_point point)
{   ba0_int_p p, t;
    bool found;
    ba0_mpz_t val, pow;
    struct ba0_mark M;

    ba0_push_another_stack ();
    ba0_record (&M);

    ba0_mpz_init (pow);
    ba0_mpz_init_set_ui (val, 1);
    for (t = 0; value != 0 && t < T->size; t++)
    {	found = false;
	for (p = 0; ! found && p < point->size; p++)
	{   if (T->rg [t].var == point->tab [p].var)
	    {	ba0_mpz_set_si (pow, point->tab [p].value);
		ba0_mpz_pow_ui (pow, pow, T->rg [t].deg);
		ba0_mpz_mul (val, val, pow);
		found = true;
	    }
	}
	if (! found)
	    BA0_RAISE_EXCEPTION (BA0_ERRALG);
    }
    ba0_pull_stack ();
    ba0_mpz_set (value, val);
    ba0_restore (&M);
}

BAV_DLL void bav_switch_ring_term (bav_term T, bav_differential_ring R)
{   ba0_int_p i;

    for (i = 0; i < T->size; i++)
	T->rg [i].var = bav_switch_ring_variable (T->rg [i].var, R);
}

BAV_DLL bool bav_depends_on_zero_derivatives_of_parameter_term 
					(bav_term T, bav_tableof_parameter P)
{   ba0_int_p i;
    for (i = 0; i < T->size; i++)
	if (bav_is_zero_derivative_of_parameter (T->rg [i].var, P))
	    return true;
    return false;
}

 
