#include "bav_parameter.h"
#include "bav_variable.h"
#include "bav_differential_ring.h"
#include "bav_global.h"
#include "bav_term.h"

#define bav_diff_string			bav_global.variable.diff_string
#define bav_custom_scanf_variable       bav_initialized_global.variable.scanf
#define bav_custom_printf_variable      bav_initialized_global.variable.printf

BAV_DLL void bav_set_settings_variable 
		(ba0_scanf_function* s, ba0_printf_function* p, char* temp)
{
    bav_custom_scanf_variable = s ? s : &bav_scanf_jet_variable;
    bav_custom_printf_variable = p ? p : &bav_printf_jet_variable;
    bav_temp_string = temp ? temp : BAV_TEMP_STRING;
}

BAV_DLL void bav_get_settings_variable
		(ba0_scanf_function** s, ba0_printf_function** p, char** temp)
{
    if (s) *s = bav_custom_scanf_variable;
    if (p) *p = bav_custom_printf_variable;
    if (temp) *temp = bav_temp_string;
}

BAV_DLL void bav_init_variable (bav_variable v)
{
    v->root = (bav_symbol)0;
    ba0_init_table ((ba0_table)&v->number);
    ba0_init_table ((ba0_table)&v->order);
    ba0_init_table ((ba0_table)&v->derivative);
}

BAV_DLL bav_variable bav_new_variable (void)
{   bav_variable v;

    v = (bav_variable)ba0_alloc (sizeof (struct bav_variable));
    bav_init_variable (v);
    return v;
}

BAV_DLL bav_variable bav_not_a_variable (void)
{
    return BAV_NOT_A_VARIABLE;
}

BAV_DLL bav_Iorder bav_order_variable (bav_variable v, bav_symbol d)
{   bav_Inumber i;

    if (d->type != bav_independent_symbol ||
	(v->root->type != bav_dependent_symbol &&
					v->root->type != bav_operator_symbol))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    i = d->derivation_index;
    return v->order.tab [i];
}

BAV_DLL bav_Iorder bav_total_order_variable (bav_variable v)
{   bav_Inumber i;
    bav_Iorder cpt;

    if (v->root->type != bav_dependent_symbol &&
		    			v->root->type != bav_operator_symbol)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    cpt = 0;
    for (i = 0; i < v->order.size; i++)
	cpt += v->order.tab [i];
    return cpt;
}

BAV_DLL bav_variable bav_diff_variable (bav_variable v, bav_symbol s)
{   ba0_int_p i;

    if (v->root->type != bav_dependent_symbol &&
					v->root->type != bav_operator_symbol)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    i = s->derivation_index;
    if (v->derivative.tab [i] == BAV_NOT_A_VARIABLE)
	v->derivative.tab [i] = bav_R_derivative (v, s);
    return v->derivative.tab [i];
}

BAV_DLL bav_variable bav_diff2_variable (bav_variable v, bav_term theta)
{   ba0_int_p i;

    for (i = 0; i < theta->size; i++)
	v = bav_diff_variable (v, theta->rg [i].var->root);
    return v;
}

BAV_DLL bav_variable bav_int_variable (bav_variable v, bav_symbol s)
{   bav_variable w;
    bav_symbol t;
    ba0_int_p i, j, k;

    i = s->derivation_index;
    if ((v->root->type != bav_dependent_symbol &&
		v->root->type != bav_operator_symbol) || v->order.tab [i] <= 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    w = bav_R_symbol_to_variable (v->root);
    for (k = 0; k < v->order.size; k++)
    {	t = bav_R_derivation_index_to_derivation (k)->root;
	if (k == i)
	    for (j = 0; j < v->order.tab [k] - 1; j++)
		w = bav_diff_variable (w, t);
	else
	    for (j = 0; j < v->order.tab [k]; j++)
		w = bav_diff_variable (w, t);
    }
    return w;
}

BAV_DLL enum bav_typeof_symbol bav_symbol_type_variable (bav_variable v)
{
    return v->root->type;
}

BAV_DLL bav_variable bav_order_zero_variable (bav_variable v)
{   
    if (v->root->type != bav_dependent_symbol &&
					v->root->type != bav_operator_symbol)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    return bav_R_symbol_to_variable (v->root);
}

BAV_DLL bav_variable bav_lcd_variable (bav_variable v, bav_variable w)
{   bav_symbol t;
    ba0_int_p j;
	    
    if ((v->root->type != bav_dependent_symbol &&
		v->root->type != bav_operator_symbol) || v->root != w->root)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    for (j = 0; j < v->order.size; j++)
    {	t = bav_R_derivation_index_to_derivation (j)->root;
	while (v->order.tab [j] < w->order.tab [j])
	{   if (v->derivative.tab [j] == BAV_NOT_A_VARIABLE)
		v->derivative.tab [j] = bav_R_derivative (v, t);
	    v = v->derivative.tab [j];
	}
    }
    return v;
}

BAV_DLL bool bav_disjoint_variables (bav_variable v, bav_variable w)
{   ba0_int_p i;

    if (v->root != w->root || v->root->type == bav_dependent_symbol)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    for (i = 0; i < v->order.size; i++)
	if (v->order.tab [i] != 0 && w->order.tab [i] != 0)
	    return false;
    return true;
}

/*
 * Returns true if v is a derivative of w
 */

BAV_DLL bool bav_is_derivative (bav_variable v, bav_variable w)
{   ba0_int_p j;

    if (v->root != w->root)
	return false;
    if (v->root->type == bav_independent_symbol)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    for (j = 0; j < v->order.size; j++)
	if (v->order.tab [j] < w->order.tab [j])
	    return false;
    return true;
}

/*
 * Returns true if v is a proper derivative of w
 */

BAV_DLL bool bav_is_proper_derivative (bav_variable v, bav_variable w)
{   ba0_int_p j;

    if (v->root != w->root || v == w)
	return false;
    if (v->root->type == bav_independent_symbol)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    for (j = 0; j < v->order.size; j++)
	if (v->order.tab [j] < w->order.tab [j])
	    return false;
    return true;
}

/*
 * Returns true is v is a derivative of w wrt d
 */

BAV_DLL bool bav_is_d_derivative (bav_variable v, bav_variable w, bav_symbol d)
{   ba0_int_p j;

    if (v->root->type == bav_independent_symbol ||
	w->root->type == bav_independent_symbol ||
	d->type != bav_independent_symbol)
        BA0_RAISE_EXCEPTION (BA0_ERRALG);

    if (v->root != w->root)
        return false;

    for (j = 0; j < v->order.size; j++)
    {   if (d->derivation_index == j)
	{   if (v->order.tab [j] != w->order.tab [j] + 1)
		return false;
	} else if (v->order.tab [j] != w->order.tab [j])
	   return false;
    }
    return true;
}

/*
 * Returns true if v is a derivative of some element of T
 */

BAV_DLL bool bav_is_derivative2 (bav_variable v, bav_tableof_variable T)
{   ba0_int_p i;
    bool found;

    for (i = 0, found = false; ! found && i < T->size; i++)
	found = bav_is_derivative (v, T->tab [i]);
    return found;
}

BAV_DLL bav_variable bav_derivation_between_derivatives 
					(bav_variable v, bav_variable w)
{   bav_Inumber j;

    if (v->root != w->root || v == w)
    	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    else if (bav_is_derivative (w, v))
    	BA0_SWAP (bav_variable, v, w);
    else if (! bav_is_derivative (v, w))
	BA0_RAISE_EXCEPTION (BA0_ERRALG);

    j = 0; 
    while (v->order.tab [j] == w->order.tab [j])
	j++;
    return bav_R_derivation_index_to_derivation (j);
}

BAV_DLL void bav_operator_between_derivatives
				(bav_term T, bav_variable v, bav_variable w)
{   bav_variable x;

    if (bav_is_derivative (v, w))
	BA0_SWAP (bav_variable, v, w);

    bav_set_term_one (T);
    bav_realloc_term (T, bav_R.ders.size);
    while (v != w) 
    {	x = bav_derivation_between_derivatives (v, w);
	bav_mul_term_variable (T, T, x, 1);
	v = bav_diff_variable (v, x->root);
    }
}

static bav_variable bav_basic_next_derivative (bav_variable v)
{   bav_variable w;
    bav_symbol x;
    bav_Iorder d, d0, i;

    d = bav_total_order_variable (v);
    x = bav_R.vars.tab [bav_R.ders.tab [0]]->root;
    d0 = bav_order_variable (v, x);
    if (d == d0)
    {	w = bav_order_zero_variable (v);
	x = bav_R.vars.tab [bav_R.ders.tab [bav_R.ders.size - 1]]->root;
	for (i = 0; i < d+1; i++)
	    w = bav_diff_variable (w, x);
    } else
    {   w = v;
	for (i = 0; i < d0; i++)
	    w = bav_int_variable (w, x);
	i = 1;
	x = bav_R.vars.tab [bav_R.ders.tab [i]]->root;
	while (bav_order_variable (w, x) == 0)
	{   i += 1;
	    x = bav_R.vars.tab [bav_R.ders.tab [i]]->root;
	}
	w = bav_int_variable (w, x);
	i -= 1;
	x = bav_R.vars.tab [bav_R.ders.tab [i]]->root;
	for (i = 0; i <= d0; i++)
	    w = bav_diff_variable (w, x);
    }
    return w;
}

BAV_DLL bav_variable bav_next_derivative 
				(bav_variable v, bav_tableof_variable T)
{   bav_variable w, x;
    ba0_int_p i;
    bool found;

    if (bav_symbol_type_variable (v) != bav_dependent_symbol ||
						bav_R.ders.size == 0)
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
/*
 * w is the next derivative of w. 
 * If it is okay, return it
 */
    w = bav_basic_next_derivative (v);
    found = false;
    for (i = 0; ! found && i < T->size; i++)
        found = bav_is_derivative (w, T->tab [i]);
    if (! found)
	return w;
/*
 * The next derivative, w, of v is not okay.
 */
    x = w;
    do
    {	x = bav_basic_next_derivative (x);
	found = false;
	for (i = 0; ! found && i < T->size; i++)
	    found = bav_is_derivative (x, T->tab [i]);
    } while (found && ! bav_is_derivative (x, w));
    return found ? BAV_NOT_A_VARIABLE : x;
}

/*
 * In the jet notation, one must beware to the fact that square brackets
 * may either be part of the symbol or stand for some list of derivations
 */

BAV_DLL void* bav_scanf_jet_variable (void* z)
{   bav_variable v;
    bav_symbol y;
    struct ba0_mark M;
    ba0_int_p i;
    ba0_indexed indexed;
    ba0_indices der_indices;
    bool has_der_indices;
    char *s;

    if (ba0_type_token_analex () != ba0_string_token)
        BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);

    ba0_record (&M);
    indexed = ba0_scanf_indexed (0, &has_der_indices, &bav_is_a_derivation);

    if (!has_der_indices)
    {	s = ba0_indexed_to_string (indexed);
	v = bav_R_string_to_variable (s);
	if (v == BAV_NOT_A_VARIABLE)
	{   (*bav_custom_unknown) (indexed);
	    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRUSY);
	}
    } else
    {	der_indices = indexed->Tindic.tab [indexed->Tindic.size-1];
	if (der_indices->po != '[')
	{   (*bav_custom_unknown) (indexed);
	    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRUSY);
	}
	indexed->Tindic.size -= 1;
	s = ba0_indexed_to_string (indexed);
	v = bav_R_string_to_variable (s);
	if (v == BAV_NOT_A_VARIABLE)
	{   if (der_indices->Tindex.size == 0)
	    {	indexed->Tindic.size += 1;
		s = ba0_indexed_to_string (indexed);
		v = bav_R_string_to_variable (s);
		if (v == BAV_NOT_A_VARIABLE)
		{   (*bav_custom_unknown) (indexed);
		    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRUSY);
		}
	    } else
	    {	(*bav_custom_unknown) (indexed);
		BA0_RAISE_PARSER_EXCEPTION (BAV_ERRUSY);
	    }
	}
	if (v->root->type != bav_dependent_symbol &&
				v->root->type != bav_operator_symbol)
	    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRBSD);
	for (i = 0; i < der_indices->Tindex.size; i++)
	{   s = ba0_indexed_to_string (der_indices->Tindex.tab [i]);
	    y = bav_R_string_to_derivation (s);
/*
 * This should not happen since has_der_indices is true
 */
	    if (y == BAV_NOT_A_SYMBOL)
		BA0_RAISE_EXCEPTION (BA0_ERRALG);
	    v = bav_diff_variable (v, y);
	}
    }
    ba0_restore (&M);
    if (z != (void*)0)
	*(bav_variable*)z = v;
    return v;
}

/*
 * Subfunction of bav_scanf_diff_variable
 * Reads non differentiated variables of the form: u(x,y)
 * where u is a dependent variable
 * x,y is the sequence of the derivations in the same order as in bav_R.ders
 *
 * If the function symbol is present in the bav_parameters table, then
 * the notation provided in that table is required.
 */

static bav_variable bav_scanf_non_diff_variable ()
{   bav_variable v;
    bav_symbol y, d;
    struct ba0_mark M;
    ba0_int_p i, k;
    ba0_indexed indexed;
    ba0_indices der_indices = (ba0_indices)0;
    bool has_der_indices, tolerated, found;
    char *s;

    if (ba0_type_token_analex () != ba0_string_token)
        BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);

    ba0_record (&M);
    indexed = ba0_scanf_indexed (0, &has_der_indices, &bav_is_a_derivation);
    if (has_der_indices)
        der_indices = indexed->Tindic.tab [indexed->Tindic.size-1];
/*
 * has_der_indices = the last sequence of indices is a list of derivations.
 * This list may be empty.
 *
 * tolerated = indexed is tolerated as a symbol i.e. it has no terminating
 *             list of derivations or an empty one (but not "()").
 */
    tolerated = !has_der_indices || 
		(has_der_indices && 
		 der_indices->po != '(' && 
		 der_indices->Tindex.size == 0);

    if (tolerated)
    {	s = ba0_indexed_to_string (indexed);
	y = bav_R_string_to_symbol (s);
	if (y == BAV_NOT_A_SYMBOL)
	{   (*bav_custom_unknown) (indexed);
	    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRUSY);
	}
    } else
    {	
	if (der_indices->po != '(')
	    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRDIF);
	indexed->Tindic.size -= 1;
	s = ba0_indexed_to_string (indexed);
	y = bav_R_string_to_symbol (s);
	if (y == BAV_NOT_A_SYMBOL)
	{   (*bav_custom_unknown) (indexed);
	    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRUSY);
	}
    }
/*
 * At this step, the symbol is recognized. It is in y.
 */
    if (y->type == bav_dependent_symbol)
    {	
/*
 * found = y is dependent, declared in bav_parameters at index k.
 * One then needs to check whether the order of the independent variables
 * following y is correct or not. Possible errors have nothing to do with
 * not recognizing a symbol.
 */
	found = bav_is_a_parameter (y, &k, &bav_parameters);
	if (found && has_der_indices)
	{   if (der_indices->Tindex.size != bav_parameters.tab [k]->dep.size)
		BA0_RAISE_PARSER_EXCEPTION (BAV_ERRDIF);
	    for (i = 0; i < der_indices->Tindex.size; i++)
	    {	s = ba0_indexed_to_string (der_indices->Tindex.tab [i]);
		d = bav_parameters.tab [k]->dep.tab [i];
		if (strcmp (s, d->ident) != 0)
		    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRDIF);
	    }
        } else if (found)
	{   if (bav_parameters.tab [k]->dep.size > 0)
		BA0_RAISE_PARSER_EXCEPTION (BAV_ERRDIF);
	} else if (!found && has_der_indices)
        {   
	    if (der_indices->Tindex.size != bav_R.ders.size)
	        BA0_RAISE_PARSER_EXCEPTION (BAV_ERRDIF);
	    for (i = 0; i < der_indices->Tindex.size; i++)
	    {   s = ba0_indexed_to_string (der_indices->Tindex.tab [i]);
		d = bav_R.syms.tab [bav_R.ders.tab [i]];
		if (strcmp (s, d->ident) != 0)
		    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRDIF);
	    }
        } else 
	{   if (bav_R.ders.size > 0)
		BA0_RAISE_PARSER_EXCEPTION (BAV_ERRDIF);
	}
    } else if (!tolerated)
	BA0_RAISE_PARSER_EXCEPTION (BAV_ERRDIF);

    ba0_restore (&M);
    v = bav_R_symbol_to_variable (y);
    return v;
}

/*
 * Reads a variable of the form: diff(diff(u(x,y),y,x)) or diff(u(x,y),x,x)
 */

static void* bav_scanf_generic_diff_variable (void* z)
{   bav_variable v;
    bav_symbol d;
    struct ba0_exception_code code;
    struct ba0_mark M;
    char* s;

    ba0_record (&M);
    s = ba0_value_token_analex ();
    if (strcmp (s, bav_diff_string) == 0)
    {   ba0_get_token_analex ();
	if (! ba0_sign_token_analex ("("))
	    BA0_RAISE_PARSER_EXCEPTION (BAV_ERRDIF);
	ba0_get_token_analex ();
	v = bav_scanf_generic_diff_variable ((void*)0);
	ba0_get_token_analex ();

	BA0_PUSH_EXCEPTION (code);
	if (ba0_exception_is_raised (code))
        {   if (ba0_mesgerr == BA0_ERROOM || ba0_mesgerr == BA0_ERRALR)
                BA0_RAISE_EXCEPTION (ba0_mesgerr);
            BA0_RAISE_EXCEPTION (BAV_ERRDIF);
        }

	while (ba0_sign_token_analex (","))
	{   ba0_get_token_analex ();
	    ba0_scanf ("%y", &d);
	    v = bav_diff_variable (v, d);
	    ba0_get_token_analex ();
	}

	if (! ba0_sign_token_analex (")"))
	    BA0_RAISE_PARSER_EXCEPTION (BA0_ERRSYN);

	ba0_pull_exception (code);
    } else
	v = bav_scanf_non_diff_variable ();

    ba0_restore (&M);

    if (z != (void*)0)
	*(bav_variable*)z = v;

    return v;
}

BAV_DLL void* bav_scanf_diff_variable (void* z)
{
    bav_diff_string = "diff";
    return bav_scanf_generic_diff_variable (z);
}

BAV_DLL void* bav_scanf_Diff_variable (void* z)
{
    bav_diff_string = "Diff";
    return bav_scanf_generic_diff_variable (z);
}

BAV_DLL void* bav_scanf_variable (void* z)
{   void* res;

    if (bav_custom_scanf_variable != (ba0_scanf_function*)0)
	res = (*bav_custom_scanf_variable) (z);
    else
    {	BA0_RAISE_EXCEPTION (BA0_ERRALG);
	res = bav_scanf_jet_variable (z);
    }
    return res;
}

/*
 * Prints a variable following the jet format
 */

BAV_DLL void bav_printf_jet_variable (void* z)
{   bav_variable v = (bav_variable)z;
    bav_variable d;
    bav_Inumber i, j;
    bool yet;

    bav_printf_symbol (v->root);
    if ((v->root->type == bav_dependent_symbol || 
		v->root->type == bav_operator_symbol) && 
			bav_total_order_variable (v) > 0)
    {	ba0_put_char ('[');
	yet = false;
	for (i = 0; i < v->order.size; i++)
	{   if (v->order.tab [i] != 0)
	    {	d = bav_R_derivation_index_to_derivation (i);
		for (j = 0; j < v->order.tab [i]; j++)
		{   if (yet) ba0_put_char (','); else yet = true;
		    bav_printf_symbol (d->root);
		}
	    }
	}
	ba0_put_char (']');
    }
}

BAV_DLL void bav_printf_jet_wesb_variable (void* z)
{   bav_variable v = (bav_variable)z;
    bav_variable d;
    bav_Inumber i, j;
    bool yet;

    bav_printf_symbol (v->root);
    if ((v->root->type == bav_dependent_symbol || 
		v->root->type == bav_operator_symbol))
    {	ba0_put_char ('[');
	yet = false;
	for (i = 0; i < v->order.size; i++)
	{   if (v->order.tab [i] != 0)
	    {	d = bav_R_derivation_index_to_derivation (i);
		for (j = 0; j < v->order.tab [i]; j++)
		{   if (yet) ba0_put_char (','); else yet = true;
		    bav_printf_symbol (d->root);
		}
	    }
	}
	ba0_put_char (']');
    }
}

/*
 * Prints a variable following the LaTeX format
 */

BAV_DLL void bav_printf_LaTeX_variable (void* z)
{   bav_variable v = (bav_variable)z;
    bav_variable d;
    bav_Inumber i, j;
    bool yet;

    bav_printf_symbol (v->root);

    if (v->root->type != bav_dependent_symbol)
	return;

    yet = false;
    for (i = 0; i < v->order.size && ! yet; i++)
	yet = v->order.tab [i] > 0;

    if (! yet)
	return;
 
    ba0_put_char ('_');
    ba0_put_char ('{');
    for (i = 0; i < v->order.size; i++)
    {	if (v->order.tab [i] != 0)
	{   d = bav_R_derivation_index_to_derivation (i);
	    for (j = 0; j < v->order.tab [i]; j++)
		bav_printf_symbol (d->root);
	}
    }
    ba0_put_char ('}');
}

/*
 * Prints a variable using the diff format
 * If a symbol is present in the bav_parameters table then a special
 * notation is applied.
 */

static void bav_printf_generic_diff_variable (void* z)
{   bav_variable v = (bav_variable)z;
    bav_variable d;
    bav_Iorder order;
    bav_Inumber i, j;
    ba0_int_p k;

    switch (v->root->type)
    {	case bav_temporary_symbol:
	case bav_independent_symbol:
	    bav_printf_symbol (v->root);
	    break;
	case bav_operator_symbol:
	    BA0_RAISE_EXCEPTION (BA0_ERRNYP);
	    break;
	case bav_dependent_symbol:
	    order = bav_total_order_variable (v);
/*
 * diff (
 */
	    if (order > 0)
	    {	ba0_put_string (bav_diff_string);
		ba0_put_char (' ');
		ba0_put_char ('(');
	    }
/*
 * u(x,y)
 * beware to the special case of non differential problems
 */
	    if (bav_is_a_parameter (v->root, &k, &bav_parameters))
	        ba0_printf ("%param", bav_parameters.tab [k]);
	    else
	    {	bav_printf_symbol (v->root);
	    	if (v->order.size > 0)
	        {   ba0_put_char ('(');
		    for (i = 0; i < v->order.size; i++)
		    {   d = bav_R_derivation_index_to_derivation (i);
		        bav_printf_symbol (d->root);
		        if (i < v->order.size - 1)
		            ba0_put_char (',');
		        else
			    ba0_put_char (')');
		    }
		}
	    }
/*
 * ,x ,x, y)
 */
	    if (order > 0)
	    {	for (i = 0; i < v->order.size; i++)
		{   if (v->order.tab [i] != 0)
		    {	d = bav_R_derivation_index_to_derivation (i);
			for (j = 0; j < v->order.tab [i]; j++)
			{    ba0_put_char (',');
			     ba0_put_char (' ');
			     bav_printf_symbol (d->root);
			}
		    }
		}
		ba0_put_char (')');
	    }
	    break;
    }
}

BAV_DLL void bav_printf_diff_variable (void* z)
{
    bav_diff_string = "diff";
    bav_printf_generic_diff_variable (z);
}

BAV_DLL void bav_printf_Diff_variable (void* z)
{
    bav_diff_string = "Diff";
    bav_printf_generic_diff_variable (z);
}

BAV_DLL void bav_printf_variable (void* z)
{
    if (ba0_LaTeX_flag)
	bav_printf_LaTeX_variable (z);
    else if (bav_custom_printf_variable != (ba0_printf_function*)0)
	(*bav_custom_printf_variable) (z);
    else 
    {   BA0_RAISE_EXCEPTION (BA0_ERRALG);
	bav_printf_jet_variable (z);
    }
}

/*
 * Returns true if the index of x is greater than that of y.
 */

BAV_DLL bool bav_gt_index_variable (void* x, void* y)
{   bav_variable v = (bav_variable)x, w = (bav_variable) y;

    return v->index > w->index;
}

BAV_DLL bool bav_gt_variable (void* x, void* y)
{   bav_variable v = (bav_variable)x, w = (bav_variable) y;

    return bav_R_variable_number (v) > bav_R_variable_number (w);
}

/*
 * For qsort. See below
 */

static int comp_variable_ascending (const void* x, const void* y)
{
    bav_variable v = *(bav_variable*)x;
    bav_variable w = *(bav_variable*)y;

    if (v == w)
	return 0;
    else if (bav_gt_variable (v, w))
	return 1;
    else
	return -1;
}

static int comp_variable_descending (const void* x, const void* y)
{
    bav_variable v = *(bav_variable*)x;
    bav_variable w = *(bav_variable*)y;

    if (v == w)
	return 0;
    else if (bav_gt_variable (v, w))
	return -1;
    else
	return 1;
}

/*
 * To be used for sorting tables of variables in normal mode
 */

BAV_DLL void bav_sort_tableof_variable 
			(bav_tableof_variable T, enum ba0_sort_mode mode)
{
    switch (mode)
    {	case ba0_descending_mode:
	    qsort (T->tab, T->size, sizeof (bav_variable), 
						&comp_variable_descending);
	    break;
    	case ba0_ascending_mode:
	    qsort (T->tab, T->size, sizeof (bav_variable), 
						&comp_variable_ascending);
	    break;
    }
}

/*
 * Used in order to create an ordering. The above function cannot be used yet.
 */

static int compare_just_created_variable (const void* x, const void* y)
{   bav_variable v = *(bav_variable*)x, w = *(bav_variable*)y;
    bav_ordering O;
    bav_Inumber num_blk_v, num_blk_w;
    ba0_int_p num_in_blk_v, num_in_blk_w;
    ba0_int_p i;
    bool v_gt_w = false, found;

    if (v == w) return 0;

    O = bav_R_ordering ();

    found = false;
    i = 0;
    while (! found && i < O->varmax.size)
    {	found = O->varmax.tab [i] == v || O->varmax.tab [i] == w;
	if (! found)
	    i++;
    }
    if (found)
    {	v_gt_w = O->varmax.tab [i] == v;
	return v_gt_w ? 1 : -1;
    }

    switch (v->root->type)
    {	case bav_operator_symbol:
	    switch (w->root->type)
	    {	case bav_operator_symbol:
		    v_gt_w = (*bav_subranking_table
			[O->operator_block.subr].inf) (w, v, 0, 0, &O->ders);
		    break;
		case bav_independent_symbol:
		case bav_dependent_symbol:
		case bav_temporary_symbol:
		    v_gt_w = true;
	    }
	    break;
	case bav_dependent_symbol:
	    switch (w->root->type)
	    {   case bav_operator_symbol:
		    v_gt_w = false;
		    break;
		case bav_dependent_symbol:
	            num_blk_v = bav_R_symbol_block_number 
						(v->root, &num_in_blk_v);
	            num_blk_w = bav_R_symbol_block_number 
						(w->root, &num_in_blk_w);
	            if (num_blk_v < num_blk_w)
		        v_gt_w = true;
		    else if (num_blk_v > num_blk_w)
			v_gt_w = false;
		    else
		        v_gt_w = (*bav_subranking_table 
			          [O->blocks.tab [num_blk_v]->subr].inf) 
                                      (w, v, 
				       num_in_blk_w, num_in_blk_v, 
				       &O->ders);
		    break;
		case bav_independent_symbol:
		case bav_temporary_symbol:
		    v_gt_w = true;
	    }
	    break;
	case bav_independent_symbol:
	    switch (w->root->type)
	    {	case bav_operator_symbol:
		case bav_dependent_symbol:
		    v_gt_w = false;
		    break;
		case bav_independent_symbol:
		    found = false;
		    i = 0;
		    while (! found && i < O->ders.size)
		    {	found = O->ders.tab [i] == v->root || 
						O->ders.tab [i] == w->root;
			if (!found)
			    i++;
		    }
		    if (i == O->ders.size)
			BA0_RAISE_EXCEPTION (BA0_ERRALG);
		    v_gt_w = O->ders.tab [i] == v->root;
		    break;
		case bav_temporary_symbol:
		    v_gt_w = true;
	    }
	    break;
	case bav_temporary_symbol:
	    switch (w->root->type)
	    {	case bav_operator_symbol:
		case bav_dependent_symbol:
		case bav_independent_symbol:
		    v_gt_w = false;
		    break;
		case bav_temporary_symbol:
		    v_gt_w = v->root->ident < w->root->ident;
		    break;
	    }
    }
    return v_gt_w ? 1 : -1;
}

static int bav_compare_just_created_variable (const void* x, const void* y)
{
#if defined (BA0_HEAVY_DEBUG)
    int a, b;
    a = compare_just_created_variable (x, y);
    b = compare_just_created_variable (y, x);
    if (a*b != -1) 
	BA0_RAISE_EXCEPTION (BA0_ERRALG);
    return a;
#else
    return compare_just_created_variable (x, y);
#endif
}

/*
 * Used by bav_differential_ring.c when numbers are not yet available.
 * Increasing order.
 */

BAV_DLL void bav_sort_tableof_just_created_variable (bav_tableof_variable T)
{
    qsort (T->tab, T->size, sizeof(bav_variable), 
					&bav_compare_just_created_variable);
}

BAV_DLL void bav_independent_variables (bav_tableof_variable T)
{   ba0_int_p i;

    ba0_reset_table ((ba0_table)T);
    ba0_realloc_table ((ba0_table)T, bav_R.ders.size);
    for (i = 0; i < bav_R.ders.size; i++)
	T->tab [T->size++] = bav_R_derivation_index_to_derivation (i);
}

BAV_DLL ba0_mint_hp bav_random_eval_variable_to_mint_hp (bav_variable v)
{   bav_variable t;
    ba0_int_p clef, k, i;

    clef = (ba0_int_p)v->root / sizeof (struct bav_symbol);
    if (v->root->type == bav_dependent_symbol || 
					v->root->type == bav_operator_symbol)
    {   for (i = 0; i < v->order.size; i++)
        {   t = bav_R_derivation_index_to_derivation (i);
	    k = v->order.tab [i] * (ba0_int_p)t / sizeof (struct bav_variable);
            clef += k;
        }
    }
    return (ba0_mint_hp)(clef % (unsigned ba0_int_p)ba0_mint_hp_module);
}

BAV_DLL bav_variable bav_switch_ring_variable 
				(bav_variable v, bav_differential_ring R)
{
    return R->vars.tab [v->index];
}

