#include "ba0.h"

/*
   The two stacks mechanism

   A stack is a big array of memory. The border between what is free and
what is not free is given by a free pointer (hence allocating is very fast).

   There always exists a "current stack".

   The programming convention is the following.

   R receives A wedge B.
   The variable R is stored in the current stack.
*/

static void wedge 
	(ba0_tableof_int_p R, ba0_tableof_int_p A, ba0_tableof_int_p B)
{   struct ba0_tableof_int_p T;
    struct ba0_mark M;
/*
   Memory for the intermediate variable T is allocated in another stack.
   One records in M the value of the free pointer of this other stack.
   This stack becomes the current stack during intermediate computations.
*/
    ba0_push_another_stack ();
    ba0_record (&M);
/*
   Observe that one cannot directly modify R for R and A might be the
   same variable !
*/
    ba0_init_table ((ba0_table)&T);
    ba0_realloc_table ((ba0_table)&T, 3);
    T.size = 3;
    T.tab [0] = A->tab [1]*B->tab [2] - A->tab [2]*B->tab [1];
    T.tab [1] = - A->tab [0]*B->tab [2] + A->tab [2]*B->tab [0];
    T.tab [2] = A->tab [0]*B->tab [1] - A->tab [1]*B->tab [0];
/*
   Intermediate computations are over.
   The intermediate variable T is copied into R, in the stack which was
   current when the function was called i.e. in the stack of R.
*/
    ba0_pull_stack ();
    ba0_set_table ((ba0_table)R, (ba0_table)&T);
/*
   Just by moving the free pointer of the other stack, the memory allocated
   to T is freed.
*/
    ba0_restore (&M);
/*
   Observe that the whole mechanism works recursively.
   The two stacks play an equivalent role.
*/
}

int main ()
{   ba0_tableof_int_p R, A, B;

    ba0_restart (0, 0);

    R = (ba0_tableof_int_p)ba0_new_table ();
    A = (ba0_tableof_int_p)ba0_new_table ();
    B = (ba0_tableof_int_p)ba0_new_table ();

    ba0_sscanf2 ("[2, 3, 4]", "%t[%d]", A);
    ba0_sscanf2 ("[3, 1, -2]", "%t[%d]", B);

    wedge (R, A, B);

    ba0_printf ("%t[%d] ^ %t[%d] = %t[%d]\n", A, B, R);
    ba0_terminate (ba0_init_level);
    return 0;
}
